#ifndef CGUL_BINARY_SEARCH_H
#define CGUL_BINARY_SEARCH_H

/**
 * Binary search an array.
 *
 * @file
 * @author Paul Serice
 * @brief binary search
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/**
 * This typedef is the interface the client must define in order to
 * perform a binary search of the search key <code>search_key</code>.
 * The comparison function should return less than zero, zero, or
 * greater than zero if <code>search_key</code> is less than, equal
 * to, or greater than the array element <code>array_element</code>
 * respectively.
 *
 * @param[in] search_key search key
 * @param[in] array_element array element
 * @return result of comparison
 */
typedef int (*cgul_binary_search__compare_t)(const void* search_key,
                                             const void* array_element);


/**
 * Generic search with the same parameter signature as stdlib's
 * <code>bsearch()</code>.  <code>v</code> is a pointer to the top of
 * your array.  <code>v_count</code> is the number of elements in
 * <code>v</code>.  <code>element_size</code> is the number of
 * contiguous bytes that comprise each element.  The
 * <code>void*</code> return value is a pointer to the element that
 * matched <code>key</code>.  If more than one element matches
 * <code>key</code>, any of the matching elements could be returned.
 * If no match was found, the return value is <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] key value to search for
 * @param[in] v array of sorted elements to search
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] element_size size of each element in <code>v</code>
 * @param[in] f comparison function
 * @return pointer to matching element or <code>NULL</code> if there
 * was no match
 */
extern CGUL_EXPORT void*
cgul_binary_search(cgul_exception_t* cex,
                   const void* key,
                   const void* v,
                   size_t v_count,
                   size_t element_size,
                   cgul_binary_search__compare_t f);

/**
 * If your array is already composed of pointers, you can directly
 * call this function.  <code>v</code> will be searched using
 * <code>f</code> in the same manner as described in
 * <code>cgul_binary_search()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] key value to search for
 * @param[in] v array of pointers to sorted elements to search
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] f comparison function
 * @return pointer to matching element or <code>NULL</code> if there
 * was no match
 */
extern CGUL_EXPORT void*
cgul_binary_search_pointers(cgul_exception_t* cex,
                            const void* key,
                            const void** v,
                            size_t v_count,
                            cgul_binary_search__compare_t f);

/**
 * Peform a binary search for the smallest element that is greater
 * than or equal to the search key <code>key</code>.  This function
 * has the same parameter signature as stdlib's
 * <code>bsearch()</code>.  <code>v</code> is a pointer to the top of
 * your array.  <code>v_count</code> is the number of elements in
 * <code>v</code>.  <code>element_size</code> is the number of
 * contiguous bytes that comprise each element.  The
 * <code>void*</code> return value is a pointer to the ceiling
 * element.  If there is more than one ceiling element, any of the
 * matching elements could be returned.  If no match was found, the
 * return value is <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] key value to search for
 * @param[in] v array of sorted elements to search
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] element_size size of each element in <code>v</code>
 * @param[in] f comparison function
 * @return pointer to matching element or <code>NULL</code> if there
 * was no match
 */
extern CGUL_EXPORT void*
cgul_binary_search_ceiling(cgul_exception_t* cex,
                           const void* key,
                           const void* v,
                           size_t v_count,
                           size_t element_size,
                           cgul_binary_search__compare_t f);

/**
 * If your array is already composed of pointers, you can directly
 * call this function.  <code>v</code> will be searched using
 * <code>f</code> in the same manner as described in
 * <code>cgul_binary_search_ceiling()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] key value to search for
 * @param[in] v array of pointers to sorted elements to search
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] f comparison function
 * @return pointer to matching element or <code>NULL</code> if there
 * was no match
 */
extern CGUL_EXPORT void*
cgul_binary_search_ceiling_pointers(cgul_exception_t* cex,
                                    const void* key,
                                    const void** v,
                                    size_t v_count,
                                    cgul_binary_search__compare_t f);

/**
 * Peform a binary search for the largest element that is less than or
 * equal to the search key <code>key</code>.  This function has the
 * same parameter signature as stdlib's <code>bsearch()</code>.
 * <code>v</code> is a pointer to the top of your array.
 * <code>v_count</code> is the number of elements in <code>v</code>.
 * <code>element_size</code> is the number of contiguous bytes that
 * comprise each element.  The <code>void*</code> return value is a
 * pointer to the floor element.  If there is more than one floor
 * element, any of the matching elements could be returned.  If no
 * match was found, the return value is <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] key value to search for
 * @param[in] v array of sorted elements to search
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] element_size size of each element in <code>v</code>
 * @param[in] f comparison function
 * @return pointer to matching element or <code>NULL</code> if there
 * was no match
 */
extern CGUL_EXPORT void*
cgul_binary_search_floor(cgul_exception_t* cex,
                         const void* key,
                         const void* v,
                         size_t v_count,
                         size_t element_size,
                         cgul_binary_search__compare_t f);

/**
 * If your array is already composed of pointers, you can directly
 * call this function.  <code>v</code> will be searched using
 * <code>f</code> in the same manner as described in
 * <code>cgul_binary_search_floor()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] key value to search for
 * @param[in] v array of pointers to sorted elements to search
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] f comparison function
 * @return pointer to matching element or <code>NULL</code> if there
 * was no match
 */
extern CGUL_EXPORT void*
cgul_binary_search_floor_pointers(cgul_exception_t* cex,
                                  const void* key,
                                  const void** v,
                                  size_t v_count,
                                  cgul_binary_search__compare_t f);

/**
 * This function performs a binary search over the range of integers
 * <code>[low, high)</code> where <code>low</code> is the _inclusive_
 * lower bound of the range and <code>high</code> is the _exclusive_
 * upper bound of the range.  For each iteration, the user-supplied
 * callback function <code>f</code> will be called with the
 * user-supplied client data <code>data</code> and the next integer to
 * test.  <code>f</code> should return less than zero, zero, or
 * greater than zero if the integer passed into <code>f</code> is less
 * than, equal to, or greater than the target index. <p>
 *
 * This function only returns whether a match was found.  If the index
 * where the match occurred is required, the client must arrange for
 * it to be returned through the client data <code>data</code>. <p>
 *
 * To perform a binary search over an array, call this function as
 * follows: <p>
 *
 * <pre>
 *     binary_search_irange(0, length, f, data);
 * </pre>
 *
 * However, to perform a binary search over the inclusive mathematical
 * range [xmin, xmax], call this function as follows: <p>
 *
 * <pre>
 *     binary_search_irange(xmin, xmax + 1, f, data);
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] low inclusive lower bound
 * @param[in] high exclusive upper bound
 * @param[in] f callback function
 * @param[in] data client data
 * @return whether a match was found
 */
extern CGUL_EXPORT int
cgul_binary_search_irange(cgul_exception_t* cex,
                          int low,
                          int high,
                          int (*f)(int, void*),
                          void* data);

CGUL_END_C


#endif  /*  CGUL_BINARY_SEARCH_H  */
