#ifndef CGUL_CRC32_CXX_H
#define CGUL_CRC32_CXX_H

/**
 * C++ wrapper for <code>cgul_crc32</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_crc32</code>
 */

#include "cgul_crc32.h"
#include "cgul_exception_cxx.h"

/**
 * This class provides the C++ bindings for C <code>cgul_crc32</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_crc32</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_crc32</code>
 */
class cgul_crc32_cxx {

public:

    /**
     * Create a new <code>cgul_crc32_cxx</code> object.  Objects
     * created with this constructor do not append the size to the
     * data stream that is CRCed.  If memory cannot be allocated, an
     * exception is thrown.
     *
     * @param[in] g generator polynomial
     * @param[in] init initial value loaded into the accumulator
     * @param[in] final_xor value to xor with the calculated crc
     * @param[in] reflected whether to use the reflected algorithm
     */
    cgul_crc32_cxx(unsigned long g,
                   unsigned long init,
                   unsigned long final_xor,
                   int reflected)
    {
        cgul_exception_t local = NULL;
        obj = cgul_crc32__new(&local, g, init, final_xor, reflected);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_crc32</code> object.  Objects created
     * with this constructor append the size to the data stream that
     * is CRCed.  If memory cannot be allocated, an exception is
     * thrown.
     *
     * @param[in] g polynomial
     * @param[in] init initial value loaded into the accumulator
     * @param[in] final_xor value to xor with the calculated crc
     * @param[in] reflected whether to use the reflected algorithm
     * @param[in] size_xor value to xor with the size before appending
     */
    cgul_crc32_cxx(unsigned long g,
                   unsigned long init,
                   unsigned long final_xor,
                   int reflected,
                   size_t size_xor)
    {
        cgul_exception_t local = NULL;
        obj = cgul_crc32__new_appends_size(&local,
                                           g,
                                           init,
                                           final_xor,
                                           reflected,
                                           size_xor);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Destructor.
     */
    virtual
    ~cgul_crc32_cxx()
    {
        if (obj) {
            cgul_crc32__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Assign the current accumulator and size of <code>rhs</code> to
     * this object.  This operation probably only makes sense if
     * this object and <code>rhs</code> have been constructed
     * identically.  If an error occurs, an exception is thrown.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    assign(cgul_crc32_cxx* rhs)
    {
        if (rhs) {
            cgul_exception_t local = NULL;
            cgul_crc32__assign(&local, obj, rhs->obj);
            if (local) {
                throw cgul_exception_cxx(local);
            }
        }
    }

    /**
     * This method completes the CRC calculation by performing the size
     * XOR and the final XOR.  You will generally call one of the update()
     * methods before calling this method.  When this method returns, the
     * internal state of the object is reset so that it is ready to
     * calculate the next CRC.  The return value is the final value for
     * the CRC.
     *
     * @return final crc value
     */
    virtual unsigned long
    crc()
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crc32__crc(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This is the method to call when you only have one block for which
     * you want to calculate a CRC.  You can also call this if you have
     * called one of the update() methods before, and you know that
     * <code>block</code> is your last block.  When this method returns,
     * the internal state of the object is reset so that it is ready to
     * calculate the next CRC.  <code>length</code> is the number of bytes
     * inside of <code>block</code> that you want to add to the CRC
     * calculation.  The return value is the final value for the CRC.
     *
     * @param[in] block block of data to add to the crc
     * @param[in] length length of <code>block</code>
     * @return final crc value
     */
    virtual unsigned long
    crc_block(const void* block,
              size_t length)
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crc32__crc_block(&local, obj, block, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This is the method to call when you want to calculate the CRC from
     * a range of bytes from one block.  You can also call this if you
     * have called one of the update() methods before, and you know that
     * <code>block</code> is your last block.  When this method returns,
     * the internal state of the object is reset so that it is ready to
     * calculate the next CRC.  <code>offset</code> is the index into
     * <code>block</code> where you want to add <code>length</code> bytes
     * to the CRC calculation.  The return value is the final value for
     * the CRC.
     *
     * @param[in] block block of data to add to the crc
     * @param[in] offset offset into <code>block</code>
     * @param[in] length length of <code>block</code>
     * @return final crc value
     */
    virtual unsigned long
    crc_range_of_block(const void* block,
                       size_t offset,
                       size_t length)
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crc32__crc_range_of_block(&local,
                                            obj,
                                            block,
                                            offset,
                                            length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This is the method to call when you only have one file for which
     * you want to calculate a CRC.  You can also call this if you have
     * called one of the update() methods before (like
     * <code>cgul_crc32_cxx::update_file()</code>), and you know that
     * <code>f</code> is your last file.  When this method returns, the
     * internal state of the object is reset so that it is ready to
     * calculate the next CRC.  The return value is the final value for
     * the CRC.
     *
     * @param[in] f input file
     * @return final crc value
     */
    virtual unsigned long
    crc_file(FILE* f)
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crc32__crc_file(&local, obj, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This is the method to call when you only have one file for which
     * you want to calculate a CRC.  You can also call this if you have
     * called one of the update() methods before (like
     * <code>cgul_crc32_cxx::update_fname()</code>), and you know that
     * <code>fname</code> is your last file.  When this method returns,
     * the internal state of the object is reset so that it is ready to
     * calculate the next CRC.  The return value is the final value for
     * the CRC.
     *
     * @param[in] fname input file name
     * @return final crc value
     */
    virtual unsigned long
    crc_fname(const char* fname)
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crc32__crc_fname(&local, obj, fname);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method updates the CRC by adding the byte <code>c</code> to
     * the CRC calculation.
     *
     * @param[in] c byte to add to crc
     */
    virtual void
    update_byte(unsigned char c)
    {
        cgul_exception_t local = NULL;
        cgul_crc32__update_byte(&local, obj, c);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method updates the CRC by adding all the bytes in
     * <code>block</code> to the calculation.
     *
     * @param[in] block block of data to add to the crc
     * @param[in] length length of <code>block</code>
     */
    virtual void
    update_block(const void* block,
                 size_t length)
    {
        cgul_exception_t local = NULL;
        cgul_crc32__update_block(&local, obj, block, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method updates the CRC by adding a range of bytes in
     * <code>block</code> to the calculation.  The range starts at
     * <code>offset</code> and extends <code>length</code> bytes.
     *
     * @param[in] block block of data to add to the crc
     * @param[in] offset offset into <code>block</code>
     * @param[in] length length of <code>block</code>
     */
    virtual void
    update_range_of_block(const void* block,
                          size_t offset,
                          size_t length)
    {
        cgul_exception_t local = NULL;
        cgul_crc32__update_range_of_block(&local, obj, block, offset, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method updates the CRC by adding all the bytes (starting at
     * the current offset) in the file <code>f</code> to the calculation.
     * If an error occurs, an exception is thrown.
     *
     * @param[in] f input file
     */
    virtual void
    update_file(FILE* f)
    {
        cgul_exception_t local = NULL;
        cgul_crc32__update_file(&local, obj, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method updates the CRC by adding all the bytes in the file
     * with name <code>fname</code> to the calculation.  If an error
     * occurs, an exception is thrown.
     *
     * @param[in] fname input file name
     */
    virtual void
    update_fname(const char* fname)
    {
        cgul_exception_t local = NULL;
        cgul_crc32__update_fname(&local, obj, fname);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Reset the object for further use.
     */
    virtual void
    reset()
    {
        cgul_exception_t local = NULL;
        cgul_crc32__reset(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * You can use this method to see the CRC lookup table that is being
     * used.  The returned value is the CRC lookup table for this object
     * as a string.  The caller is responsible for calling free() on the
     * string that is returned.
     *
     * @return crc lookup table as a string
     */
    virtual char*
    get_table_as_string() const
    {
        char* rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crc32__get_table_as_string(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the underlying <code>cgul_crc32</code> object.
     *
     * @return underlying object
     */
    virtual cgul_crc32_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_crc32</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_crc32__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_crc32_t
    take_obj()
    {
        cgul_crc32_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_crc32_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_crc32__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_crc32_cxx();
    cgul_crc32_cxx(const cgul_crc32_cxx&);
    cgul_crc32_cxx& operator=(const cgul_crc32_cxx&);

    /* Underlying <code>cgul_crc32</code> object. */
    cgul_crc32_t obj;

};

#endif  /*  CGUL_CRC32_CXX_H  */
