#ifndef CGUL_CRC32_INCR_CXX_H
#define CGUL_CRC32_INCR_CXX_H

/**
 * C++ wrapper for <code>cgul_crc32_incr</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_crc32_incr</code>
 */

#include "cgul_crc32_incr.h"
#include "cgul_exception_cxx.h"

/**
 * This class provides the C++ bindings for C <code>cgul_crc32_incr</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_crc32_incr</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_crc32_incr</code>
 */
class cgul_crc32_incr_cxx {

public:

    /**
     * Create a new <code>cgul_crc32_incr_cxx</code> object.  Objects
     * created with this constructor do not append the size to the
     * data stream that is CRCed.  The client is responsible for
     * calling <code>delete()</code> on the object returned.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] g generator polynomial
     * @param[in] init initial value loaded into the accumulator
     * @param[in] final_xor value to xor with the calculated crc
     * @param[in] reflected whether to use the reflected algorithm
     * @see cgul_crc32__new()
     */
    cgul_crc32_incr_cxx(unsigned long g,
                        unsigned long init,
                        unsigned long final_xor,
                        int reflected)
    {
        cgul_exception_t local = NULL;
        obj = cgul_crc32_incr__new(&local, g, init, final_xor, reflected);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_crc32_incr_cxx</code> object.  Objects
     * created with this constructor append the size to the data
     * stream that is CRCed.  The client is responsible for calling
     * <code>delete()</code> on the object returned.  If an error
     * occurs, an exception is thrown.
     *
     * @param[in] g polynomial
     * @param[in] init initial value loaded into the accumulator
     * @param[in] final_xor value to xor with the calculated crc
     * @param[in] reflected whether to use the reflected algorithm
     * @param[in] size_xor value to xor with the size before appending
     * @see cgul_crc32__new_appends_size()
     */
    cgul_crc32_incr_cxx(unsigned long g,
                        unsigned long init,
                        unsigned long final_xor,
                        int reflected,
                        size_t size_xor)
    {
        cgul_exception_t local = NULL;
        obj = cgul_crc32_incr__new_appends_size(
                  &local, g, init, final_xor, reflected, size_xor);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Destructor.
     */
    virtual
    ~cgul_crc32_incr_cxx()
    {
        if (obj) {
            cgul_crc32_incr__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Assign the current state of of <code>rhs</code> to
     * <code>this</code>.  This operation probably only makes sense if
     * <code>this</code> and <code>rhs</code> have been constructed
     * identically.  If an error occurs, an exception is thrown.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    assign(const cgul_crc32_incr_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_crc32_incr__assign(&local, obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return;
    }

    /**
     * This method completes the CRC calculation by performing the size
     * XOR and the final XOR.  Unlike <code>cgul_crc32_cxx::crc()</code>, this
     * method does not reset the CRC-32 calculation.  Instead, the current
     * CRC-32 calculation can be continued with subsequent calls to
     * <code>cgul_crc32_incr_cxx::update_block()</code>.  The return value is
     * the current value for the entire CRC-32 calculation.  If an error
     * occurs, an exception is thrown.
     *
     * @return final crc value
     */
    virtual unsigned long
    crc()
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crc32_incr__crc(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method updates the CRC by adding all the bytes in
     * <code>block</code> to the calculation.
     *
     * @param[in] block block of data to add to the crc
     * @param[in] length length of <code>block</code>
     */
    virtual void
    update_block(const void* block,
                 size_t length)
    {
        cgul_exception_t local = NULL;
        cgul_crc32_incr__update_block(&local, obj, block, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return;
    }

    /**
     * This method updates the CRC by adding all the bytes (starting at
     * the current offset) in the file <code>f</code> to the calculation.
     * If an error occurs, an exception is thrown.
     *
     * @param[in] f input file
     */
    virtual void
    update_file(FILE* f)
    {
        cgul_exception_t local = NULL;
        cgul_crc32_incr__update_file(&local, obj, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method updates the CRC by adding all the bytes in the file
     * with name <code>fname</code> to the calculation.  If an error
     * occurs, an exception is thrown.
     *
     * @param[in] fname input file name
     */
    virtual void
    update_fname(const char* fname)
    {
        cgul_exception_t local = NULL;
        cgul_crc32_incr__update_fname(&local, obj, fname);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Reset to the initial internal state.
     */
    virtual void
    reset()
    {
        cgul_exception_t local = NULL;
        cgul_crc32_incr__reset(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return;
    }

    /**
     * Get the underlying <code>cgul_crc32_incr</code> object.
     *
     * @return underlying object
     */
    virtual cgul_crc32_incr_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_crc32_incr</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_crc32_incr__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_crc32_incr_t
    take_obj()
    {
        cgul_crc32_incr_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_crc32_incr_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_crc32_incr__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_crc32_incr_cxx();
    cgul_crc32_incr_cxx(const cgul_crc32_incr_cxx& rhs);
    cgul_crc32_incr_cxx& operator=(const cgul_crc32_incr_cxx& rhs);

    /* Underlying <code>cgul_crc32_incr</code> object. */
    cgul_crc32_incr_t obj;

};

#endif  /*  CGUL_CRC32_INCR_CXX_H  */
