#ifndef CGUL_DEFENSIVE_H
#define CGUL_DEFENSIVE_H

/**
 * Defensive addition operations that add two positive addends and
 * throw an exception if overflow is detected.
 *
 * @file
 * @author Paul Serice
 * @brief defensive addition
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/**
 * This function detects whether the <code>char</code> type is
 * <code>signed</code> or <code>unsigned</code> and forwards the call
 * to either <code>cgul_defensive__add_signed_char()</code> or
 * <code>cgul_defensive__add_unsigned_char()</code> whichever is
 * correct.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x1 first positive addend
 * @param[in] x2 second positive addend
 * @return x1 + x2
 * @see cgul_defensive__add_signed_char
 * @see cgul_defensive__add_unsigned_char
 */
extern CGUL_EXPORT char
cgul_defensive__add_char(cgul_exception_t* cex,
                         char x1,
                         char x2);

/**
 * This function returns the sum of the positive addends
 * <code>x1</code> and <code>x2</code>.  If overflow or negative
 * addends are detected, an exception is thrown, and <code>0</code> is
 * returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x1 first positive addend
 * @param[in] x2 second positive addend
 * @return x1 + x2
 */
extern CGUL_EXPORT signed char
cgul_defensive__add_signed_char(cgul_exception_t* cex,
                                signed char x1,
                                signed char x2);

/**
 * This function returns the sum of the addends <code>x1</code> and
 * <code>x2</code>.  If overflow is detected, an exception is thrown,
 * and <code>0</code> is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x1 first addend
 * @param[in] x2 second addend
 * @return x1 + x2
 */
extern CGUL_EXPORT unsigned char
cgul_defensive__add_unsigned_char(cgul_exception_t* cex,
                                  unsigned char x1,
                                  unsigned char x2);

/**
 * This function returns the sum of the positive addends
 * <code>x1</code> and <code>x2</code>.  If overflow or negative
 * addends are detected, an exception is thrown, and <code>0</code> is
 * returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x1 first positive addend
 * @param[in] x2 second positive addend
 * @return x1 + x2
 */
extern CGUL_EXPORT short
cgul_defensive__add_short(cgul_exception_t* cex,
                          short x1,
                          short x2);

/**
 * This function returns the sum of the addends <code>x1</code> and
 * <code>x2</code>.  If overflow is detected, an exception is thrown,
 * and <code>0</code> is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x1 first addend
 * @param[in] x2 second addend
 * @return x1 + x2
 */
extern CGUL_EXPORT unsigned short
cgul_defensive__add_unsigned_short(cgul_exception_t* cex,
                                   unsigned short x1,
                                   unsigned short x2);

/**
 * This function returns the sum of the positive addends
 * <code>x1</code> and <code>x2</code>.  If overflow or negative
 * addends are detected, an exception is thrown, and <code>0</code> is
 * returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x1 first positive addend
 * @param[in] x2 second positive addend
 * @return x1 + x2
 */
extern CGUL_EXPORT int
cgul_defensive__add_int(cgul_exception_t* cex,
                        int x1,
                        int x2);

/**
 * This function returns the sum of the addends <code>x1</code> and
 * <code>x2</code>.  If overflow is detected, an exception is thrown,
 * and <code>0</code> is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x1 first addend
 * @param[in] x2 second addend
 * @return x1 + x2
 */
extern CGUL_EXPORT unsigned int
cgul_defensive__add_unsigned(cgul_exception_t* cex,
                             unsigned int x1,
                             unsigned int x2);

/**
 * This function returns the sum of the positive addends
 * <code>x1</code> and <code>x2</code>.  If overflow or negative
 * addends are detected, an exception is thrown, and <code>0</code> is
 * returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x1 first positive addend
 * @param[in] x2 second positive addend
 * @return x1 + x2
 */
extern CGUL_EXPORT long
cgul_defensive__add_long(cgul_exception_t* cex,
                         long x1,
                         long x2);

/**
 * This function returns the sum of the addends <code>x1</code> and
 * <code>x2</code>.  If overflow is detected, an exception is thrown,
 * and <code>0</code> is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x1 first addend
 * @param[in] x2 second addend
 * @return x1 + x2
 */
extern CGUL_EXPORT unsigned long
cgul_defensive__add_unsigned_long(cgul_exception_t* cex,
                                  unsigned long x1,
                                  unsigned long x2);

/**
 * This function returns the sum of the addends <code>x1</code> and
 * <code>x2</code>.  If overflow is detected, an exception is thrown,
 * and <code>0</code> is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x1 first addend
 * @param[in] x2 second addend
 * @return x1 + x2
 */
extern CGUL_EXPORT size_t
cgul_defensive__add_size_type(cgul_exception_t* cex,
                              size_t x1,
                              size_t x2);

CGUL_END_C

#endif  /*  CGUL_DEFENSIVE_H  */
