#ifndef CGUL_HASH_NODE_CXX_H
#define CGUL_HASH_NODE_CXX_H

/**
 * C++ wrapper for <code>cgul_hash_node</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_hash_node</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_hash_node.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_hash_node</code> objects.  The main purpose of this
 * class is to convert the C-style function calls and exception
 * handling in <code>cgul_hash_node</code> into C++-style function
 * calls and exception handling.  This class must not be extended.
 *
 * @brief C++ bindings for <code>cgul_hash_node</code>
 */
class cgul_hash_node_cxx {

public:

    /*
     * IMPLEMENTATION NOTE:
     * ===================
     *
     * There are no real cgul_hash_node_cxx objects.  Instead, this
     * class is carefully implemented so that "cgul_hash_node"
     * objects (not "cgul_hash_node_cxx" objects) can be directly
     * used to invoke the methods of this class.
     *
     * This works because this class has no associated data _and_ no
     * virtual methods which would cause a vtable lookup, and vtables
     * are typically embedded in the object itself.
     *
     * In summary, do _not_ add data or virtual functions to this
     * class.
     */

    friend class cgul_hash_cxx;

    /**
     * Get the key for this node.
     *
     * @return key
     */
    const void*
    get_key() const
    {
        const void* rv = NULL;
        cgul_hash_node_t n = (cgul_hash_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_hash_node__get_key(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the value for this node.
     *
     * @return value
     */
    void*
    get_value() const
    {
        void* rv = NULL;
        cgul_hash_node_t n = (cgul_hash_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_hash_node__get_value(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the value for this node.
     *
     * @param[in] value new value
     */
    void
    set_value(const void* value)
    {
        cgul_hash_node_t n = (cgul_hash_node_t)this;
        cgul_exception_t local = NULL;
        cgul_hash_node__set_value(&local, n, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the previous node in insertion order.  This method is a
     * synonym for <code>get_older()</code>.  You can use this method
     * to iterate over the hash in reverse insertion order.  When
     * there are no more nodes, <code>NULL</code> is returned.  You
     * can use <code>cgul_hash::get_back()</code> to get the last node
     * in a hash in order to start iterating.
     *
     * @return previous node
     * @see get_older()
     */
    cgul_hash_node_cxx*
    get_prev() const
    {
        cgul_hash_node_t rv = NULL;
        cgul_hash_node_t n = (cgul_hash_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_hash_node__get_prev(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_hash_node_cxx*)rv;
    }

    /**
     * Get the next node in insertion order.  This method is a synonym
     * for <code>get_younger()</code>.  You can use this method to
     * iterate over the hash in insertion order.  When there are no
     * more nodes, <code>NULL</code> is returned.  You can use
     * <code>cgul_hash::get_front()</code> to get the first node in a
     * cgul_hash in order to start iterating.
     *
     * @return previous node
     * @see get_younger()
     */
    cgul_hash_node_cxx*
    get_next() const
    {
        cgul_hash_node_t rv = NULL;
        cgul_hash_node_t n = (cgul_hash_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_hash_node__get_next(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_hash_node_cxx*)rv;
    }

    /**
     * Return the next older node.  Together with
     * <code>get_younger()</code>, this method lets you traverse the
     * hash in chronological order. If there is no older node, this
     * method returns <code>NULL</code>. <p>
     *
     * The following example shows how to iterate over the entire hash
     * in reverse chronological order.  Notice that you have to start
     * with the youngest node when calling <code>get_older()</code>:
     *
     * <pre>
     *     cgul_hash_node_cxx* n = h->get_youngest();
     *     for ( ; n ; n = n->get_older()) {
     *         ...
     *     }
     * </pre>
     *
     * @return older node
     */
    cgul_hash_node_cxx*
    get_older() const
    {
        cgul_hash_node_t rv = NULL;
        cgul_hash_node_t n = (cgul_hash_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_hash_node__get_older(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_hash_node_cxx*)rv;
    }

    /**
     * Return the next younger node.  Together with
     * <code>get_older()</code>, this method lets you traverse the
     * hash in chronological order. If there is no younger node, this
     * method returns <code>NULL</code>. <p>
     *
     * The following example shows how to iterate over the entire hash
     * in chronological order.  Notice that you have to start with the
     * oldest node when calling <code>get_younger()</code>:
     *
     * <pre>
     *     cgul_hash_node_cxx* n = h->get_oldest();
     *     for ( ; n ; n = n->get_younger()) {
     *         ...
     *     }
     * </pre>
     *
     * @return younger node
     */
    cgul_hash_node_cxx*
    get_younger() const
    {
        cgul_hash_node_t rv = NULL;
        cgul_hash_node_t n = (cgul_hash_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_hash_node__get_younger(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_hash_node_cxx*)rv;
    }

private:

    /* Hide these. */
    cgul_hash_node_cxx();
    cgul_hash_node_cxx(const cgul_hash_node_cxx&);
    ~cgul_hash_node_cxx();
    cgul_hash_node_cxx& operator=(const cgul_hash_node_cxx&);

};

#endif  /*  CGUL_HASH_NODE_CXX_H  */
