#ifndef CGUL_LIBC_CXX_H
#define CGUL_LIBC_CXX_H

/**
 * C++ wrapper for <code>cgul_libc</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_libc</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_libc.h"

/**
 * This class provides the C++ bindings for C <code>cgul_libc</code>.
 * The main purpose of this class is to convert the C-style function
 * calls and exception handling in <code>cgul_libc</code> into
 * C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_libc</code>
 */
class cgul_libc_cxx {

public:

    /**
     * This method lets you test out-of-memory exception handling by
     * forcing <code>cgul_libc__calloc()</code>,
     * <code>cgul_libc__malloc()</code>, and
     * <code>cgul_libc__realloc()</code> to skip memory allocation and
     * immediately return <code>NULL</code>.  You should set
     * <code>out_of_memory</code> immediately before the allocation call
     * you want to test and clear it immediately after.
     *
     * @param[in] out_of_memory whether to simulate out-of-memory condition
     */
    static void
    set_out_of_memory(int out_of_memory)
    {
        cgul_exception_t local = NULL;
        cgul_libc__set_out_of_memory(&local, out_of_memory);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Wrapper for <code>calloc()</code>.  On success, a pointer to a
     * block of memory that is large enough to hold an array of
     * <code>element_count</code> elements is returned.  The block of
     * memory is initialized with all zeros, and the client must call
     * <code>free()</code> on the value returned.  On failure, an
     * exception is thrown, and <code>NULL</code> is returned.
     *
     * @param[in] element_count number of elements to allocate space for
     * @param[in] element_size size of one element
     * @return pointer to a block of memory
     */
    static void*
    calloc(size_t element_count,
           size_t element_size)
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__calloc(&local, element_count, element_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>malloc()</code>.  On success, a pointer to a
     * block of memory that is <code>new_size</code> bytes long is
     * returned uninitialized, and the client must call
     * <code>free()</code> on the value returned.  On failure, an
     * exception is thrown, and <code>NULL</code> is returned.
     *
     * @param[in] new_size new size for the block of memory to return
     * @return point to a block of memory
     */
    static void*
    malloc(size_t new_size)
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__malloc(&local, new_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>realloc()</code>.  On success, a pointer to
     * memory that holds <code>new_size</code> bytes is returned, and the
     * client must eventually call <code>free()</code> on the value
     * returned.  The client must not call <code>free()</code> on
     * <code>orig_pointer</code> if this function returns
     * successfully. <p>
     *
     * If an error occurs, an exception will be thrown, and the client
     * will still need to call <code>free()</code> on
     * <code>orig_pointer</code>.  (Notice that this is different
     * behavior from <code>realloc()</code> which returns
     * <code>NULL</code> if an error occurs.)
     *
     * @param[in] orig_pointer original pointer
     * @param[in] new_size new size for the pointer that is returned
     * @return pointer to a block of memory
     */
    static void*
    realloc(void* orig_pointer,
            size_t new_size)
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__realloc(&local, orig_pointer, new_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>memset()</code>.  Set a region of memory pointed
     * to by <code>buf</code> by repeatedly copying <code>c</code>.
     *
     * @param[in] buf buffer where memory will be set
     * @param[in] c byte repeatedly used to set the memory region
     * @param[in] n buffer size
     * @return <code>buf</code>
     */
    static void*
    memset(void *buf,
           int c,
           size_t n)
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__memset(&local, buf, c, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>memcpy()</code>.  Copy the first <code>n</code>
     * bytes from the <code>src</code> buffer to the <code>dst</code>
     * buffer.  It is important that the <code>src</code> and
     * <code>dst</code> buffers do not overlap.  If they do, you should
     * use <code>memmove()</code> instead.
     *
     * @param[in] dst destination buffer
     * @param[in] src source buffer
     * @param[in] n buffer size
     * @return <code>dst</code>
     */
    static void*
    memcpy(void* dst,
           const void* src,
           size_t n)
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__memcpy(&local, dst, src, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>memmove()</code>.  Copy the first <code>n</code>
     * bytes from the <code>src</code> buffer to the <code>dst</code>
     * buffer.  The <code>src</code> and <code>dst</code> buffers may
     * overlap.
     *
     * @param[in] dst destination buffer
     * @param[in] src source buffer
     * @param[in] n buffer size
     * @return <code>dst</code>
     */
    static void*
    memmove(void* dst,
            const void* src,
            size_t n)
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__memmove(&local, dst, src, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>fopen()</code>.  It opens the file with name
     * <code>fname</code> using the mode specified by
     * <code>mode</code>.  If an error occurs, an exception is thrown.
     *
     * @param[in] fname file name
     * @param[in] mode open file mode
     * @return opened file
     */
    static FILE*
    fopen(const char* fname,
          const char* mode)
    {
        FILE* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__fopen(&local, fname, mode);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>freopen()</code>.  It reopens <code>f</code>
     * so that the underlying file descriptor points to
     * <code>fname</code> opened using <code>mode</code>.  If an error
     * occurs, an exception is thrown.
     *
     * @param[in] fname file name
     * @param[in] mode open file mode
     * @param[in] f file object to reopen
     * @return opened file
     */
    static FILE*
    freopen(const char* fname,
            const char* mode,
            FILE* f)
    {
        FILE* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__freopen(&local, fname, mode, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }
    
    /**
     * Wrapper for <code>fclose()</code>.  It closes <code>f</code>.  If
     * an error occurs, an exception is thrown.  You should be careful
     * using this function because you will almost always want to close
     * files if the main part of your code throws an exception.  Thus, if
     * the stack is unwinding, to avoid this function from overwriting the
     * current exception, you should probably call <code>fclose()</code>
     * directly instead of calling this function.
     *
     * @param[in] f file object to close
     */
    static void
    fclose(FILE *f)
    {
        cgul_exception_t local = NULL;
        cgul_libc__fclose(&local, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Wrapper for <code>fflush()</code>.  It flushes the file
     * <code>f</code>, or if <code>f</code> is <code>NULL</code>, it
     * flushes all open files.  If an error occurs, an exception is
     * thrown.
     *
     * @param[in] f file object to close
     */
    static void
    fflush(FILE* f)
    {
        cgul_exception_t local = NULL;
        cgul_libc__fflush(&local, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Wrapper for <code>fread()</code>.  It reads
     * <code>element_count</code> elements of size
     * <code>element_size</code> each into <code>buf</code> from
     * <code>f</code>.  On success, the number of elements read (which is
     * not necessarily the same as the number of bytes read) is returned.
     * If EOF is reached, a byte count of 0 is returned.  On failure,
     * an exception is thrown.
     *
     * @param[in] buf destination buffer
     * @param[in] element_size size of each element to read
     * @param[in] element_count number of elements to read
     * @param[in] f file from which to read
     * @return number of elements read
     */
    static size_t
    fread(void* buf,
          size_t element_size,
          size_t element_count,
          FILE* f)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__fread(&local, buf, element_size, element_count, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>fwrite()</code>.  It writes
     * <code>element_count</code> elements of size
     * <code>element_size</code> each from <code>buf</code> into
     * <code>f</code>.  On success, the number of elements written (which
     * is not necessarily the same as the number of bytes written) is
     * returned.  On failure, an exception is thrown.
     *
     * @param[in] buf source buffer
     * @param[in] element_size size of each element to write
     * @param[in] element_count number of elements to write
     * @param[in] f file to which to write
     * @return number of elements written
     * @see fwrite_block()
     */
    static size_t
    fwrite(const void* buf,
           size_t element_size,
           size_t element_count,
           FILE* f)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__fwrite(&local, buf, element_size, element_count, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>cgul_libc__fwrite_block()</code>.  It writes
     * the block <code>block</code> of size <code>block_size</code>
     * bytes to the output file <code>fout</code>.  Unlike
     * <code>fwrite()</code>, the write is restarted if interrupted.
     * This function returns the total number of bytes written.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] block block
     * @param[in] block_size size of block in bytes
     * @param[in] fout output file
     * @return total number of bytes written
     * @see fwrite()
     */
    static size_t
    fwrite_block(const void* block,
                 size_t block_size,
                 FILE* fout)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__fwrite_block(&local, block, block_size, fout);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Load the file <code>fin</code> into memory saving a copy of the
     * starting location of the file in memory to <code>*ptr</code> and
     * its length to <code>*length</code> if <code>length</code> is not
     * <code>NULL</code>.  The caller is responsible for calling
     * <code>free()</code> on <code>*ptr</code>.  If an error occurs, an
     * exception is thrown.
     *
     * @param[in] fin input file
     * @param[out] ptr pointer to in-memory copy of file
     * @param[out] length length in bytes of the in-memory copy of the file
     * @see save_file()
     */
    static void
    load_file(FILE* fin,
              char** ptr,
              size_t* length)
    {
        cgul_exception_t local = NULL;
        cgul_libc__load_file(&local, fin, ptr, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Load the file with name <code>fname</code> into memory saving a
     * copy of the starting location of the file in memory to
     * <code>*ptr</code> and its length to <code>*length</code> if
     * <code>length</code> is not <code>NULL</code>.  The caller is
     * responsible for calling <code>free()</code> on <code>*ptr</code>.
     * If an error occurs, an exception is thrown.
     *
     * @param[in] fname name of input file
     * @param[out] ptr pointer to in-memory copy of file
     * @param[out] length length in bytes of the in-memory copy of the file
     * @see save_fname()
     */
    static void
    load_fname(const char* fname,
               char** ptr,
               size_t* length)
    {
        cgul_exception_t local = NULL;
        cgul_libc__load_fname(&local, fname, ptr, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Save the binary block of memory <code>block</code> extending for
     * <code>block_size</code> bytes to the output file <code>fout</code>.
     * If an error occurs, an exception is thrown.
     *
     * @param[in] fout output file
     * @param[in] block block
     * @param[in] block_size size of block in bytes
     * @see load_file()
     */
    static void
    save_file(FILE* fout,
              const char* block,
              size_t block_size)
    {
        cgul_exception_t local = NULL;
        cgul_libc__save_file(&local, fout, block, block_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Save the binary block of memory <code>block</code> extending for
     * <code>block_size</code> bytes to the output file with name
     * <code>fname</code>.  If an error occurs, an exception is thrown.
     *
     * @param[in] fname name of the output file
     * @param[in] block block
     * @param[in] block_size size of block in bytes
     * @see load_fname()
     */
    static void
    save_fname(const char* fname,
               const char* block,
               size_t block_size)
    {
        cgul_exception_t local = NULL;
        cgul_libc__save_fname(&local, fname, block, block_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Copy everything from <code>fin</code> to <code>fout</code>.  If an
     * error occurs, an exception is thrown.  This is not a wrapper around
     * any libc function.
     *
     * @param[in] fin input file
     * @param[in] fout output file
     */
    static void
    copy_file(FILE* fin,
              FILE* fout)
    {
        cgul_exception_t local = NULL;
        cgul_libc__copy_file(&local, fin, fout);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Wrapper for <code>ftell()</code>.  On success, the file position of
     * <code>f</code> is returned.  On failure, an exception is thrown.
     * 
     * @param[in] f file
     * @return file position
     */
    static long
    ftell(FILE* f)
    {
        long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__ftell(&local, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>fseek()</code>.  On success, the file is
     * positioned according to the values of <code>offset</code> and
     * <code>whence</code> where <code>offset</code> is relative to the
     * beginning of the file, the current position, or the end of the file
     * if <code>whence</code> is <code>SEEK_SET</code>,
     * <code>SEEK_CUR</code>, or <code>SEEK_END</code> respectively.  On
     * failure, an exception is thrown.
     * 
     * @param[in] f file
     * @param[in] offset offset
     * @param[in] whence position to which offset is relative
     */
    static void
    fseek(FILE* f,
          long offset,
          int whence)
    {
        cgul_exception_t local = NULL;
        cgul_libc__fseek(&local, f, offset, whence);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Wrapper for <code>rewind()</code> to rewind <code>f</code>.  An
     * exception is thrown if <code>f</code> is not seekable.
     *
     * @param[in] f file
     */
    static void
    rewind(FILE* f)
    {
        cgul_exception_t local = NULL;
        cgul_libc__rewind(&local, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Wrapper for <code>fgetpos()</code>.  On success, the current file
     * position for <code>f</code> is stored in pos <code>pos</code>.  On
     * failure, an exception is thrown.
     *
     * @param[in] f file
     * @param[out] pos current file position
     */
    static void
    fgetpos(FILE* f,
            fpos_t* pos)
    {
        cgul_exception_t local = NULL;
        cgul_libc__fgetpos(&local, f, pos);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Wrapper for <code>fsetpos()</code>.  On success, the current file
     * position for <code>f</code> is set to <code>pos</code>.  On
     * failure, an exception is thrown.
     *
     * @param[in] f file
     * @param[out] pos current file position
     */
    static void
    fsetpos(FILE* f,
            fpos_t* pos)
    {
        cgul_exception_t local = NULL;
        cgul_libc__fsetpos(&local, f, pos);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * Wrapper for <code>tmpfile()</code>.  On success, a file opened
     * for reading and writting is returned.  On failure, an exception
     * is thrown.  When opening a file using glibc-2.3.4,
     * <code>strace</code> returned the following which says the file
     * is opened without race conditions and is immediately unlinked
     * so that the file is removed even if your program exits
     * unexpectedly:
     *
     * <pre>
     *     open("/tmp/tmpfGQda9F", O_RDWR|O_CREAT|O_EXCL, 0600) = 3
     *     unlink("/tmp/tmpfGQda9F")               = 0
     * </pre>
     *
     * @return temporary file
     */
    static FILE*
    tmpfile()
    {
        FILE* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__tmpfile(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>vprintf()</code>.  On success, the number of
     * characters printed to <code>stdout</code> is returned.  On
     * failure, an exception is thrown.
     *
     * @param[in] format format string
     * @param[in] ... variable argument list that matches <code>format</code>
     * @return number of characters read
     */
    static int
    printf(const char* format,
           ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
        __attribute__ ((format (printf, 1, 2)))
#endif
    {
        int rv = -1;
        cgul_exception_t local = NULL;
        va_list args;
        assert(format);
        va_start(args, format);
        rv = cgul_libc__vprintf(&local, format, args);
        va_end(args);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>vprintf()</code>.  On success, the number of
     * characters printed to <code>stdout</code> is returned.  On failure,
     * an exception is thrown.
     *
     * @param[in] format format string
     * @param[in] args variable argument list that matches <code>format</code>
     * @return number of characters read
     */
    static int
    vprintf(const char* format,
            va_list args)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__vprintf(&local, format, args);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>vfprintf()</code>.  On success, the number of
     * characters printed to <code>f</code> is returned.  On failure,
     * an exception is thrown.
     *
     * @param[in] f output file
     * @param[in] format format string
     * @param[in] ... variable argument list that matches <code>format</code>
     * @return number of characters read
     */
    static int
    fprintf(FILE* f,
            const char* format,
            ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
        __attribute__ ((format (printf, 2, 3)))
#endif
    {
        int rv = -1;
        cgul_exception_t local = NULL;
        va_list args;
        va_start(args, format);
        rv = cgul_libc__vfprintf(&local, f, format, args);
        va_end(args);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>vfprintf()</code>.  On success, the number of
     * characters printed to <code>f</code> is returned.  On failure,
     * an exception is thrown.
     *
     * @param[in] f output file
     * @param[in] format format string
     * @param[in] args variable argument list that matches <code>format</code>
     * @return number of characters read
     */
    static int
    vfprintf(FILE* f,
             const char* format,
             va_list args)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__vfprintf(&local, f, format, args);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>fgetc()</code>.  Return the next character from
     * <code>f</code>.  If the EOF is reached, <code>EOF</code> is
     * returned.  If an error occurs, an exception is thrown.
     *
     * @param[in] f input file
     * @return character read from <code>f</code>
     */
    static int
    fgetc(FILE* f)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__fgetc(&local, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>fputc()</code>.  Cast <code>c</code> to an
     * <code>unsigned char</code> and write it to <code>f</code>.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] c character to write
     * @param[in] f output file
     */
    static void
    fputc(int c,
          FILE* f)
    {
        cgul_exception_t local = NULL;
        cgul_libc__fputc(&local, c, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Wrapper for <code>fgets()</code>.  Read a string from
     * <code>f</code>.  The string will be returned in <code>buf</code>
     * and will be null terminated.  The string is limited to
     * <code>buf_size - 1</code> or the first '\\n' character.  On
     * success, <code>buf</code> is returned.  If EOF is reached,
     * <code>NULL</code> is returned.  If an error occurs, an exception
     * is thrown.
     *
     * @param[in] buf buffer where string is stored
     * @param[in] buf_size size of <code>buf</code>
     * @param[in] f input file
     * @return <code>buf</code> on success or <code>NULL</code> on EOF
     */
    static char*
    fgets(char* buf,
          int buf_size,
          FILE* f)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__fgets(&local, buf, buf_size, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>fputs</code>.  Write <code>s</code> to
     * <code>f</code>.  If an error occurs, an exception is thrown.
     *
     * @param[in] s string
     * @param[in] f file to which <code>s</code> is written
     */
    static void
    fputs(const char* s,
          FILE* f)
    {
        cgul_exception_t local = NULL;
        cgul_libc__fputs(&local, s, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Wrapper for <code>puts</code>.  It writes the string <code>s</code>
     * to <code>stdout</code>.  If an error occurs, an exception is thrown.
     *
     * @param[in] s string to write
     */
    static void
    puts(const char* s)
    {
        cgul_exception_t local = NULL;
        cgul_libc__puts(&local, s);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Wrapper for <code>strlen()</code> to determine the string length of
     * <code>s</code>.
     *
     * @param[in] s string
     * @return length of <code>s</code>
     */
    static size_t
    strlen(const char* s)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__strlen(&local, s);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>strcpy()</code> to copy <code>src</code> to
     * <code>dst</code>.
     *
     * @param[in] dst destination string
     * @param[in] src source string
     * @return <code>dst</code>
     */
    static char*
    strcpy(char* dst,
           const char* src)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__strcpy(&local, dst, src);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method implements <code>strncpy()</code> functionality except
     * it always <code>null</code> terminates the result.
     *
     * @param[in] dst destination string
     * @param[in] dst_size size in bytes of <code>dst</code>
     * @param[in] src source string
     * @return <code>dst</code>
     */
    static char*
    strncpy(char* dst,
            const char* src,
            size_t dst_size)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__strncpy(&local, dst, src, dst_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>strcpy()</code> to concatenate <code>src</code>
     * with <code>dst</code>.
     *
     * @param[in] dst destination string
     * @param[in] src source string
     * @return <code>dst</code>
     */
    static char*
    strcat(char* dst,
           const char *src)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__strcat(&local, dst, src);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Wrapper for <code>strcmp()</code>.  It compares <code>s1</code> with
     * <code>s2</code> and returns less than zero if <code>s1</code> is
     * less than <code>s2</code>.  It returns greater than zero if
     * <code>s1</code> is greater than <code>s2</code>, and it returns
     * zero if <code>s1</code> is equal to <code>s2</code>.
     *
     * @param s1 first string to compare
     * @param s2 second string to compare
     * @return result of comparison
     */
    static int
    strcmp(const char* s1,
           const char* s2)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__strcmp(&local, s1, s2);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return a duplicate of <code>s</code>.  The client is
     * responsible for calling <code>free()</code> on the pointer
     * returned.  If an error occurs, an exception is thrown.  This is
     * not a wrapper around <code>strdup()</code> because
     * <code>strdup()</code> is not portable.
     *
     * @param[in] s string to duplicate
     * @return duplicate string
     */
    static char*
    strdup(const char* s)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__strdup(&local, s);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function is a thin wrapper around
     * <code>cgul_libc__vsnprintf()</code>.
     *
     * @param[in] dst destination string
     * @param[in] dst_size size of <code>dst</code>
     * @param[in] format format string
     * @param[in] ... variable length arguments that match <code>format</code>
     * @return length of the string copied to <code>dst</code>
     * @see cgul_libc__vsnprintf
     */
    static int
    snprintf(char* dst,
             long dst_size,
             const char* format,
             ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
        __attribute__ ((format (printf, 3, 4)))
#endif
    {
        int rv = -1;
        cgul_exception_t local = NULL;
        va_list args;
        va_start(args, format);
        rv = cgul_libc__vsnprintf(&local, dst, dst_size, format, args);
        va_end(args);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function is not a wrapper around <code>snprintf()</code>
     * because <code>snprintf()</code> is not generally portable.
     * Instead, this function uses the sprintf()-like implementation
     * provided by <code>cgul_string__append_vsprintf()</code>. <p>
     *
     * This function expands <code>format</code> and writes up to a
     * maximum of <code>dst_size - 1</code> bytes to <code>dst</code>
     * truncating the expanded string if necessary.  It always
     * null-terminates <code>dst</code>.  On success, the entire length of
     * the expanded string is returned even if <code>dst</code> only holds
     * a truncated version of that string.  On failure, an exception is
     * thrown. <p>
     *
     * NOTE: If you pass in <code>0</code> for <code>dst_size</code> and
     * (optionally) <code>NULL</code> for <code>dst</code>, this function
     * will not attempt to alter <code>dst</code>, but it will return the
     * full length of the expanded format string.  You can use this return
     * value to allocate the right amount of space so that you can turn
     * around and call <code>cgul_libc__vsprintf()</code>.  However, you
     * are probably better off calling <code>cgul_libc__asprintf()</code>
     * which does the allocation and format string expansion in a single
     * call.
     *
     * @param[in] dst destination string
     * @param[in] dst_size size of <code>dst</code>
     * @param[in] format format string
     * @param[in] args variable length arguments that match <code>format</code>
     * @return length of the string copied to <code>dst</code>
     * @see cgul_libc__asprintf
     */
    static int
    vsnprintf(char* dst,
              long dst_size,
              const char* format,
              va_list args)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__vsnprintf(&local, dst, dst_size, format, args);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function is a thin wrapper around
     * <code>cgul_libc__vasprintf()</code>. <p>
     *
     * @param[in] format format string
     * @param[in] ... variable arguments that match <code>format</code>
     * @return newly allocated and formatted string
     */
    static char*
    asprintf(const char* format,
             ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
        __attribute__ ((format (printf, 1, 2)))
#endif
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        va_list args;
        va_start(args, format);
        rv = cgul_libc__vasprintf(&local, format, args);
        va_end(args);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function mimics the BSD vasprintf() function by allocating
     * enough memory before calling sprintf() to guarantee that no buffer
     * overflows occur.  The caller is responsible for calling free() on
     * the pointer returned.  If this method fails, an exception is thrown.
     *
     * @param[in] format format string
     * @param[in] args variable arguments that match <code>format</code>
     * @return newly allocated and formatted string
     */
    static char*
    vasprintf(const char* format,
              va_list args)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__vasprintf(&local, format, args);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the base name of <code>name</code>.  It returns everything
     * after the last slash or the last backslash.  If no slash or
     * backslash can be found, it simply returns <code>name</code>.  The
     * client is responsible for calling <code>free()</code> on the
     * pointer returned.  If memory cannot be allocated, an exception is
     * thrown.
     *
     * @param[in] name name of file
     * @return base name of <code>name</code>
     */
    static char*
    basename(const char* name)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__basename(&local, name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the directory name of <code>name</code>.  It returns everything
     * before the last slash or the last backslash.  If no slash or
     * backslash can be found, it simply returns <code>"."</code>.  The
     * client is responsible for calling <code>free()</code> on the
     * pointer returned.  If memory cannot be allocated, an exception is
     * thrown.
     *
     * @param[in] name name of file
     * @return directory name of <code>name</code>
     */
    static char*
    dirname(const char* name)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__dirname(&local, name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the file extension of the base name of <code>name</code>.  It
     * returns everything after and including the last dot in the base
     * name of <code>name</code>.  If no dot can be found, it simply
     * returns an empty string.  The client is responsible for calling
     * <code>free()</code> on the pointer returned.  If memory cannot be
     * allocated, an exception is thrown. <p>
     *
     * For example, the base name of "/foo/bar.d/" is "bar.d" (despite the
     * trailing path separator).  Thus, the extension is ".d".  This
     * differs from how the Tcl "file extension" command works which
     * returns an empty string in this case (because of the path
     * separator). <p>
     *
     * @param[in] name name of file
     * @return file extension of <code>name</code>
     * @see rootname()
     */
    static char*
    extension(const char* name)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__extension(&local, name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the root name of <code>name</code>.  This function joins the
     * directory name and base name of <code>name</code> after stripping
     * off the extension.  The client is responsible for calling
     * <code>free()</code> on the pointer returned.  If memory cannot be
     * allocated, an exception is thrown.
     *
     * For example, the base name of "/foo/bar.d/" is "bar.d" (despite the
     * trailing path separator).  Thus, the extension is ".d" and the root
     * name is "/foo/bar".  This differs from how the Tcl "file extension"
     * command works which returns "/foo/bar.d/" n this case (because of
     * the path separator). <p>
     *
     * @param[in] name name of file
     * @return root name of <code>name</code>
     */
    static char*
    rootname(const char* name)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_libc__rootname(&local, name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method strips the UTF-8 Byte-Order Mark (BOM) from
     * <code>f</code>.  This method returns <code>1</code> if the BOM is
     * stripped; otherwise, it returns <code>0</code>.  If an error
     * occurs, an exception is thrown. <p>
     *
     * IMPORTANT: Because of the way this method is written, it can only
     * be called on files that allow random access.  Do not try to use it
     * on <code>stdin</code> for example.
     *
     * @param[in] f file
     * @return whether the utf8 byte order mark was stripped
     */
    static int
    strip_utf8_byte_order_mark(FILE* f)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_libc__strip_utf8_byte_order_mark(&local, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

};

#endif  /*  CGUL_LIBC_CXX_H  */
