#ifndef CGUL_MERGE_SORT_H
#define CGUL_MERGE_SORT_H

/**
 * Merge sort an array.
 *
 * @file
 * @author Paul Serice
 * @brief merge sort
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/**
 * This typedef is the interface the client must define in order for
 * <code>cgul_merge_sort()</code> or
 * <code>cgul_merge_sort_pointers()</code> to sort an array.  When
 * defining the comparison function, it should compare <code>e1</code>
 * to <code>e2</code> and return less than zero, zero, or greater than
 * zero if <code>e1</code> is less than, equal to, or greater than
 * <code>e2</code>.
 *
 * @param[in] e1 first element
 * @param[in] e2 second element
 * @return result of comparison
 */
typedef int (*cgul_merge_sort__compare_t)(const void* e1,
                                          const void* e2);

/**
 * Generic in-place sort with the same parameter signature as stdlib's
 * qsort().  <code>v</code> is a pointer to the top of your array.
 * <code>v_count</code> is the number of elements in <code>v</code>.
 * <code>element_size</code> is the number of contiguous bytes that
 * comprise each element.  <code>cf</code> is the comparison function.
 * It will be passed pointers to two elements.  It must return less
 * than zero, zero, or greater than zero if the first element is less
 * than, equal to, or greater than the second element.  This function
 * has to internally allocate memory to perform its task.  If
 * out-of-memory occurs, an exception is thrown, and the array is
 * returned partially sorted. <p>
 *
 * The main draw back to using merge sort when sorting arrays is that
 * it needs to create a copy of the original array in order to work
 * efficiently.  This particular implementation actually requires 1.5
 * copies of the original array at peak memory usage.  If you are
 * short on memory, you can either use cgul_merge_sort_iteratively()
 * which only requires 1 copy or <code>cgul_list__sort()</code> which
 * requires no extra copies because it operates on linked lists which
 * have efficient removal and insertion operations. <p>
 *
 * If you need to constantly maintain a sorted list,
 * <code>cgul_rbtree</code> or <code>cgul_multimap</code> are usually
 * better solutions. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] element_size size of each element in <code>v</code> (in bytes)
 * @param[in] cf comparison function
 */
extern CGUL_EXPORT void
cgul_merge_sort(cgul_exception_t* cex,
                void* v,
                size_t v_count,
                size_t element_size,
                cgul_merge_sort__compare_t cf);

/**
 * Generic in-place sort with the same parameter signature as stdlib's
 * qsort().  <code>v</code> is a pointer to the top of your array.
 * <code>v_count</code> is the number of elements in <code>v</code>.
 * <code>element_size</code> is the number of contiguous bytes that
 * comprise each element.  <code>cf</code> is the comparison function.
 * It will be passed pointers to two elements.  It must return less
 * than zero, zero, or greater than zero if the first element is less
 * than, equal to, or greater than the second element.  This function
 * has to internally allocate memory to perform its task.  If
 * out-of-memory occurs, an exception is thrown, and the array is
 * returned partially sorted. <p>
 *
 * There are three main differences between this function and
 * <code>cgul_merge_sort()</code>.  First,
 * <code>cgul_merge_sort()</code> is about 30% faster than this
 * function.  Second, <code>cgul_merge_sort()</code> is a recursive
 * implementation that consumes more stack space than this function
 * which is an iterative implementation.  Third,
 * <code>cgul_merge_sort()</code> uses 50% more memory than this
 * implementation because it requires 1.5 extra copies of
 * <code>v</code>; whereas, this implementation only requires 1 extra
 * copy.  If in doubt, use <code>cgul_merge_sort()</code> instead of
 * this function. <p>
 *
 * If you need to constantly maintain a sorted list,
 * <code>cgul_rbtree</code> or <code>cgul_multimap</code> are usually
 * better solutions. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] element_size size of each element in <code>v</code> (in bytes)
 * @param[in] cf comparison function
 */
extern CGUL_EXPORT void
cgul_merge_sort_iteratively(cgul_exception_t* cex,
                            void* v,
                            size_t v_count,
                            size_t element_size,
                            cgul_merge_sort__compare_t cf);

/**
 * If your array is already composed of pointers, you can directly
 * call this function.  <code>v</code> will be sorted using
 * <code>cf</code> the same as described for
 * <code>cgul_merge_sort()</code>.  If out-of-memory occurs, an
 * exception is thrown, and the array is returned partially
 * sorted. <p>
 *
 * The main draw back to using merge sort when sorting arrays is that
 * it needs to create a copy of the original array in order to work
 * efficiently.  This particular implementation actually requires 1.5
 * copies of the original array at peak memory usage.  If you are
 * short on memory, you can either use
 * <code>cgul_merge_sort_pointers_iteratively()</code> which only
 * requires 1 copy or <code>cgul_list__sort()</code> which requires no
 * extra copies because it operates on linked list which have
 * efficient removal and insertion operations. <p>
 *
 * If you need to constantly maintain a sorted list,
 * <code>cgul_rbtree</code> or <code>cgul_multimap</code> are usually
 * better solutions. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of pointers to elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] cf comparison function
 */
extern CGUL_EXPORT void
cgul_merge_sort_pointers(cgul_exception_t* cex,
                         void** v,
                         size_t v_count,
                         cgul_merge_sort__compare_t cf);

/**
 * If your array is already composed of pointers, you can directly
 * call this function.  <code>v</code> will be sorted using
 * <code>cf</code> the same as described for
 * <code>cgul_merge_sort_iteratively()</code>.  If out-of-memory
 * occurs, an exception is thrown, and the array is returned partially
 * sorted. <p>
 *
 * There are three main differences between this function and
 * <code>cgul_merge_sort_pointers()</code>.  First,
 * <code>cgul_merge_sort_pointers()</code> is about 30% faster than
 * this function.  Second, <code>cgul_merge_sort_pointers()</code> is
 * a recursive implementation that consumes more stack space than this
 * function which is an iterative implementation.  Third,
 * <code>cgul_merge_sort_pointers()</code> uses 50% more memory than
 * this implementation because it requires 1.5 extra copies of
 * <code>v</code>; whereas, this implementation only requires 1 extra
 * copy.  If in doubt, use <code>cgul_merge_sort_pointers()</code>
 * instead of this function. <p>
 *
 * If you need to constantly maintain a sorted list,
 * <code>cgul_rbtree</code> or <code>cgul_multimap</code> are usually
 * better solutions. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of pointers to elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] cf comparison function
 */
extern CGUL_EXPORT void
cgul_merge_sort_pointers_iteratively(cgul_exception_t* cex,
                                     void** v,
                                     size_t v_count,
                                     cgul_merge_sort__compare_t cf);

CGUL_END_C

#endif  /*  CGUL_MERGE_SORT_H  */
