#ifndef CGUL_STREAM_READER_CXX_H
#define CGUL_STREAM_READER_CXX_H

/**
 * C++ wrapper for <code>cgul_stream_reader</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_stream_reader</code>
 */

#include "cgul_stream_reader.h"
#include "cgul_block_reader_cxx.h"
#include "cgul_wstring_cxx.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_stream_reader</code> objects.  The main purpose of this
 * class is to convert the C-style function calls and exception
 * handling in <code>cgul_stream_reader</code> into C++-style function
 * calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_stream_reader</code>
 */
class cgul_stream_reader_cxx {

public:

    /**
     * Type definition for decoding characters from blocks of bytes.
     *
     * @see cgul_stream_reader__decode_t
     */
    typedef cgul_stream_reader__decode_t decode_t;

    /**
     * Simple wrapper for <code>cgul_stream_reader__decode_utf8()</code>.
     *
     * @see cgul_stream_reader__decode_utf8()
     */
    static void
    decode_utf8(cgul_exception_t* cex,
                unsigned char* block_encoded,
                size_t size_encoded,
                size_t* index_encoded,
                cgul_wchar_t* block_decoded,
                size_t size_decoded,
                size_t* count_decoded)
    {
        /* Because this static C++ function is being used as a C
         * callback, we must let the C-style exception pass through
         * unmodified back to the underly C function that made the
         * callback. */
        cgul_stream_reader__decode_utf8(
            cex, block_encoded, size_encoded, index_encoded,
            block_decoded, size_decoded, count_decoded);
    }

    /**
     * Simple wrapper for <code>cgul_stream_reader__decode_ascii()</code>.
     *
     * @see cgul_stream_reader__decode_ascii()
     */
    static void
    decode_ascii(cgul_exception_t* cex,
                 unsigned char* block_encoded,
                 size_t size_encoded,
                 size_t* index_encoded,
                 cgul_wchar_t* block_decoded,
                 size_t size_decoded,
                 size_t* count_decoded)
    {
        /* Because this static C++ function is being used as a C
         * callback, we must let the C-style exception pass through
         * unmodified back to the underly C function that made the
         * callback. */
        cgul_stream_reader__decode_ascii(
            cex, block_encoded, size_encoded, index_encoded,
            block_decoded, size_decoded, count_decoded);
    }

    /**
     * Create a new <code>cgul_stream_reader_cxx</code> instance that
     * by default uses <code>cgul_stream_reader__decode_utf8</code> to
     * decode blocks of UTF-8 encoded bytes into a stream of UTF-32
     * characters.  After this method returns,
     * <code>add_block_reader()</code> should be called and
     * <code>set_decoder()</code> may be called before calling
     * <code>read()</code>.  The client is responsible for deleting
     * the pointer returned.  If an error occurs, an exception is
     * thrown.
     */
    cgul_stream_reader_cxx()
    {
        cgul_exception_t local = NULL;
        /* Create the object. */
        obj = cgul_stream_reader__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_stream_reader_cxx</code> instance that
     * by default uses <code>cgul_stream_reader__decode_utf8</code> to
     * decode blocks of UTF-8 encoded bytes from the block reader
     * <code>br</code> into a stream of UTF-32 characters.  After this
     * method returns, <code>set_decoder()</code> may be called before
     * calling <code>read()</code>.  This class does not take
     * ownership of the block reader <code>br</code> so the client is
     * responsible for deleting it when this class is no longer using
     * it.  The client is responsible for deleting the pointer
     * returned.  If an error occurs, an exception is thrown.
     *
     * @param[in] br block reader to add
     */
    cgul_stream_reader_cxx(cgul_block_reader_cxx& br)
    {
        cgul_exception_t local = NULL;
        /* Create the object. */
        obj = cgul_stream_reader__new_from_block_reader(
                  &local, br.as_block_reader());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_stream_reader_cxx</code> instance that
     * by default uses <code>cgul_stream_reader__decode_utf8</code> to
     * decode blocks of UTF-8 encoded bytes from the block reader
     * <code>br</code> into a stream of UTF-32 characters.  After this
     * method returns, <code>set_decoder()</code> may be called before
     * calling <code>read()</code>.  This class does not take
     * ownership of the block reader <code>br</code> so the client is
     * responsible for deleting it when this class is no longer using
     * it.  The client is responsible for deleting the pointer
     * returned.  If an error occurs, an exception is thrown.
     *
     * @param[in] br block reader to add
     */
    cgul_stream_reader_cxx(cgul_block_reader_t br)
    {
        cgul_exception_t local = NULL;
        /* Create the object. */
        obj = cgul_stream_reader__new_from_block_reader(&local, br);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method deletes the instance freeing all internally
     * allocated resources.  Clients must not attempt to use the
     * instance after this method returns.
     */
    virtual
    ~cgul_stream_reader_cxx()
    {
        if (obj) {
            cgul_stream_reader__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Reset the instance so it can be used with a new set of block
     * readers.  This clears the end of file (EOF) file, resets the
     * line and column numbers to <code>1</code>, and clears the queue
     * of block readers.
     */
    virtual void
    reset()
    {
        cgul_exception_t local = NULL;
        cgul_stream_reader__reset(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Read the next encoded character from the stream and return the
     * result as <code>*next_char</code> provided <code>next_char</code>
     * is not <code>NULL</code>.  This methods advances the character
     * stream.  If successful, <code>1</code> is returned.  If the end of
     * file is reached, <code>0</code> is returned and
     * <code>*next_char</code> is set to <code>CGUL_WCHAR__NUL</code>.  If
     * an error occurs, <code>0</code> is returned,
     * <code>*next_char</code> is set to <code>CGUL_WCHAR__NUL</code>, and
     * an exception is thrown.
     *
     * @param[out] next_char next character
     * @return whether the read completed successfully
     */
    virtual int
    read(cgul_wchar_t* next_char)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__read(&local, obj, next_char);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return whether the end of file (EOF) has been reached.  Calling
     * <code>reset()</code> with clear the EOF flag.  Putting a
     * character back will also clear the EOF flag.
     *
     * @return whether the end of file has been reached
     */
    virtual int
    eof() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__eof(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return whether line and column number tracking is enabled.  By
     * default, line and column number tracking is enabled.
     *
     * @return whether line and column number tracking is enabled
     */
    virtual int
    get_line_and_column_tracking() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__get_line_and_column_tracking(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether line and column number tracking is enabled.  By
     * default, line and column number tracking is enabled.  Disabling
     * line and column number tracking results in about a 10% performance
     * increase.
     *
     * @param[in] is_tracking_enabled whether line and column number
     * tracking is enabled
     */
    virtual void
    set_line_and_column_tracking(int is_tracking_enabled)
    {
        cgul_exception_t local = NULL;
        cgul_stream_reader__set_line_and_column_tracking(&local,
                                                         obj,
                                                         is_tracking_enabled);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the line number for the current character (i.e., the
     * character that will be returned by the next
     * <code>read()</code>).  The line number is 1-based (i.e., the
     * first line is assigned the number 1). <p>
     *
     * @note If line and column number tracking has been disabled,
     * <code>0</code> is returned.  (See see
     * cgul_stream_reader__set_line_and_column_tracking().)
     *
     * @return current line number
     * @see get_line_number_previous()
     * @see set_line_and_column_tracking()
     */
    virtual size_t
    get_line_number_current() const
    {
        size_t rv;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__get_line_number_current(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the column number for the current character (i.e., the
     * character that will be returned by the next
     * <code>read()</code>).  The column number is 1-based (i.e., the
     * first column is assigned the number 1).
     *
     * Even though the column number is 1-based, it is possible for
     * this method to return <code>0</code>.  This indicates that
     * <code>putback()</code> was called with an EOL character which
     * moved the cursor backward to the end of the previous line.
     * Unfortunately, there is no good way to know the number of
     * columns in the previous line, and because the return value is
     * an unsigned <code>size_t</code>, the column numbers cannot be
     * made negative without causing other problems.  To this end,
     * this class latches the column number to <code>0</code> when
     * retreating past an EOL character.  It stays latched until the
     * stream advances past an EOL character. <p>
     *
     * @note If line and column number tracking has been disabled,
     * <code>0</code> is returned.  (See see
     * cgul_stream_reader__set_line_and_column_tracking().)
     *
     * @return current column number or <code>0</code>
     * @see get_column_number_previous()
     * @see set_line_and_column_tracking()
     */
    virtual size_t
    get_column_number_current() const
    {
        size_t rv;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__get_column_number_current(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the line number for the previous character (i.e., the
     * character that was returned by the previous call to
     * <code>read()</code>).  The line number is 1-based (i.e., the
     * first line is assigned the number 1).  If <code>0</code> is
     * returned, it means the first character has not yet been
     * read. <p>
     *
     * In the unlike case that the column number is latched (see
     * comments for <code>get_column_number_previous()</code>), it is
     * more likely that the previous character is on the same line as
     * the current character so the current line is returned. <p>
     *
     * @note If line and column number tracking has been disabled,
     * <code>0</code> is returned.  (See see
     * cgul_stream_reader__set_line_and_column_tracking().)
     *
     * @return current line number
     * @see get_line_number_current()
     * @see set_line_and_column_tracking()
     */
    virtual size_t
    get_line_number_previous() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__get_line_number_previous(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the column number for the previous character (i.e., the
     * character that was returned by the previous call to
     * <code>cgul_stream_reader__read()</code>) or <code>0</code> if no
     * character has been read yet.  The column number is 1-based (i.e.,
     * the first column is assigned the number 1).
     *
     * Even though the column number is 1-based, it is possible for this
     * method to return <code>0</code>.  This indicates that the first
     * character has not yet been read or that
     * <code>cgul_stream_read__putback()</code> was called with an EOL
     * character which moved the cursor backward to the end of the
     * previous line.  In the latter case, unfortunately there is no good
     * way to know the number of columns in the previous line, and because
     * the return value is an unsigned <code>size_t</code>, the column
     * numbers cannot be made negative without causing other problems.  To
     * this end, there is an internal stack of line widths to fill in this
     * missing information, but the stack is limited in size.  When the
     * stack is empty, this class latches the column number to
     * <code>0</code> when retreating past an EOL character.  It stays
     * latched until the stream advances past an EOL character.  As a
     * practical matter, this should rarely if ever happen. <p>
     *
     * @note If line and column number tracking has been disabled,
     * <code>0</code> is returned.  (See see
     * cgul_stream_reader__set_line_and_column_tracking().)
     *
     * @return current column number or <code>0</code>
     * @see get_column_number_current()
     * @see set_line_and_column_tracking()
     */
    virtual size_t
    get_column_number_previous() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__get_column_number_previous(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return whether end-of-line (EOL) character sequences are
     * normalized.  If normalization is enabled, "\r" and "\r\n" sequences
     * are normalized to "\n".  All other characters pass though without
     * replacement.
     *
     * @return whether EOLs are normalized
     */
    virtual int
    get_normalize_eol() const
    {
        int rv;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__get_normalize_eol(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set whether end-of-line character sequences are normalized.  If
     * normalization is enabled, "\r" and "\r\n" sequences are normalized
     * to "\n".  All other characters passed though without replacement.
     *
     * @param[in] normalize whether to normalize EOLs
     */
    virtual void
    set_normalize_eol(int normalize)
    {
        cgul_exception_t local = NULL;
        cgul_stream_reader__set_normalize_eol(&local, obj, normalize);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Add the block reader <code>br</code> to the queue of block readers.
     * The block reader must end on an encoded character boundary (i.e.,
     * encoded characters must not span across block readers).  Characters
     * are streamed from the front block reader on the queue.  When end of
     * file is reach for the front block reader, it is removed from the
     * queue, and characters are then streamed from the next block reader
     * in the queue.  Once a block reader is added to the queue, it should
     * not be modified externally.  If an error occurs, an exception is
     * thrown.
     *
     * @param[in] br block reader to add
     * @see cgul_block_reader
     * @see cgul_block_reader_file
     * @see cgul_block_reader_memory
     */
    virtual void
    add_block_reader(cgul_block_reader_cxx& br)
    {
        cgul_exception_t local = NULL;
        cgul_stream_reader__add_block_reader(&local,
                                             obj,
                                             br.as_block_reader());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Set the block reader to <code>br</code>.  If <code>br</code> is
     * <code>NULL</code>, the current block reader will no longer be used
     * allowing the client to delete the block reader immediately.  Also
     * if <code>br</code> is <code>NULL</code>, the line and column
     * numbers will be reset to indicate the start of a new input stream.
     * This method should be called whenever a new file has been loaded
     * into the *block* reader so this class will know to discard the
     * unused portion of the final block that it read from the old file.
     *
     * @param[in] br block reader to add
     * @see cgul_block_reader
     * @see cgul_block_reader_file
     * @see cgul_block_reader_memory
     */
    virtual void
    add_block_reader(cgul_block_reader_t br)
    {
        cgul_exception_t local = NULL;
        cgul_stream_reader__add_block_reader(&local, obj, br);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the size of the block used to read from the block reader.
     *
     * @return size of block used to read from the block reader
     */
    virtual size_t
    get_block_size() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__get_block_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the block size <code>block_size</code> used to read from
     * the block reader.  If a block reader has been added, the
     * block size can only be increased.  If the block size is not
     * large enough to decode any character in the input stream, the
     * decoder will throw an exception when <code>read()</code>,
     * <code>peek()</code>, or <code>accept()</code> (or related
     * methods) is called.  If successful, <code>1</code> is returned;
     * otherwise, <code>0</code> is returned.  If an error occurs,
     * <code>0</code> is returned, and an exception is thrown.
     *
     * @param[in] block_size block size
     * @return whether the block size was set
     */
    virtual int
    set_block_size(size_t block_size)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__set_block_size(&local, obj, block_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the decoder used to decode characters from the block reader.
     *
     * @return decoder
     * @see cgul_stream_reader__decode_t
     */
    virtual cgul_stream_reader_cxx::decode_t
    get_decoder() const
    {
        decode_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__get_decoder(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the decoder <code>decoder</code> used to decode characters from
     * the block reader.
     *
     * @param[in] decoder decoder
     * @see cgul_stream_reader__decode_t
     */
    virtual void
    set_decoder(cgul_stream_reader_cxx::decode_t decoder)
    {
        cgul_exception_t local = NULL;
        cgul_stream_reader__set_decoder(&local, obj, decoder);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the presentation name for the stream reader.  If a
     * presentation name has been explicitly set using
     * <code>set_presentation_name()</code>, it is returned.  If no
     * presentation name has been explicitly set and only one block
     * reader has been added, the presentation name for the block
     * reader is returned (and should not be cached by the client).
     * If no presentation name has been explicitly set and more than
     * one block reader has been added, <code>"MULTIPART"</code> is
     * returned.  If no block reader has been set, an exception is
     * thrown.
     *
     * @return presentation name
     */
    virtual const char*
    get_presentation_name() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_stream_reader__get_presentation_name(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the presentation name for the stream reader. If an error
     * occurs, an exception is thrown.
     *
     * @param[in] presentation_name presentation name
     */
    virtual void
    set_presentation_name(const char* presentation_name)
    {
        cgul_exception_t local = NULL;
        cgul_stream_reader__set_presentation_name(
            &local, obj, presentation_name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the underlying <code>cgul_stream_reader</code> object.
     *
     * @return underlying object
     */
    virtual cgul_stream_reader_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_stream_reader</code> object.
     * This means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_stream_reader_t
    take_obj()
    {
        cgul_stream_reader_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_stream_reader_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_stream_reader__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_stream_reader_cxx(const cgul_stream_reader_cxx&);

    cgul_stream_reader_cxx&
    operator=(const cgul_stream_reader_cxx&);

    /* Underlying <code>cgul_stream_reader</code> object. */
    cgul_stream_reader_t obj;

};

#endif  /*  CGUL_STREAM_READER_CXX_H  */
