#ifndef CGUL_UNICODE_CXX_H
#define CGUL_UNICODE_CXX_H

/**
 * C++ wrapper for <code>cgul_unicode</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_unicode</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_unicode.h"
#include "cgul_string_cxx.h"
#include "cgul_hstring_cxx.h"
#include "cgul_wstring_cxx.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_unicode</code> external functions.  The main purpose of
 * this class is to convert the C-style function calls and exception
 * handling in <code>cgul_unicode</code> into C++-style function
 * calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_unicode</code>
 */
class cgul_unicode_cxx {

public:

    /**
     * UTF-8 to UTF-32 string conversion. <p>
     *
     * This functions converts the UTF-8 multi-byte string
     * <code>utf8</code> to a UTF-32 wide character string and appends the
     * result to <code>utf32</code>.  If a leading byte-order mark (BOM)
     * is present and <code>skip_bom</code> is set, the BOM is
     * automatically removed. <p>
     *
     * Note that BOM characters can be embedded inside of UTF-8 strings.
     * It is only the leading BOM that gets removed if
     * <code>skip_bom</code> is set.  This is necessary because any other
     * BOM is supposed to be interpreted as a "zero-width no-break space"
     * that's only purpose is to prevent the formation of ligatures.  So,
     * if you are converting an entire file by calling this function more
     * than once, you should set <code>skip_bom</code> the first time this
     * function is called, but clear <code>skip_bom</code> all the other
     * times this function is called. <p>
     *
     * To guard against reading off the end of <code>utf8</code>, its
     * length must be given in <code>utf8_length</code>.  The value of
     * <code>utf8_length</code> is the total number of bytes in the
     * <code>utf8</code> string. <p>
     *
     * If an error occurs, an exception is thrown.  Errors can occur
     * because <code>utf8</code> is not a valid UTF-8 encoding or because
     * memory could not be allocated for the string that holds the UTF-32
     * conversion.  If an error occurs, the value of <code>utf32</code> is
     * undefined. <p>
     *
     * @param[in] utf8 UTF-8 string
     * @param[in] utf8_length length of <code>utf8</code> in bytes
     * @param[in] skip_bom whether the leading byte-order mark should be skipped
     * @param[out] utf32 conversion of <code>utf8</code> to UTF-32
     */
    static void
    mbstowcs(const char* utf8,
             size_t utf8_length,
             int skip_bom,
             cgul_wstring_cxx& utf32)
    {
        cgul_exception_t local = NULL;
        cgul_unicode__mbstowcs(&local,
                               utf8,
                               utf8_length,
                               skip_bom,
                               utf32.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * UTF-32 to UTF-8 string conversion. <p>
     *
     * This functions converts the wide-character string
     * <code>utf32</code> from UTF-32 to UTF-8 and appends the result to
     * the multi-byte string <code>utf8</code>.  <p>
     *
     * If an error occurs, an exception is thrown.  Errors can occur
     * because <code>utf32</code> is not a valid UTF-32 encoding or
     * because memory could not be allocated for the string that holds the
     * UTF-8 conversion.  If an error occurs, the value of
     * <code>utf8</code> is undefined. <p>
     *
     * @param[in] utf32 UTF-32 string
     * @param[in] utf32_length length of <code>utf32</code> in wide characters
     * @param[in] skip_bom whether the leading byte-order mark should be skipped
     * @param[out] utf8 conversion of <code>utf32</code> to UTF-8
     */
    static void
    wcstombs(const cgul_wchar_t* utf32,
             size_t utf32_length,
             int skip_bom,
             cgul_string_cxx& utf8)
    {
        cgul_exception_t local = NULL;
        cgul_unicode__wcstombs(&local,
                               utf32,
                               utf32_length,
                               skip_bom,
                               utf8.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * UTF-8 to UTF-16 string conversion. <p>
     *
     * This functions converts the UTF-8 multi-byte string
     * <code>utf8</code> to a UTF-16 wide character string and appends the
     * result to <code>utf16</code>.  If a leading byte-order mark (BOM)
     * is present and <code>skip_bom</code> is set, the BOM is
     * automatically removed. <p>
     *
     * Note that BOM characters can be embedded inside of UTF-8 strings.
     * It is only the leading BOM that gets removed if
     * <code>skip_bom</code> is set.  This is necessary because any other
     * BOM is supposed to be interpreted as a "zero-width no-break space"
     * that's only purpose is to prevent the formation of ligatures.  So,
     * if you are converting an entire file by calling this function more
     * than once, you should set <code>skip_bom</code> the first time this
     * function is called, but clear <code>skip_bom</code> all the other
     * times this function is called. <p>
     *
     * To guard against reading off the end of <code>utf8</code>, its
     * length must be given in <code>utf8_length</code>.  The value of
     * <code>utf8_length</code> is the total number of bytes in the
     * <code>utf8</code> string. <p>
     *
     * If an error occurs, an exception is thrown.  Errors can occur
     * because <code>utf8</code> is not a valid UTF-8 encoding or because
     * memory could not be allocated for the string that holds the UTF-16
     * conversion.  If an error occurs, the value of <code>utf16</code> is
     * undefined. <p>
     *
     * @param[in] utf8 UTF-8 string
     * @param[in] utf8_length length of <code>utf8</code> in bytes
     * @param[in] skip_bom whether the leading byte-order mark should be skipped
     * @param[out] utf16 conversion of <code>utf8</code> to UTF-16
     */
    static void
    mbstohcs(const char* utf8,
             size_t utf8_length,
             int skip_bom,
             cgul_hstring_cxx& utf16)
    {
        cgul_exception_t local = NULL;
        cgul_unicode__mbstohcs(&local,
                               utf8,
                               utf8_length,
                               skip_bom,
                               utf16.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * UTF-16 to UTF-8 string conversion. <p>
     *
     * This functions converts the UTF-16 wide-character string
     * <code>utf16</code> to a UTF-8 multi-byte string and appends the
     * result to <code>utf8</code>.  If a leading byte-order mark (BOM) is
     * present and <code>skip_bom</code> is set, the BOM is automatically
     * removed. <p>
     *
     * Note that BOM characters can be embedded inside of UTF-16 strings.
     * It is only the leading BOM that gets removed if
     * <code>skip_bom</code> is set.  This is necessary because any other
     * BOM is supposed to be interpreted as a "zero-width no-break space"
     * that's only purpose is to prevent the formation of ligatures.  So,
     * if you are converting an entire file by calling this function more
     * than once, you should set <code>skip_bom</code> the first time this
     * function is called, but clear <code>skip_bom</code> all the other
     * times this function is called. <p>
     *
     * To guard against reading off the end of <code>utf16</code>, its
     * length must be given in <code>utf16_length</code>.  The value of
     * <code>utf16_length</code> is the total number of
     * <code>cgul_hchar_t</code> elements in the <code>utf16</code>
     * string; this value can be obtained by calling
     * <code>cgul_hchar_cxx::hcslen()</code> or
     * <code>cgul_hstring_cxx::get_length()</code>.  Basically, surrogate
     * pairs count as two; everything else counts as one. <p>
     *
     * If an error occurs, an exception is thrown.  Errors can occur
     * because <code>utf16</code> is not a valid UTF-16 encoding or because
     * memory could not be allocated for the string that holds the UTF-8
     * conversion.  If an error occurs, the value of <code>utf8</code> is
     * undefined. <p>
     *
     * @param[in] utf16 UTF-16 string
     * @param[in] utf16_length length of <code>utf16</code> in
     * <code>cgul_hchar_t</code> elements
     * @param[in] skip_bom whether the leading byte-order mark should be skipped
     * @param[out] utf8 conversion of <code>utf16</code> to UTF-8
     */
    static void
    hcstombs(const cgul_hchar_t* utf16,
             size_t utf16_length,
             int skip_bom,
             cgul_string_cxx& utf8)
    {
        cgul_exception_t local = NULL;
        cgul_unicode__hcstombs(&local,
                               utf16,
                               utf16_length,
                               skip_bom,
                               utf8.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * UTF-16 to UTF-32 string conversion. <p>
     *
     * This functions converts the UTF-16 wide-character string
     * <code>utf16</code> to a UTF-32 wide-character string and appends
     * the result to <code>utf32</code>.  If a leading byte-order mark
     * (BOM) is present and <code>skip_bom</code> is set, the BOM is
     * automatically removed. <p>
     *
     * Note that BOM characters can be embedded inside of UTF-16 strings.
     * It is only the leading BOM that gets removed if
     * <code>skip_bom</code> is set.  This is necessary because any other
     * BOM is supposed to be interpreted as a "zero-width no-break space"
     * that's only purpose is to prevent the formation of ligatures.  So,
     * if you are converting an entire file by calling this function more
     * than once, you should set <code>skip_bom</code> the first time this
     * function is called, but clear <code>skip_bom</code> all the other
     * times this function is called. <p>
     *
     * To guard against reading off the end of <code>utf16</code>, its
     * length must be given in <code>utf16_length</code>.  The value of
     * <code>utf16_length</code> is the total number of
     * <code>cgul_hchar_t</code> elements in the <code>utf16</code>
     * string; this value can be obtained by calling
     * <code>cgul_hchar_cxx::hcslen()</code> or
     * <code>cgul_hstring_cxx::get_length()</code>.  Basically, surrogate
     * pairs count as two; everything else counts as one. <p>
     *
     * If an error occurs, an exception is thrown.  Errors can occur
     * because <code>utf16</code> is not a valid UTF-16 encoding or because
     * memory could not be allocated for the string that holds the UTF-32
     * conversion.  If an error occurs, the value of <code>utf32</code> is
     * undefined. <p>
     *
     * @param[in] utf16 UTF-16 string
     * @param[in] utf16_length length of <code>utf16</code> in
     * <code>cgul_hchar_t</code> elements
     * @param[in] skip_bom whether the leading byte-order mark should be skipped
     * @param[out] utf32 conversion of <code>utf16</code> to UTF-32
     */
    static void
    hcstowcs(const cgul_hchar_t* utf16,
             size_t utf16_length,
             int skip_bom,
             cgul_wstring_cxx& utf32)
    {
        cgul_exception_t local = NULL;
        cgul_unicode__hcstowcs(&local,
                               utf16,
                               utf16_length,
                               skip_bom,
                               utf32.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * UTF-32 to UTF-16 string conversion. <p>
     *
     * This functions converts the wide-character string
     * <code>utf32</code> from UTF-32 to UTF-16 and appends the result to
     * the wide-character string <code>utf16</code>.  <p>
     *
     * If an error occurs, an exception is thrown.  Errors can occur
     * because <code>utf32</code> is not a valid UTF-32 encoding or
     * because memory could not be allocated for the string that holds the
     * UTF-16 conversion.  If an error occurs, the value of
     * <code>utf16</code> is undefined. <p>
     *
     * @param[in] utf32 UTF-32 string
     * @param[in] utf32_length length of <code>utf32</code> in wide characters
     * @param[in] skip_bom whether the leading byte-order mark should be skipped
     * @param[out] utf16 conversion of <code>utf32</code> to UTF-16
     */
    static void
    wcstohcs(const cgul_wchar_t* utf32,
             size_t utf32_length,
             int skip_bom,
             cgul_hstring_cxx& utf16)
    {
        cgul_exception_t local = NULL;
        cgul_unicode__wcstohcs(&local,
                               utf32,
                               utf32_length,
                               skip_bom,
                               utf16.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * UTF-8 to UTF-32 character conversion. <p>
     *
     * This method converts exactly one UTF-8 multi-byte character to a
     * UTF-32 wide character and returns it.  The multiple bytes that
     * comprise the UTF-8 character start at <code>utf8[*index]</code>. <p>
     *
     * In order to make iterating over an entire UTF-8 string easy, when
     * this method returns, the value of <code>*index</code> is updated to
     * point to the beginning of the next character. <p>
     *
     * To guard against reading off the end of <code>utf8</code>, its
     * length must be given in <code>utf8_length</code>.  The value of
     * <code>utf8_length</code> is the total number of bytes in the
     * <code>utf8</code> string.  It is not the number of bytes remaining
     * in the string. <p>
     *
     * NOTE: This method throws an exception if <code>utf8_length</code>
     * is 0.  Thus, the caller must verify that <code>utf8</code> is not
     * an empty string before calling this method! <p>
     *
     * If an error occurs, <code>CGUL_WCHAR__NUL</code> is returned, and
     * an exception is thrown. <p>
     *
     * @param[in] utf8 UTF-8 string
     * @param[in] utf8_length length of <code>utf8</code> in bytes
     * @param[in,out] index index into <code>utf8</code>
     * @return next UTF-32 character
     */
    static cgul_wchar_t
    mbtowc(const char* utf8,
           size_t utf8_length,
           size_t* index)
    {
        cgul_wchar_t rv = CGUL_WCHAR__NUL;
        cgul_exception_t local = NULL;
        rv = cgul_unicode__mbtowc(&local, utf8, utf8_length, index);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * UTF-32 to UTF-8 character conversion. <p>
     *
     * This method converts exactly one UTF-32 wide character to a UTF-8
     * multi-byte character sequence and returns it in <code>buf</code> as
     * a C-style string that is <code>NUL</code> terminated.  The UTF-32
     * character that gets converted is passed in as <code>wc</code>.  As
     * an added convenience, this method returns the number of bytes
     * written to <code>buf</code> excluding the <code>NUL</code>
     * terminator. <p>
     *
     * The caller is responsible for making sure that <code>buf</code>
     * points to a buffer that can hold at least
     * <code>CGUL_WCHAR__MB_LEN_MAX + 1</code> bytes. <p>
     *
     * If an error occurs, zero is returned to the caller, an empty string
     * is returned in <code>buf</code>, and an exception is thrown.
     * Errors can occur, for example, if <code>wc</code> is larger than
     * 0x10ffff which is the largest valid Unicode code point. <p>
     *
     * @param[in] utf32 UTF-32 wide character
     * @param[out] buf buffer that holds the resulting UTF-8 multi-byte sequence
     * @return number of bytes written to <code>buf</code>
     */
    static size_t
    wctomb(const cgul_wchar_t utf32,
           char* buf)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_unicode__wctomb(&local, utf32, buf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * UTF-8 to UTF-16 character conversion. <p>
     *
     * This method converts exactly one UTF-8 multi-byte character to a
     * UTF-16 wide character sequence and returns it in <code>buf</code>
     * as a <code>cgul_hchar_t</code> string that is <code>NUL</code>
     * terminated.  As an added convenience, this method returns the
     * number of <code>cgul_hchar_t</code> elements written to
     * <code>buf</code> excluding the <code>NUL</code> terminator. <p>
     *
     * The multiple bytes that comprise the UTF-8 character start at
     * <code>utf8[*index]</code>.  In order to make iterating over an
     * entire UTF-8 string easy, when this method returns, the value of
     * <code>*index</code> is updated to point to the beginning of the
     * next character. <p>
     *
     * To guard against reading off the end of <code>utf8</code>, its
     * length must be given in <code>utf8_length</code>.  The value of
     * <code>utf8_length</code> is the total number of bytes in the
     * <code>utf8</code> string.  It is not the number of bytes remaining
     * in the string. <p>
     *
     * The caller is responsible for making sure that <code>buf</code>
     * points to a buffer that can hold at least three
     * <code>cgul_hchar_t</code> characters.  This accounts for the worst
     * case which requires a surrogate pair and a <code>NUL</code>
     * terminator. <p>
     *
     * If an error occurs, zero is returned to the caller, an empty string
     * is returned in <code>buf</code>, and an exception is thrown. <p>
     *
     * NOTE: This method throws an exception if <code>utf8_length</code>
     * is 0.  Thus, the caller must verify that <code>utf8</code> is not
     * an empty string before calling this method! <p>
     *
     * @param[in] utf8 UTF-8 string
     * @param[in] utf8_length length of <code>utf8</code> in bytes
     * @param[in,out] index index into <code>utf8</code>
     * @param[out] buf buffer that holds the resulting UTF-16 multi-byte sequence
     * @return number of <code>cgul_hchar_t</code> elements written to
     * <code>buf</code>
     */
    static size_t
    mbtohc(const char* utf8,
           size_t utf8_length,
           size_t* index,
           cgul_hchar_t* buf)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_unicode__mbtohc(&local, utf8, utf8_length, index, buf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * UTF-16 to UTF-8 character conversion. <p>
     *
     * This method converts exactly one UTF-16 wide character sequence to
     * a UTF-8 multi-byte character sequence and returns it in
     * <code>buf</code> as a C-style string that is <code>NUL</code>
     * terminated.  As an added convenience, this method returns the
     * number of bytes written to <code>buf</code> excluding the
     * <code>NUL</code> terminator. <p>
     *
     * The multiple <code>cgul_hcahr_t</code> elements that comprise the
     * UTF-16 wide character sequence start at <code>utf16[*index]</code>.
     * In order to make iterating over an entire UTF-16 string easy, when
     * this method returns, the value of <code>*index</code> is updated to
     * point to the beginning of the next character. <p>
     *
     * To guard against reading off the end of <code>utf16</code>, its
     * length must be given in <code>utf16_length</code>.  The value of
     * <code>utf16_length</code> is the total number of
     * <code>cgul_hchar_t</code> elements in the <code>utf16</code>
     * string; this value can be obtained by calling
     * <code>cgul_hchar_cxx::hcslen()</code> or
     * <code>cgul_hstring_cxx::get_length()</code>.  Basically, surrogate
     * pairs count as two; everything else counts as one.  Note that the
     * value is not the number of characters remaining in the string; it's
     * the total number of characters in the string. <p>
     *
     * The caller is responsible for making sure that <code>buf</code>
     * points to a buffer that can hold at least
     * <code>CGUL_WCHAR__MB_LEN_MAX + 1</code> bytes. <p>
     *
     * If an error occurs, zero is returned to the caller, an empty string
     * is returned in <code>buf</code>, and an exception is thrown. <p>
     *
     * NOTE: This method throws an exception if <code>utf16_length</code>
     * is 0.  Thus, the caller must verify that <code>utf16</code> is not
     * an empty string before calling this method! <p>
     *
     * @param[in] utf16 UTF-8 string
     * @param[in] utf16_length length of <code>utf16</code> in
     * <code>cgul_hchar_t</code> elements
     * @param[in,out] index index into <code>utf16</code>
     * @param[out] buf buffer that holds the resulting UTF-8 multi-byte sequence
     * @return number of bytes written to <code>buf</code>
     */
    static size_t
    hctomb(const cgul_hchar_t* utf16,
           size_t utf16_length,
           size_t* index,
           char* buf)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_unicode__hctomb(&local, utf16, utf16_length, index, buf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * UTF-16 to UTF-32 character conversion. <p>
     *
     * This method converts exactly one UTF-16 multi-byte character to a
     * UTF-32 wide character and returns it.  The multiple bytes that
     * comprise the UTF-16 character start at <code>utf16[*index]</code>. <p>
     *
     * In order to make iterating over an entire UTF-16 string easy, when
     * this method returns, the value of <code>*index</code> is updated to
     * point to the beginning of the next character. <p>
     *
     * To guard against reading off the end of <code>utf16</code>, its
     * length must be given in <code>utf16_length</code>.  The value of
     * <code>utf16_length</code> is the total number of
     * <code>cgul_hchar_t</code> elements in the <code>utf16</code>
     * string; this value can be obtained by calling
     * <code>cgul_hchar_cxx::hcslen()</code> or
     * <code>cgul_hstring_cxx::get_length()</code>.  Basically, surrogate
     * pairs count as two; everything else counts as one.  Note that the
     * value is not the number of characters remaining in the string; it's
     * the total number of characters in the string. <p>
     *
     * NOTE: This method throws an exception if <code>utf16_length</code>
     * is 0.  Thus, the caller must verify that <code>utf16</code> is not
     * an empty string before calling this method! <p>
     *
     * If an error occurs, <code>CGUL_HCHAR__NUL</code> is returned, and
     * an exception is thrown. <p>
     *
     * @param[in] utf16 UTF-16 string
     * @param[in] utf16_length length of <code>utf16</code> in
     * <code>cgul_hchar_t</code> elements
     * @param[in,out] index index into <code>utf16</code>
     * @return next UTF-32 character
     */
    static cgul_wchar_t
    hctowc(const cgul_hchar_t* utf16,
           size_t utf16_length,
           size_t* index)
    {
        cgul_wchar_t rv = CGUL_WCHAR__NUL;
        cgul_exception_t local = NULL;
        rv = cgul_unicode__hctowc(&local, utf16, utf16_length, index);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * UTF-32 to UTF-16 character conversion. <p>
     *
     * This method converts exactly one UTF-32 wide character to a UTF-16
     * multi-byte character sequence and returns it in <code>buf</code> as
     * a <code>cgul_hchar_t</code> string that is <code>NUL</code>
     * terminated.  The UTF-32 character that gets converted is passed in
     * as <code>wc</code>.  As an added convenience, this method returns
     * the number of bytes written to <code>buf</code> excluding the
     * <code>NUL</code> terminator. <p>
     *
     * The caller is responsible for making sure that <code>buf</code>
     * points to a buffer that can hold at least three
     * <code>cgul_hchar_t</code> characters.  This accounts for the worst
     * case which requires a surrogate pair and a <code>NUL</code>
     * terminator. <p>
     *
     * If an error occurs, zero is returned to the caller, an empty string
     * is returned in <code>buf</code>, and an exception is thrown.
     * Errors can occur, for example, if <code>wc</code> is larger than
     * 0x10ffff which is the largest valid Unicode code point. <p>
     *
     * @param[in] utf32 UTF-32 wide character
     * @param[out] buf buffer that holds the resulting UTF-16 multi-byte sequence
     * @return number of bytes written to <code>buf</code>
     */
    static size_t
    wctohc(const cgul_wchar_t utf32,
           cgul_hchar_t* buf)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_unicode__wctohc(&local, utf32, buf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method counts the number of Unicode characters (not bytes) in
     * the UTF-8 string <code>utf8</code> and returns the result.  If an
     * error occurs decoding <code>utf8</code>, <code>0</code> is
     * returned, and an exception is thrown. <p>
     *
     * If a leading byte-order mark (BOM) is present and
     * <code>skip_bom</code> is set, the leading BOM will not be included
     * in the character count. <p>
     *
     * To guard against reading off the end of <code>utf8</code>, its
     * length must be given in <code>utf8_length</code>.  The value of
     * <code>utf8_length</code> is the total number of bytes in the
     * <code>utf8</code> string. <p>
     *
     * @param[in] utf8 UTF-8 string
     * @param[in] utf8_length length of <code>utf8</code> in bytes
     * @param[in] skip_bom whether to skip the byte-order mark
     * @return number of Unicode characters in <code>utf8</code>
     */
    static size_t
    get_char_count(const char* utf8,
                   size_t utf8_length,
                   int skip_bom)

    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_unicode__get_char_count(&local, utf8, utf8_length, skip_bom);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method counts the number of Unicode characters (not bytes and
     * not <code>cgul_hchar_t</code> elements) in the UTF-16 string
     * <code>utf16</code> and returns the result.  If an error occurs
     * decoding <code>utf16</code>, <code>0</code> is returned, and an
     * exception is thrown. <p>
     *
     * If a leading byte-order mark (BOM) is present and
     * <code>skip_bom</code> is set, the leading BOM will not be included
     * in the character count. <p>
     *
     * To guard against reading off the end of <code>utf16</code>, its
     * length must be given in <code>utf16_length</code>.  The value of
     * <code>utf16_length</code> is the total number of
     * <code>cgul_hchar_t</code> elements in the <code>utf16</code>
     * string; this value can be obtained by calling
     * <code>cgul_hchar_cxx::hcslen()</code> or
     * <code>cgul_hstring_cxx::get_length()</code>.  Basically, surrogate
     * pairs count as two; everything else counts as one. <p>
     *
     * @param[in] utf16 UTF-16 string
     * @param[in] utf16_length length of <code>utf16</code> in
     * <code>cgul_hchar_t</code> elements
     * @param[in] skip_bom whether to skip the byte-order mark
     * @return number of Unicode characters in <code>utf16</code>
     */
    static size_t
    get_hchar_count(const cgul_hchar_t* utf16,
                    size_t utf16_length,
                    int skip_bom)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_unicode__get_hchar_count(&local,
                                           utf16,
                                           utf16_length,
                                           skip_bom);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method counts the number of Unicode characters (not bytes) in
     * the UTF-32 string <code>utf32</code> and returns the result.  If an
     * error occurs decoding <code>utf32</code>, <code>0</code> is
     * returned, and an exception is thrown. <p>
     *
     * If a leading byte-order mark (BOM) is present and
     * <code>skip_bom</code> is set, the leading BOM will not be included
     * in the character count. <p>
     *
     * @param[in] utf32 UTF-32 string
     * @param[in] skip_bom whether to skip the byte-order mark
     * @return number of Unicode characters in <code>utf32</code>
     */
    static size_t
    get_wchar_count(const cgul_wchar_t* utf32,
                    int skip_bom)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_unicode__get_wchar_count(&local, utf32, skip_bom);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

};


#endif  /*  CGUL_UNICODE_CXX_H  */
