#ifndef CGUL_VARSUB_H
#define CGUL_VARSUB_H

/**
 * Perform variable substitution on a string.
 *
 * @file
 * @author Paul Serice
 * @brief perform variable substituion
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_string.h"
#include "cgul_convspec.h"

CGUL_BEGIN_C

/**
 * This typedef defines the lookup function used by
 * <code>cgul_varsub()</code> to substitute a variable with its value.
 * When <code>cgul_varsub()</code> encounters a variable, it passes
 * the name of the variable to the lookup function in
 * <code>varname</code>.  The lookup function should then append the
 * value associated with <code>varname</code> to the destination
 * string <code>dst</code>. <p>
 *
 * Note that the <code>dst</code> that is passed into the lookup
 * function is the same <code>dst</code> that is passed into
 * <code>cgul_varsub()</code>.  By altering <code>dst</code>, the
 * lookup function is directly altering the result that will be
 * returned to the caller. <p>
 *
 * The lookup function should return <code>1</code> if the lookup was
 * performed and substituted for <code>varname</code>.  On the other
 * hand, if the lookup failed, <code>0</code> should be returned so
 * <code>cgul_varsub()</code> will know to insert <code>varname</code>
 * into <code>dst</code> literally. <p>
 *
 * If an error occurs during the execution of the lookup function, it
 * should use <code>cex</code> to throw a c-style exception. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] varname variable name
 * @param[in,out] dst destination string
 * @param[in] data client data
 * @return whether the lookup was successful
 * @see cgul_varsub()
 */
typedef int
(*cgul_varsub__lookup_t)(cgul_exception_t* cex,
                         const char* varname,
                         cgul_string_t dst,
                         void* data);

/**
 * This typedef defines the lookup function used by
 * <code>cgul_varsub_formatted()</code> to substitute the variable
 * <code>varname</code> with its value using an optional
 * printf()-style conversion specification <code>cs</code>.  The
 * lookup function should not try to delete <code>cs</code>. <p>
 *
 * If <code>cgul_varsub_formatted()</code> does not find a conversion
 * specification or if it detects that there is a problem with the
 * conversion specification, it will set <code>cs</code> to
 * <code>NULL</code>.  Thus, the lookup function must check to make
 * sure <code>cs</code> is not <code>NULL</code> before attempting to
 * dereference it. <p>
 *
 * <code>cgul_varsub_formatted()</code> considers conversion
 * specifications like "%*s", "%.*s", or "%*.*s" to be invalid because
 * they serve no good purpose in this context, and they have negative
 * security implications.  The lookup function will be called with
 * <code>cs</code> set to <code>NULL</code> if they are detected. <p>
 *
 * Thus to validate <code>cs</code>, the lookup function primarily
 * needs to verify that the type of the conversion specification
 * matches the type of the value that is being appended.  After this
 * is verified, it should then be safe to call
 * <code>cgul_convspec__serialize()</code> to convert
 * <code>cs</code> to a format string that can then be used with
 * <code>cgul_string__append_sprintf()</code> in order to append
 * <code>varname</code> to <code>dst</code>. <p>
 *
 * Note that the <code>dst</code> that is passed into the lookup
 * function is the same <code>dst</code> that is passed into
 * <code>cgul_varsub_formatted()</code>.  By altering
 * <code>dst</code>, the lookup function is directly altering the
 * result that will be returned to the caller. <p>
 *
 * The lookup function should return <code>1</code> if the lookup was
 * performed and substituted for <code>varname</code>.  On the other
 * hand, if the lookup failed, <code>0</code> should be returned so
 * <code>cgul_varsub_formatted()</code> will know to insert
 * <code>varname</code> into <code>dst</code> literally. <p>
 *
 * If an error occurs during the execution of the lookup function, it
 * should use <code>cex</code> to throw a c-style exception. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] varname variable name
 * @param[in] cs conversion specification or <code>NULL</code>
 * @param[in,out] dst destination string
 * @param[in] data client data
 * @return whether the lookup was successful
 * @see cgul_varsub_formatted()
 */
typedef int
(*cgul_varsub_formatted__lookup_t)(cgul_exception_t* cex,
                                   const char* varname,
                                   cgul_convspec_t cs,
                                   cgul_string_t dst,
                                   void* data);

/**
 * This function performs variable substitution on the string
 * <code>src</code> appending the result to <code>dst</code>.  If an
 * error occurs, an exception is thrown.  If an exception is thrown,
 * <code>dst</code> will contain a partial result. <p>
 *
 * The syntax for variable names embedded in <code>src</code> should
 * follow the same syntax that is used by the Bourne shell.  For
 * example if name="Mr. President", "Hello, $name!" or "Hello,
 * ${name}!" will yield "Hello, Mr. President!" after basic variable
 * substitution is performed. <p>
 *
 * Any character can be escaped by prefixing it with the backslash
 * character.  The following conventional escape sequences produce the
 * conventional results: <p>
 *
 * <pre>
 *     "\a", "\b", "\t", "\n", "\v", "\f", "\r", "\\", "\$"
 * </pre>
 *
 * The client is responsible for providing the lookup function
 * <code>f</code> that maps variable names to values.  <code>f</code>
 * is free to treat the variable names creatively.  By placing the
 * variable names in braces, it is possible to embed extra information
 * in the source string <code>src</code> because the name is
 * everything between the braces and is not limited to alphanumeric or
 * underscore characters.  <code>f</code> just needs to be written to
 * handle the extra information.  (See
 * <code>cgul_varsub_formatted()</code> for an example of how to take
 * advantage of this feature.) <p>
 *
 * When using braces to demarcate variable names, it is worth knowing
 * that braces nest, and escape sequences are enabled.  If you escape
 * a brace, it will be treated like any other character at the current
 * brace level.  An escaped brace will not increase or decrease the
 * current brace level, and it will not match any other brace. <p>
 *
 * <code>data</code> will be passed to <code>f</code> as the client
 * data.  Typically, <code>data</code> will hold the data needed by
 * <code>f</code> to perform the lookup. <p>
 *
 * See tests/main_varsub.c for a simple example of how to write a
 * lookup function <code>f</code>.  See the private
 * <code>cgul_varsub_formatted__lookup_cb()</code> for a more advanced
 * example of how to write a lookup function <code>f</code> that takes
 * advantage of the ability to embed extra information in the variable
 * names. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] src source string
 * @param[out] dst destination string
 * @param[in] f lookup function
 * @param[in] data client data
 * @see cgul_varsub_formatted()
 * @see cgul_varsub__lookup_t
 */
extern CGUL_EXPORT void
cgul_varsub(cgul_exception_t* cex,
            const char* src,
            cgul_string_t dst,
            cgul_varsub__lookup_t f,
            void* data);

/**
 * This function performs variable substitution on the string
 * <code>src</code> appending the result to <code>dst</code>.  If an
 * optional printf()-style conversion specification is embedded in the
 * variable name, it will be used to format the variable when
 * appending to <code>dst</code>.  If an error occurs, an exception is
 * thrown.  If an exception is thrown, <code>dst</code> will contain a
 * partial result. <p>
 *
 * As explained in the comments for <code>cgul_varsub()</code>, it
 * understands the Bourne shell convention of placing braces around
 * the name of a variable.  When braces are used, everything between
 * the braces is treat as the literal name of the variable including
 * non-alphanumeric characters.  This makes it possible to embed an
 * optional printf()-style conversion specification in the name of the
 * variable. <p>
 *
 * This function registers a local callback with
 * <code>cgul_varsub()</code> so that <code>cgul_varsub()</code> can
 * parse the source string <code>src</code>.  Each time
 * <code>cgul_varsub()</code> detects a variable name, the local
 * callback is invoked.  The local callback then splits the string
 * into the real variable name and an option printf()-style conversion
 * specification which is then passed to <code>f</code>. <p>
 *
 * For example, if <code>src</code> is set to "${key:%32s}",
 * <code>cgul_varsub()</code> will call the local callback with the
 * variable name set to "key:%32s".  The local callback will split the
 * string into the real variable name ("key") and its conversion
 * specification ("%32s").  The local callback will then call
 * <code>f</code> passing in the real variable name and its conversion
 * specification. <p>
 *
 * If the optional conversion specification is missing or invalid,
 * <code>f</code> will be called with the conversion specification set
 * to <code>NULL</code>.  Conversion specifications like "%*s",
 * "%.*s", or "%*.*s" are considered to be invalid because they serve
 * no good purpose in this context, and they have negative security
 * implications.  <p>
 *
 * Any character can be escaped by prefixing it with the backslash
 * character.  The following conventional escape sequences produce the
 * conventional results: <p>
 *
 * <pre>
 *     "\a", "\b", "\t", "\n", "\v", "\f", "\r", "\\", "\$"
 * </pre>
 *
 * <code>data</code> will be passed to <code>f</code> as the client
 * data.  Typically, <code>data</code> will hold the data needed by
 * <code>f</code> to perform the lookup. <p>
 *
 * See tests/main_varsub_formatted.c for an example of how to write
 * the lookup function <code>f</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] src source string
 * @param[out] dst destination string
 * @param[in] f lookup function
 * @param[in] data client data
 * @see cgul_varsub()
 * @see cgul_varsub_formatted__lookup_t
 */
extern CGUL_EXPORT void
cgul_varsub_formatted(cgul_exception_t* cex,
                      const char* src,
                      cgul_string_t dst,
                      cgul_varsub_formatted__lookup_t f,
                      void* data);

CGUL_END_C

#endif  /*  CGUL_VARSUB_H  */
