#include "cgul_bdf_glyph_p.h"

/**
 * Allocate a new <code>glyph->bitmap</code> array using
 * <code>glyph->bytes_wide</code> and <code>glyph->bb_height</code> to
 * calculate the size of the new array.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 */
static void
cgul_bdf_glyph__allocate_bitmap(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph)
{
    size_t bytes_total = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    /* If necessary, free the previous underlying bitmap data. */
    if (glyph->bitmap) {
        free(glyph->bitmap);
        glyph->bitmap = NULL;
    }

    /* Calculate the size of the array needed to hold the entire glyph. */
    bytes_total = glyph->bytes_wide * glyph->bb_height;

    /* Allocate a new bitmap that matches the new dimensions.  The new
     * bitmap has all of its pixels unset. */
    if (bytes_total) {
        glyph->bitmap = (unsigned char*)cgul_libc__calloc(cex, 1, bytes_total);
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

/**
 * This function accepts a value in <code>x</code> between 0 and 15
 * (inclusive) and returns the ASCII character that represents the hex
 * digit for that value.  The user is responsible for making sure
 * <code>x</code> is in the allowed range.  If an error occurs, a
 * <code>NUL</code> character is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] x value returned as a hex digit
 * @return hex digit
 */
static char
cgul_bdf_glyph__four_bits_as_hex(cgul_exception_t* cex, 
                                 int x)
{
    char rv = '\0';

    assert(cex);
    if (*cex) {
        goto out;
    }

    if ((x < 0) || (x > 15)) {
        *cex = cgul_exception__new("cgul_bdf_glyph__four_bits_as_hex_digit: "
                                   "Unable to convert to hex digit: %d",
                                   x);
        goto out;
    }

    switch (x) {
    case 0x0:
        rv = '0';
        break;
    case 0x1:
        rv = '1';
        break;
    case 0x2:
        rv = '2';
        break;
    case 0x3:
        rv = '3';
        break;
    case 0x4:
        rv = '4';
        break;
    case 0x5:
        rv = '5';
        break;
    case 0x6:
        rv = '6';
        break;
    case 0x7:
        rv = '7';
        break;
    case 0x8:
        rv = '8';
        break;
    case 0x9:
        rv = '9';
        break;
    case 0xA:
        rv = 'A';
        break;
    case 0xB:
        rv = 'B';
        break;
    case 0xC:
        rv = 'C';
        break;
    case 0xD:
        rv = 'D';
        break;
    case 0xE:
        rv = 'E';
        break;
    case 0xF:
        rv = 'F';
        break;
    default:
        *cex = cgul_exception__new("Internal Error: "
                                   "cgul_bdf_glyph__four_bits_as_hex: "
                                   "Unexpected switch value.");
        goto out;
    }

 out:
    return rv;
}

cgul_bdf_glyph_t
cgul_bdf_glyph__new(cgul_exception_t* cex)
{
    cgul_bdf_glyph_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the object. */
    rv = (cgul_bdf_glyph_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up if an exception is thrown. */
    if (*cex && rv) {
        cgul_bdf_glyph__delete(rv);
        rv = NULL;
    }

    return rv;
}

/**
 * This method holds the common code needed for releasing all
 * internally allocated resources without deleting the object itself.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 */
static void
cgul_bdf_glyph__release_resources(cgul_exception_t* cex,
                                  cgul_bdf_glyph_t glyph)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    /* bitmap */
    if (glyph->bitmap) {
        free(glyph->bitmap);
        glyph->bitmap = NULL;
    }

    /* descriptive_name */
    if (glyph->descriptive_name) {
        free(glyph->descriptive_name);
        glyph->descriptive_name = NULL;
    }

 out:
    return;
}

void
cgul_bdf_glyph__delete(cgul_bdf_glyph_t glyph)
{
    cgul_exception_t local = NULL;

    if (glyph) {

        /* Release all internally allocated resources. */
        cgul_bdf_glyph__release_resources(&local, glyph);

        /* Free the object itself. */
        free(glyph);
    }
}

void
cgul_bdf_glyph__copy(cgul_exception_t* cex,
                     cgul_bdf_glyph_t dst,
                     cgul_bdf_glyph_t src)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(src);
    assert(dst);

    /* Don't do anything if src and dst are the same glyph. */
    if (src == dst) {
        goto out;
    }

    /* Release all of the internally allocated resources for "dst". */
    cgul_bdf_glyph__release_resources(cex, dst);

    /* Do a shallow copy first. */
    cgul_libc__memcpy(cex, dst, src, sizeof(*src));

    /* Reset the dst pointers that the src object owns. */
    dst->descriptive_name = NULL;
    dst->bitmap = NULL;

    /* Copy the descriptive name. */
    dst->descriptive_name = cgul_libc__strdup(cex, src->descriptive_name);

    /* Allocate a new bitmap. */
    cgul_bdf_glyph__allocate_bitmap(cex, dst);
    if (*cex) {
        goto out;
    }

    /* Copy the bitmap. */
    cgul_libc__memcpy(cex,
                      dst->bitmap,
                      src->bitmap,
                      src->bytes_wide * src->bb_height);

 out:
    return;
}

const char*
cgul_bdf_glyph__get_descriptive_name(cgul_exception_t* cex,
                                     cgul_bdf_glyph_t glyph)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->descriptive_name;

 out:
    return rv;
}

cgul_wchar_t
cgul_bdf_glyph__get_encoding(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph)
{
    cgul_wchar_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->encoding;

 out:
    return rv;
}

unsigned int
cgul_bdf_glyph__get_ascent(cgul_exception_t* cex,
                           cgul_bdf_glyph_t glyph)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    /* Calculate the ascent for the normal case. */
    rv = (int)glyph->bb_height + glyph->bb_y_offset;

    /* It is possible that bb_y_offset is negative and has a larger
     * magnitude than bb_height.  In this case, there is no ascent. */
    if (rv < 0) {
        rv = 0;
    }

 out:
    return (unsigned int)rv;
}

unsigned int
cgul_bdf_glyph__get_descent(cgul_exception_t* cex,
                            cgul_bdf_glyph_t glyph)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    /* The descent is zero unless the bounding box has a negative y offset. */
    if (glyph->bb_y_offset < 0) {
        rv = -glyph->bb_y_offset;
    }

 out:
    return (unsigned int)rv;
}

void
cgul_bdf_glyph__set_descriptive_name(cgul_exception_t* cex,
                                     cgul_bdf_glyph_t glyph,
                                     const char* descriptive_name)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert descriptive_name.  (See below.) */
    assert(glyph);

    /* The user is allowed to disable the descriptive name by passing
     * in NULL. */
    if (descriptive_name == NULL) {
        if (glyph->descriptive_name) {
            free(glyph->descriptive_name);
            glyph->descriptive_name = NULL;
            goto out;
        }
    }

    /* Otherwise, we save a copy of the descriptive name. */
    glyph->descriptive_name = cgul_libc__strdup(cex, descriptive_name);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_bdf_glyph__set_encoding(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             cgul_wchar_t encoding)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    glyph->encoding = encoding;

 out:
    return;
}

unsigned int
cgul_bdf_glyph__get_swidth_x(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->swidth_x;

 out:
    return rv;
}

void
cgul_bdf_glyph__set_swidth_x(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             unsigned int swidth_x)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    glyph->swidth_x = swidth_x;

 out:
    return;
}

unsigned int
cgul_bdf_glyph__get_swidth_y(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->swidth_y;

 out:
    return rv;
}

void
cgul_bdf_glyph__set_swidth_y(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             unsigned int swidth_y)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    glyph->swidth_y = swidth_y;

 out:
    return;
}

unsigned int
cgul_bdf_glyph__get_dwidth_x(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->dwidth_x;

 out:
    return rv;
}

void
cgul_bdf_glyph__set_dwidth_x(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             unsigned int dwidth_x)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    glyph->dwidth_x = dwidth_x;

 out:
    return;
}

unsigned int
cgul_bdf_glyph__get_dwidth_y(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->dwidth_y;

 out:
    return rv;
}

void
cgul_bdf_glyph__set_dwidth_y(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             unsigned int dwidth_y)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    glyph->dwidth_y = dwidth_y;

 out:
    return;
}

unsigned int
cgul_bdf_glyph__get_bb_width(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->bb_width;

 out:
    return rv;
}

void
cgul_bdf_glyph__set_bb_width(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             unsigned int bb_width)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    /* Save the new width of the bounding box. */
    glyph->bb_width = bb_width;

    /* The dots that make up the glyph are stored as a bitmap padded
     * out to the nearest byte.  So, calculate the total number of
     * bytes needed to hold one scan line of the glyph. */
    glyph->bytes_wide = (bb_width + 7) / 8;

    /* Allocate a new bitmap using the new dimensions. */
    cgul_bdf_glyph__allocate_bitmap(cex, glyph);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

unsigned int
cgul_bdf_glyph__get_bb_height(cgul_exception_t* cex,
                              cgul_bdf_glyph_t glyph)
{
    unsigned int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->bb_height;

 out:
    return rv;
}

void
cgul_bdf_glyph__set_bb_height(cgul_exception_t* cex,
                              cgul_bdf_glyph_t glyph,
                              unsigned int bb_height)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    /* Save the new height of the bounding box. */
    glyph->bb_height = bb_height;

    /* Allocate a new bitmap using the new dimensions. */
    cgul_bdf_glyph__allocate_bitmap(cex, glyph);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

int
cgul_bdf_glyph__get_bb_x_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->bb_x_offset;

 out:
    return rv;
}

void
cgul_bdf_glyph__set_bb_x_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph,
                                int bb_x_offset)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    glyph->bb_x_offset = bb_x_offset;

 out:
    return;
}

int
cgul_bdf_glyph__get_bb_y_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->bb_y_offset;

 out:
    return rv;
}

void
cgul_bdf_glyph__set_bb_y_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph,
                                int bb_y_offset)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    glyph->bb_y_offset = bb_y_offset;

 out:
    return;
}

int
cgul_bdf_glyph__get_dot(cgul_exception_t* cex,
                        cgul_bdf_glyph_t glyph,
                        unsigned int x,
                        unsigned int y)
{
    int rv = 0;
    unsigned int byte_x = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    /* Because the bits are packed into "glyph->bitmap", we have to
     * map the bit "x" to the correct byte.  Also, because x values
     * start at zero, we need to add one in order to get the math
     * right. */
    byte_x = x / 8;

    /* Make sure the bitmap array is allocated and that byte_x and
     * y are within bounds. */
    if (   glyph->bitmap
        && (byte_x < glyph->bytes_wide)
        && (y < glyph->bb_height))
    {
        /* Determine if this dot is set. */
        if (glyph->bitmap[y * glyph->bytes_wide + byte_x] & (1 << (x % 8))) {
            rv = 1;
        }
    }

 out:
    return rv;
}

void
cgul_bdf_glyph__set_dot(cgul_exception_t* cex,
                        cgul_bdf_glyph_t glyph,
                        unsigned int x,
                        unsigned int y,
                        int is_set)
{
    unsigned int byte_x = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    /* Because the bits are packed into "bitmap", we have to map
     * the bit "x" to the correct byte. */
    byte_x = x / 8;

    /* Make sure the bitmap array is allocated and that byte_x and
     * y are within bounds. */
    if (   glyph->bitmap
        && (byte_x < glyph->bytes_wide)
        && (y < glyph->bb_height))
    {
        if (is_set) {
            glyph->bitmap[y * glyph->bytes_wide + byte_x] |= (1 << (x % 8));
        } else {
            glyph->bitmap[y * glyph->bytes_wide + byte_x] &= ~(1 << (x % 8));
        }
    }

 out:
    return;
}

size_t
cgul_bdf_glyph__get_bytes_wide(cgul_exception_t* cex,
                               cgul_bdf_glyph_t glyph)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->bytes_wide;

 out:
    return rv;
}

unsigned long
cgul_bdf_glyph__get_file_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph)
{
    unsigned long rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    rv = glyph->file_offset;

 out:
    return rv;
}

void
cgul_bdf_glyph__set_file_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph,
                                unsigned long int offset)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);

    glyph->file_offset = offset;

 out:
    return;
}

void
cgul_bdf_glyph__append_bitmap_to_string(cgul_exception_t* cex,
                                        cgul_bdf_glyph_t glyph,
                                        cgul_string_t s)
{
    unsigned char b = '\0';
    char cout = '\0';
    unsigned int i = 0;
    unsigned int j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);
    assert(s);

    /* Iterate over the bitmap. */
    for (i = 0 ; i < glyph->bb_height ; ++i) {
        for (j = 0 ; j < glyph->bytes_wide ; ++j) {

            /* Get this byte, but ... */
            b = glyph->bitmap[i * glyph->bytes_wide + j];

            /* ... It was easier to write get_dot() and set_dot()
             * using the least significant bit to represent the
             * left-most pixel.  This means "b" is the mirror
             * image of what we actually need to output.  So, flip
             * it around. */
            b =   ((b & 0x01) << 7)
                | ((b & 0x02) << 5)
                | ((b & 0x04) << 3)
                | ((b & 0x08) << 1)
                | ((b & 0x10) >> 1)
                | ((b & 0x20) >> 3)
                | ((b & 0x40) >> 5)
                | ((b & 0x80) >> 7);

            /* Output the first character of this byte. */
            cout = cgul_bdf_glyph__four_bits_as_hex(cex, (b >> 4) & 0x0f);
            cgul_string__append_char(cex, s, cout);

            /* Output the second character of this byte. */
            cout = cgul_bdf_glyph__four_bits_as_hex(cex, b & 0x0f);
            cgul_string__append_char(cex, s, cout);
            if (*cex) {
                goto out;
            }
        }
        cgul_string__append_char(cex, s, '\n');
    }

 out:
    return;
}

extern void
cgul_bdf_glyph__write_file(cgul_exception_t* cex,
                           cgul_bdf_glyph_t glyph,
                           FILE* f)
{
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);
    assert(f);

    /* Allocate a temporary string. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* It is legal for glyph->descriptive_name to be NULL.  A NULL
     * value means this character does not have a descriptive name.
     * In that case, we will generate one for it. */
    if (glyph->descriptive_name) {
        cgul_libc__fprintf(cex,
                           f,
                           "STARTCHAR %s\n",
                           glyph->descriptive_name);
    } else {
        cgul_libc__fprintf(cex,
                           f,
                           "STARTCHAR char-%lu\n",
                           (unsigned long)glyph->encoding);
    }

    /* Draw the comment. */
    cgul_string__set_value(cex, s, "");
    cgul_bdf_glyph__sdraw(cex,
                          glyph,
                          s,
                          "COMMENT                         ",
                          NULL,
                          '*',
                          ' ');
    cgul_libc__fprintf(cex, f, "%s", cgul_string__get_value(cex, s));

    cgul_libc__fprintf(cex,
                       f,
                       "ENCODING %lu\n",
                       (unsigned long)glyph->encoding);

    cgul_libc__fprintf(cex,
                       f,
                       "SWIDTH %u %u\n",
                       glyph->swidth_x,
                       glyph->swidth_y);

    cgul_libc__fprintf(cex,
                       f,
                       "DWIDTH %u %u\n",
                       glyph->dwidth_x,
                       glyph->dwidth_y);

    cgul_libc__fprintf(cex,
                       f,
                       "BBX %u %u %d %d\n",
                       glyph->bb_width,
                       glyph->bb_height,
                       glyph->bb_x_offset,
                       glyph->bb_y_offset);

    cgul_libc__fprintf(cex,
                       f,
                       "BITMAP\n");

    /* Append the bitmap. */
    cgul_string__set_value(cex, s, "");
    cgul_bdf_glyph__append_bitmap_to_string(cex, glyph, s);
    cgul_libc__fprintf(cex, f, "%s", cgul_string__get_value(cex,s));

    cgul_libc__fprintf(cex,
                       f,
                       "ENDCHAR\n");

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

void
cgul_bdf_glyph__fdraw(cgul_exception_t* cex,
                      cgul_bdf_glyph_t glyph,
                      FILE* f,
                      const char* prefix,
                      const char* suffix,
                      char c_active,
                      char c_inactive)
{
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);
    assert(f);

    /* Allocate a temporary string. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Draw to the temporary string. */
    cgul_bdf_glyph__sdraw(cex, glyph, s, prefix, suffix, c_active, c_inactive);

    /* Output the string. */
    cgul_libc__fprintf(cex, f, "%s", cgul_string__get_value(cex, s));

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

void
cgul_bdf_glyph__sdraw(cgul_exception_t* cex,
                      cgul_bdf_glyph_t glyph,
                      cgul_string_t s,
                      const char* prefix,
                      const char* suffix,
                      char c_active,
                      char c_inactive)
{
    unsigned int i = 0;
    unsigned int j = 0;
    int dot = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(glyph);
    assert(s);

    /* Iterate over the bitmap. */
    for (i = 0 ; i < glyph->bb_height ; ++i) {

        /* Append the prefix. */
        if (prefix) {
            cgul_string__append(cex, s, prefix);
        }

        /* Append the glyph as ASCII. */
        for (j = 0 ; j < glyph->bb_width ; ++j) {
            dot = cgul_bdf_glyph__get_dot(cex, glyph, j, i);
            if (dot) {
                cgul_string__append_char(cex, s, c_active);
            } else {
                cgul_string__append_char(cex, s, c_inactive);
            }
        }

        /* Append the suffix. */
        if (suffix) {
            cgul_string__append(cex, s, suffix);
        }

        /* Terminate this scan line. */
        cgul_string__append_char(cex, s, '\n');

        /* Check for errors. */
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}
