#include "cgul_list_p.h"
#include "cgul_libc.h"

/* Forward declarations. */
static void
cgul_list__insert_after_node__common(cgul_exception_t* cex,
                                     cgul_list_t l,
                                     cgul_list_node_t n1,
                                     cgul_list_node_t n2);

/**
 * This function inserts the node <code>n</code> into the empty list
 * <code>l</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n node to insert
 */
static void
cgul_list__insert_into_empty_list(cgul_exception_t* cex,
                                  cgul_list_t l,
                                  cgul_list_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);
    assert(n);
    assert(l->size == 0);

    /* Insert n into l. */
    n->next = NULL;
    n->prev = NULL;
    l->front = n;
    l->back = n;
    ++(l->size);

 out:
    return;
}

/**
 * This method inserts the node <code>n2</code> before the node
 * <code>n1</code> in the list <code>l</code>.  If <code>n1</code> is
 * <code>NULL</code>, <code>n2</code> is inserted before the
 * <code>NULL</code> at the end of the list.  <code>n2</code> must not
 * be in the list already.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n1 insert before this node
 * @param[in] n2 node to insert
 */
static void
cgul_list__insert_before_node__common(cgul_exception_t* cex,
                                      cgul_list_t l,
                                      cgul_list_node_t n1,
                                      cgul_list_node_t n2)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* n1 is allowed to be NULL. */
    assert(l);
    assert(n2);

    /* If n1 is NULL, one of the following special cases apply. */
    if (n1 == NULL) {
        /* If the list is empty insert into an empty list. */
        if (l->size == 0) {
            cgul_list__insert_into_empty_list(cex, l, n2);
            goto out;
        }
        /* Otherwise, insert before the NULL at the end of the list.
         * In other words, insert after the back node. */
        cgul_list__insert_after_node__common(cex, l, l->back, n2);
        goto out;
    }

    /* Nothing to do if the two nodes are the same object. */
    if (n1 == n2) {
        goto out;
    }

    /* Insert n2 before n1. */
    if (n1->prev) {
        n1->prev->next = n2;
    } else {
        l->front = n2;
    }
    n2->prev = n1->prev;
    n2->next = n1;
    n1->prev = n2;
    ++(l->size);

 out:
    return;
}

/**
 * This method inserts the node <code>n2</code> after the node
 * <code>n1</code> in the list <code>l</code>.  If <code>n1</code> is
 * <code>NULL</code>, <code>n2</code> is inserted after the
 * <code>NULL</code> at the beginning of the list.  <code>n2</code>
 * must not be in the list already.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n1 insert after this node
 * @param[in] n2 node to insert
 */
static void
cgul_list__insert_after_node__common(cgul_exception_t* cex,
                                     cgul_list_t l,
                                     cgul_list_node_t n1,
                                     cgul_list_node_t n2)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* n1 is allowed to be NULL. */
    assert(l);
    assert(n2);

    /* If n1 is NULL, one of the following special cases apply. */
    if (n1 == NULL) {
        /* If the list is empty insert into an empty list. */
        if (l->size == 0) {
            cgul_list__insert_into_empty_list(cex, l, n2);
            goto out;
        }
        /* Otherwise, insert after the NULL at the beginning of the
         * list.  In other words, insert before the front node. */
        cgul_list__insert_before_node__common(cex, l, l->front, n2);
        goto out;
    }

    /* Nothing to do if the two nodes are the same object. */
    if (n1 == n2) {
        goto out;
    }

    /* Insert n2 after n1. */
    if (n1->next) {
        n1->next->prev = n2;
    } else {
        l->back = n2;
    }
    n2->next = n1->next;
    n2->prev = n1;
    n1->next = n2;
    ++(l->size);

 out:
    return;
}

/**
 * This method removes the node <code>n</code> from the list
 * <code>l</code>.  The caller is responsible for putting
 * <code>n</code> back on <code>l->cache</code> when <code>n</code> is
 * no longer in use. <p>
 *
 * Do not confuse this method with the public
 * <code>cgul_list__remove_node()</code> method which eventually calls
 * this method through <code>cgul_list__free_node_range()</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n node to remove
 */
static void
cgul_list__remove_node__common(cgul_exception_t* cex,
                               cgul_list_t l,
                               cgul_list_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);
    assert(n);

    /* Remove n from l. */
    if (n->prev) {
        n->prev->next = n->next;
    } else {
        l->front = n->next;
    }
    if (n->next) {
        n->next->prev = n->prev;
    } else {
        l->back = n->prev;
    }

    --(l->size);

 out:
    return;
}

/**
 * This method removes and frees all nodes between <code>n1</code> and
 * <code>n2</code> (inclusive) from the list <code>l</code>.  If
 * <code>values_cache</code> is not <code>NULL</code>, the user's
 * values are put back on the cache.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n1 node to delete first
 * @param[in] n2 node to delete last
 * @param[in] values_cache values cache
 */
static void
cgul_list__free_node_range(cgul_exception_t* cex,
                           cgul_list_t l,
                           cgul_list_node_t n1,
                           cgul_list_node_t n2,
                           cgul_cache_t values_cache)
{
    cgul_list_node_t curr = NULL;
    cgul_list_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);


    /* Iterate over the linked list remove all the nodes between n1
     * and n2 inclusively. */
    for (curr = n1 ; curr ; curr = next) {
        /* Get "next" while "curr" is still valid. */
        next = curr->next;
        /* Remove curr from l. */
        cgul_list__remove_node__common(cex, l, curr);
        /* Put the value back if values_cache is not NULL. */
        if (values_cache) {
            cgul_cache__put(cex, values_cache, (void*)curr->value);
        }
        /* Put curr back. */
        cgul_cache__put(cex, l->cache, curr);
        /* Break out if we just removed the last node. */
        if (curr == n2) {
            break;
        }
    }

 out:
    return;
}

cgul_list_t
cgul_list__new(cgul_exception_t* cex)
{
    cgul_list_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_list_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    rv->cache =
        cgul_cache__new(
            cex,
            (cgul_cache__constructor_t)&cgul_list_node__new,
            (cgul_cache__destructor_t)&cgul_list_node__delete,
            (cgul_cache__restructor_t)&cgul_list_node__initialize);
    if (*cex) {
        goto out;
    }

 out:

    if (*cex && rv) {
        cgul_list__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_list__delete(cgul_list_t l)
{
    cgul_exception_t local = NULL;

    if (l) {
        if (l->cache) {
            /* Delete all the nodes.  This might move some of the
             * nodes to the cache.  Thus, this must be done before
             * deleting the cache. */
            cgul_list__free_node_range(&local, l, l->front, l->back, NULL);
            cgul_cache__delete(l->cache);
        }

        /* Delete the object itself. */
        free(l);
    }
}

void
cgul_list__free_values(cgul_list_t l)
{
    cgul_list_node_t curr = NULL;

    if (l) {
        /* Iterate over each node. */
        for (curr = l->front ; curr ; curr = curr->next) {
            /* Free each value. */
            free((void*)curr->value);
        }
    }
}

unsigned long int
cgul_list__get_cache_size(cgul_exception_t* cex,
                          cgul_list_t l)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    rv = cgul_cache__get_size(cex, l->cache);

 out:
    return rv;
}

void
cgul_list__set_cache_size(cgul_exception_t* cex,
                          cgul_list_t l,
                          unsigned long int size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    cgul_cache__set_size(cex, l->cache, size);

 out:
    return;
}

int
cgul_list__is_empty(cgul_exception_t* cex,
                    cgul_list_t l)
{
    int rv = 1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    rv = (l->size == 0);

 out:
    return rv;
}

cgul_list_node_t
cgul_list__get_front(cgul_exception_t* cex,
                     cgul_list_t l)
{
    cgul_list_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    rv = l->front;

 out:
    return rv;
}

cgul_list_node_t
cgul_list__get_back(cgul_exception_t* cex,
                    cgul_list_t l)
{
    cgul_list_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    rv = l->back;

 out:
    return rv;
}

cgul_list_node_t
cgul_list__push_front(cgul_exception_t* cex,
                      cgul_list_t l,
                      const void* value)
{
    cgul_list_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    rv = (cgul_list_node_t)cgul_cache__get(cex, l->cache);
    if (*cex) {
        goto out;
    }

    rv->value = value;
    if (l->front == NULL) {
        assert(l->back == NULL);
        l->front = rv;
        l->back = rv;
    } else {
        l->front->prev = rv;
        rv->next = l->front;
        l->front = rv;
    }
    ++(l->size);

 out:
    return rv;
}

void*
cgul_list__pop_front(cgul_exception_t* cex,
                     cgul_list_t l)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    if (l->front) {
        rv = (void*)l->front->value;
        cgul_list__free_node_range(cex, l, l->front, l->front, NULL);
    }

 out:
    return rv;
}

cgul_list_node_t
cgul_list__push_back(cgul_exception_t* cex,
                     cgul_list_t l,
                     const void* value)
{
    cgul_list_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    rv = (cgul_list_node_t)cgul_cache__get(cex, l->cache);
    if (*cex) {
        goto out;
    }

    rv->value = value;
    if (l->back == NULL) {
        assert(l->front == NULL);
        l->front = rv;
        l->back = rv;
    } else {
        l->back->next = rv;
        rv->prev = l->back;
        l->back = rv;
    }
    ++(l->size);

 out:
    return rv;
}

void*
cgul_list__pop_back(cgul_exception_t* cex,
                    cgul_list_t l)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    if (l->back) {
        rv = (void*)l->back->value;
        cgul_list__free_node_range(cex, l, l->back, l->back, NULL);
    }

 out:
    return rv;
}

void
cgul_list__move_before_node(cgul_exception_t* cex,
                            cgul_list_t l,
                            cgul_list_node_t n1,
                            cgul_list_node_t n2)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);
    assert(n2);

    /* Nothing to do if n1 and n2 are the same node. */
    if (n1 == n2) {
        goto out;
    }

    /* Remove n2 from the list. */
    cgul_list__remove_node__common(cex, l, n2);

    /* Reinsert n2 before n1. */
    cgul_list__insert_before_node__common(cex, l, n1, n2);

 out:
    return;
}

void
cgul_list__move_after_node(cgul_exception_t* cex,
                           cgul_list_t l,
                           cgul_list_node_t n1,
                           cgul_list_node_t n2)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);
    assert(n2);

    /* Nothing to do if n1 and n2 are the same node. */
    if (n1 == n2) {
        goto out;
    }

    /* Remove n2 from the list. */
    cgul_list__remove_node__common(cex, l, n2);

    /* Reinsert n2 after n1. */
    cgul_list__insert_after_node__common(cex, l, n1, n2);

 out:
    return;
}

cgul_list_node_t
cgul_list__insert_before_node(cgul_exception_t* cex,
                              cgul_list_t l,
                              cgul_list_node_t n,
                              const void* value)
{
    cgul_list_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* n is allowed to be NULL. */
    assert(l);

    /* Get a new node from the cache. */
    rv = (cgul_list_node_t)cgul_cache__get(cex, l->cache);
    if (*cex) {
        goto out;
    }

    /* Set the node's value. */
    rv->value = value;

    /* Insert rv before n. */
    cgul_list__insert_before_node__common(cex, l, n, rv);

 out:
    return rv;
}

cgul_list_node_t
cgul_list__insert_after_node(cgul_exception_t* cex,
                             cgul_list_t l, 
                             cgul_list_node_t n,
                             const void* value)
{
    cgul_list_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* n is allowed to be NULL. */
    assert(l);

    /* Get a new node from the cache. */
    rv = (cgul_list_node_t)cgul_cache__get(cex, l->cache);
    if (*cex) {
        goto out;
    }

    /* Set the node's value. */
    rv->value = value;

    /* Insert rv after n. */
    cgul_list__insert_after_node__common(cex, l, n, rv);

 out:
    return rv;
}

cgul_list_node_t
cgul_list__insert_sorted(cgul_exception_t* cex,
                         cgul_list_t l,
                         const void* value,
                         cgul_list__compare_t f)
{
    cgul_list_node_t rv = NULL;
    cgul_list_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);
    assert(f);

    /* Iterate over the list in reverse order looking for the first
     * value (in reverse order) in the list that is less than or equal
     * to "value".  It is done this way so that calling this function
     * repeatedly usually yields the same result as calling
     * cgul_list__sort() which uses merge sort which is a stable sort.
     * The main assumption here is that user wants the sort to be
     * stable relative to insertion order.  If this is not the case,
     * the user can simply define "f" to do what is needed. */
    for (curr = l->back ; curr ; curr = curr->prev) {
        if ((*f)(curr->value, value) <= 0) {
            break;
        }
    }

    /* Insert "value" before "curr".  Note that this works even if
     * "curr" is NULL because cgul_list__insert_after_node() treats
     * NULL as the beginning of the list. */
    rv = cgul_list__insert_after_node(cex, l, curr, value);

 out:
    return rv;
}

void
cgul_list__sort(cgul_exception_t* cex,
                cgul_list_t l,
                cgul_list__compare_t f)
{
    unsigned long int i = 0;
    unsigned long int i_stop = 0;
    unsigned long int j1 = 0;
    unsigned long int j2 = 0;
    unsigned long int partition_size = 1;
    cgul_list_node_t tmp = NULL;
    cgul_list_node_t n1 = NULL;
    cgul_list_node_t n2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);
    assert(f);

    /* Nothing to do if the list has zero or one element. */
    if (l->size <= 1) {
        goto out;
    }

    /* This implementation of merge sort is bottom-up.  Each iteration
     * of the outer loop is one iteration over the entire list.  We
     * start with a parition size of 1 and merge adjacent partitions.
     * We then keep doubling the size of the partition until the
     * partition is at least as large as the entire list effectively
     * sorting the list. */
    for ( ; partition_size < l->size ; partition_size <<= 1) {

        /* Reset. */
        n2 = l->front;  /* node in second partition */
        i = 0;          /* index of current node in second partition */

        /* Iterate over the pairs of partitions merging each pair. */
        while (n2) {

            /* Move n1 to the start of the next "first" partition. */
            n1 = n2;

            /* Move n2 to the start of the next "second" partition. */
            i_stop = i + partition_size;
            for ( ; n2 && (i < i_stop) ; ++i) {
                n2 = n2->next;
            }

            /* An odd number of partitions means the last partition
             * does not have another partition with which to pair.
             * This is ok though because the last partition is still
             * sorted and eventually the even number of partitions
             * above will merge into an odd number of partitions.  At
             * that time the last partition will get paired and
             * merged.
             *
             * We can detect whether there are an odd number of
             * partitions because n2 reaches the end of the list (in
             * the for() loop above) before it reaches the start of
             * the "second" partition. */
            if (!n2) {
                continue;
            }

            /* Calculate the start of the next partition. */
            i_stop = i + partition_size;

            /* Merge the partitions.  j1 counts the number of times n1
             * has been advanced.  j2 counts the number of times n2
             * has been advanced.  Note that "i" continues to hold the
             * index of n2 relative to the entire list.  It only gets
             * incremented when n2 advances which occurs only after
             * moving n2 to the "first" partition. */
            j1 = 0;
            j2 = 0;
            while (n2 && (j1 < partition_size) && (j2 < partition_size)) {
                /* Compare the values of the nodes in each partition.
                 * If the value in the "first" partition is larger
                 * than the value in the "second" partition, insert
                 * the node from the "second" partition before the
                 * node in the "first" partition. */
                if ((*f)(n1->value, n2->value) > 0) {
                    /* Get the next node while we still can. */
                    tmp = n2->next;
                    /* Remove n2. */
                    cgul_list__remove_node__common(cex, l, n2);
                    /* Insert n2 before n1. */
                    cgul_list__insert_before_node__common(cex, l, n1, n2);
                    /* Advance n2 to the next node. */
                    n2 = tmp;
                    ++i;
                    ++j2;
                } else {
                    /* Advance n1. */
                    n1 = n1->next;
                    ++j1;
                }
            }

            /* It is possible to exhaust all the values in the "first"
             * partition without exhausting all the values in the
             * "second" partition.  In this case, we need to
             * fast-forward n2 to the start of the next partition. */
            for ( ; n2 && (i < i_stop) ; ++i) {
                n2 = n2->next;
            }
        }
    }

 out:
    return;
}

void
cgul_list__merge(cgul_exception_t* cex,
                 cgul_list_t l1,
                 cgul_list_t l2,
                 cgul_list__compare_t f)
{
    cgul_list_node_t n1 = NULL;
    cgul_list_node_t n2 = NULL;
    cgul_list_node_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l1);
    assert(l2);
    assert(f);

    /*
     * NOTE: As described in the comments for this method, it assumes
     * l1 and l2 are individually sorted.  This means we just need to
     * do a single pass merge.
     */

    /* Initialize n1 and n2 to be the front nodes. */
    n1 = l1->front;
    n2 = l2->front;

    /* Iterate over the lists. */
    while (n1 && n2) {
        /* Compare the current values in each list.  If the value in
         * the first list is greater than the value in the second
         * list, remove the node from the second list and insert it
         * before the node in the first list. */
        if ((*f)(n1->value, n2->value) > 0) {
            /* Get the next node while we still can. */
            tmp = n2->next;
            /* Remove n2 from l2. */
            cgul_list__remove_node__common(cex, l2, n2);
            /* Insert n2 before n1 in l1. */
            cgul_list__insert_before_node__common(cex, l1, n1, n2);
            /* Advance n2. */
            n2 = tmp;
        } else {
            /* Advance n1. */
            n1 = n1->next;
        }
    }

    /* Exhaust n2.  This is necessary because the loop above only
     * moves nodes from n2 to n1 if the n1 value is larger.  So, if n2
     * ends with larger values, they will still be in n2. */
    while (n2) {
        /* Get the next node while we still can. */
        tmp = n2->next;
        /* Remove n2 from l2. */
        cgul_list__remove_node__common(cex, l2, n2);
        /* Insert n2 at the end of l1. */
        cgul_list__insert_after_node__common(cex, l1, l1->back, n2);
        /* Advance n2. */
        n2 = tmp;
    }

 out:
    return;
}

void
cgul_list__reverse(cgul_exception_t* cex,
                   cgul_list_t l)
{
    cgul_list_node_t curr = NULL;
    cgul_list_node_t next = NULL;
    cgul_list_node_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    /* Iterate over each node swapping "next" and "prev". */
    for (curr = l->front ; curr ; curr = next) {
        next = curr->next;
        curr->next = curr->prev;
        curr->prev = next;
    }

    /* Swap the "front" and "back" nodes. */
    tmp = l->front;
    l->front = l->back;
    l->back = tmp;

 out:
    return;
}

void
cgul_list__clear(cgul_exception_t* cex,
                 cgul_list_t l,
                 cgul_cache_t values_cache)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    cgul_list__free_node_range(cex, l, l->front, l->back, values_cache);

 out:
    return;
}

void
cgul_list__remove_node(cgul_exception_t* cex,
                       cgul_list_t l,
                       cgul_list_node_t n,
                       void** value_out)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);
    assert(n);

    /*
     * NOTE: If you are just moving a node around in the list try to
     * use cgul_list__remove_node__common() first.
     */

    /* Get the value while n is still valid. */
    if (value_out) {
        *value_out = (void*)n->value;
    }

    /* Remove n from the list. */
    cgul_list__free_node_range(cex, l, n, n, NULL);

 out:
    return;
}

void
cgul_list__remove_range(cgul_exception_t* cex,
                        cgul_list_t l,
                        cgul_list_node_t n_begin,
                        cgul_list_node_t n_end,
                        cgul_cache_t values_cache)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    if (n_begin == NULL) {
        n_begin = l->front;
    }
    cgul_list__free_node_range(cex, l, n_begin, n_end, values_cache);

 out:
    return;
}

int
cgul_list__remove_value(cgul_exception_t* cex,
                        cgul_list_t l,
                        const void* value,
                        cgul_cache_t values_cache)
{
    int rv = 0;
    cgul_list_node_t curr = NULL;
    cgul_list_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    /* Iterate over the list removing all nodes with a value of "value". */
    for (curr = l->front ; curr ; curr = next) {
        /* Get next while we still can. */
        next = curr->next;
        if (value == curr->value) {
            cgul_list__free_node_range(cex, l, curr, curr, values_cache);
            rv = 1;
        }
    }

 out:
    return rv;
}


unsigned long int
cgul_list__get_size(cgul_exception_t* cex,
                    cgul_list_t l)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);

    rv = l->size;

 out:
    return rv;
}

void
cgul_list__swap(cgul_exception_t* cex,
                cgul_list_t l1,
                cgul_list_t l2)
{
    struct cgul_list tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l1);
    assert(l2);

    /* Nothing to swap if they point to the same objects. */
    if (l1 == l2) {
        goto out;
    }

    /* A shallow copy, by design, is exactly what is needed. */
    tmp = *l1;
    *l1 = *l2;
    *l2 = tmp;

 out:
    return;
}

void
cgul_list__foldl(cgul_exception_t* cex,
                 cgul_list_t l,
                 cgul_list__fold_t f,
                 void* data)
{
    int f_rv = 0;
    cgul_list_node_t curr = NULL;
    cgul_list_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);
    assert(f);

    /* Iterate over the nodes in the list. */
    for (curr = l->front ; curr ; curr = next) {
        /* Get the next node now in case f deletes curr. */
        next = curr->next;
        /* Call the combining function. */
        f_rv = (*f)(cex, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_list__foldr(cgul_exception_t* cex,
                 cgul_list_t l,
                 cgul_list__fold_t f,
                 void* data)
{
    int f_rv = 0;
    cgul_list_node_t curr = NULL;
    cgul_list_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(l);
    assert(f);

    /* Iterate over the nodes in the list. */
    for (curr = l->back ; curr ; curr = prev) {
        /* Get the prev node now in case f deletes curr. */
        prev = curr->prev;
        /* Call the combining function. */
        f_rv = (*f)(cex, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_list__traverse(cgul_exception_t* cex,
                    cgul_list_t l,
                    cgul_list__traverse_t f,
                    void* data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_list__traverse_range(cex, l, NULL, NULL, f, data);

 out:
    return;
}

void
cgul_list__traverse_range(cgul_exception_t* cex,
                          cgul_list_t l,
                          cgul_list_node_t first,
                          cgul_list_node_t last,
                          cgul_list__traverse_t f,
                          void* data)
{
    int f_rv = 0;
    cgul_list_node_t curr = NULL;
    cgul_list_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert(first) nor assert(last) because a NULL value
     * indicate that we should begin or end with the first or last
     * node of the entire tree respectively. */
    assert(l);
    assert(f);

    if (first == NULL) {
        first = l->front;
    }

    if (last == NULL) {
        last = l->back;
    }

    /* In order to provide "f" with a context in which it is safe for
     * "f" to delete curr, you have to be careful to dereference
     * curr before calling "f". */
    for (curr = first ; curr ; curr = next) {
        next = curr->next;
        f_rv = (*f)(cex, l, curr, data);
        if (*cex) {
            goto out;
        }
        /* We are done if f returned false or if we just processed the
         * last node. */
        if (!f_rv || (curr == last)) {
            break;
        }
    }

 out:
    return;
}
