#include "cgul_mime_p.h"
#include "cgul_defensive.h"
#include "cgul_libc.h"

/**
 * The block size used by <code>cgul_mime</code>.  Could be overridden
 * by <code>cgul_mime__parse__minimum_block_size()</code> or
 * overridden if the user specifies a block in memory that holds the
 * MIME message to be parsed.
 */
static size_t CGUL_MIME__BLOCK_SIZE = 16384;

/**
 * Default EOL sequence to use if the correct EOL sequence cannot be
 * determined.
 */
static const char* CGUL_MIME__DEFAULT_EOL = "\r\n";

/**
 * Return the first EOL sequence in the block <code>block</code> that
 * extends for <code>block_size</code> bytes.  If no EOL sequence can
 * be found, <code>default_eol</code> is returned.  It is permissible
 * for <code>default_eol</code> to be <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] block block
 * @param[in] block_size size of <code>block</code> in bytes
 * @param[in] default_eol default EOL sequence to return if none can be found
 * @return EOL sequence
 */
static const char*
cgul_mime__probe_eol(cgul_exception_t* cex,
                     const char* block,
                     size_t block_size,
                     const char* default_eol)
{
    char c = '\0';
    size_t i = 0;
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block);

    /* Iterate over each character in "block". */
    for (i = 0 ; i < block_size ; ++i) {

        /* Get the current character. */
        c = block[i];

        /* Test for "\r\n". */
        if (   (c == '\r')
            && (i + 1 < block_size)
            && (block[i + 1] == '\n'))
        {
            rv = "\r\n";
            break;

        /* Test for "\r". */
        } else if (c == '\r') {
            rv = "\r";
            break;

        /* Test for "\n". */
        } else if (c == '\n') {
            rv = "\n";
            break;
        }

    }

    /* If the correct EOL sequence could not be determined, return the
     * default value provided by the user. */
    if (!rv) {
        rv = default_eol;
    }

 out:
    return rv;
}

/**
 * This function refills the internal buffer <code>ctx->buf</code>.
 * There are two modes of operation depending on whether the client
 * registered a general "producer" function that pages in a MIME
 * message block-by-block from an external source or on whether the
 * client specifically passed in a block of memory that already holds
 * the entire MIME message to parse. <p>
 *
 * In the first case, <code>ctx->own_buf</code> is set to true because
 * the parser needs an internal buffer that is used to page in blocks
 * of the MIME message from the external source, and the point of this
 * function is to refill the buffer by shifting the remaining active
 * bytes from the back of the buffer (starting with and including the
 * byte indexed by <code>ctx->i</code>) to the front of the buffer.
 * It then invokes <code>ctx->produce</code> to page in more bytes to
 * fill the remainder of the buffer.  Finally, it updates
 * <code>ctx->i</code> and <code>ctx->acount</code> to reflect the new
 * contents of the buffer. <p>
 *
 * In the second case, <code>ctx->own_buf</code> is set to false
 * because the user supplied the value for <code>ctx->buf</code> which
 * points to the entire MIME message already resident in memory, and
 * the point of this function is to be a no-op so that we can
 * efficiently parse any MIME message regardless of whether the MIME
 * message originates with an external source or internal memory. <p>
 *
 * Regardless of the mode of operation, before calling this function,
 * <code>ctx->i</code> must be set to point one-beyond the last
 * consumed byte in the buffer <code>ctx->buf</code>.  So if you are
 * calling this function after <code>cgul_substring__search()</code>
 * was unable to find a match, <code>ctx->i</code> should be set as
 * follows where <code>ss_length</code> is the length of the substring
 * in bytes: <p>
 *
 * <pre>
 *     ctx->i = ctx->acount - ss_length + 1;
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx <code>cgul_mime__parse__context</code> instance
 * @return count of new bytes read into the buffer
 */
static size_t
cgul_mime__refill(cgul_exception_t* cex,
                  cgul_mime__parse__context_t ctx)
{
    size_t i = 0;
    size_t j = 0;
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);
    assert(ctx->i <= ctx->acount);

    /* If the user supplied the buffer (i.e., we are parsing a file
     * that already resides in memory), don't do anything because
     * there simply are no more bytes.  To check for EOF, just check
     * if (ctx->i >= ctx->acount).  This is true no matter what mode
     * of operation we are in. */
    if (!ctx->own_buf) {
        goto out;
    }

    /* Throw an exception if the parser is asking for more bytes when
     * the buffer is already full. */
    if ((ctx->i == 0) && (ctx->acount >= ctx->buf_size)) {
        *cex = cgul_exception__new("cgul_mime__refill:%d: "
                                   "internal error: "
                                   "attempt to refill a full buffer",
                                   __LINE__);
        goto out;
    }

    /* Make room for loading more bytes from the MIME message by
     * shifting the remaining bytes at the back of the buffer to the
     * front of the buffer. */
    if (ctx->i > 0) {
        /* Shift. */
        for (i = 0, j = ctx->i ; j < ctx->acount ; ++i, ++j) {
            ctx->buf[i] = ctx->buf[j];
        }
        /* Adjust acount and i.  Order is important here. */
        ctx->acount = ctx->acount - ctx->i;
        ctx->i = 0;
    }

    /* Fill the rest of the buffer with bytes loaded by the producer
     * that was registered by the user when cgul_mime__parse() was
     * first called. */
    rv = ctx->produce(cex,
                      &ctx->buf[ctx->acount],
                      ctx->buf_size - ctx->acount,
                      ctx->produce_data);
    if (*cex) {
        goto out;
    }

    /* Update "acount" to reflect the new bytes that were loaded
     * into the buffer. */
    ctx->acount += rv;

 out:
    return rv;
}

/**
 * This function frees all internally allocated resources for the
 * function context <code>ctx</code>.  The caller must not attempt to
 * dereference <code>ctx</code> after calling this function.
 *
 * @param[in] ctx <code>cgul_mime__split__context</code> instance
 */
static void
cgul_mime__split__context__delete(cgul_mime__split__context_t ctx)
{
    if (ctx) {

        /* fout */
        if (ctx->fout) {
            fclose(ctx->fout);
        }

        /* tmp */
        if (ctx->tmp) {
            cgul_string__delete(ctx->tmp);
        }

        /* Free the instance itself. */
        free(ctx);
    }
}

/**
 * Create a new context for use with the following functions: <p>
 *
 * <pre>
 *     cgul_mime__split_memory()
 *     cgul_mime__split_file()
 *     cgul_mime__split_fname()
 * </pre>
 *
 * If an error occurs, <code>NULL</code> is returned, and an exception
 * is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] dname directory name where output files will be created
 * @param[in] pname name for the preamble file
 * @param[in] hname base name for header files
 * @param[in] bname base name for body files
 * @param[in] ename name for the epilogue file
 * @return new <code>cgul_mime__split__context</code> instance
 */
static cgul_mime__split__context_t
cgul_mime__split__context__new(cgul_exception_t* cex,
                               const char* dname,
                               const char* pname,
                               const char* hname,
                               const char* bname,
                               const char* ename)
{
    cgul_mime__split__context_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    rv = (cgul_mime__split__context_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /*
     * Do not assert any of the file names.  If they are NULL, it
     * means the associated files should not be created.
     */

    /* Initialize. */
    rv->dname = dname;
    rv->pname = pname;
    rv->hname = hname;
    rv->bname = bname;
    rv->ename = ename;
    rv->prev_section = CGUL_MIME__SECTION_INVALID;
    rv->fout = NULL;
    rv->tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially initialized instances. */
    if (*cex && rv) {
        cgul_mime__split__context__delete(rv);
        rv = NULL;
    }

    return rv;
}

/**
 * This is a helper function for <code>cgul_mime__split__cb()</code>.
 * It is called as part of the callback registered with
 * <code>cgul_mime__parse()</code>.  It writes the blocks from the
 * MIME preamble of the multipart MIME file to disk.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 * @param[in] block block
 * @param[in] block_size block size
 */
static void
cgul_mime__split__write_preamble(cgul_exception_t* cex,
                                 cgul_mime__split__context_t ctx,
                                 char* block,
                                 size_t block_size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);
    assert(block);

    /* Do not write anything if the user did not specify a name. */
    if (!ctx->pname) {
        goto out;
    }

    /* Open a new file if this is the first block of the preamble. */
    if (ctx->prev_section != CGUL_MIME__SECTION_PREAMBLE) {

        /* Close any previously open file. */
        if (ctx->fout) {
            fclose(ctx->fout);
            ctx->fout = NULL;
        }

        /* Open the new file. */
        ctx->fout = cgul_libc__fopen(cex, ctx->pname, "wb");
        if (*cex) {
            goto out;
        }
    }

    /* Write this block to file. */
    cgul_libc__fwrite_block(cex, block, block_size, ctx->fout);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * This is a helper function for <code>cgul_mime__split__cb()</code>.
 * It is called as part of the callback registered with
 * <code>cgul_mime__parse()</code>.  It writes the blocks from the
 * MIME header for this part of the multipart MIME file to disk.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 * @param[in] part MIME part of the multipart message
 * @param[in] block block
 * @param[in] block_size block size
 */
static void
cgul_mime__split__write_header(cgul_exception_t* cex,
                               cgul_mime__split__context_t ctx,
                               unsigned long int part,
                               char* block,
                               size_t block_size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);
    assert(block);

    /* Do not write anything if the user did not specify a name. */
    if (!ctx->hname) {
        goto out;
    }

    /* Open a new file if this is the first block of the header. */
    if (ctx->prev_section != CGUL_MIME__SECTION_HEADER) {

        /* Close any previously open file. */
        if (ctx->fout) {
            fclose(ctx->fout);
            ctx->fout = NULL;
        }

        /* Generate the name of the new file. */
        cgul_string__set_value(cex, ctx->tmp, "");
        cgul_string__append_sprintf(cex,
                                    ctx->tmp,
                                    "%s/%s-%03lu",
                                    ctx->dname,
                                    ctx->hname,
                                    part);
        if (*cex) {
            goto out;
        }

        /* Open the new file. */
        ctx->fout = cgul_libc__fopen(cex,
                                     cgul_string__get_value(cex, ctx->tmp),
                                     "wb");
        if (*cex) {
            goto out;
        }
    }

    /* Write this block to file. */
    cgul_libc__fwrite_block(cex, block, block_size, ctx->fout);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * This is a helper function for <code>cgul_mime__split__cb()</code>.
 * It is called as part of the callback registered with
 * <code>cgul_mime__parse()</code>.  It writes the blocks from the
 * MIME body for this part of the multipart MIME file to disk.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 * @param[in] part MIME part of the multipart message
 * @param[in] block block
 * @param[in] block_size block size
 */
static void
cgul_mime__split__write_body(cgul_exception_t* cex,
                             cgul_mime__split__context_t ctx,
                             unsigned long int part,
                             char* block,
                             size_t block_size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);
    assert(block);

    /* Do not write anything if the user did not specify a name. */
    if (!ctx->bname) {
        goto out;
    }

    /* Open a new file if this is the first block of the body. */
    if (ctx->prev_section != CGUL_MIME__SECTION_BODY) {

        /* Close any previously open file. */
        if (ctx->fout) {
            fclose(ctx->fout);
            ctx->fout = NULL;
        }

        /* Generate the name of the new file. */
        cgul_string__set_value(cex, ctx->tmp, "");
        cgul_string__append_sprintf(cex,
                                    ctx->tmp,
                                    "%s/%s-%03lu",
                                    ctx->dname,
                                    ctx->bname,
                                    part);
        if (*cex) {
            goto out;
        }

        /* Open the new file. */
        ctx->fout = cgul_libc__fopen(cex,
                                     cgul_string__get_value(cex, ctx->tmp),
                                     "wb");
        if (*cex) {
            goto out;
        }
    }

    /* Write this block to file. */
    cgul_libc__fwrite_block(cex, block, block_size, ctx->fout);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * This is a helper function for <code>cgul_mime__split__cb()</code>.
 * It is called as part of the callback registered with
 * <code>cgul_mime__parse()</code>.  It writes the blocks from the
 * MIME epilogue of the multipart MIME file to disk.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 * @param[in] block block
 * @param[in] block_size block size
 */
static void
cgul_mime__split__write_epilogue(cgul_exception_t* cex,
                                 cgul_mime__split__context_t ctx,
                                 char* block,
                                 size_t block_size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);
    assert(block);

    /* Do not write anything if the user did not specify a name. */
    if (!ctx->ename) {
        goto out;
    }

    /* Open a new file if this is the first block of the epilogue. */
    if (ctx->prev_section != CGUL_MIME__SECTION_EPILOGUE) {

        /* Close any previously open file. */
        if (ctx->fout) {
            fclose(ctx->fout);
            ctx->fout = NULL;
        }

        /* Open the new file. */
        ctx->fout = cgul_libc__fopen(cex, ctx->ename, "wb");
        if (*cex) {
            goto out;
        }
    }

    /* Write this block to file. */
    cgul_libc__fwrite_block(cex, block, block_size, ctx->fout);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Callback used by <code>cgul_mime__split_memory()</code> and
 * <code>cgul_mime__split_file()</code> to split a MIME multipart file
 * into its various parts and write them to file.
 *
 * @param[in,out] cex c-style exception
 * @param[in] part MIME part of the multipart message
 * @param[in] section MIME section where the block originates
 * @param[in] block block
 * @param[in] block_size block size
 * @param[in] data client data
 * @return whether to continue
 */
static int
cgul_mime__split__cb(cgul_exception_t* cex,
                     unsigned long int part,
                     cgul_mime__section_t section,
                     char* block,
                     size_t block_size,
                     void* data)
{
    int rv = 0;
    cgul_mime__split__context_t ctx = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Recover the context. */
    ctx = (cgul_mime__split__context_t)data;

    /* Switch based on the current section. */
    switch (section) {

    /* Preamble */
    case CGUL_MIME__SECTION_PREAMBLE:
        cgul_mime__split__write_preamble(cex, ctx, block, block_size);
        if (*cex) {
            goto out;
        }
        break;

    /* Header */
    case CGUL_MIME__SECTION_HEADER:
        cgul_mime__split__write_header(cex, ctx, part, block, block_size);
        if (*cex) {
            goto out;
        }
        break;

    /* Body */
    case CGUL_MIME__SECTION_BODY:
        cgul_mime__split__write_body(cex, ctx, part, block, block_size);
        if (*cex) {
            goto out;
        }
        break;

    /* Epilogue */
    case CGUL_MIME__SECTION_EPILOGUE:
        cgul_mime__split__write_epilogue(cex, ctx, block, block_size);
        if (*cex) {
            goto out;
        }
        break;

    /* EOF */
    case CGUL_MIME__SECTION_EOF:
        if (ctx->fout) {
            fclose(ctx->fout);
            ctx->fout = NULL;
        }
        break;

    /* Error */
    default:
        *cex = cgul_exception__new("cgul_mime__split__cb:%d: "
                                   "unexpected section: %d",
                                   __LINE__,
                                   (int)section);
        goto out;
    }

    /* Remember the section we just processed. */
    ctx->prev_section = section;

    /* Continue parsing. */
    rv = 1;

 out:
    return rv;
}

/**
 * Return the minimum block size needed when reading the MIME file
 * given the boundary string <code>boundary</code> and the EOL
 * sequence <code>eol</code> which can be <code>NULL</code>. <p>
 *
 * There is a small chicken-or-the-egg problem in that the correct
 * minimum block size depends on the EOL sequence, but if the user
 * passes in <code>NULL</code> for <code>eol</code> we need to already
 * know the block size so we can read a block from the file in order
 * to automatically determine the correct EOL sequence. <p>
 *
 * To break this cycle, we use <code>strlen(eol)</code> if
 * <code>eol</code> is not <code>NULL</code>, and we use 2 if
 * <code>eol</code> is <code>NULL</code> because we know that
 * <code>cgul_mime</code> only automatically detects "\r\n", "\n", or
 * "\r" so it is never larger than 2. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] boundary boundary string
 * @param[in] eol EOL sequence
 * @return minimum block size
 */
static size_t
cgul_mime__parse__minimum_block_size(cgul_exception_t* cex,
                                     const char* boundary,
                                     const char* eol)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "eol".  It is allowed to be NULL. */
    assert(boundary);

    /* Determine the minimum block size to accommodate the boundary
     * string.  The block size just needs to be larger than the the
     * largest substring, i.e., "b_final".  For example, to generate
     * "b_final" when "eol" is "\r\n", we prepend "\r\n--" and append
     * "--" or 6 extra characters. */
    if (eol) {
        /* Use the user's "eol". */
        rv = cgul_defensive__add_size_type(cex, strlen(boundary), strlen(eol));
    } else {
        /* The "eol" automatically detected by cgul_mime is never more
         * than 2 characters. */
        rv = cgul_defensive__add_size_type(cex, strlen(boundary), 2);
    }

    /* Add in 4 for the dashes and 1 because the buffer needs to be
     * larger than the substring. */
    rv = cgul_defensive__add_size_type(cex, rv, 5);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

/**
 * This function frees all internally allocated resources for the
 * function context <code>ctx</code>.  The caller must not attempt to
 * dereference <code>ctx</code> after calling this function.
 *
 * @param[in] ctx <code>cgul_mime__parse__context</code> instance
 */
static void
cgul_mime__parse__context__delete(cgul_mime__parse__context_t ctx)
{
    if (ctx) {

        /* buf */
        if (ctx->own_buf && ctx->buf) {
            free(ctx->buf);
        }

        /* two_eols */
        if (ctx->two_eols) {
            cgul_substring__delete(ctx->two_eols);
        }

        /* b_final */
        if (ctx->b_final) {
            cgul_substring__delete(ctx->b_final);
        }

        /* b */
        if (ctx->b) {
            cgul_substring__delete(ctx->b);
        }

        /* Free the object itself. */
        free(ctx);
    }
}

/**
 * Create a new context for use with the
 * <code>cgul_mime__parse()</code> family of function calls.  This
 * function is responsible for reading in the first block of data by
 * calling <code>produce</code> if <code>produce</code> is not
 * <code>NULL</code>.  If <code>eol</code> is <code>NULL</code>, this
 * function probes the first block of data looking for the correct EOL
 * sequence to use and will set the <code>eol</code> member of the
 * returned context accordingly.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * For parsing a MIME message from an external source, the user should
 * pass in a value for <code>produce</code> that is not
 * <code>NULL</code>.  The <code>produce</code> function will be
 * called each time more bytes are needed from the MIME message.  For
 * parsing a MIME message that is already resident in memory, the user
 * should pass in a pointer to the MIME message in <code>block</code>
 * along with the block size <code>block_size</code>.  If both
 * <code>produce</code> and <code>block</code> are set, an exception
 * is thrown. <p>
 *
 * The remaining parameters are the same as describe in the comments
 * for <code>cgul_mime__parse()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] boundary MIME boundary
 * @param[in] eol EOL sequence
 * @param[in] produce callback function to produce MIME blocks
 * @param[in] produce_data client data for <code>produce</code>
 * @param[in] consume callback function to consume parsed blocks
 * @param[in] consume_data client data for <code>consume</code>
 * @param[in] block block
 * @param[in] block_size block size
 * @return new <code>cgul_mime__parse__context</code> instance
 */
static cgul_mime__parse__context_t
cgul_mime__parse__context__new(cgul_exception_t* cex,
                               const char* boundary,
                               const char* eol,
                               cgul_mime__produce_t produce,
                               void* produce_data,
                               cgul_mime__consume_t consume,
                               void* consume_data,
                               char* block,
                               size_t block_size)
{
    size_t probe_size = 0;
    size_t min_block_size = 0;
    cgul_string_t tmp = NULL;
    cgul_mime__parse__context_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "produce", "eol", or "block".  They are allowed
     * to be NULL. */
    assert(boundary);
    assert(consume);

    /* Make sure that "produce" and "block" are not both set. */
    if (produce && block) {
        *cex = cgul_exception__new(
                   "cgul_mime__parse__context__new:%d: "
                   "internal error: multiple data sources",
                   __LINE__);
        goto out;
    }

    /* Allocate the object itself. */
    rv = (cgul_mime__parse__context_t)
             cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Allocate most members. */
    rv->b = cgul_substring__new(cex);
    rv->b_final = cgul_substring__new(cex);
    rv->two_eols = cgul_substring__new(cex);
    if (*cex) {
        goto out;
    }

    /* Set up the mode of operation depending on whether the user
     * passed in a block that holds the MIME message already resident
     * in memory or passed in a function pointer that will be used to
     * produce blocks from the MIME message from an external source.
     * The rest of the code will use "own_buf" to check which mode of
     * opertion is in effect; fortunately, this is only necessary in a
     * couple of places.  The vast majority of the code runs
     * independent of the mode of operation. */
    if (block) {
        rv->buf_size = block_size;
        rv->buf = block;
        rv->i = 0;
        rv->acount = block_size;
        rv->rcount = block_size;
        rv->own_buf = 0;
    } else {
        assert(produce);
        min_block_size =
            cgul_mime__parse__minimum_block_size(cex, boundary, eol);
        rv->buf_size = CGUL_MAX(CGUL_MIME__BLOCK_SIZE, min_block_size);
        rv->buf = (char*)cgul_libc__malloc(cex, rv->buf_size);
        rv->i = 0;
        rv->acount = 0;
        rv->rcount = 0;
        rv->own_buf = 1;
    }

    /* Propagate exceptions. */
    if (*cex) {
        goto out;
    }

    /* Allocate the temporary string. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Remember the "produce" callback function and its client data. */
    rv->produce = produce;
    rv->produce_data = produce_data;

    /* Remember the "consume" callback function and its client data. */
    rv->consume = consume;
    rv->consume_data = consume_data;

    /* Read the first block from the MIME file into memory.  It is
     * safe to call cgul_mime__refill() even at this early stage in
     * the life of "rv" because cgul_mime__refill() only needs "buf",
     * "buf_size", "i", and "produce" to be initialized. */
    do {
        rv->rcount = cgul_mime__refill(cex, rv);
        if (*cex) {
            goto out;
        }
    } while((rv->rcount > 0) && (rv->acount < rv->buf_size));

    /* Remember the EOL sequence or automatically determine it from
     * the block that was just loaded into memory. */
    if (eol) {
        rv->eol = eol;
    } else {
        if (rv->own_buf) {
            probe_size = CGUL_MIN(rv->buf_size, rv->acount);
        } else {
            probe_size = rv->buf_size;
        }
        rv->eol = cgul_mime__probe_eol(
                      cex, rv->buf, probe_size, CGUL_MIME__DEFAULT_EOL);
    }

    /* Determine the length of the EOL sequence. */
    rv->eol_length = strlen(rv->eol);

    /* Generate a substring to search for two back-to-back EOL sequences. */
    cgul_string__set_value(cex, tmp, rv->eol);
    cgul_string__append(cex, tmp, rv->eol);
    cgul_substring__set_value(cex,
                              rv->two_eols,
                              cgul_string__get_value(cex, tmp),
                              cgul_string__get_length(cex, tmp));
    if (*cex) {
        goto out;
    }

    /* Generate the substring to search for the boundary string.  The
     * boundary string is required to start with "--", but the
     * boundary is never specified in the "Content-Type" header with
     * the implicit "--" prefix.  So, we have to insert it here.
     *
     * The only time an EOL sequence is not required to precede the
     * boundary string is when the boundary string appears on the
     * first line of the file.  The code below will handle this as a
     * special case.  For the general, case we want to be able to
     * search for the leading EOL sequence so prepend it.
     *
     * It is tempting to append a trailing EOL sequence, but the MIME
     * standard allows for optional whitespace between the boundary
     * string and the trailing EOL sequence. */
    cgul_string__set_value(cex, tmp, rv->eol);
    cgul_string__append(cex, tmp, "--");
    cgul_string__append(cex, tmp, boundary);
    cgul_substring__set_value(cex,
                              rv->b,
                              cgul_string__get_value(cex, tmp),
                              cgul_string__get_length(cex, tmp));
    if (*cex) {
        goto out;
    }

    /* Generate the substring to search for the final boundary string.
     * The final boundary string is the same as the boundary string
     * (currently held in "tmp") except it has a trailing "--"
     * sequence. */
    cgul_string__append(cex, tmp, "--");
    cgul_substring__set_value(cex,
                              rv->b_final,
                              cgul_string__get_value(cex, tmp),
                              cgul_string__get_length(cex, tmp));
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially initialized instances. */
    if (*cex && rv) {
        cgul_mime__parse__context__delete(rv);
        rv = NULL;
    }
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

/**
 * This function is a helper for <code>cgul_mime__parse()</code> which
 * is broken into many sub-procedures of which this function is just
 * one.  This function is used as the special-case parser of the
 * preamble which is the part of the file that comes before the first
 * boundary delimiter.  Various values in the function context
 * <code>ctx</code> will be changed so that the other sub-procedures
 * will know where to pick up.  If an error occurs, an exception is
 * thrown. <p>
 *
 * The special case it detects is when the boundary delimiter appears
 * as the first thing on the first line of the file.  In this case and
 * only in this case, is it permissible for the boundary delimiter to
 * appear without a preceding EOL sequence. The general case is
 * handled by <code>cgul_mime__parse_preamble__general_case()</code>. <p>
 *
 * On entry, <code>ctx->i</code> points to the first byte of the first
 * buffer. <p>
 *
 * On return, <code>ctx->i</code> points to just beyond the EOL
 * sequence for the start of the first boundary string that indicates
 * both the end of the preamble and the start of the first header.  It
 * is done this way so that <code>cgul_mime__parse_header()</code> can
 * use the same code to detect both the general case (where the
 * boundary is preceded by an EOL sequence) and the special case
 * (where the boundary appears on the first line without a preceding
 * EOL sequence). <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 * @return whether the special case boundary delimiter was found
 */
static int
cgul_mime__parse_preamble__special_case(
    cgul_exception_t* cex,
    cgul_mime__parse__context_t ctx)
{
    int rv = 0;
    size_t block_size = 0;
    size_t ss_length = 0;
    const char* ss = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);

    /* Get the substring and its length. */
    ss = cgul_substring__get_value(cex, ctx->b);
    ss_length = cgul_substring__get_length(cex, ctx->b);

    /* Adjust ss and ss_length to skip passed the leading EOL sequence. */
    ss += ctx->eol_length;
    ss_length -= ctx->eol_length;

    /* Make sure the upcoming call to strncmp() does not attempt to
     * read off the end of the buffer. */
    while (ctx->i + ss_length > ctx->acount) {

        /* Stitch the next block into the buffer. */
        ctx->rcount = cgul_mime__refill(cex, ctx);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (ctx->rcount == 0) {
            *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                       "Unable to find start of header %lu",
                                       __LINE__,
                                       ctx->part + 1);
            goto out;
        }
    }

    /* Compare the boundary string to the beginning of the first
     * line.  We check above to make sure ctx->buf has at least
     * ss_length bytes. */
    if (strncmp(ctx->buf, ss, ss_length) != 0) {
        goto out;
    }

    /* Success. */
    rv = 1;

    /* Callback letting the client know the preamble is empty. */
    block_size = 0;
    ctx->done = !ctx->consume(cex,
                              0,  /* part */
                              CGUL_MIME__SECTION_PREAMBLE,
                              ctx->buf,
                              block_size,
                              ctx->consume_data);
    if (*cex || ctx->done) {
        goto out;
    }

    /*
     * NOTE: Because the boundary string is the first thing on the
     * first line, it is not preceded with an EOL sequence so ctx->i
     * is already properly positioned for the upcoming call to
     * cgul_mime__parse__find_header().
     */

 out:
    return rv;
}

/**
 * This function is a helper for <code>cgul_mime__parse()</code>
 * which is broken into many sub-procedures of which this function is
 * just one.  This function is used as the general-case parser for the
 * preamble which is the part of the file that comes before the first
 * boundary delimiter.  Various values in the function context
 * <code>ctx</code> will be changed so that the other sub-procedures
 * will know where to pick up.  If an error occurs, an exception is
 * thrown. <p>
 *
 * The general case for finding the boundary delimiter is when it
 * appears with a preceding EOL sequence which is true for all
 * boundary delimiters except when the boundary delimiter appears as
 * the first thing on the first line. <p>
 *
 * On entry, <code>ctx->i</code> points to the first byte of the first
 * buffer. <p>
 *
 * On return, <code>ctx->i</code> points to just beyond the EOL
 * sequence that precedes the first boundary string that indicates
 * both the end of the preamble and the start of the first header.  It
 * is done this way so <code>cgul_mime__parse_header()</code> can use
 * the same code to detect both the general case where the boundary is
 * preceded by an EOL sequence and the special case where the boundary
 * appears on the first line without a preceding EOL sequence. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 */
static void
cgul_mime__parse_preamble__general_case(
    cgul_exception_t* cex,
    cgul_mime__parse__context_t ctx)
{
    size_t block_size = 0;
    size_t ss_length = 0;
    const char* ptr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);

    /* Get the length of the substring. */
    ss_length = cgul_substring__get_length(cex, ctx->b);

    for (;;) {

        /* Refill the buffer if we cannot search for a full substring. */
        while (ctx->i + ss_length > ctx->acount) {

            /* Stitch the next block into the buffer. */
            ctx->rcount = cgul_mime__refill(cex, ctx);
            if (*cex) {
                goto out;
            }

            /* EOF */
            if (ctx->rcount == 0) {
                *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                           "Unable to find start of header %lu",
                                           __LINE__,
                                           ctx->part + 1);
                goto out;
            }
        }

        /* Search for the MIME boundary string. */
        ptr = cgul_substring__search(cex,
                                     ctx->b,
                                     &ctx->buf[ctx->i],
                                     ctx->acount - ctx->i);

        /* Callback with the part of the preamble we just searched. */
        if (ptr) {
            block_size = ptr - &ctx->buf[ctx->i];
        } else {
            /* Remember that cgul_substring__search() stops searching
             * when the remaining bytes in the buffer are fewer than
             * in the substring.  So we do not pass the entire
             * remainder of the block to the callback; instead, we
             * just pass the part that has been searched. */
            block_size = ctx->acount - ctx->i - ss_length + 1;
        }
        ctx->done = !ctx->consume(cex,
                                  0,  /* part */
                                  CGUL_MIME__SECTION_PREAMBLE,
                                  &ctx->buf[ctx->i],
                                  block_size,
                                  ctx->consume_data);
        if (*cex || ctx->done) {
            goto out;
        }

        /* Found the MIME boundary.  Advance "i" passed the leading
         * EOL sequence because it is required for the upcoming call
         * to cgul_mime__parse__find_header() and break out of
         * the loop. */
        if (ptr) {
            ctx->i = ptr - &ctx->buf[ctx->i] + ctx->eol_length;
            break;
        }

        /* Advance "i" to the one-beyond the last byte consumed by the
         * failed substring search. */
        ctx->i = ctx->acount - ss_length + 1;
    }

 out:
    return;
}

/**
 * This function is a helper for <code>cgul_mime__parse()</code> which
 * is broken into many sub-procedures of which this function is just
 * one.  This function is used to find the start of the next header.
 * Various values in the function context <code>ctx</code> will be
 * changed so that the other sub-procedures will know where to pick
 * up.  If an error occurs, an exception is thrown. <p>
 *
 * This function verifies the boundary delimiter that was found by
 * either <code>cgul_mime__parse_preamble</code> or
 * <code>cgul_mime__parse_body</code> immediately before calling this
 * function.  It then skips the optional horizontal white space that
 * the MIME standard allows between the end of the boundary string and
 * the terminating EOL sequence.  It then verifies that there is in
 * fact a terminating EOL sequence. <p>
 *
 * On entry, <code>ctx->i</code> must point to just beyond the leading
 * EOL sequence of the boundary string, i.e., <code>ctx->i</code>
 * should skip the leading EOL sequence and point to the "--" prefix.
 * The reason is that this function needs to detect both the general
 * case where the boundary is preceded by an EOL sequence and the
 * special case where the boundary appears on the first line without a
 * preceding EOL sequence. <p>
 *
 * On return, <code>ctx->i</code> points to the start of the EOL
 * sequence that follows the boundary line.  This is done so
 * <code>cgul_mime__parse__find_header_stop()</code> can search
 * in all cases (including in the empty header case) for the
 * back-to-back EOL sequences that mark the end of the header. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 */
static void
cgul_mime__parse__find_header_start(cgul_exception_t* cex,
                                    cgul_mime__parse__context_t ctx)
{
    size_t ss_length = 0;
    const char* ss = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);

    /* Get the substring and its length. */
    ss = cgul_substring__get_value(cex, ctx->b);
    ss_length = cgul_substring__get_length(cex, ctx->b);

    /* Adjust ss and ss_length to skip passed the leading EOL sequence. */
    ss += ctx->eol_length;
    ss_length -= ctx->eol_length;

    /* Make sure the upcoming call to strncmp() does not attempt to
     * read off the end of the buffer. */
    while (ctx->i + ss_length > ctx->acount) {

        /* Stitch the next block into the buffer. */
        ctx->rcount = cgul_mime__refill(cex, ctx);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (ctx->rcount == 0) {
            *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                       "Unable to find start of header %lu",
                                       __LINE__,
                                       ctx->part + 1);
            goto out;
        }
    }

    /* Make sure "i" actually points to the boundary string.  We
     * checked immediately above to make sure strncmp() does not read
     * off the end of the buffer. */
    if (strncmp(&ctx->buf[ctx->i], ss, ss_length) != 0) {
        *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                   "Unable to find start of header %lu",
                                   __LINE__,
                                   ctx->part + 1);
        goto out;
    }

    /* Move i to the end of the boundary string. */
    ctx->i += ss_length;

    /* Carefully skip optional white space. */
    for (;;) {

        /* Reload if all the bytes in the buffer are exhausted. */
        while (ctx->i >= ctx->acount) {

            /* Stitch the next block into the buffer. */
            ctx->rcount = cgul_mime__refill(cex, ctx);
            if (*cex) {
                goto out;
            }

            /* EOF */
            if (ctx->rcount == 0) {
                *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                           "Unable to find start of header %lu",
                                           __LINE__,
                                           ctx->part + 1);
                goto out;
            }
        }

        /* Break on non-horizontal whitespace. */
        if ((ctx->buf[ctx->i] != ' ') && (ctx->buf[ctx->i] != '\t')) {
            break;
        }

        /* Increment i. */
        ++ctx->i;
    }

    /* Make sure the upcoming call to strncmp() does not attempt to
     * read off the end of the buffer. */
    while (ctx->i + ctx->eol_length > ctx->acount) {

        /* Stitch the next block into the buffer. */
        ctx->rcount = cgul_mime__refill(cex, ctx);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (ctx->rcount == 0) {
            *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                       "Unable to find start of header %lu",
                                       __LINE__,
                                       ctx->part + 1);
            goto out;
        }
    }

    /* Make sure the EOL sequence is at the end of the boundary line.
     * We checked immediately above to make sure strncmp() doesn't
     * read off the end of the buffer. */
    if (strncmp(&ctx->buf[ctx->i], ctx->eol, ctx->eol_length) != 0) {
        *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                   "Unable to find start of header %lu",
                                   __LINE__,
                                   ctx->part + 1);
        goto out;
    }

    /*
     * NOTE: Leave "i" pointing to the start of the EOL sequence that
     * follows the boundary line.  This is done so
     * cgul_mime__parse__find_header_stop() can search in all
     * cases (including in the empty header case) for the back-to-back
     * EOL sequences that mark the end of the header.
     */

 out:
    return;
}

/**
 * This function is a helper for <code>cgul_mime__parse()</code>
 * which is broken into many sub-procedures of which this function is
 * just one.  This function is used to find the end of the current
 * header calling the client's callback with blocks from the header
 * along the way.  Various values in the function context
 * <code>ctx</code> will be changed so that the other sub-procedures
 * will know where to pick up.  If an error occurs, an exception is
 * thrown. <p>
 *
 * On entry, <code>ctx->i</code> must point to the start of the EOL
 * sequence that follows the boundary line.  This is done so this
 * function can search in all cases (including in the empty header
 * case) for the back-to-back EOL sequences that mark the end of the
 * header. <p>
 *
 * On return, <code>ctx->i</code> points at the start of the second of
 * the back-to-back EOL sequences that indicate the end of the header
 * and the start of the body which is required by
 * <code>cgul_mime__parse_body()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 */
static void
cgul_mime__parse__find_header_stop(cgul_exception_t* cex,
                                   cgul_mime__parse__context_t ctx)
{
    int first_pass = 1;
    size_t block_size = 0;
    size_t ss_length = 0;
    const char* ptr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);

    /* Get the length of the substring. */
    ss_length = cgul_substring__get_length(cex, ctx->two_eols);

    /* Find the end of the header.  Note that we start the search at
     * the EOL sequence borrowed from the line that holds the boundary
     * string.  We do this in order to detect empty headers which
     * otherwise would terminate with just a single EOL sequence
     * instead of back-to-back EOL sequences.  The cost of doing this
     * is just that we have to adjust "i" immediately after performing
     * the first search so that it points to the true start of the
     * header. */
    for (;;) {

        /* Refill the buffer if we cannot search for a full substring. */
        while (ctx->i + ss_length > ctx->acount) {

            /* Stitch the next block into the buffer. */
            ctx->rcount = cgul_mime__refill(cex, ctx);
            if (*cex) {
                goto out;
            }

            /* EOF */
            if (ctx->rcount == 0) {
                *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                           "Unable to find end of header.  "
                                           "Unexpected end of header %lu",
                                           __LINE__,
                                           ctx->part + 1);
                goto out;
            }
        }

        /* Search for back-to-back EOL sequences. */
        ptr = cgul_substring__search(cex,
                                     ctx->two_eols,
                                     &ctx->buf[ctx->i],
                                     ctx->acount - ctx->i);

        /* If this is the first pass, now that we've finished the
         * substring search, we need to move "i" passed the EOL
         * sequence that we borrowed so that "i" points to the true
         * start of the header. */
        if (first_pass) {

            /* Disable special first-pass processing. */
            first_pass = 0;

            /* Advance ctx->i pass the EOL sequence we borrowed. */
            ctx->i += ctx->eol_length;

            /* If the search failed and there are not enough characters
             * in the buffer to handle the new value of ctx->i, try again. */
            if (!ptr && (ctx->i + ss_length > ctx->acount)) {
                continue;
            }
        }

        /* Callback with the part of the header we just searched. */
        if (ptr) {
            /* Note that if a substring match is found, "ptr" needs to
             * be adjusted because the search above finds the first
             * EOL sequence in the pair of back-to-back EOL sequences,
             * but the first EOL sequence is the EOL separator for the
             * previous header line so the header actually ends at
             * (ptr + eol_length). */
            ptr += ctx->eol_length;
            block_size = ptr - &ctx->buf[ctx->i];
        } else {
            /* Remember that cgul_substring__search() stops searching
             * when the remaining bytes in the buffer are fewer than
             * in the substring.  So we do not pass the entire
             * remainder of the block to the callback; instead, we
             * just pass the part that has been searched. */
            block_size = ctx->acount - ctx->i - ss_length + 1;
        }
        ctx->done = !ctx->consume(cex,
                                  ctx->part + 1,
                                  CGUL_MIME__SECTION_HEADER,
                                  &ctx->buf[ctx->i],
                                  block_size,
                                  ctx->consume_data);
        if (*cex || ctx->done) {
            goto out;
        }

        /* Found the back-to-back EOL sequences.  Adjust "i" leaving
         * it pointing at the start of the second of the back-to-back
         * EOL sequences. */
        if (ptr) {
            ctx->i = ptr - ctx->buf;
            break;
        }

        /* Advance "i" to one-beyond the last byte consumed by the
         * failed substring search. */
        ctx->i = ctx->acount - ss_length + 1;
    }

 out:
    return;
}

/**
 * This function is a helper for <code>cgul_mime__parse()</code>
 * which is broken into many sub-procedures of which this function is
 * just one.  This function is used to find the end of the current
 * body calling the client's callback with blocks from the body along
 * the way.  Various values in the function context <code>ctx</code>
 * will be changed so that the other sub-procedures will know where to
 * pick up.  If an error occurs, an exception is thrown. <p>
 *
 * On entry, <code>ctx->i</code> must point at the start of the second
 * of the back-to-back EOL sequences that indicate the end of the
 * header and the start of the body making it possible to detect empty
 * bodies. <p>
 *
 * On return, <code>ctx->i</code> points to just beyond the EOL
 * sequence that precedes the next boundary string, i.e.,
 * <code>ctx->i</code> skips the leading EOL sequence and points to
 * the "--" prefix as required by
 * <code>cgul_mime__parse_header()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 */
static void
cgul_mime__parse__find_body_stop(cgul_exception_t* cex,
                                 cgul_mime__parse__context_t ctx)
{
    int first_pass = 1;
    int already_skipped_eol = 0;
    size_t block_size = 0;
    size_t ss_length = 0;
    const char* ptr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);

    /* Get the length of the substring. */
    ss_length = cgul_substring__get_length(cex, ctx->b);

    /* Find the end of the body.  Note that we start the search at the
     * EOL sequence borrowed from the second of the back-to-back EOL
     * sequences that ends the header section.  We do this because "b"
     * has a leading EOL sequence, and without borrowing the EOL
     * sequence, we would not be able to detect the case where the
     * boundary string immediately follows the back-to-back EOL
     * sequences.  The cost of doing this is just that we have to
     * adjust "i" immediately after performing the first search so
     * that it points to the true start of the body. */
    for (;;) {

        /* Refill the buffer if we cannot search for a full substring. */
        while (ctx->i + ss_length > ctx->acount) {

            /* Stitch the next block into the buffer. */
            ctx->rcount = cgul_mime__refill(cex, ctx);
            if (*cex) {
                goto out;
            }

            /* EOF */
            if (ctx->rcount == 0) {
                *cex = cgul_exception__new("cgul_mime__parse:%d "
                                           "Unable to find end of body.  "
                                           "Unexpected end of body %lu",
                                           __LINE__,
                                           ctx->part + 1);
                goto out;
            }
        }

        /* Search for the MIME boundary string. */
        ptr = cgul_substring__search(cex,
                                     ctx->b,
                                     &ctx->buf[ctx->i],
                                     ctx->acount - ctx->i);

        /* If this is the first pass, now that we've finished the
         * substring search, we need to move "i" passed the EOL
         * sequence that we borrowed so that "i" points to the true
         * start of the body.  We also need to adjust ptr if it
         * matched at the front of the string. */
        if (first_pass) {

            /* Disable special first-pass processing. */
            first_pass = 0;

            /* Advance ctx->i pass the EOL sequence we borrowed. */
            ctx->i += ctx->eol_length;

            /* If the search failed and there are not enough characters
             * in the buffer to handle the new value of ctx->i, try again. */
            if (!ptr && (ctx->i + ss_length > ctx->acount)) {
                continue;
            }

            /* If the search succeeded and it matched the EOL sequence
             * we borrowed, advance ptr passed the borrowed EOL sequence. */
            if (ptr && (ptr < &ctx->buf[ctx->i])) {
                ptr += ctx->eol_length;
                already_skipped_eol = 1;  /* We just forced ptr passed EOL. */
            }
        }

        /* Callback with the part of the header we just searched. */
        if (ptr) {
            block_size = ptr - &ctx->buf[ctx->i];
        } else {
            /* Remember that cgul_substring__search() stops searching
             * when the remaining bytes in the buffer are fewer than
             * in the substring.  So we do not pass the entire
             * remainder of the block to the callback; instead, we
             * just pass the part that has been searched. */
            block_size = ctx->acount - ctx->i - ss_length + 1;
        }
        ctx->done = !ctx->consume(cex,
                                  ctx->part + 1,
                                  CGUL_MIME__SECTION_BODY,
                                  &ctx->buf[ctx->i],
                                  block_size,
                                  ctx->consume_data);
        if (*cex || ctx->done) {
            goto out;
        }

        /* Found the MIME boundary.  Adjust "i" leaving it pointing to
         * the MIME boundary except it skips passed the leading EOL
         * sequence as required by the upcoming call to
         * cgul_mime__parse_header(). */
        if (ptr) {
            ctx->i = ptr - ctx->buf;
            if (!already_skipped_eol) {
                ctx->i += ctx->eol_length;
            }
            break;
        }

        /* Advance "i" to the one-beyond the last byte consumed by the
         * failed substring search. */
        ctx->i = ctx->acount - ss_length + 1;
    }

 out:
    return;
}

/**
 * This function is a helper for <code>cgul_mime__parse()</code>
 * which is broken into many sub-procedures of which this function is
 * just one.  This function returns true if the current boundary
 * string is the final boundary string (as denoted by the special "--"
 * suffix that the other boundary strings do not have); otherwise,
 * this function returns false.  Various values in the function
 * context <code>ctx</code> will be changed so that the other
 * sub-procedures will know where to pick up.  If an error occurs, an
 * exception is thrown. <p>
 *
 * On entry, <code>ctx->i</code> must point to just beyond the leading
 * EOL sequence of the boundary string, i.e., <code>ctx->i</code>
 * should skip the leading EOL sequence and point to the "--"
 * prefix. <p>
 *
 * On return, this function does not change <code>ctx->i</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 * return whether the current boundary is the final boundary
 */
static int
cgul_mime__parse__is_final_boundary(cgul_exception_t* cex,
                                    cgul_mime__parse__context_t ctx)
{
    size_t ss_length = 0;
    const char* ss = NULL;
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);

    /* Get the substring and its length. */
    ss = cgul_substring__get_value(cex, ctx->b_final);
    ss_length = cgul_substring__get_length(cex, ctx->b_final);

    /* Adjust ss and ss_length to skip passed the leading EOL sequence. */
    ss += ctx->eol_length;
    ss_length -= ctx->eol_length;

    /* Make sure the upcoming call to strncmp() does not attempt to
     * read off the end of the buffer. */
    while (ctx->i + ss_length > ctx->acount) {

        /* Stitch the next block into the buffer. */
        ctx->rcount = cgul_mime__refill(cex, ctx);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (ctx->rcount == 0) {
            *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                       "Unable to find final MIME separator.  "
                                       "Unexpected end of file.",
                                       __LINE__);
            goto out;
        }
    }

    /* Compare the final boundary string to the string at the current
     * offset in the buffer. We checked immediately above to make sure
     * strncmp() doesn't read off the end of the buffer. */
    if (strncmp(&ctx->buf[ctx->i], ss, ss_length) == 0) {
        rv = 1;
    }

 out:
    return rv;
}

/**
 * This function is a helper for <code>cgul_mime__parse()</code>
 * which is broken into many sub-procedures of which this function is
 * just one.  This function is used to parse the epilogue which is the
 * part of the file that comes after the last boundary delimiter.
 * Various values in the function context <code>ctx</code> will be
 * changed so that the other sub-procedures will know where to pick
 * up.  If an error occurs, an exception is thrown. <p>
 *
 * On entry, <code>ctx->i</code> must point to just beyond the leading
 * EOL sequence of the final boundary string, i.e., <code>ctx->i</code>
 * should skip the leading EOL sequence and point to the "--" prefix. <p>
 *
 * On return, <code>ctx->i</code> is undefined. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 */
static void
cgul_mime__parse_epilogue(cgul_exception_t* cex,
                          cgul_mime__parse__context_t ctx)
{
    int sent_epilogue_block = 0;
    size_t block_size = 0;
    size_t ss_length = 0;
    const char* ss = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ctx);

    /* Get the substring length. */
    ss = cgul_substring__get_value(cex, ctx->b_final);
    ss_length = cgul_substring__get_length(cex, ctx->b_final);

    /* Adjust ss_length to skip passed the leading EOL sequence. */
    ss += ctx->eol_length;
    ss_length -= ctx->eol_length;

    /* Make sure the upcoming call to strncmp() does not attempt to
     * read off the end of the buffer. */
    while (ctx->i + ss_length > ctx->acount) {

        /* Stitch the next block into the buffer. */
        ctx->rcount = cgul_mime__refill(cex, ctx);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (ctx->rcount == 0) {
            *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                       "Unable to verify final boundary.",
                                       __LINE__);
            goto out;
        }
    }

    /* Make sure "i" actually points to the final boundary string. */
    if (strncmp(&ctx->buf[ctx->i], ss, ss_length) != 0) {
        *cex = cgul_exception__new("cgul_mime__parse:%d: "
                                   "Unable to verify final boundary.",
                                   __LINE__);
        goto out;
    }

    /* Move i to the end of the boundary string. */
    ctx->i += ss_length;

    /* Carefully skip optional white space. */
    for (;;) {

        /* Reload if all the bytes in the buffer are exhausted. */
        while (ctx->i >= ctx->acount) {

            /* Read the next block. */
            ctx->rcount = cgul_mime__refill(cex, ctx);
            if (*cex) {
                goto out;
            }

            /* EOF */
            if (ctx->rcount == 0) {
                goto eof;
            }
        }

        /* Break on non-horizontal whitespace. */
        if ((ctx->buf[ctx->i] != ' ') && (ctx->buf[ctx->i] != '\t')) {
            break;
        }

        /* Increment i. */
        ++ctx->i;
    }

    /* Make sure the upcoming call to strncmp() does not attempt to
     * read off the end of the buffer. */
    while (ctx->i + ctx->eol_length > ctx->acount) {

        /* Stitch the next block into the buffer. */
        ctx->rcount = cgul_mime__refill(cex, ctx);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (ctx->rcount == 0) {
            goto eof;
        }
    }

    /* Skip the optional EOL sequence at the end of the boundary line
     * if it exists.  We checked immediately above to make sure
     * strncmp() doesn't read off the end of the buffer. */
    if (strncmp(&ctx->buf[ctx->i], ctx->eol, ctx->eol_length) == 0) {
        ctx->i += ctx->eol_length;
    }

    /* Iterate over the remaining blocks in the MIME file sending each
     * block to the client's callback function. */
    for (;;) {

        /* Reload if all the bytes in the buffer are exhausted. */
        while (ctx->i >= ctx->acount) {

            /* Stitch the next block into the buffer. */
            ctx->rcount = cgul_mime__refill(cex, ctx);
            if (*cex) {
                goto out;
            }

            /* EOF */
            if (ctx->rcount == 0) {
                goto eof;
            }
        }

        /* Callback with this block of the epilogue. */
        block_size = ctx->acount - ctx->i;
        ctx->done = !ctx->consume(cex,
                                  0,  /* part */
                                  CGUL_MIME__SECTION_EPILOGUE,
                                  &ctx->buf[ctx->i],
                                  block_size,
                                  ctx->consume_data);
        if (*cex || ctx->done) {
            goto out;
        }

        /* Advance i. */
        ctx->i += block_size;

        /* Remember that we've sent at least one epligue block. */
        sent_epilogue_block = 1;
    }

 eof:

    /* Always send CGUL_MIME__SECTION_EPILOGUE at least once even if
     * it is just an empty block. */
    if (!sent_epilogue_block) {
        block_size = 0;
        ctx->done = !ctx->consume(cex,
                                  0,  /* part */
                                  CGUL_MIME__SECTION_EPILOGUE,
                                  ctx->buf,
                                  block_size,
                                  ctx->consume_data);
        if (*cex || ctx->done) {
            goto out;
        }
    }

    /* Callback with EOF. */
    block_size = 0;
    ctx->done = !ctx->consume(cex,
                              0,  /* part */
                              CGUL_MIME__SECTION_EOF,
                              ctx->buf,
                              block_size,
                              ctx->consume_data);
    if (*cex || ctx->done) {
        goto out;
    }

 out:
    return;
}

/**
 * This function is a helper for <code>cgul_mime__parse()</code>
 * which is broken into many sub-procedures of which this function is
 * just one.  This function is used to parse the preamble which is the
 * part of the file that comes before the first boundary delimiter.
 * Various values in the function context <code>ctx</code> will be
 * changed so that the other sub-procedures will know where to pick
 * up.  If an error occurs, an exception is thrown. <p>
 *
 * On entry, <code>ctx->i</code> points to the first byte of the first
 * buffer. <p>
 *
 * On return, <code>ctx->i</code> points to just beyond the EOL
 * sequence for the start of the first boundary string that indicates
 * both the end of the preamble and the start of the first header.  It
 * is done this way so that <code>cgul_mime__parse_header()</code> can
 * use the same code to detect both the general case (where the
 * boundary is preceded by an EOL sequence) and the special case
 * (where the boundary appears on the first line without a preceding
 * EOL sequence). <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 */
static void
cgul_mime__parse_preamble(cgul_exception_t* cex,
                          cgul_mime__parse__context_t ctx)
{
    int special = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Handle the special case where the first boundary string is the
     * first thing on the first line.  This is a special case because
     * this is the only time a boundary string is allowed to appear in
     * a MIME file without a leading EOL sequence. */
    special = cgul_mime__parse_preamble__special_case(cex, ctx);
    if (*cex || ctx->done) {
        goto out;
    }
    if (special) {
        goto out;
    }

    /* Handle the general case stitching together blocks from the MIME
     * file while searching for the first boundary string. */
    cgul_mime__parse_preamble__general_case(cex, ctx);
    if (*cex || ctx->done) {
        goto out;
    }

 out:
    return;
}

/**
 * This function is a helper for <code>cgul_mime__parse()</code>
 * which is broken into many sub-procedures of which this function is
 * just one.  This function is used to parse the header for the
 * current MIME part.  Various values in the function context
 * <code>ctx</code> will be changed so that the other sub-procedures
 * will know where to pick up.  If an error occurs, an exception is
 * thrown. <p>
 *
 * This function does not need to search for the boundary string that
 * indicates the start of the header because
 * <code>cgul_mime__parse_preamble()</code> already
 * searches for the first boundary which indicates both the end of the
 * preamble and the start of the first header.  Similarly,
 * <code>cgul_mime__parse_body()</code> finds the next
 * boundary which indicates both the end of the current body and the
 * start of the next header. <p>
 *
 * On entry, <code>ctx->i</code> must point to just beyond the EOL
 * sequence that precedes the boundary string, i.e., <code>ctx->i</code>
 * should skip the leading EOL sequence and point to the "--" prefix.
 * The reason is that this function needs to detect both the general
 * case where the boundary is preceded by an EOL sequence and the
 * special case where the boundary appears on the first line without a
 * preceding EOL sequence. <p>
 *
 * On return, <code>ctx->i</code> points at the start of the second of
 * the back-to-back EOL sequences that indicate the end of the header
 * and the start of the body which is required by
 * <code>cgul_mime__parse_body()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 */
static void
cgul_mime__parse_header(cgul_exception_t* cex,
                        cgul_mime__parse__context_t ctx)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Find the start of the header. */
    cgul_mime__parse__find_header_start(cex, ctx);
    if (*cex || ctx->done) {
        goto out;
    }

    /* Find the end of the header. */
    cgul_mime__parse__find_header_stop(cex, ctx);
    if (*cex || ctx->done) {
        goto out;
    }

 out:
    return;
}

/**
 * This function is a helper for <code>cgul_mime__parse()</code>
 * which is broken into many sub-procedures of which this function is
 * just one.  This function is used to parse the body for the current
 * MIME part.  Various values in the function context <code>ctx</code>
 * will be changed so that the other sub-procedures will know where to
 * pick up.  If an error occurs, an exception is thrown.
 *
 * On entry, <code>ctx->i</code> must point at the start of the second
 * of the back-to-back EOL sequences that indicate the end of the
 * header and the start of the body making it possible to detect empty
 * bodies. <p>
 *
 * On return, <code>ctx->i</code> points to just beyond the EOL
 * sequence that precedes the next boundary string, i.e.,
 * <code>ctx->i</code> skips the leading EOL sequence and points to
 * the "--" prefix as required by
 * <code>cgul_mime__parse_header()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] ctx function context
 */
static void
cgul_mime__parse_body(cgul_exception_t* cex,
                      cgul_mime__parse__context_t ctx)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /*
     * Note that there is no cgul_mime__parse__find_body_start()
     * because cgul_mime__parse_header() (via
     * cgul_mime__parse__find_header_stop()) finds the end of the
     * header which is necessarily the beginning of the body.
     */

    /* Find the end of the body. */
    cgul_mime__parse__find_body_stop(cex, ctx);
    if (*cex || ctx->done) {
        goto out;
    }

 out:
    return;
}

/**
 * This is the common code used for parsing a multipart MIME message
 * held either in a general, external data source or held in memory.
 * For general parsing, the <code>produce</code> callback is not
 * <code>NULL</code> and is invoked whenever the parser needs to
 * refill its internal buffer which requires copying memory into the
 * buffer typically with an I/O routine like <code>fread()</code>. <p>
 *
 * The same scheme that is used for general parsing could be used for
 * parsing MIME messages resident in memory, but it would be
 * inefficient to copy memory into the internal buffer when the parser
 * could just read it directly.  To enable efficient parsing of
 * memory, set <code>block</code> to the beginning of the MIME message
 * that extends for <code>block_size</code> bytes. <p>
 *
 * If both <code>produce</code> and <code>block</code> are set an
 * exception is thrown (via
 * <code>cgul_mime__parse__context__new()</code>). <p>
 *
 * The remaining parameters are the same as describe in the comments
 * for <code>cgul_mime__parse()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] boundary MIME boundary
 * @param[in] eol EOL sequence
 * @param[in] produce callback function to produce MIME blocks
 * @param[in] produce_data client data for <code>produce</code>
 * @param[in] consume callback function to consume parsed blocks
 * @param[in] consume_data client data for <code>consume</code>
 * @param[in] block block
 * @param[in] block_size block size
 * @return total number of parts
 */
static unsigned long int
cgul_mime__parse__common(cgul_exception_t* cex,
                         const char* boundary,
                         const char* eol,
                         cgul_mime__produce_t produce,
                         void* produce_data,
                         cgul_mime__consume_t consume,
                         void* consume_data,
                         char* block,
                         size_t block_size)
{
    unsigned long int rv = 0;
    cgul_mime__parse__context_t ctx = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the function context.  If "produce" is set, this reads
     * the first block into ctx->buf, and if eol == NULL, calls
     * cgul_mime__probe_eol() to try to automatically determine the
     * correct EOL sequence to use.  If "block" is set, this scans the
     * entire block for the correct EOL sequence. */
    ctx = cgul_mime__parse__context__new(cex,
                                         boundary,
                                         eol,
                                         produce,
                                         produce_data,
                                         consume,
                                         consume_data,
                                         block,
                                         block_size);
    if (*cex) {
        goto out;
    }

    /* Parse the preamble. */
    cgul_mime__parse_preamble(cex, ctx);
    if (*cex || ctx->done) {
        goto out;
    }

    /* Iterate over each part. */
    for (;;)  {

        /* Parse the header for this MIME part. */
        cgul_mime__parse_header(cex, ctx);
        if (*cex || ctx->done) {
            goto out;
        }

        /* Parse the body for this MIME part. */
        cgul_mime__parse_body(cex, ctx);
        if (*cex || ctx->done) {
            goto out;
        }

        /* Increment the "part" count. */
        ctx->part = cgul_defensive__add_unsigned_long(cex, ctx->part, 1);
        if (*cex) {
            goto out;
        }

        /* Update the return value. */
        rv = ctx->part;

        /* Break out if this is the final boundary which is denoted by
         * the same boundary as above but with the special trail "--"
         * suffix. */
        if (cgul_mime__parse__is_final_boundary(cex, ctx)) {
            break;
        }
    }

    /* Parse the epilogue. */
    cgul_mime__parse_epilogue(cex, ctx);
    if (*cex || ctx->done) {
        goto out;
    }

 out:

    /* Clean up. */
    if (ctx) {
        cgul_mime__parse__context__delete(ctx);
    }

    return rv;
}

unsigned long int
cgul_mime__parse(cgul_exception_t* cex,
                 const char* boundary,
                 const char* eol,
                 cgul_mime__produce_t produce,
                 void* produce_data,
                 cgul_mime__consume_t consume,
                 void* consume_data)
{
    unsigned long int rv= 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Parse. */
    rv = cgul_mime__parse__common(cex,
                                  boundary,
                                  eol,
                                  produce,
                                  produce_data,
                                  consume,
                                  consume_data,
                                  NULL,    /* block      */
                                  0);      /* block_size */
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

unsigned long int
cgul_mime__parse_memory(cgul_exception_t* cex,
                        char* block,
                        size_t block_size,
                        const char* boundary,
                        const char* eol,
                        cgul_mime__consume_t consume,
                        void* consume_data)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Parse the memory. */
    rv = cgul_mime__parse__common(cex,
                                  boundary,
                                  eol,
                                  NULL,     /* produce */
                                  NULL,     /* produce_data */
                                  consume,
                                  consume_data,
                                  block,
                                  block_size);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

unsigned long int
cgul_mime__parse_file(cgul_exception_t* cex,
                      FILE* fin,
                      const char* boundary,
                      const char* eol,
                      cgul_mime__consume_t consume,
                      void* consume_data)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Parse the file. */
    rv = cgul_mime__parse(cex,
                          boundary,
                          eol,
                          (cgul_mime__produce_t)cgul_libc__fread_block,
                          fin,
                          consume,
                          consume_data);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

unsigned long int
cgul_mime__parse_fname(cgul_exception_t* cex,
                       const char* fname,
                       const char* boundary,
                       const char* eol,
                       cgul_mime__consume_t consume,
                       void* consume_data)
{
    unsigned long int rv = 0;
    FILE* fin = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fname);

    /* Open the file. */
    fin = cgul_libc__fopen(cex, fname, "rb");
    if (*cex) {
        goto out;
    }

    /* Parse the file. */
    rv = cgul_mime__parse_file(cex,
                               fin,
                               boundary,
                               eol,
                               consume,
                               consume_data);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (fin) {
        fclose(fin);
    }

    return rv;
}

unsigned long int
cgul_mime__split_memory(cgul_exception_t* cex,
                        char* block,
                        size_t block_size,
                        const char* dname,
                        const char* pname,
                        const char* hname,
                        const char* bname,
                        const char* ename,
                        const char* boundary,
                        const char* eol)
{
    unsigned long int rv = 0;
    cgul_mime__split__context_t ctx = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    ctx = cgul_mime__split__context__new(
              cex, dname, pname, hname, bname, ename);
    if (*cex) {
        goto out;
    }

    /* Split the memory. */
    cgul_mime__parse_memory(cex,
                            block,
                            block_size,
                            boundary,
                            eol,
                            cgul_mime__split__cb,
                            ctx);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (ctx) {
        cgul_mime__split__context__delete(ctx);
    }

    return rv;
}

unsigned long int
cgul_mime__split_file(cgul_exception_t* cex,
                      FILE* fin,
                      const char* dname,
                      const char* pname,
                      const char* hname,
                      const char* bname,
                      const char* ename,
                      const char* boundary,
                      const char* eol)
{
    unsigned long int rv = 0;
    cgul_mime__split__context_t ctx = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    ctx = cgul_mime__split__context__new(
              cex, dname, pname, hname, bname, ename);
    if (*cex) {
        goto out;
    }

    /* Split the file. */
    cgul_mime__parse_file(cex,
                          fin,
                          boundary,
                          eol,
                          cgul_mime__split__cb,
                          ctx);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (ctx) {
        cgul_mime__split__context__delete(ctx);
    }

    return rv;
}

unsigned long int
cgul_mime__split_fname(cgul_exception_t* cex,
                       const char* fname,
                       const char* dname,
                       const char* pname,
                       const char* hname,
                       const char* bname,
                       const char* ename,
                       const char* boundary,
                       const char* eol)
{
    unsigned long int rv = 0;
    cgul_mime__split__context_t ctx = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    ctx = cgul_mime__split__context__new(
              cex, dname, pname, hname, bname, ename);
    if (*cex) {
        goto out;
    }

    /* Split the file. */
    cgul_mime__parse_fname(cex,
                           fname,
                           boundary,
                           eol,
                           cgul_mime__split__cb,
                           ctx);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (ctx) {
        cgul_mime__split__context__delete(ctx);
    }

    return rv;
}
