#include "cgul_multimap_p.h"
#include "cgul_rbtree_p.h"
#include "cgul_libc.h"
#include "cgul_list.h"
#include "cgul_rbtree_node.h"

/**
 * This method is the callback portion of the
 * <code>cgul_multimap__clear()</code> convenience method that is
 * designed to remove all the nodes of a <code>cgul_multimap</code>
 * _and_ to put back all the keys and values to their respective
 * <code>cgul_cache</code> caches.  If the keys cache is
 * <code>NULL</code>, the key will not be put back or otherwise freed.
 * The same applies for the values cache. <p>
 *
 * This method should not be invoked directly.  Instead, it should be
 * invoked as a callback function for
 * <code>cgul_multimap__traverse()</code>.  When you invoke
 * <code>cgul_multimap__traverse()</coded>, you should do so in such a
 * way that <code>client_data</code> holds a
 * <code>cgul_multimap__clear_cb_data_t</code>.  Typically, your code
 * will work with <code>cgul_multimap__clear()</code> or
 * <code>cgul_multimap__remove_range()</code> instead of with this
 * method.
 *
 * @param[in] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] n multimap node
 * @param[in] client_data client data
 * @return whether to continue traversing the tree
 */
static int
cgul_multimap__clear_cb(cgul_exception_t* cex,
                        cgul_multimap_t mm,
                        cgul_multimap_node_t n,
                        void* client_data)
{
    int rv = 0;
    void* key = NULL;
    void* value = NULL;
    unsigned long int remaining = 0;
    cgul_multimap__clear_cb_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(n);
    assert(client_data);

    data = (cgul_multimap__clear_cb_data_t)client_data;

    /* Remove the node. */
    remaining = cgul_multimap__remove_node(cex, mm, n, &key, &value);
    data->rv = 1;

    /* The value can be put back at any time. */
    if (data->values_cache && value) {
        cgul_cache__put(cex, data->values_cache, value);  
    }

    /* Only after all values are removed is it safe to put back the key. */
    if (data->keys_cache && key && (remaining == 0)) {
        cgul_cache__put(cex, data->keys_cache, key);
    }

    rv = 1;

 out:
    return rv;
}

/**
 * This method holds common code needed by
 * <code>cgul_multimap__insert()</code> and
 * <code>cgul_multimap__insert_with_hint()</code>.  Once those methods
 * have used the "key" to find or create the <code>tree_node</code>
 * that holds the list into which <code>value</code> needs to be
 * inserted, they call this method to do the insertion. <p>
 *
 * If <code>node</code> is not <code>NULL</code> it holds the newly
 * inserted node upon completion.  The return value is the count of
 * how many values in the multimap are now associated with
 * <code>key</code>.  If an error occurs, <code>0</code> is returned,
 * and an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] mm <code>cgul_multimap</code> instance
 * @param[in] tree_node tree node that holds the key and list of values
 * @param[in] value value to insert
 * @param[out] node newly inserted node
 * @return count of values now associated with the key
 */
static unsigned long int
cgul_multimap__insert_value(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            cgul_rbtree_node_t tree_node,
                            void* value,
                            cgul_multimap_node_t* node)
{
    unsigned long rv = 0;
    int is_mm_node_inserted = 0;
    cgul_list_t l = NULL;
    cgul_multimap_node_t mm_node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tree_node);

    /* Assume failure. */
    if (node) {
        *node = NULL;
    }

    /* Get the list into which we need to insert "value." */
    l = (cgul_list_t)cgul_rbtree_node__get_value(cex, tree_node);
    assert(l);

    /* Create a new multimap node for insertion into the linked list. */
    mm_node = cgul_multimap_node__new(cex);
    if (*cex) {
        goto out;
    }

    /* Insert the mm_node into the list before initializing it.  We do
     * it this way because mm_node->list_node cannot be initialized
     * until after insertion.  Also notice that we always put the
     * youngest node at the front of the list.  This means the oldest
     * node is at the back of the list. */
    cgul_list__push_front(cex, l, mm_node);
    if (*cex) {
        goto out;
    }
    is_mm_node_inserted = 1;

    /* Initialize mm_node. */
    mm_node->value = value;
    mm_node->tree_node = tree_node;
    mm_node->list_node = cgul_list__get_front(cex, l);

    /* Increment the size and check for overflow. */
    ++mm->size;
    if (mm->size == 0) {
        /* Restore the original size. */
        --mm->size;
        /* Throw an exception. */
        *cex = cgul_exception__new(
                   "cgul_multimap__insert_value: overflow detected");
        goto out;
    }

    /*
     * Success.
     */

    /* Update "youngest" because we just inserted a new node.  Also
     * remember to set the "younger" pointer of the node that was
     * previously the youngest and the "older" pointer of the node
     * that is now the youngest. */
    if (mm->youngest) {
        mm->youngest->younger = mm_node;
        mm_node->older = mm->youngest;
    }
    mm->youngest = mm_node;

    /* If "oldest" is NULL, it means the multimap was empty and we
     * just inserted the first node which is both the youngest and
     * oldest node. */
    if (mm->oldest == NULL) {
        mm->oldest = mm_node;
    }

    /* Set return values. */
    rv = cgul_list__get_size(cex, l);
    if (node) {
        *node = mm_node;
    }

 out:

    /* Rollback on error. */
    if (*cex) {
        cgul_exception_t local = NULL;
        if (is_mm_node_inserted) {
            /* Removed the node from the list. */
            cgul_list__remove_node(&local, l, mm_node->list_node, NULL);
        }
        if (mm_node) {
            /* Delete the node. */
            cgul_multimap_node__delete(mm_node);
        }
    }

    return rv;
}

/**
 * This method calls <code>cgul_multimap_node__delete()</code> on each
 * node in <code>mm</code>.  It does not bother to maintain invariants
 * during the process because this function should only be called from
 * <code>cgul_multimap__delete()</code>.  Because this method is only
 * called from the destructor, it is intentional that a
 * <code>cgul_exception</code> instance is not passed to this method.
 *
 * @param[in] t <code>cgul_rbtree</code> instance
 */
static void
cgul_multimap__free_all_nodes(cgul_multimap_t mm)
{
    cgul_multimap_node_t curr = NULL;
    cgul_multimap_node_t next = NULL;

    assert(mm);

    /* Iterate over all the nodes in the multimap using the embedded
     * older/younger linked list.  This makes it possible to iterate
     * even if the user has already called free() on all the keys in
     * the multimap. */
    for (curr = mm->youngest ; curr ; curr = next) {
        /* Get the next node while we still can. */
        next = curr->older;
        /* Delete the current node. */
        cgul_multimap_node__delete(curr);
    }

    return;
}

/**
 * This method calls <code>cgul_list__delete()</code> on each list
 * that is held as a value in the underlying red-black tree.  It does
 * not bother to maintain invariants during the process because this
 * function should only be called from
 * <code>cgul_multimap__delete()</code>.  Because this method is only
 * called from the destructor, it is intentional that a
 * <code>cgul_exception</code> instance is not passed to this method.
 *
 * @param[in] t <code>cgul_rbtree</code> instance
 */
static void
cgul_multimap__free_all_lists(cgul_multimap_t mm)
{
    cgul_rbtree_node_t curr = NULL;

    assert(mm);
    assert(mm->tree);

    /* Iterate over all the nodes in the underlying red-black tree
     * using its embedded older/younger linked list.  This makes it
     * possible to iterate even if the user has already called free()
     * on all the keys in the underlying red-black tree. */
    for (curr = mm->tree->youngest ; curr ; curr = curr->older) {
        cgul_list__delete((cgul_list_t)curr->value);
    }

    return;
}

cgul_multimap_t
cgul_multimap__new(cgul_exception_t* cex,
                   cgul_multimap__compare_t compare)
{
    cgul_multimap_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(compare);

    rv = (cgul_multimap_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    rv->tree = cgul_rbtree__new(cex, compare);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_multimap__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_multimap__delete(cgul_multimap_t mm)
{
    if (mm) {

        /* tree */
        if (mm->tree) {
            /* Free all nodes in the multimap. */
            cgul_multimap__free_all_nodes(mm);
            /* Delete all the lists in the underlying red-black tree. */
            cgul_multimap__free_all_lists(mm);
            /* Delete the tree itself. */
            cgul_rbtree__delete(mm->tree);
        }

        /* Free the object itself. */
        free(mm);

    }
}

void
cgul_multimap__free_keys(cgul_multimap_t mm)
{
    cgul_rbtree_node_t curr = NULL;

    if (mm && mm->tree) {
        /* Iterate over each node. */
        for (curr = mm->tree->oldest ; curr ; curr = curr->younger) {
            /* Free the key. */
            free((void*)curr->key);
        }
    }
}

void
cgul_multimap__free_values(cgul_multimap_t mm)
{
    cgul_multimap_node_t curr = NULL;

    if (mm) {
        /* Iterate over each node. */
        for (curr = mm->oldest ; curr ; curr = curr->younger) {
            /* Free the value. */
            free((void*)curr->value);
        }
    }
}

int
cgul_multimap__is_empty(cgul_exception_t* cex,
                        cgul_multimap_t mm)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (mm->size == 0);

 out:
    return rv;
}

unsigned long int
cgul_multimap__insert(cgul_exception_t* cex,
                      cgul_multimap_t mm,
                      void* key,
                      void* value,
                      cgul_multimap_node_t* node)
{
    int insert_rv = 0;
    unsigned long rv = 0;
    cgul_list_t l = NULL;
    cgul_rbtree_node_t tree_node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(mm->tree);

    /* Reset the return node in case an exception is thrown. */
    if (node) {
        *node = NULL;
    }

    /* Try to insert "key" into mm->tree. */
    insert_rv = cgul_rbtree__insert(cex, mm->tree, key, &tree_node);
    if (*cex) {
        goto out;
    }

    /* Test if this is a new or existing key. */
    if (insert_rv) {
        /* This is a new key.  Create a list for it. */
        l = cgul_list__new(cex);
        if (*cex) {
            goto out;
        }
        /* Set the list as the value for the key in the red-black tree. */
        cgul_rbtree_node__set_value(cex, tree_node, l);
    }

    /* Try to insert "value" into the list. */
    rv = cgul_multimap__insert_value(cex, mm, tree_node, value, node);
    if (*cex) {
        goto out;
    }

 out:

    /* Rollback on error. */
    if (*cex && insert_rv) {
        cgul_exception_t local = NULL;
        if (l) {
            /* Delete the list. */
            cgul_list__delete(l);
            l = NULL;
        }
        /* Undo the insertion of key into mm->tree. */
        cgul_rbtree__remove(&local, mm->tree, key, NULL, NULL);
    }

    return rv;
}

extern unsigned long int
cgul_multimap__insert_with_hint(cgul_exception_t* cex,
                                cgul_multimap_t mm,
                                void* key,
                                void* value,
                                cgul_multimap_node_t hint,
                                cgul_multimap_node_t* node)
{
    unsigned long int rv = 0;
    int compare_rv = 0;
    void* key_for_hint = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(mm->tree);
    assert(mm->tree->compare);

    /* Reset the return node in case an exception is thrown. */
    if (node) {
        *node = NULL;
    }

    /* If "hint" is NULL, just pass through to the normal insert. */
    if (!hint) {
        rv = cgul_multimap__insert(cex, mm, key, value, node);
        goto out;
    }

    /* Compare "key" with the key for "hint."  If the keys are not
     * equal, just pass through to the normal insert. */
    key_for_hint = cgul_rbtree_node__get_key(cex, hint->tree_node);
    compare_rv = (*mm->tree->compare)(key, key_for_hint);
    if (compare_rv != 0) {
        rv = cgul_multimap__insert(cex, mm, key, value, node);
        goto out;
    }

    /* If you reach here, it is because "key" is equal to
     * "key_for_hint."  So, we just need to push a new
     * cgul_multimap_node onto the same linked list that holds
     * "hint." */
    rv = cgul_multimap__insert_value(cex, mm, hint->tree_node, value, node);
    if (*cex) {
        goto out;
    }

 out:

    return rv;
}

unsigned long int
cgul_multimap__find(cgul_exception_t* cex,
                    cgul_multimap_t mm,
                    void* key,
                    int older_is_greater,
                    cgul_multimap_node_t* node)
{
    unsigned long int rv = 0;
    cgul_list_t l = NULL;
    cgul_list_node_t list_node = NULL;
    cgul_rbtree_node_t tree_node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(mm->tree);
    assert(node);

    /* Assume failure. */
    if (node) {
        *node = NULL;
    }

    /* Lookup key in the red-black tree. */
    tree_node = cgul_rbtree__find(cex, mm->tree, key);
    if (!tree_node) {
        /* The key is not in the tree. */
        goto out;
    }

    /* Get the linked list that holds the cgul_multimap_node objects. */
    l = (cgul_list_t)cgul_rbtree_node__get_value(cex, tree_node);
    assert(l);

    /* Get the smallest node on the list depending on
     * older_is_greater.  The older nodes are at the back of the list.
     * If older nodes are greater than newer nodes, the newer nodes
     * are the smaller ones, and they are at the front of the list and
     * vice versa. */
    if (older_is_greater) {
        list_node = cgul_list__get_front(cex, l);
    } else {
        list_node = cgul_list__get_back(cex, l);
    }

    assert(list_node);

    /* Success. */
    if (node) {
        *node = (cgul_multimap_node_t)cgul_list_node__get_value(cex, list_node);
    }
    rv = cgul_list__get_size(cex, l);

 out:
    return rv;
}

unsigned long int
cgul_multimap__find_floor(cgul_exception_t* cex,
                          cgul_multimap_t mm,
                          void* search_key,
                          int older_is_greater,
                          cgul_multimap_node_t* node)
{
    unsigned long int rv = 0;
    cgul_list_t l = NULL;
    cgul_list_node_t list_node = NULL;
    cgul_rbtree_node_t tree_node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "node" which is allowed to be NULL. */
    assert(mm);
    assert(mm->tree);

    /* Assume failure. */
    if (node) {
        *node = NULL;
    }

    /* Lookup the floor node in the red-black tree. */
    tree_node = cgul_rbtree__find_floor(cex, mm->tree, search_key);
    if (!tree_node) {
        /* No floor node exists (i.e., the search key is less than all
         * other keys in the). */
        goto out;
    }

    /* Get the linked list that holds the cgul_multimap_node objects. */
    l = (cgul_list_t)cgul_rbtree_node__get_value(cex, tree_node);
    assert(l);

    /* The smallest node is returned so that forward iteration will
     * include all the values for the ceiling node.  The older nodes
     * are at the back of the list. */
    if (older_is_greater) {
        list_node = cgul_list__get_front(cex, l);
    } else {
        list_node = cgul_list__get_back(cex, l);
    }

    assert(list_node);

    /* Success. */
    if (node) {
        *node = (cgul_multimap_node_t)cgul_list_node__get_value(cex, list_node);
    }
    rv = cgul_list__get_size(cex, l);

 out:
    return rv;
}

unsigned long int
cgul_multimap__find_ceiling(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            void* search_key,
                            int older_is_greater,
                            cgul_multimap_node_t* node)
{
    unsigned long int rv = 0;
    cgul_list_t l = NULL;
    cgul_list_node_t list_node = NULL;
    cgul_rbtree_node_t tree_node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "node" which is allowed to be NULL. */
    assert(mm);
    assert(mm->tree);

    /* Assume failure. */
    if (node) {
        *node = NULL;
    }

    /* Lookup the ceiling node in the red-black tree. */
    tree_node = cgul_rbtree__find_ceiling(cex, mm->tree, search_key);
    if (!tree_node) {
        /* No ceiling node exists (i.e., the search key is greater
         * than all other keys in the). */
        goto out;
    }

    /* Get the linked list that holds the cgul_multimap_node objects. */
    l = (cgul_list_t)cgul_rbtree_node__get_value(cex, tree_node);
    assert(l);

    /* The largest node is returned so that reverse iteration will
     * include all the values for the ceiling node.  The older nodes
     * are at the back of the list. */
    if (older_is_greater) {
        list_node = cgul_list__get_back(cex, l);
    } else {
        list_node = cgul_list__get_front(cex, l);
    }

    assert(list_node);

    /* Success. */
    if (node) {
        *node = (cgul_multimap_node_t)cgul_list_node__get_value(cex, list_node);
    }
    rv = cgul_list__get_size(cex, l);

 out:
    return rv;
}

int
cgul_multimap__remove(cgul_exception_t* cex,
                      cgul_multimap_t mm,
                      void* key,
                      cgul_cache_t keys_cache,
                      cgul_cache_t values_cache)
{
    int rv = 0;
    cgul_list_t l = NULL;
    cgul_rbtree_node_t tree_node = NULL;
    cgul_list_node_t first_list_node = NULL;
    cgul_list_node_t last_list_node = NULL;
    cgul_multimap_node_t first_mm_node = NULL;
    cgul_multimap_node_t last_mm_node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(mm->tree);

    /* Lookup "key" in the red-black tree. */
    tree_node = cgul_rbtree__find(cex, mm->tree, key);
    if (!tree_node) {
        /* The key doesn't exist in the tree. */
        goto out;
    }

    /* Get the list of cgul_multimap_nodes. */
    l = (cgul_list_t)cgul_rbtree_node__get_value(cex, tree_node);
    assert(l);

    /* Get the first multimap node. */
    first_list_node =
        (cgul_list_node_t)cgul_list__get_front(cex, l);
    first_mm_node =
        (cgul_multimap_node_t)cgul_list_node__get_value(cex, first_list_node);

    /* Get the last multimap node. */
    last_list_node =
        (cgul_list_node_t)cgul_list__get_back(cex, l);
    last_mm_node =
        (cgul_multimap_node_t)cgul_list_node__get_value(cex, last_list_node);

    /* Remove the nodes. */
    cgul_multimap__remove_range(cex,
                                mm,
                                first_mm_node,
                                last_mm_node,
                                1, /* older_is_greater */
                                keys_cache,
                                values_cache);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

unsigned long int
cgul_multimap__remove_node(cgul_exception_t* cex,
                           cgul_multimap_t mm,
                           cgul_multimap_node_t node,
                           void** key_out,
                           void** value_out)
{
    unsigned long int rv = 0;
    cgul_list_t l = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(node);
    assert(node->tree_node);
    assert(node->list_node);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Get the list for this node. */
    l = (cgul_list_t)cgul_rbtree_node__get_value(cex, node->tree_node);
    assert(l);

    /* Remove the multimap node from the list. */
    cgul_list__remove_node(cex, l, node->list_node, NULL);

    /* Get the number of values that remain for this key. */
    rv = cgul_list__get_size(cex, l);

    /* If this key maps to no more values, remove it and its
     * associated list completely from the red-black tree. */
    if (rv == 0) {
        /* Remove the cgul_rbtree_node. */
        cgul_rbtree__remove_node(cex, mm->tree, node->tree_node, key_out, NULL);
        /* Delete the list. */
        cgul_list__delete(l);
        l = NULL;
    }

    /* Return the value to the user. */
    if (value_out) {
        *value_out = cgul_multimap_node__get_value(cex, node);
    }

    /* While "node" is still valid, update the "oldest" and "youngest"
     * pointers for the entire multimap. */
    if (node == mm->oldest) {
        mm->oldest = node->younger;
    }
    if (node == mm->youngest) {
        mm->youngest = node->older;
    }

    /* While "node" is still valid, update the "older/younger" linked
     * list that runs through all the nodes. */
    if (node->older) {
        node->older->younger = node->younger;
    }
    if (node->younger) {
        node->younger->older = node->older;
    }

    /* Delete the node. */
    cgul_multimap_node__delete(node);

    /* Decrement the size. */
    --mm->size;

 out:
    return rv;
}

int
cgul_multimap__remove_range(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            cgul_multimap_node_t first,
                            cgul_multimap_node_t last,
                            int older_is_greater,
                            cgul_cache_t keys_cache,
                            cgul_cache_t values_cache)
{
    int rv = 0;
    struct cgul_multimap__clear_cb_data data;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    data.keys_cache = keys_cache;
    data.values_cache = values_cache;
    data.rv = 0;

    cgul_multimap__traverse_range(cex,
                                  mm,
                                  first,
                                  last,
                                  older_is_greater,
                                  &cgul_multimap__clear_cb,
                                  &data);

    rv = data.rv;

 out:
    return rv;
}

int
cgul_multimap__remove_front(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            int older_is_greater,
                            void** key_out,
                            void** value_out)
{
    int rv = 0;
    cgul_multimap_node_t front = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Get the front node. */
    front = cgul_multimap__get_front(cex, mm, older_is_greater);
    if (!front) {
        goto out;
    }

    /* Remove the node. */
    cgul_multimap__remove_node(cex, mm, front, key_out, value_out);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
cgul_multimap__remove_back(cgul_exception_t* cex,
                           cgul_multimap_t mm,
                           int older_is_greater,
                           void** key_out,
                           void** value_out)
{
    int rv = 0;
    cgul_multimap_node_t back = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Get the back node. */
    back = cgul_multimap__get_back(cex, mm, older_is_greater);
    if (!back) {
        goto out;
    }

    /* Remove the node. */
    cgul_multimap__remove_node(cex, mm, back, key_out, value_out);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

cgul_multimap_node_t
cgul_multimap__get_front(cgul_exception_t* cex,
                         cgul_multimap_t mm,
                         int older_is_greater)
{
    cgul_list_t l = NULL;
    cgul_list_node_t list_node = NULL;
    cgul_multimap_node_t rv = NULL;
    cgul_rbtree_node_t front_tree_node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    /* Get the front node of the red-black tree. */
    front_tree_node = cgul_rbtree__get_front(cex, mm->tree);
    if (!front_tree_node) {
        /* The tree is empty.  Therefore, the multimap is empty. */
        goto out;
    }

    /* Get the list that holds the values for the front tree node. */
    l = (cgul_list_t)cgul_rbtree_node__get_value(cex, front_tree_node);

    /* Get the correct list node. */
    if (older_is_greater) {
        list_node = cgul_list__get_front(cex, l);
    } else {
        list_node = cgul_list__get_back(cex, l);
    }

    assert(list_node);

    /* Get the front multimap node. */
    rv = (cgul_multimap_node_t)cgul_list_node__get_value(cex, list_node);

 out:
    return rv;
}

cgul_multimap_node_t
cgul_multimap__get_back(cgul_exception_t* cex,
                        cgul_multimap_t mm,
                        int older_is_greater)
{
    cgul_list_t l = NULL;
    cgul_list_node_t list_node = NULL;
    cgul_multimap_node_t rv = NULL;
    cgul_rbtree_node_t back_tree_node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    /* Get the back node of the red-black tree. */
    back_tree_node = cgul_rbtree__get_back(cex, mm->tree);
    if (!back_tree_node) {
        /* The tree is empty.  Therefore, the multimap is empty. */
        goto out;
    }

    /* Get the list that holds the values for the back tree node. */
    l = (cgul_list_t)cgul_rbtree_node__get_value(cex, back_tree_node);

    /* Get the correct list node. */
    if (older_is_greater) {
        list_node = cgul_list__get_back(cex, l);
    } else {
        list_node = cgul_list__get_front(cex, l);
    }

    assert(list_node);

    /* Get the back multimap node. */
    rv = (cgul_multimap_node_t)cgul_list_node__get_value(cex, list_node);

 out:
    return rv;
}

cgul_multimap_node_t
cgul_multimap__get_oldest(cgul_exception_t* cex,
                          cgul_multimap_t mm)
{
    cgul_multimap_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    rv = mm->oldest;

 out:
    return rv;
}

void
cgul_multimap__set_oldest(cgul_exception_t* cex,
                          cgul_multimap_t mm,
                          cgul_multimap_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(mm->oldest);
    assert(n);

    /* There is nothing to do if n is already the oldest. */
    if (n == mm->oldest) {
        goto out;
    }

    /* If n is currently the youngest, it will not be when this method
     * returns.  In this case, we need to remember which node is the
     * new youngest node. */
    if (n == mm->youngest) {
        mm->youngest = n->older;
    }

    /* Carefully remove n from the list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Insert n as the oldest node.  We assert mm->oldest above because
     * any non-empty multimap has an oldest node, and because n is a
     * node in mm, mm is non-empty. */
    mm->oldest->older = n;
    n->younger = mm->oldest;
    n->older = NULL;

    /* Remember that n is now the oldest node. */
    mm->oldest = n;

 out:
    return;
}

cgul_multimap_node_t
cgul_multimap__get_youngest(cgul_exception_t* cex,
                            cgul_multimap_t mm)
{
    cgul_multimap_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    rv = mm->youngest;

 out:
    return rv;
}

void
cgul_multimap__set_youngest(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            cgul_multimap_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(mm->youngest);
    assert(n);

    /* There is nothing to do if n is already the youngest. */
    if (n == mm->youngest) {
        goto out;
    }

    /* If n is currently the oldest, it will not be when this method
     * returns.  In this case, we need to remember which node is the
     * new oldest node. */
    if (n == mm->oldest) {
        mm->oldest = n->younger;
    }

    /* Carefully remove n from the list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Insert n as the youngest node.  We assert mm->youngest above
     * because any non-empty multimap has an youngest node, and
     * because n is a node in mm, mm is non-empty. */
    mm->youngest->younger = n;
    n->older = mm->youngest;
    n->younger = NULL;

    /* Remember that n is now the youngest node. */
    mm->youngest = n;

 out:
    return;
}

void
cgul_multimap__clear(cgul_exception_t* cex,
                     cgul_multimap_t mm,
                     cgul_cache_t keys_cache,
                     cgul_cache_t values_cache)
{
    struct cgul_multimap__clear_cb_data data;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    data.keys_cache = keys_cache;
    data.values_cache = values_cache;
    data.rv = 0;

    cgul_multimap__traverse(cex,
                            mm,
                            1, /* is_greater_older */
                            &cgul_multimap__clear_cb,
                            &data);

 out:
    return;
}

unsigned long int
cgul_multimap__get_size(cgul_exception_t* cex,
                        cgul_multimap_t mm)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    rv = mm->size;

 out:
    return rv;
}

cgul_rbtree_t
cgul_multimap__get_tree(cgul_exception_t* cex,
                        cgul_multimap_t mm)
{
    cgul_rbtree_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);

    rv = mm->tree;

 out:
    return rv;
}

void
cgul_multimap__swap(cgul_exception_t* cex,
                    cgul_multimap_t mm1,
                    cgul_multimap_t mm2)
{
    struct cgul_multimap tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm1);
    assert(mm2);

    /* Nothing to swap if they point to the same objects. */
    if (mm1 == mm2) {
        goto out;
    }

    /* A shallow copy, by design, is exactly what is needed. */
    tmp = *mm1;
    *mm1 = *mm2;
    *mm2 = tmp;

 out:
    return;
}

void
cgul_multimap__foldl_keys(cgul_exception_t* cex,
                          cgul_multimap_t mm,
                          cgul_multimap__fold_key_t f,
                          void* data)
{
    int f_rv = 0;
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(mm->tree);
    assert(f);

    /* Iterate over the nodes in the underlying red-black tree.  Each
     * node in the underlying red-black tree maps one key to many
     * values. */
    for (curr = mm->tree->front ; curr ; curr = next) {
        /* Get the next node now in case f deletes curr. */
        next = cgul_rbtree_node__get_next(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_multimap__foldr_keys(cgul_exception_t* cex,
                          cgul_multimap_t mm,
                          cgul_multimap__fold_key_t f,
                          void* data)
{
    int f_rv = 0;
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(mm->tree);
    assert(f);

    /* Iterate over the nodes in the underlying red-black tree.  Each
     * node in the underlying red-black tree maps one key to many
     * values. */
    for (curr = mm->tree->back ; curr ; curr = prev) {
        /* Get the previous node now in case f deletes curr. */
        prev = cgul_rbtree_node__get_prev(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_multimap__foldl_values(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            int older_is_greater,
                            cgul_multimap__fold_value_t f,
                            void* data)
{
    int f_rv = 0;
    cgul_multimap_node_t curr = NULL;
    cgul_multimap_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(f);

    /* Iterate over the nodes in the multimap. */
    curr = cgul_multimap__get_front(cex, mm, older_is_greater);
    for ( ; curr; curr = next) {
        /* Get the next node now in case f deletes curr. */
        next = cgul_multimap_node__get_next(cex, curr, older_is_greater);
        /* Call the combining function. */
        f_rv = (*f)(cex, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_multimap__foldr_values(cgul_exception_t* cex,
                            cgul_multimap_t mm,
                            int older_is_greater,
                            cgul_multimap__fold_value_t f,
                            void* data)
{
    int f_rv = 0;
    cgul_multimap_node_t curr = NULL;
    cgul_multimap_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(f);

    /* Iterate over the nodes in the multimap. */
    curr = cgul_multimap__get_back(cex, mm, older_is_greater);
    for ( ; curr ; curr = prev) {
        /* Get the previous node now in case f deletes curr. */
        prev = cgul_multimap_node__get_prev(cex, curr, older_is_greater);
        /* Call the combining function. */
        f_rv = (*f)(cex, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_multimap__foldl_pairs(cgul_exception_t* cex,
                           cgul_multimap_t mm,
                           int older_is_greater,
                           cgul_multimap__fold_pair_t f,
                           void* data)
{
    int f_rv = 0;
    cgul_multimap_node_t curr = NULL;
    cgul_multimap_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(f);

    /* Iterate over the nodes in the multimap. */
    curr = cgul_multimap__get_front(cex, mm, older_is_greater);
    for ( ; curr; curr = next) {
        /* Get the next node now in case f deletes curr. */
        next = cgul_multimap_node__get_next(cex, curr, older_is_greater);
        /* Call the combining function. */
        assert(curr->tree_node);
        f_rv = (*f)(cex, curr->tree_node->key, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_multimap__foldr_pairs(cgul_exception_t* cex,
                           cgul_multimap_t mm,
                           int older_is_greater,
                           cgul_multimap__fold_pair_t f,
                           void* data)
{
    int f_rv = 0;
    cgul_multimap_node_t curr = NULL;
    cgul_multimap_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(mm);
    assert(f);

    /* Iterate over the nodes in the multimap. */
    curr = cgul_multimap__get_back(cex, mm, older_is_greater);
    for ( ; curr ; curr = prev) {
        /* Get the previous node now in case f deletes curr. */
        prev = cgul_multimap_node__get_prev(cex, curr, older_is_greater);
        /* Call the combining function. */
        assert(curr->tree_node);
        f_rv = (*f)(cex, curr->tree_node->key, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_multimap__traverse(cgul_exception_t* cex,
                        cgul_multimap_t mm,
                        int older_is_greater,
                        cgul_multimap__traverse_t f,
                        void* data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_multimap__traverse_range(cex,
                                  mm,
                                  NULL,
                                  NULL,
                                  older_is_greater,
                                  f,
                                  data);

 out:
    return;
}

void
cgul_multimap__traverse_range(cgul_exception_t* cex,
                              cgul_multimap_t mm,
                              cgul_multimap_node_t first,
                              cgul_multimap_node_t last,
                              int older_is_greater,
                              cgul_multimap__traverse_t f,
                              void* data)
{
    int f_rv = 0;
    cgul_multimap_node_t curr = NULL;
    cgul_multimap_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert(first) nor assert(last) because NULL values
     * indicate that we should begin or end with the first or last
     * node of the entire tree respectively. */
    assert(mm);
    assert(f);

    if (first == NULL) {
        first = cgul_multimap__get_front(cex, mm, older_is_greater);
    }

    if (last == NULL) {
        last = cgul_multimap__get_back(cex, mm, older_is_greater);
    }

    /* In order to provide "f" with a context in which it is safe for
     * "f" to delete curr, you have to be careful to dereference
     * curr before calling "f". */
    for (curr = first ; curr ; curr = next) {
        next = cgul_multimap_node__get_next(cex, curr, older_is_greater);
        f_rv = (*f)(cex, mm, curr, data);
        if (*cex) {
            goto out;
        }
        /* We are done if f returned false or if we just processed the
         * last node. */
        if (!f_rv || (curr == last)) {
            break;
        }
    }

 out:
    return;
}
