#include "cgul_random_cmwc32_p.h"
#include "cgul_int.h"
#include "cgul_libc.h"
#include "cgul_random_lcg32.h"
#ifndef CGUL_UINT64
#include "cgul_multiply64.h"
#endif

/*
 * This class implements a Complementary-Multiply-With-Carry PRNG with
 * the following parameters:
 *
 *     a = 18782
 *     b = 2^32 - 1
 *     r = 4096
 *
 * A good discussion of the theory behind this PRNG can be found in
 * the first article of the following tome:
 *
 *     http://tbf.coe.wayne.edu/jmasm/vol2_no1.pdf
 */

/**
 * The size of the vector that will hold the PRNs until it is their
 * turn to be multiplied.  If you change this value, you must also
 * change the value for <code>cgul_random_cmwc32__mask</code> to
 * match.  Note that the size of the vector is the same thing as "r"
 * when picking the parameters for the CMWC PRNG.
 */
static size_t cgul_random_cmwc32__v_size = 4096;

/**
 * The mask used to keep the index into the vector in bounds.  This
 * works because the size of the vector is a power of two.  Remember
 * to update this value to match any changes made to
 * <code>cgul_random_cmwc32__v_size</code>.
 */
static size_t cgul_random_cmwc32__mask = 0xfff;

/**
 * The "a" parameters for the CMWC PRNG.  If you change this value
 * verify that <code>cgul_random_cmwc32__initial_carry</code> is still
 * bound by (0 <= initial_carry < a).  If you forget, a sanity check
 * will throw an exception at runtime.
 */
static cgul_uint32_t cgul_random_cmwc32__a = 18782U;

/**
 * The initial carry.  This value is bound by (0 <= initial_carry <
 * a).  If you change "a", remember to verify that this value is
 * within bounds.  If you forget, a sanity check will throw an
 * exception at runtime.
 */
static cgul_uint32_t cgul_random_cmwc32__initial_carry = 7789U;

cgul_random_cmwc32_t
cgul_random_cmwc32__new(cgul_exception_t* cex)
{
    cgul_uint32_t seed = 0;
    cgul_random_cmwc32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a new object seeded with time(). */
    seed = (cgul_uint32_t)time(NULL);
    rv = cgul_random_cmwc32__new_from_seed(cex, seed);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

cgul_random_cmwc32_t
cgul_random_cmwc32__new_from_seed(cgul_exception_t* cex,
                                  cgul_uint32_t seed)
{
    unsigned int i = 0;
    unsigned int seed_count = 0;
    cgul_uint32_t* seeds = NULL;
    cgul_random_lcg32_t lcg32 = NULL;
    cgul_random_cmwc32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the temporary cgul_random_lcg32 instance that will be
     * used to generate the seeds. */
    lcg32 = cgul_random_lcg32__new_from_seed(cex, seed);
    if (*cex) {
        goto out;
    }

    /* Allocate the seed array and get its size. */
    cgul_random_cmwc32__allocate_seed_array(cex, &seeds, &seed_count);
    if (*cex) {
        goto out;
    }

    /* Initialize the seed array. */
    for (i = 0 ; i < seed_count ; ++i) {
        seeds[i] = cgul_random_lcg32__next(cex, lcg32);
    }

    /* Allocate the new object from the seed array. */
    rv = cgul_random_cmwc32__new_from_seed_array(cex, seeds);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (seeds) {
        free(seeds);
    }
    if (lcg32) {
        cgul_random_lcg32__delete(lcg32);
    }

    return rv;
}

cgul_random_cmwc32_t
cgul_random_cmwc32__new_from_seed_array(cgul_exception_t* cex,
                                        cgul_uint32_t* seeds)
{
    size_t i = 0;
    cgul_random_cmwc32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(seeds);

    /* Allocate the object. */
    rv = (cgul_random_cmwc32_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Allocate the vector that will hold the PRNs until they are
     * multiplied. */
    rv->v = (cgul_uint32_t*)cgul_libc__calloc(cex,
                                              cgul_random_cmwc32__v_size,
                                              sizeof(*rv->v));
    if (*cex) {
        goto out;
    }

    /* Copy the caller's seed values into our circular array.  Note
     * that the comments for this method recommend that the caller use
     * allocate_seed_array() to allocate "seeds" so it should be the
     * correct size. */
    for (i = 0 ; i < cgul_random_cmwc32__v_size ; ++i) {
        rv->v[i] = seeds[i];
    }

    /* Initialize the index into v. */
    rv->v_index = 0;

    /* Initialize the carry.  The initial value is constrained by the
     * following:
     *
     *     0 <= carry < a
     *
     * If you set carry = 0, you will likely experience a large number
     * of leading zeros.  Dr. Marsaglia likes c = 123.
     */
    rv->carry = cgul_random_cmwc32__initial_carry;

    /* Sanity check. */
    if (rv->carry >= cgul_random_cmwc32__a) {
        *cex = cgul_exception__new("cgul_random_cmwc32__new_from_seed: "
                                   "initial carry is out of bounds");
        goto out;
    }

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        cgul_random_cmwc32__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_random_cmwc32__delete(cgul_random_cmwc32_t r)
{
    if (r) {

        /* v */
        if (r->v) {
            free(r->v);
        }

        /* Free the object itself. */
        free(r);
    }
}

void
cgul_random_cmwc32__allocate_seed_array(cgul_exception_t* cex,
                                        cgul_uint32_t** v,
                                        unsigned int* v_count)
{
    size_t count = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v_count);

     /* Number of elements in the array. */
    count = cgul_random_cmwc32__v_size;

   /* Allocate the seed array. */
    if (v) {
        *v = (cgul_uint32_t*)cgul_libc__calloc(cex, count, sizeof(**v));
        if (*cex) {
            goto out;
        }
    }

    /* Return the count. */
    if (v_count) {
        *v_count = count;
    }

 out:
    return;
}

cgul_uint32_t
cgul_random_cmwc32__next(cgul_exception_t* cex,
                         cgul_random_cmwc32_t r)
{
    cgul_uint32_t rv = 0;
    cgul_uint32_t quotient = 0;
    cgul_uint32_t remainder = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);

    /*
     * The recurrence implemented by this method is as follows:
     *
     *     X  =  (b - 1) - [(a * X      + carry) mod (2^32 - 1)]
     *      n                     n - r
     *
     * where a, b, and r are defined at the top of this file and
     * "carry" is just the whole-number part from the previous
     * multiply.  Also note that it is mod (2^32 - 1), not mod 2^32 so
     * we have to work a little extra to convert the computers
     * arithmetic into something we can use.
     */


    /*
     * NOTE: A 32-bit CMWC PRNG requires multiplication resulting in a
     * 64-bit product.  The code below uses the CGUL_UINT64 macro to
     * conditionally use native 64-bit multiplication if it is
     * available.  If not, the code uses cgul_multiply64() which is
     * about 2.5 times slower.
     */


    /* Multiply and carry _and_ divide by 2^32.  Note that these
     * values for the quotient and remainder are not quite right for
     * our calculation because we need our math to be mod (2^32 - 1),
     * but these values are mod 2^32.  They will be tweaked as
     * explained below. */
#ifdef CGUL_UINT64
    {
        /* Multiply and add "carry". */
        cgul_uint64_t prod =
              (cgul_uint64_t)cgul_random_cmwc32__a
            * (cgul_uint64_t)r->v[r->v_index]
            + (cgul_uint64_t)r->carry;

        /* Divide by 2^32. */
        quotient = (cgul_uint32_t)(prod >> 32);
        remainder = (cgul_uint32_t)prod;
    }
#else
    {
        /* Multiply and implicitly divide by 2^32. */
        cgul_multiply64(cex,
                        cgul_random_cmwc32__a,
                        r->v[r->v_index],
                        &quotient,
                        &remainder);

        /* Add "carry".  This could cause a mathematical carry.
         * Remember that this is still mod 2^32 arithmetic at this
         * point. */
        remainder += r->carry;
        if (remainder < r->carry) {
            ++quotient;
        }
    }
#endif

    /*
     * Tweak the "quotient" and "remainder" that resulted from
     * division by 2^32 so that they accurately reflect division by
     * (2^32 - 1) instead.
     *
     * This is both common sense and simple algebra, but the algebra
     * is easier to show.  Divide by 2^32 and rely on the Chinese
     * Remainder Theorem to write the result as follows:
     *
     *             x = a * 2^32 + b
     *
     * Now just do simply algebraic manipulations to convert from
     * division by 2^32 to division by (2^32 - 1):
     *
     *             x = a * 2^32 + b
     *
     *         x - b = a * 2^32
     *
     *     x - b - a = a * 2^32 - a
     *
     *     x - b - a = a * (2^32 - 1)
     *
     *             x = a * (2^32 - 1) + a + b
     *
     * Rewriting the equation above in terms of the variable names
     * used in this function gives
     *
     *     x = quotient * (2^32 - 1) + quotient + remainder
     *
     * REMEMBER, that the (quotient + remainder) term could cause a
     * carry that you will also need to deal with.
     */

    /* Adjust the remainder and check for carry (by detecting
     * overflow).  Note that if you get a carry, it affects the
     * quotient _and_ the remainder.  It affects the quotient for the
     * usual reason.  It affects the remainder, because the overflow
     * is relative to base 2^32.  In other words, 0xffffffff and 0x0
     * are the same value mod (2^32 - 1) meaning the 0x0 was
     * accidentally counted twice meaning the remainder is one smaller
     * than it should be.  */
    remainder += quotient;
    if (remainder < quotient) {
        ++quotient;
        ++remainder;
    }

    /* Normalize (0xffffffff mod (2^32 - 1)) values to always be 0x0. */
    if (remainder == 0xffffffff) {
        ++quotient;
        remainder = 0;
    }

    /* Remember the carry for the next iteration. */
    r->carry = quotient;

    /*
     * Return the PRN for this iteration described by the recurrence:
     *
     *     X  =  (b - 1) - [(a * X      + carry) mod (2^32 - 1)]
     *      n                     n - r
     */
    rv = 0xfffffffe - remainder;

    /* Remember the PRN so we can use it on a subsequent iteration. */
    r->v[r->v_index] = rv;

    /* Advance the circular index. */
    r->v_index = (r->v_index + 1) & cgul_random_cmwc32__mask;

 out:
    return rv;
}

cgul_uint32_t
cgul_random_cmwc32__next_in_range(cgul_exception_t* cex,
                                  cgul_random_cmwc32_t r,
                                  cgul_uint32_t range_max)
{
    cgul_uint32_t rv = 0;
    cgul_uint32_t upper_limit = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (range_max == 0) {
        *cex = cgul_exception__new("cgul_random_cmwc32__next_in_range: "
                                   "exclusive range_max is zero");
        goto out;
    }

    /* Use integer arithmetic to find the largest multiple of
     * range_max that is less than the maximum random value that could
     * be produced by the PRNG. */
    upper_limit = 0xffffffff - (0xffffffff % range_max);

    /* Iterate the PRNG until it returns a raw value that is less than
     * the largest multiple of range_max.  In effect, we are just
     * discarding the upper few PRNs so that the mod operation is
     * fair.
     *
     * Note that upper_limit is an _exclusive_ upper bound because if
     * it were included (upper_limit % range) would necessarily be
     * zero (because "upper_limit" is defined to be a multiple of
     * "range") which would cause there to be one more potential zero
     * than any other value that could be returned.  This is why
     * "greater than or equal to" is used rather than just "greater
     * than." */
    do {
        rv = cgul_random_cmwc32__next(cex, r);
    } while (rv >= upper_limit);

    /* Return a PRN in the user's range that is bias-free. */
    rv %= range_max;

 out:
    return rv;
}

double
cgul_random_cmwc32__next_in_unit(cgul_exception_t* cex,
                                 cgul_random_cmwc32_t r)
{
    double rv = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the next PRN and divide by 2^32 by multiplying by the
     * precomputed reciprocal of 2^32. */
    rv = cgul_random_cmwc32__next(cex, r) * CGUL_INVERSE_2_32;

 out:
    return rv;
}
