#include "cgul_random_mt32_p.h"
#include "cgul_libc.h"

/**
 * The size of the vector that will hold the PRNs until it is their
 * turn to be combined.
 */
static size_t cgul_random_mt32__v_size = 624;

cgul_random_mt32_t
cgul_random_mt32__new(cgul_exception_t* cex)
{
    cgul_uint32_t seed = 0;
    cgul_random_mt32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a new object seeded with time(). */
    seed = (cgul_uint32_t)time(NULL);
    rv = cgul_random_mt32__new_from_seed(cex, seed);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

cgul_random_mt32_t
cgul_random_mt32__new_from_seed(cgul_exception_t* cex,
                                cgul_uint32_t seed)
{
    size_t i = 0;
    unsigned int seed_count = 0;
    cgul_uint32_t* seeds = NULL;
    cgul_random_mt32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the seed array and get its size. */
    cgul_random_mt32__allocate_seed_array(cex, &seeds, &seed_count);
    if (*cex) {
        goto out;
    }

    /* Initialize the seed array.  The initialization sequence is
     * prescribed by the MT19937 standard and should guarantee that at
     * least one of the seed values is nonzero. */
    seeds[0] = seed;
    for (i = 1 ; i < seed_count ; ++i) {
        seeds[i] = 0x6c078965 * (seeds[i - 1] ^ (seeds[i - 1] >> 30)) + i;
    }

    /* Allocate the new object from the seed array. */
    rv = cgul_random_mt32__new_from_seed_array(cex, seeds);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (seeds) {
        free(seeds);
    }

    return rv;
}

cgul_random_mt32_t
cgul_random_mt32__new_from_seed_array(cgul_exception_t* cex,
                                      cgul_uint32_t* seeds)
{
    size_t i = 0;
    int has_nonzero_seed = 0;
    cgul_random_mt32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(seeds);

    /* Allocate the object. */
    rv = (cgul_random_mt32_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Allocate the vector that will hold the PRNs until they are
     * randomly selected. */
    rv->v = (cgul_uint32_t*)cgul_libc__calloc(cex,
                                              cgul_random_mt32__v_size,
                                              sizeof(*rv->v));
    if (*cex) {
        goto out;
    }

    /* Copy the caller's seed values into our circular array.  At
     * least one of the seed values must be nonzero.  Note that the
     * comments for this method recommend that the caller use
     * allocate_seed_array() to allocate "seeds" so it should be the
     * correct size. */
    for (i = 0 ; i < cgul_random_mt32__v_size ; ++i) {
        rv->v[i] = seeds[i];
        if (rv->v[i]) {
            has_nonzero_seed = 1;
        }
    }

    /* Throw an exception if all seeds are zero. */
    if (!has_nonzero_seed) {
        *cex = cgul_exception__new(
                   "cgul_random_mt32__new_from_seed_array: "
                   "At least one seed must be nonzero.");
        goto out;
    }

    /* Initialize the index into v to indicate that the buffer is exhausted. */
    rv->v_index = cgul_random_mt32__v_size;

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        cgul_random_mt32__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_random_mt32__delete(cgul_random_mt32_t r)
{
    if (r) {

        /* v */
        if (r->v) {
            free(r->v);
        }

        /* Free the object itself. */
        free(r);
    }
}

void
cgul_random_mt32__allocate_seed_array(cgul_exception_t* cex,
                                      cgul_uint32_t** v,
                                      unsigned int* v_count)
{
    size_t count = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v_count);

    /* Number of elements in the array. */
    count = cgul_random_mt32__v_size;

    /* Allocate the seed array. */
    if (v) {
        *v = (cgul_uint32_t*)cgul_libc__calloc(cex, count, sizeof(**v));
        if (*cex) {
            goto out;
        }
    }

    /* Return the count. */
    if (v_count) {
        *v_count = count;
    }

 out:
    return;
}

cgul_uint32_t
cgul_random_mt32__next(cgul_exception_t* cex,
                       cgul_random_mt32_t r)
{
    cgul_uint32_t rv = 0;
    cgul_uint32_t tmp = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);

    /* Generate the next batch of 624 numbers if the previous batch is
     * has been consumed. */
    if (r->v_index >= cgul_random_mt32__v_size) {
        unsigned int i = 0;
        cgul_uint32_t* v = r->v;  /* alias */

#if 0
        /* Unoptimized version that uses the mod operator directly. */
        for (i = 0 ; i < cgul_random_mt32__v_size ; ++i) {
            /* Calculate the next untempered PRN. */
            tmp = (v[i] & 0x80000000)
                | (v[(i + 1)  % cgul_random_mt32__v_size] & 0x7fffffff);
            v[i] = v[(i + 397) % cgul_random_mt32__v_size] ^ (tmp >> 1);
            if (tmp & 0x1) {
                v[i] = v[i] ^ 0x9908b0df;
            }
        }
#elif 0
        /* Replace mod operator (above) with subtraction. */
        unsigned int i_plus_1 = 0;
        unsigned int i_plus_397 = 0;
        for (i = 0 ; i < cgul_random_mt32__v_size ; ++i) {
            /* Calculate (i + 1 mod 624). */
            i_plus_1 = i + 1;
            if (i_plus_1 >= cgul_random_mt32__v_size) {
                i_plus_1 -= cgul_random_mt32__v_size;
            }
            /* Calculate (i + 397 mod 624). */
            i_plus_397 = i + 397;
            if (i_plus_397 >= cgul_random_mt32__v_size) {
                i_plus_397 -= cgul_random_mt32__v_size;
            }
            /* Calculate the next untempered PRN. */
            tmp = (v[i] & 0x80000000) | (v[i_plus_1] & 0x7fffffff);
            v[i] = v[i_plus_397] ^ (tmp >> 1);
            if (tmp & 0x1) {
                v[i] = v[i] ^ 0x9908b0df;
            }
        }
#else
        /* Replace the if() conditions (above) by splitting the for()
         * loop based on when (i + 397) and (i + 1) start to wrap and
         * by using the "final_xor" array to lookup whether to do the
         * final xor. */
        cgul_uint32_t final_xor[] = {0, 0x9908b0df};
        unsigned int mod_397_delta = 397 - cgul_random_mt32__v_size;
        unsigned int i_max = cgul_random_mt32__v_size - 397;
        for (i = 0 ; i < i_max ; ++i) {
            /* Calculate the next untempered PRN. */
            tmp = (v[i] & 0x80000000) | (v[i + 1] & 0x7fffffff);
            v[i] = v[i + 397] ^ (tmp >> 1) ^ final_xor[tmp & 0x1];
        }

        /* (i + 397) starts to wrap. */
        i_max = cgul_random_mt32__v_size - 1;
        for ( ; i < i_max ; ++i) {
            /* Calculate the next untempered PRN. */
            tmp = (v[i] & 0x80000000) | (v[i + 1] & 0x7fffffff);
            v[i] = v[i + mod_397_delta] ^ (tmp >> 1) ^ final_xor[tmp & 0x1];
        }

        /* (i + 1) starts to wrap.  This is the last iteration so
         * there is no need for a for() loop. */
        tmp = (v[i] & 0x80000000) | (v[0] & 0x7fffffff);
        v[i] = v[i + mod_397_delta] ^ (tmp >> 1) ^ final_xor[tmp & 0x1];
#endif

        /* Reset the index. */
        r->v_index = 0;
    }

    /* Temper the PRN. */
    rv = r->v[r->v_index];
    rv ^= (rv >> 11);
    rv ^= ((rv << 7) & 0x9d2c5680);
    rv ^= ((rv << 15) & 0xefc60000);
    rv ^= (rv >> 18);

    /* Advance the circular index. */
    ++r->v_index;

 out:
    return rv;
}

cgul_uint32_t
cgul_random_mt32__next_in_range(cgul_exception_t* cex,
                                cgul_random_mt32_t r,
                                cgul_uint32_t range_max)
{
    cgul_uint32_t rv = 0;
    cgul_uint32_t upper_limit = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (range_max == 0) {
        *cex = cgul_exception__new("cgul_random_mt32__next_in_range: "
                                   "exclusive range_max is zero");
        goto out;
    }

    /* Use integer arithmetic to find the largest multiple of
     * range_max that is less than the maximum random value that could
     * be produced by the PRNG. */
    upper_limit = 0xffffffff - (0xffffffff % range_max);

    /* Iterate the PRNG until it returns a raw value that is less than
     * the largest multiple of range_max.  In effect, we are just
     * discarding the upper few PRNs so that the mod operation is
     * fair.
     *
     * Note that upper_limit is an _exclusive_ upper bound because if
     * it were included (upper_limit % range) would necessarily be
     * zero (because "upper_limit" is defined to be a multiple of
     * "range") which would cause there to be one more potential zero
     * than any other value that could be returned.  This is why
     * "greater than or equal to" is used rather than just "greater
     * than." */
    do {
        rv = cgul_random_mt32__next(cex, r);
    } while (rv >= upper_limit);

    /* Return a PRN in the user's range that is bias-free. */
    rv %= range_max;

 out:
    return rv;
}

double
cgul_random_mt32__next_in_unit(cgul_exception_t* cex,
                               cgul_random_mt32_t r)
{
    double rv = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the next PRN and divide by 2^32 by multiplying by the
     * precomputed reciprocal of 2^32. */
    rv = cgul_random_mt32__next(cex, r) * CGUL_INVERSE_2_32;

 out:
    return rv;
}
