#include "cgul_shuffle.h"
#include "cgul_libc.h"

/**
 * This macro is the entry point to the common code used by all the
 * shuffle functions.  It performs an in-place Fisher-Yates shuffle
 * on the array <code>com_v</code> holding <code>com_v_count</code>
 * of any integer type, pointer type, or any type that can be
 * shallow-copied <code>com_v_element_type</code>.  The quality of
 * the shuffle depends entirely on the quality of the random numbers
 * returned by the <code>com_next_in_range</code> callback which is
 * passed the PRNG <code>com_prng</code> object unmodified.
 * Because this function assumes a 32-bit PRNG,
 * <code>com_v_count</code> can indicate no more than (2^32 - 1)
 * elements.  If an error occurs, an exception is thrown. <p>
 *
 * This macro uses the "com" prefix (short for "common") to help avoid
 * namespace collisions because C macros are not hygienic. <p>
 *
 * @param[in, out] com_cex c-style exception
 * @param[in] com_v array of elements to shuffle
 * @param[in] com_v_count number of elements in <code>v</code>
 * @param[in] com_v_element_type type of the elements in <code>com_v</code>
 * @param[in] com_next_in_range PRNG callback
 * @param[in] prng PRNG
 */
#define cgul_shuffle__common(com_cex,                                   \
                             com_v,                                     \
                             com_v_count,                               \
                             com_v_element_type,                        \
                             com_next_in_range,                         \
                             com_prng)                                  \
do {                                                                    \
    /* Use the "com_lc" prefix for our local variables. */              \
    size_t com_lc_i = 0;                                                \
    size_t com_lc_back_index = 0;                                       \
    size_t com_lc_picked_index = 0;                                     \
    com_v_element_type com_lc_tmp;                                      \
                                                                        \
    assert(com_cex);                                                    \
    if (*(com_cex)) {                                                   \
        goto com_lc_out;                                                \
    }                                                                   \
                                                                        \
    assert(com_v);                                                      \
    assert(com_next_in_range);                                          \
                                                                        \
    /* Make sure com_v_count is within bounds.  (The                 */ \
    /* (com_v_count > 0) is here just to avoid warnings on 32-bit    */ \
    /* machines from compilers that check that both branches of if() */ \
    /* statements are reachable.) */                                    \
    if (((com_v_count) > 0) && ((com_v_count) > (size_t)0xffffffff)) {  \
        *(com_cex) = cgul_exception__new(                               \
             "cgul_shuffle_ulong: "                                     \
             "array length exceeds 32-bit limit");                      \
        goto com_lc_out;                                                \
    }                                                                   \
                                                                        \
    /* Randomly pick an element from an ever-shrinking array and     */ \
    /* store it at the back ("inactive") part of the array.  Note    */ \
    /* that we do not iterate when com_lc_i = 1 because arrays of    */ \
    /* length 1 are already shuffled.  This is true even when        */ \
    /* starting with larger arrays because the Fisher-Yates          */ \
    /* algorithm shrinks the size of the array by one on each        */ \
    /* iteration until we are left with just an array of length 1    */ \
    /* which is already sorted.  Also note that "com_lc_i" is the    */ \
    /* current length of the array, not the back index.              */ \
    for (com_lc_i = (com_v_count) ; com_lc_i > 1 ; --com_lc_i) {        \
                                                                        \
        /* The back index of the ever-shrinking array.  This is the  */ \
        /* last index that is still "active."                        */ \
        com_lc_back_index = com_lc_i - 1;                               \
                                                                        \
        /* Get the next random number in the range 0 to back_index   */ \
        /* (inclusive).                                              */ \
        com_lc_picked_index = (com_next_in_range)((com_cex),            \
                                                  (com_prng),           \
                                                  com_lc_i);            \
        if (*(com_cex)) {                                               \
            goto com_lc_out;                                            \
        }                                                               \
                                                                        \
        /* The picked index should only index into the "active" part */ \
        /* of the array.                                             */ \
        assert(com_lc_picked_index <= com_lc_back_index);               \
                                                                        \
        /* Move the picked element to the back of the array if it    */ \
        /* isn't already there.  Note that this simultaneously moves */ \
        /* the element at the back of the array that was not picked  */ \
        /* to the "active" part of the array where it can still be   */ \
        /* randomly picked on subsequent iterations. */                 \
        if (com_lc_picked_index < com_lc_back_index) {                  \
            com_lc_tmp = (com_v)[com_lc_back_index];                    \
            (com_v)[com_lc_back_index] = (com_v)[com_lc_picked_index];  \
            (com_v)[com_lc_picked_index] = com_lc_tmp;                  \
        }                                                               \
                                                                        \
    }                                                                   \
                                                                        \
 com_lc_out: ;                                                          \
} while (0)

void
cgul_shuffle(cgul_exception_t* cex,
             void* v,
             size_t v_count,
             size_t element_size,
             cgul_shuffle__next_in_range_t next_in_range,
             void* prng)
{
    void* scratch = NULL;
    void** pointers_array = NULL;
    char* tmp = NULL;
    size_t i = 0;
    size_t v_size = v_count * element_size;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(next_in_range);

    /* Nothing to do if the array is effectively empty or has
     * exactly one element so it is already sorted. */
    if ((v_count <= 1) || (element_size <= 0)) {
        goto out;
    }

    /*
     * The general idea is to shuffle pointers that point into a
     * copy of "v" under the assumption that copying pointers during
     * a shuffle will be more efficient than copying entire
     * structures during the shuffle.  After shuffling the pointers,
     * we will copy individual elements from the copy back into the
     * original "v" based upon the shuffled pointers.
     */

    /* Create an exact copy of "v" and put it in "scratch". */
    scratch = cgul_libc__malloc(cex, v_size);
    if (*cex) {
        goto out;
    }
    memcpy(scratch, v, v_size);

    /* Generate an array of pointers into the scratch area. */
    pointers_array = (void**)cgul_libc__malloc(cex, sizeof(void*) * v_count);
    if (*cex) {
        goto out;
    }
    for (i = 0, tmp = (char*)scratch;
         i < v_count;
         ++i, tmp += element_size)
    {
        pointers_array[i] = tmp;
    }

    /* Shuffle the pointers. */
    cgul_shuffle_pointers(cex,
                          pointers_array,
                          v_count,
                          next_in_range,
                          prng);
    if (*cex) {
        goto out;
    }

    /* Copy the elements back into "v" based upon the shuffled pointers. */
    for (i = 0, tmp = (char*)v ;
         i < v_count ;
         ++i, tmp += element_size)
    {
        memcpy(tmp, pointers_array[i], element_size);
    }

 out:

    /* Clean up. */
    if (pointers_array) {
        free(pointers_array);
    }
    if (scratch) {
        free(scratch);
    }
}

void
cgul_shuffle_pointers(cgul_exception_t* cex,
                      void** v,
                      size_t v_count,
                      cgul_shuffle__next_in_range_t next_in_range,
                      void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         void*,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_schar(cgul_exception_t* cex,
                   signed char* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         signed char,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_uchar(cgul_exception_t* cex,
                   unsigned char* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         unsigned char,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_short(cgul_exception_t* cex,
                   short int* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         short int,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_ushort(cgul_exception_t* cex,
                    unsigned short int* v,
                    size_t v_count,
                    cgul_shuffle__next_in_range_t next_in_range,
                    void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         unsigned short int,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_int(cgul_exception_t* cex,
                 int* v,
                 size_t v_count,
                 cgul_shuffle__next_in_range_t next_in_range,
                 void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         int,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_uint(cgul_exception_t* cex,
                  unsigned int* v,
                  size_t v_count,
                  cgul_shuffle__next_in_range_t next_in_range,
                  void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         unsigned int,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_long(cgul_exception_t* cex,
                  long int* v,
                  size_t v_count,
                  cgul_shuffle__next_in_range_t next_in_range,
                  void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         long int,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_ulong(cgul_exception_t* cex,
                   unsigned long int* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         unsigned long int,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_int8_t(cgul_exception_t* cex,
                    cgul_int8_t* v,
                    size_t v_count,
                    cgul_shuffle__next_in_range_t next_in_range,
                    void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         cgul_int8_t,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_uint8(cgul_exception_t* cex,
                   cgul_uint8_t* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         cgul_uint8_t,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_int16_t(cgul_exception_t* cex,
                     cgul_int16_t* v,
                     size_t v_count,
                     cgul_shuffle__next_in_range_t next_in_range,
                     void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         cgul_int16_t,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_uint16(cgul_exception_t* cex,
                    cgul_uint16_t* v,
                    size_t v_count,
                    cgul_shuffle__next_in_range_t next_in_range,
                    void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         cgul_uint16_t,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_int32_t(cgul_exception_t* cex,
                     cgul_int32_t* v,
                     size_t v_count,
                     cgul_shuffle__next_in_range_t next_in_range,
                     void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         cgul_int32_t,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_shuffle_uint32(cgul_exception_t* cex,
                    cgul_uint32_t* v,
                    size_t v_count,
                    cgul_shuffle__next_in_range_t next_in_range,
                    void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         cgul_uint32_t,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

#ifdef CGUL_INT64
void
cgul_shuffle_int64_t(cgul_exception_t* cex,
                     cgul_int64_t* v,
                     size_t v_count,
                     cgul_shuffle__next_in_range_t next_in_range,
                     void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         cgul_int64_t,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}
#endif /* CGUL_INT64 */

#ifdef CGUL_UINT64
void
cgul_shuffle_uint64(cgul_exception_t* cex,
                    cgul_uint64_t* v,
                    size_t v_count,
                    cgul_shuffle__next_in_range_t next_in_range,
                    void* prng)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_shuffle__common(cex,
                         v,
                         v_count,
                         cgul_uint64_t,
                         next_in_range,
                         prng);
    if (*cex) {
        goto out;
    }

 out:
    return;
}
#endif /* CGUL_UINT64 */
