#include "cgul_stanza_p.h"
#include "cgul_libc.h"
#include "cgul_rbtree.h"
#include "cgul_split.h"
#include "cgul_string.h"
#include "cgul_vector.h"

/**
 * This method verifies that all mandatory keys are present in the
 * stanza.  If any one of the mandatory keys is missing, an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 */
static void
cgul_stanza__verify_mandatory_keys(cgul_exception_t* cex,
                                   cgul_stanza_t stanza)
{
    void* mandatory_key = NULL;
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    if (stanza->mandatory_keys) {
        /* Iterate over the nodes in the mandatory_keys tree. */
        curr = cgul_rbtree__get_front(cex, stanza->mandatory_keys);
        for ( ; curr ; curr = cgul_rbtree_node__get_next(cex, curr)) {
            /* Find the mandatory key for the current node. */
            mandatory_key = cgul_rbtree_node__get_key(cex, curr);
            /* Try to find the mandatory key in the stanza's tree. */
            n = cgul_rbtree__find(cex, stanza->tree, mandatory_key);
            /* If the mandatory key is missing, throw an exception. */
            if (n == NULL) {
                *cex = cgul_exception__new(
                           "%s:%lu: The stanza above is "
                           "missing mandatory key: %s",
                           cgul_crlf_file__get_fname(cex, stanza->f),
                           stanza->line_number_current,
                           (char*)mandatory_key);
                goto out;
            }
        }
    }

 out:
    return;
}

/**
 * This method handles the case where the key is the special "+" key.
 * The special "+" key signals that <code>value</code> should be
 * appended to the previous key's value as multi-line text.  If there
 * is no previous key, an exception is thrown.  If there is not enough
 * memory to hold the new value, an exception is thrown and
 * <code>stanza->tree</code> is left unchanged.
 *
 * @param[in] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] value value
 */
static void
cgul_stanza__append_to_previous_value(cgul_exception_t* cex,
                                      cgul_stanza_t stanza,
                                      char* value)
{
    size_t new_value_length = 0;
    size_t prev_value_length = 0;
    char* new_value = NULL;
    char* prev_value = NULL;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);
    assert(value);

    /* If there is no previous key, throw an exception. */
    if (stanza->prev_key == NULL) {
        *cex = cgul_exception__new(
                   "%s:%lu: no previous key",
                   cgul_crlf_file__get_fname(cex, stanza->f),
                   stanza->line_number_current);
        goto out;
    }

    /* Get the node that holds the key/value pair to which "value" is
     * to be appended. */
    n = cgul_rbtree__find(cex, stanza->tree, stanza->prev_key);
    if (n == NULL) {
        *cex = cgul_exception__new(
                   "%s:%lu: Internal Error: unable to find previous key",
                   cgul_crlf_file__get_fname(cex, stanza->f),
                   stanza->line_number_current);
        goto out;
    }

    /* Get the previous value. */
    prev_value = (char*)cgul_rbtree_node__get_value(cex, n);
    if (prev_value == NULL) {
        *cex = cgul_exception__new(
                   "%s:%lu: Internal Error: unable to find previous value",
                   cgul_crlf_file__get_fname(cex, stanza->f),
                   stanza->line_number_current);
        goto out;
    }

    /*
     * Determine the length of the new string which will be the
     * following:
     *
     *     new_value = prev_value + '\n' + value + '\0'
     */
    prev_value_length = strlen(prev_value);
    new_value_length =  prev_value_length + strlen(value) + 2;

    /* Increase the capacity of "prev_value" so it can hold the new string. */
    new_value = (char*)realloc(prev_value, new_value_length);
    if (new_value == NULL) {
        *cex = cgul_exception__get_out_of_memory_exception();
        goto out;
    }

    /* Append "\n" and value to the previous value.  We use strcpy()
     * instead of strcat() because it is more efficient. */
    new_value[prev_value_length] = '\n';
    strcpy(&new_value[prev_value_length + 1], value);

    /* Replace the previous value with the new value. */
    cgul_rbtree_node__set_value(cex, n, new_value);

 out:
    return;
}

/**
 * This method inserts <code>key</code> and <code>value</code> into
 * <code>stanza->tree</code>.  <p>
 *
 * If <code>key</code> is already present in the tree, it throws an
 * exception if duplicate keys are not allowed.  If duplicate keys are
 * allowed, it replaces the old value with the new value. <p>
 *
 * You have to pass in <code>key</code> as a pointer to a pointer
 * because it will be freed if a duplicate key is found because we
 * leave the key that is already in the tree in place and free the new
 * key that is a duplicate. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in,out] key key
 * @param[in] value value
 */
static void
cgul_stanza__insert_new_key_value_pair(cgul_exception_t* cex,
                                       cgul_stanza_t stanza,
                                       char** key,
                                       char* value)
{
    int insert_rv = 0;
    void* tmp = NULL;
    cgul_rbtree_node_t n = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);
    assert(key);
    assert(*key);
    assert(value);

    /* If the user has passed us a list of allowed keys, make sure
     * "key" is allowed. */
    if (stanza->allowed_keys) {
        n = cgul_rbtree__find(cex, stanza->allowed_keys, *key);
        if (n == NULL) {
            *cex = cgul_exception__new(
                       "%s:%lu: key not allowed: %s",
                       cgul_crlf_file__get_fname(cex, stanza->f),
                       stanza->line_number_current,
                       *key);
            goto out;
        }
    }

    /* Try to insert the key into the tree. */
    insert_rv = cgul_rbtree__insert(cex, stanza->tree, *key, &n);
    if (*cex) {
        goto out;
    }

    /* If the insert failed, it is because this is a duplicate key. */
    if (insert_rv == 0) {

        /* Throw an exception if duplicate keys are not allowed. */
        if (!stanza->allow_duplicate_keys) {
            *cex = cgul_exception__new(
                       "%s:%lu: duplicate key: %s",
                       cgul_crlf_file__get_fname(cex, stanza->f),
                       stanza->line_number_current,
                       *key);
            goto out;
        }

        /* Free this key (because the insert of a new key failed).
         * This is why "key" must be passed in as a pointer to a
         * pointer; otherwise, the calling function would lose track
         * of whether "key" should be freed when an exception is
         * thrown. */
        free(*key);
        *key = NULL;

        /* Free the value from the previous duplicate key.  The new
         * value will be added below. */
        tmp = cgul_rbtree_node__get_value(cex, n);
        if (tmp) {
            free(tmp);
            tmp = NULL;
        }

    } else {

        /* Remember that a new key was inserted. */
        cgul_vector__push_back(cex, stanza->keys, *key);
        if (*cex) {
            goto out;
        }

    }

    /* Set the value. */
    cgul_rbtree_node__set_value(cex, n, value);

 out:

    /* Rollback the insert if an exception is thrown afterward. */
    if (*cex && insert_rv) {
        cgul_rbtree__remove(&local, stanza->tree, *key, NULL, NULL);
    }

    return;
}

/**
 * This method splits <code>line</code> into a key/value pair and
 * inserts the pair into <code>stanza->tree</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] line line
 * @param[in] line_length line length
 */
static void
cgul_stanza__process_line(cgul_exception_t* cex,
                          cgul_stanza_t stanza,
                          const char* line,
                          size_t line_length)
{
    int is_multi_line_append = 0;
    char* key = NULL;
    char* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);
    assert(line);

    /* Split the line putting a copy of the key in "key" and a copy of
     * the value in "value". */
    cgul_stanza__split_line(cex,
                            line,
                            line_length,
                            cgul_crlf_file__get_fname(cex, stanza->f),
                            stanza->line_number_current,
                            &key,
                            &value);
    if (*cex) {
        goto out;
    }

    /* If "key" is a "+", it means "value" should be appended to the
     * previous key's value in order to generate multi-line text. */
    if (strcmp(key, "+") == 0) {

        /* Remember that this is a multi-line append. */
        is_multi_line_append = 1;

        /* Append "value" to the previous key. */
        cgul_stanza__append_to_previous_value(cex, stanza, value);
        if (*cex) {
            goto out;
        }

    } else {

        /* Because "key" is not the special "+" key, it is treated as
         * a new key, and we need to insert this key/value pair into
         * the tree. */
        cgul_stanza__insert_new_key_value_pair(cex, stanza, &key, value);
        if (*cex) {
            goto out;
        }

        /* Keep track of key we just inserted in case the next key is
         * the special "+" key.  Note that you should never free
         * prev_key as it is just a pointer to the real key which was
         * just inserted into "tree". */
        stanza->prev_key = key;

    }

 out:

    /* We always need to free the "+" key because it is never saved in
     * the "tree".  We also always need to free its value because it
     * is appended to the previous value and never saved in the tree
     * directly. */
    if (is_multi_line_append) {
        free(key);
        key = NULL;
        free(value);
        value = NULL;
    }

    /* Clean up if an exception is thrown. */
    if (*cex) {
        if (value) {
            free(value);
        }
        if (key) {
            free(key);
        }
    }

    return;
}

cgul_stanza_t
cgul_stanza__new(cgul_exception_t* cex)
{
    cgul_stanza_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the cgul_stanza object. */
    rv = (cgul_stanza_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Create the tree that holds the key/value pair. */
    rv->tree = cgul_rbtree__new(cex, (cgul_rbtree__compare_t)strcmp);
    if (*cex) {
        goto out;
    }

    /* Create the vector that is used to remember the order that the
     * keys appeared in the file.  Note that "keys" does not own any
     * of the keys.  Instead, the keys are owned by the main
     * structure, "tree". */
    rv->keys = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

 out:

    /* Free partially constructed objects. */
    if (*cex && rv) {
        cgul_stanza__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_stanza_t
cgul_stanza__new_from_fname(cgul_exception_t* cex,
                            const char* fname,
                            int is_block_buffered)
{
    cgul_stanza_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fname);

    /* Construct a default object. */
    rv = cgul_stanza__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the file. */
    cgul_stanza__open_fname(cex, rv, fname, is_block_buffered);
    if (*cex) {
        goto out;
    }

 out:

    /* Free partially constructed objects. */
    if (*cex && rv) {
        cgul_stanza__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_stanza_t
cgul_stanza__new_from_file(cgul_exception_t* cex,
                           FILE* f,
                           int is_block_buffered)
{
    cgul_stanza_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    /* Construct a default object. */
    rv = cgul_stanza__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the file. */
    cgul_stanza__open_file(cex, rv, f, is_block_buffered);
    if (*cex) {
        goto out;
    }

 out:

    /* Free partially constructed objects. */
    if (*cex && rv) {
        cgul_stanza__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_stanza_t
cgul_stanza__new_from_memory(cgul_exception_t* cex,
                             const char* buf,
                             size_t buf_size)
{
    cgul_stanza_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(buf);

    /* Construct a default object. */
    rv = cgul_stanza__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the memory buffer. */
    cgul_stanza__open_memory(cex, rv, buf, buf_size);
    if (*cex) {
        goto out;
    }

 out:

    /* Free partially constructed objects. */
    if (*cex && rv) {
        cgul_stanza__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_stanza_t
cgul_stanza__new_from_crlf_file(cgul_exception_t* cex,
                                cgul_crlf_file_t crlf_file)
{
    cgul_stanza_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Construct a default object. */
    rv = cgul_stanza__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the file. */
    cgul_stanza__open_crlf_file(cex, rv, crlf_file);
    if (*cex) {
        goto out;
    }

 out:

    /* Free partially constructed objects. */
    if (*cex && rv) {
        cgul_stanza__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_stanza__delete(cgul_stanza_t stanza)
{
    cgul_exception_t local = NULL;

    if (stanza) {

        /* Close the file if one is open. */
        cgul_stanza__close(&local, stanza);

        /* Remove the "keys" vector which does not own any of the
         * keys.  Instead, they are owned by the main structure,
         * "tree". */
        if (stanza->keys) {
            cgul_vector__delete(stanza->keys);
            stanza->keys = NULL;
        }

        /* Free all key/value pairs in the tree and the tree itself. */
        if (stanza->tree) {
            /* Free all key/value pairs. */
            cgul_stanza__clear_tree(&local, stanza->tree);
            /* Free the tree itself. */
            cgul_rbtree__delete(stanza->tree);
            stanza->tree = NULL;
        }

        /* Free the object itself. */
        free(stanza);
    }
}

void
cgul_stanza__open_fname(cgul_exception_t* cex,
                        cgul_stanza_t stanza,
                        const char* fname,
                        int is_block_buffered)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);
    assert(fname);

    /* Close the currently open stanza source (if any). */
    cgul_stanza__close(cex, stanza);

    /* Allocate the new cgul_crlf_file object. */
    stanza->f = cgul_crlf_file__new_from_fname(cex, fname, is_block_buffered);
    if (*cex) {
        goto out;
    }

    /* Remember that we own f. */
    stanza->we_own_f = 1;

 out:

    return;
}

void
cgul_stanza__open_file(cgul_exception_t* cex,
                       cgul_stanza_t stanza,
                       FILE* f,
                       int is_block_buffered)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);
    assert(f);

    /* Close the currently open stanza source (if any). */
    cgul_stanza__close(cex, stanza);

    /* Allocate the new cgul_crlf_file object. */
    stanza->f = cgul_crlf_file__new_from_file(cex, f, is_block_buffered);
    if (*cex) {
        goto out;
    }

    /* Remember that we own f. */
    stanza->we_own_f = 1;

 out:
    return;
}

void
cgul_stanza__open_memory(cgul_exception_t* cex,
                         cgul_stanza_t stanza,
                         const char* buf,
                         size_t buf_size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);
    assert(buf);

    /* Close the currently open stanza source (if any). */
    cgul_stanza__close(cex, stanza);

    /* Allocate the new cgul_crlf_file object. */
    stanza->f = cgul_crlf_file__new_from_memory(cex, buf, buf_size);
    if (*cex) {
        goto out;
    }

    /* Remember that we own f. */
    stanza->we_own_f = 1;

 out:
    return;
}

void
cgul_stanza__open_crlf_file(cgul_exception_t* cex,
                            cgul_stanza_t stanza,
                            cgul_crlf_file_t crlf_file)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);
    assert(crlf_file);

    /* Close the currently open stanza source (if any). */
    cgul_stanza__close(cex, stanza);

    /* Remember the file. */
    stanza->f = crlf_file;

    /* Remember that we do not own f. */
    stanza->we_own_f = 0;

 out:
    return;
}

void
cgul_stanza__close(cgul_exception_t* cex,
                   cgul_stanza_t stanza)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    /* Clear "keys" and "tree" of the previous stanza and reset prev_key. */
    if (stanza->keys) {
        cgul_vector__clear(cex, stanza->keys);
    }

    /* Clear the tree that holds the key/value pairs. */
    if (stanza->tree) {
        cgul_stanza__clear_tree(cex, stanza->tree);
    }

    /* Delete f if we own it. */
    if (stanza->f && stanza->we_own_f) {
        cgul_crlf_file__delete(stanza->f);
    }
    stanza->f = NULL;
    stanza->we_own_f = 0;

    /* Reset other member variables. */
    stanza->line_number_current = 0;
    stanza->line_number_starting = 0;
    stanza->prev_key = NULL;

 out:
    return;
}

const char*
cgul_stanza__get_fname(cgul_exception_t* cex,
                       cgul_stanza_t stanza)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    /* Make sure a file has been opened. */
    if (!stanza->f) {
        *cex = cgul_exception__new("cgul_stanza__get_fname: "
                                   "file not open");
        goto out;
    }

    rv = cgul_crlf_file__get_fname(cex, stanza->f);

 out:
    return rv;
}


void
cgul_stanza__all_one_stanza(cgul_exception_t* cex,
                            cgul_stanza_t stanza,
                            int all_one_stanza)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    stanza->all_one_stanza = all_one_stanza;

 out:
    return;
}

void
cgul_stanza__allow_duplicate_keys(cgul_exception_t* cex,
                                  cgul_stanza_t stanza,
                                  int allow_duplicate_keys)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    stanza->allow_duplicate_keys = allow_duplicate_keys;

 out:
    return;
}

void
cgul_stanza__set_allowed_keys(cgul_exception_t* cex,
                              cgul_stanza_t stanza,
                              cgul_rbtree_t allowed_keys)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    /* Note that it is legal for allowed_keys to be NULL. */
    stanza->allowed_keys = allowed_keys;

 out:
    return;
}

void
cgul_stanza__set_mandatory_keys(cgul_exception_t* cex,
                                cgul_stanza_t stanza,
                                cgul_rbtree_t mandatory_keys)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    /* Note that it is legal for mandatory_keys to be NULL. */
    stanza->mandatory_keys = mandatory_keys;

 out:
    return;
}

int
cgul_stanza__load_next(cgul_exception_t* cex,
                       cgul_stanza_t stanza)
{
    int rv = 0;
    size_t line_length = 0;
    const char* line = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    /* Make sure a file has been opened. */
    if (!stanza->f) {
        *cex = cgul_exception__new("cgul_stanza__load_next: "
                                   "file not open");
        goto out;
    }

    /* Clear "keys" and "tree" of the previous stanza and reset prev_key. */
    cgul_vector__clear(cex, stanza->keys);
    cgul_stanza__clear_tree(cex, stanza->tree);
    stanza->prev_key = NULL;

    /* Remember the starting line number. */
    stanza->line_number_starting = stanza->line_number_current + 1;

    /*
     * Skip leading comments and blank lines.  We have to do this
     * because the previous time this method was called we stopped
     * processing the stanza at the first blank line or EOF.  So, when
     * this method is called again, we have to make sure all the blank
     * lines and comments are skipped until we find the start of the
     * next stanza.
     */
    for (;;) {

        /* Get the next line. */
        ++stanza->line_number_current;
        line = cgul_crlf_file__get_line(cex, stanza->f);
        if (*cex) {
            goto out;
        }

        /* If EOF, goto out. */
        if (line == NULL) {
            goto out;
        }

        /* Calculate the length of the line. */
        line_length = strlen(line);

        /* Skip blank lines and comments. */
        if (   cgul_stanza__is_blank_line(cex, line, line_length)
            || cgul_stanza__is_comment_line(cex, line, line_length))
        {
            continue;
        }

        /* The line is not a comment and it is not empty, we are ready
         * to start. */
        break;
    }

    /*
     * Read all key/value pairs in the stanza skipping all comments.
     * Note that the stanza ends with the first blank line or EOF.  If
     * all_one_stanza is set, both comments and blank lines are
     * skipped, and the stanza ends only with EOF.
     */
    for (;;) {

        /*
         * If we are here, it means we have a raw line to try to
         * insert.  Yes, almost all of the rest of this method is just
         * for carefully reading the file.
         */
        cgul_stanza__process_line(cex, stanza, line, line_length);
        if (*cex) {
            goto out;
        }

        /* Get the next line skipping comments or exiting if EOF is
         * reached or the line is blank.  If all_one_stanza is set, we
         * get the next line and skip both comments and blank lines
         * exiting only when EOF is reached. */
        for (;;) {

            /* Get the next line. */
            ++stanza->line_number_current;
            line = cgul_crlf_file__get_line(cex, stanza->f);
            if (*cex) {
                goto out;
            }

            /* Success if EOF. */
            if (line == NULL) {
                goto break_out;
            }

            /* Calculate the length of the line. */
            line_length = strlen(line);

            /* Skip comments. */
            if (cgul_stanza__is_comment_line(cex, line, line_length)) {
                continue;
            }

            /* Blank lines. */
            if (cgul_stanza__is_blank_line(cex, line, line_length)) {
                /* Skip blank lines if all_on_stanza is set. */
                if (stanza->all_one_stanza) {
                    continue;
                }

                /* End of stanza. */
                goto break_out;
            }

            break;
        }

    }

    /* The cleanest way to get out of the embedded loop above is to
     * just "goto break_out". */
 break_out:

    /* We now need to test to make sure all mandatory keys are present. */
    cgul_stanza__verify_mandatory_keys(cex, stanza);
    if (*cex) {
        goto out;
    }

    /* Success. */
    rv = 1;

 out:

    return rv;
}

const char*
cgul_stanza__lookup(cgul_exception_t* cex,
                    cgul_stanza_t stanza,
                    const char* key)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);
    assert(key);

    /* Make sure a file has been opened. */
    if (!stanza->f) {
        *cex = cgul_exception__new("cgul_stanza__lookup: "
                                   "file not open");
        goto out;
    }

    rv = cgul_stanza__lookup_tree(cex, stanza->tree, key);

 out:
    return rv;
}

char*
cgul_stanza__take_value(cgul_exception_t* cex,
                        cgul_stanza_t stanza,
                        const char* key)
{
    char* rv = NULL;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);
    assert(key);

    /* Make sure a file has been opened. */
    if (!stanza->f) {
        *cex = cgul_exception__new("cgul_stanza__take_value: "
                                   "file not open");
        goto out;
    }

    /* Search the tree to determine if the key is present. */
    n = cgul_rbtree__find(cex, stanza->tree, key);
    if (n) {

        /* If the key is present, return the value. */
        rv = (char*)cgul_rbtree_node__get_value(cex, n);

        /* The client is taking ownership of the pointer.  So, we set
         * our copy of the value to NULL so that
         * cgul_stanza__clear_tree() will not try to free it. */
        cgul_rbtree_node__set_value(cex, n, NULL);

    }


 out:
    return rv;
}

cgul_rbtree_t
cgul_stanza__get_tree(cgul_exception_t* cex,
                      cgul_stanza_t stanza)
{
    cgul_rbtree_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    rv = stanza->tree;

 out:
    return rv;
}

cgul_rbtree_t
cgul_stanza__take_tree(cgul_exception_t* cex,
                       cgul_stanza_t stanza)
{
    cgul_rbtree_t tmp = NULL;
    cgul_rbtree_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    /* Try to allocate the new tree. */
    tmp = cgul_rbtree__new(cex, (cgul_rbtree__compare_t)strcmp);
    if (*cex) {
        goto out;
    }

    /* Success. */
    rv = stanza->tree;
    stanza->tree = tmp;

 out:
    return rv;
}

/**
 * This method clears the <code>tree</code> by removing and freeing
 * all key/value pairs.
 *
 * @param[in] cex c-style exception
 * @param[in] tree tree to empty
 */
void
cgul_stanza__clear_tree(cgul_exception_t* cex,
                        cgul_rbtree_t tree)
{
    void* key = NULL;
    void* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tree);

    /* Just keep removing from the front until there are no more nodes. */
    while (cgul_rbtree__remove_front(cex, tree, &key, &value)) {
        free(key);
        if (value) {
            free(value);
        }
    }

 out:
    return;
}

const char*
cgul_stanza__lookup_tree(cgul_exception_t* cex,
                         cgul_rbtree_t tree,
                         const char* key)
{
    const char* rv = NULL;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tree);
    assert(key);

    /* Search the tree to determine if the key is present. */
    n = cgul_rbtree__find(cex, tree, key);
    if (n) {
        /* If the key is present, return the value. */
        rv = (const char*)cgul_rbtree_node__get_value(cex, n);
    }


 out:
    return rv;
}

cgul_vector_t
cgul_stanza__get_keys(cgul_exception_t* cex,
                      cgul_stanza_t stanza)
{
    cgul_vector_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    rv = stanza->keys;

 out:
    return rv;
}

cgul_vector_t
cgul_stanza__copy_keys(cgul_exception_t* cex,
                       cgul_stanza_t stanza)
{
    size_t i = 0;
    size_t key_count = 0;
    char* key = NULL;
    const char** keys = NULL;
    cgul_vector_t rv = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    /* Create the vector to return. */
    rv = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the key count. */
    key_count = cgul_vector__get_size(cex, stanza->keys);

    /* Get the keys. */
    keys = (const char**)cgul_vector__get_array(cex, stanza->keys);

    /* Iterate over the keys. */
    for (i = 0 ; i < key_count ; ++i) {

        /* Copy the key. */
        key = cgul_libc__strdup(cex, keys[i]);
        if (*cex) {
            goto out;
        }

        /* Add the key to the back of the new vector. */
        cgul_vector__push_back(cex, rv, key);
        if (*cex) {
            goto out;
        }
    }

 out:

    /* Rollback the operation if an exception occurs. */
    if (*cex && rv) {
        /* Free each key. */
        while ((key = (char*)cgul_vector__pop_back(&local, rv)) != NULL) {
            free(key);
        }
        /* Delete the vector itself. */
        cgul_vector__delete(rv);
        rv = NULL;
    }

    return rv;
}

unsigned long int
cgul_stanza__get_line_number(cgul_exception_t* cex,
                             cgul_stanza_t stanza)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(stanza);

    /* Make sure a file has been opened. */
    if (!stanza->f) {
        *cex = cgul_exception__new("cgul_stanza__get_line_number: "
                                   "file not open");
        goto out;
    }

    rv = stanza->line_number_starting;

 out:
    return rv;
}

int
cgul_stanza__is_blank_line(cgul_exception_t* cex,
                           const char* line,
                           size_t line_length)
{
    int rv = 1;
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(line);

    for (i = 0 ; i < line_length ; ++i) {
        if (!isspace((unsigned char)line[i])) {
            rv = 0;
            break;
        }
    }

 out:
    return rv;
}

int
cgul_stanza__is_comment_line(cgul_exception_t* cex,
                             const char* line,
                             size_t line_length)
{
    int rv = 0;
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(line);

    for (i = 0 ; i < line_length ; ++i) {
        if (!isspace((unsigned char)line[i])) {
            if (line[i] == '#') {
                rv = 1;
            }
            break;
        }
    }

 out:
    return rv;
}

void
cgul_stanza__split_line(cgul_exception_t* cex,
                        const char* line,
                        size_t line_length,
                        const char* fname,
                        unsigned long int line_number_current,
                        char** key,
                        char** value)
{
    size_t i = 0;
    size_t key_start = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(line);

    /* Reset *key and *value. */
    if (key) {
        *key = NULL;
    }
    if (value) {
        *value = NULL;
    }

    /* Skip leading white space. */
    for (i = 0 ; i < line_length ; ++i) {
        if (!isspace((unsigned char)line[i])) {
            break;
        }
    }

    /* Remember where the key starts. */
    key_start = i;

    /* Find the first colon which marks the end of the key. */
    for ( ; i < line_length ; ++i) {
        /* Find the first colon. */
        if (line[i] == ':') {
            break;
        }
    }

    /* If the key is empty or does not exist at all, thrown an exception. */
    if ((i == key_start) || (i >= line_length)) {
        *cex = cgul_exception__new("%s:%lu: Unable to find key on line: %s",
                                   (fname ? fname : "-"),
                                   line_number_current,
                                   line);
        goto out;
    }

    if (key) {
        /* Null-terminate the key.  Note that the colon character will
         * be restored below. */
        assert(line[i] == ':');
        ((char*)line)[i] = '\0';
        
        /* Duplicate the key. */
        *key = cgul_libc__strdup(cex, &line[key_start]);

        /* Restore the colon character before checking for an exception. */
        ((char*)line)[i] = ':';

        /* Check for an exception. */
        if (*cex) {
            goto out;
        }
    }

    /* The character after the first colon can optionally be a space. */
    if (line[i + 1] == ' ') {
        ++i;
    }

    /* Duplicate the value. */
    if (value) {
        *value = cgul_libc__strdup(cex, &line[++i]);
        if (*cex) {
            goto out;
        }
    }
     
 out:

    /* Clean up if an exception is thrown. */
    if (*cex) {
        if (value && *value) {
            free(*value);
            *value = NULL;
        }
        if (key && *key) {
            free(*key);
            *key = NULL;
        }
    }

    return;
}

void
cgul_stanza__print(cgul_exception_t* cex,
                   FILE* f,
                   const char* key,
                   const char* value)
{
    unsigned long int i = 0;
    unsigned long int line_count = 0;
    char* line = NULL;
    const char* plus_key_cstr = NULL;
    cgul_string_t plus_key = NULL;
    cgul_vector_t lines = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(key);
    assert(value);

    /* Allocate the vector of lines. */
    lines = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Split "value" into lines. */
    cgul_split(cex, value, "\n", 1, lines);
    if (*cex) {
        goto out;
    }

    /* Get the count of lines. */
    line_count = cgul_vector__get_size(cex, lines);

    /* Print the first line. */
    if (line_count == 0) {
        cgul_libc__fprintf(cex, f, "%s:\n", key);
    } else {
        line = (char*)cgul_vector__at(cex, lines, 0);
        cgul_libc__fprintf(cex, f, "%s: %s\n", key, line);
    }

    /* Check if there are more lines. */
    if (line_count > 1) {

        /* Allocate the special "+" key and pad it so it is
         * right-aligned with "key". */
        plus_key = cgul_string__new_from_cstring(cex, "+");
        cgul_string__pad_right(cex, plus_key, strlen(key), ' ');
        if (*cex) {
            goto out;
        }

        /* Get "plus_key" as a C-style string. */
        plus_key_cstr = cgul_string__get_value(cex, plus_key);

        /* Print the remaining lines. */
        for (i = 1 ; i < line_count ; ++i) {
            line = (char*)cgul_vector__at(cex, lines, i);
            cgul_libc__fprintf(cex, f, "%s: %s\n", plus_key_cstr, line);
            if (*cex) {
                goto out;
            }
        }
    }

 out:

    /*
     * Clean up.
     */

    /* plus_key */
    if (plus_key) {
        cgul_string__delete(plus_key);
    }

    /* lines */
    if (lines) {
        /* Free each line. */
        for (;;) {
            line = (char*)cgul_vector__pop_back(&local, lines);
            if (!line) {
                break;
            }
            free(line);
        }
        /* Free the vector itself. */
        cgul_vector__delete(lines);
    }

    return;
}
