/*
 * This class implements a Ternary Search Trie.  See the following URL
 * for more details:
 *
 *     https://en.wikipedia.org/wiki/Ternary_search_tree
 *
 * The main thing to keep in mind is that sort order is determined by
 * the following increasing list of priorities:
 *
 *    1) n->left
 *    2) n->key (i.e., found it)
 *    3) n->middle
 *    4) n->right
 */

#include "cgul_trie_p.h"
#include "cgul_trie_node_p.h"
#include "cgul_trie_node_with_indexing_p.h"
#include "cgul_int.h"
#include "cgul_libc.h"
#include "cgul_random_lfg32.h"
#include "cgul_shuffle.h"
#include "cgul_vector.h"

/************************************************************************
 * Trie Utilities
 ************************************************************************/

/**
 * Return the terminating node for the key <code>key</code> given the
 * starting node <code>start</code>.  The key should be specified
 * relative to the starting node.  To search the entire trie, use
 * <code>trie->root</code> as the starting node. <p>
 *
 * If <code>insert_missing_nodes</code> is true, missing nodes are
 * inserted thereby allowing the terminating node to be reached which
 * is needed when inserting new keys.  If
 * <code>insert_missing_nodes</code> is false, no new nodes are
 * inserted. <p>
 *
 * If the terminating node is found, it is returned; otherwise
 * <code>NULL</code> is returned.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * @note The caller is responsible for checking the key of the
 * returned node to determine whether the returned node maps to a
 * string that has been inserted or whether the returned node is just
 * along the path to a different string that has been inserted. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] start starting node
 * @param[in] key search key relative to the starting node
 * @param[in] insert_missing_nodes whether to insert missing nodes
 * @return the terminating node or <code>NULL</code> if not found
 * @see cgul_trie__find_longest_prefix_of()
 */
static cgul_trie_node_t
cgul_trie__find_terminus(cgul_exception_t* cex,
                         cgul_trie_t trie,
                         cgul_trie_node_t start,
                         const char* key,
                         int insert_missing_nodes)
{
    size_t i = 0;
    cgul_trie_node_t rv = NULL;
    cgul_trie_node_t* curr = NULL;
    cgul_trie_node_t prev = NULL;
    cgul_trie_node_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(trie->cache);
    assert(start);

    /*
     * Sanity check.
     */
    if (!key) {
        goto out;
    }

    /*
     * NOTE #1:
     * =======
     *
     * This method returns the terminating node for "key" whether or
     * not "key" is actually stored in the trie.  The caller *must*
     * check "rv->key" to make sure the terminating node is associated
     * with an actual key if that is why this function was called.
     *
     *
     *
     * NOTE #2:
     * =======
     *
     * An extra level of indirection is introduced for "curr" so we
     * can navigate the trie *and* overwrite the NULL leaf nodes when
     * "insert_missing_nodes" is true.
     */

    /*
     * Special Case: Empty strings map directly to the root node.  We
     * cannot enter the while() loop below because the character in
     * trie->root is not valid for the root node (which represents the
     * empty string so it doesn't really have a character).
     */
    if (start == trie->root) {
        /* If "key" is the empty string, return the root node. */
        if (key[0] == 0) {
            rv = trie->root;
            goto out;
        }
    }

    /* Otherwise, the empty string is along the path to all other
     * strings so take the middle branch. */
    prev = start;
    curr = &start->middle;

    /*
     * General Case: Non-empty strings.
     */
    i = 0;
    while (key[i]) {

        if (*curr) {

            /*
             * Navigate the trie.
             */

            /* The left and right branches are for other characters
             * that might be at the some position in "key."  The
             * middle branch is for characters at the next position in
             * "key", and we only take the middle branch when we have
             * a match at the current position in "key." */
            if (key[i] < (*curr)->ch) {
                prev = *curr;
                curr = &(*curr)->left;
            } else if (key[i] > (*curr)->ch) {
                prev = *curr;
                curr = &(*curr)->right;
            } else {
                /* If this is the last character in "key", (*curr)
                 * points to the terminating node. */
                if (key[i + 1] == 0) {
                    /* Success. */
                    rv = *curr;
                    goto out;
                }
                /* Otherwise, take the middle branch and move to the
                 * next character in the "key" because we have a match
                 * for the *current character in "key." */
                ++i;
                prev = *curr;
                curr = &(*curr)->middle;
            }

        } else if (insert_missing_nodes) {

            /*
             * Add missing nodes just-in-time.
             */

            /* Otherwise, create the missing node. */
            tmp = (cgul_trie_node_t)cgul_cache__get(cex, trie->cache);
            if (*cex) {
                goto out;
            }
            tmp->ch = key[i];
            tmp->parent = prev;

            /* Insert the missing node. */
            *curr = tmp;

            /* The trie now owns the new node. */
            tmp = NULL;

            /* Return to the top of the loop with (*curr) now not
             * being NULL.  This will cause the first branch of the
             * if() statement to be taken which will continue
             * navigating the trie. */

        } else {

            /*
             * Node is missing.
             */

            goto out;
        }

    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_exception_t local = NULL;
        cgul_cache__put(&local, trie->cache, tmp);
    }

    return rv;
}

/**
 * Walk up the trie <code>trie</code> incrementing the count for each
 * parent of the node <code>node</code> by the amount
 * <code>delta</code>.  Note that <code>delta</code> can be negative
 * allowing the count to be decremented.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] node starting node
 * @param[in] delta amount by which to increment each parent's count
 */
static void
cgul_trie__increment_parent_counts_by(cgul_exception_t* cex,
                                      cgul_trie_t trie,
                                      cgul_trie_node_t node,
                                      unsigned long int delta)
{
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_with_indexing_t parent_with_indexing = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Sanity checks. */
    if (!trie->is_indexed) {
        goto out;
    }
    if (!node) {
        goto out;
    }

    /* Walk up the trie incrementing the count variable associated
     * with the left, middle, or right subtree depend on whether curr
     * is in the left, middle, or right subtree respectively.  Note
     * that the counts are for the number of *keys* in a subtree, not
     * necessarily the number of nodes (because in a Ternary Search
     * Trie not all nodes map to keys). */
    for (curr = node ; curr && curr->parent ; curr = curr->parent) {
        /* Down cast. */
        parent_with_indexing = (cgul_trie_node_with_indexing_t)curr->parent;
        if (curr == curr->parent->left) {
            parent_with_indexing->lcount += delta;
        } else if (curr == curr->parent->middle) {
            parent_with_indexing->mcount += delta;
        } else if (curr == curr->parent->right) {
            parent_with_indexing->rcount += delta;
        }
    }

 out:
    return;
}

/**
 * Return a copy of the array <code>arr</code> that holds
 * <code>arr_size</code> pointers.  The client is responsible for
 * calling <code>free()</code> on the pointer returned.  If an error
 * occurs, <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] arr array
 * @param[in] arr_size number of pointers in the array
 * @return copy of the array
 */
static void**
cgul_trie__duplicate_pointers_array(cgul_exception_t* cex,
                                    void** arr,
                                    size_t arr_size)
{
    void** rv = NULL;
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    rv = (void**)cgul_libc__calloc(cex, arr_size, sizeof(*arr));
    if (*cex) {
        goto out;
    }

    /* Copy the pointers. */
    for (i = 0 ; i < arr_size ; ++i) {
        rv[i] = arr[i];
    }

 out:

    /* Clean up. */
    if (*cex && rv) {
        free(rv);
        rv = NULL;
    }

    return rv;
}

/**
 * Shuffle the key/value pair of arrays <code>keys</code> and
 * <code>values</code> where the key at <code>keys[i]</code>
 * corresponds to the value at <code>values[i]</code>.  Both arrays
 * are shuffled using the same stream of pseudo-random numbers so that
 * after the shuffle, it is still true that the key at
 * <code>keys[i]</code> corresponds to the value at
 * <code>values[i]</code> even though both arrays are now shuffled.
 *
 * @param[in,out] cex c-style exception
 * @param[in] keys keys
 * @param[in] values values
 * @param[in] kv_size number of key/value pairs (i.e., size of the arrays)
 */
static void
cgul_trie__shuffle_key_value_arrays(cgul_exception_t* cex,
                                    const char** keys,
                                    void** values,
                                    size_t kv_size)
{
    cgul_uint32_t seed = 0;
    cgul_random_lfg32_t prng1 = NULL;
    cgul_random_lfg32_t prng2 = NULL;

    /* Do not assert "keys" or "values" which could be NULL.  If an
     * array is NULL, it will not be shuffled. */
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (kv_size == 0) {
        goto out;
    }

    /* Get the shared seed for our PRNGs. */
    seed = (cgul_uint32_t)time(NULL);

    /* Allocate identical PRNGs so that the key at keys[i] corresponds
     * to the value at values[i] even after sorting because the sort
     * of each array will using the same PRN sequence. */
    prng1 = cgul_random_lfg32__new_from_seed(cex, seed);
    prng2 = cgul_random_lfg32__new_from_seed(cex, seed);
    if (*cex) {
        goto out;
    }

    /* Sort the keys. */
    if (keys) {
        cgul_shuffle_pointers(
            cex,
            (void**)keys,
            kv_size,
            (cgul_shuffle__next_in_range_t)cgul_random_lfg32__next_in_range,
            prng1);
    }

    /* Sort the values. */
    if (values) {
        cgul_shuffle_pointers(
            cex,
            values,
            kv_size,
            (cgul_shuffle__next_in_range_t)cgul_random_lfg32__next_in_range,
            prng2);
    }

 out:

    /* Clean up. */
    if (prng2) {
        cgul_random_lfg32__delete(prng2);
    }
    if (prng1) {
        cgul_random_lfg32__delete(prng1);
    }

    return;
}

/**
 * This method is the callback portion of the
 * <code>cgul_trie__clear()</code> convenience method that is designed
 * to remove all the nodes of a <code>cgul_trie</code> _and_ to put
 * free all the keys and values.  If the function for freeing the keys
 * <code>NULL</code>, no attempt will be made to free the key.  The
 * same applies for the value. <p>
 *
 * This method should not be invoked directly.  Instead, it should be
 * invoked as a callback function for
 * <code>cgul_trie__traverse()</code>.  When you invoke
 * <code>cgul_trie__traverse()</coded>, you should do so in such a way
 * that <code>client_data</code> holds a
 * <code>cgul_trie__clear_cb_data_t</code>.  Typically, your code will
 * work with <code>cgul_trie__clear()</code> or
 * <code>cgul_trie__remove_range()</code> instead of with this method.
 *
 * @param[in] cex c-style exception
 * @param[in] trie trie
 * @param[in] n trie node
 * @param[in] client_data client data
 * @return whether to continue traversing the trie
 */
static int
cgul_trie__clear_cb(cgul_exception_t* cex,
                    cgul_trie_t trie,
                    cgul_trie_node_t n,
                    void* client_data)
{
    int rv = 0;
    char* key = NULL;
    void* value = NULL;
    cgul_trie__clear_cb_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(n);
    assert(client_data);

    data = (cgul_trie__clear_cb_data_t)client_data;

    /* Remove the node. */
    cgul_trie__remove_node(cex, trie, n, &key, &value);
    data->rv = 1;

    /* Only after the node is removed is it safe to put back the key. */
    if (data->keys_cache && key) {
        cgul_cache__put(cex, data->keys_cache, key);
    }

    /* The value can be put back at any time. */
    if (data->values_cache && value) {
        cgul_cache__put(cex, data->values_cache, value);  
    }

    rv = 1;

 out:
    return rv;
}

/**
 * This method calls <code>cgul_cache__put()</code> on each node in
 * the trie <code>trie</code>.  It does not bother to remove the node
 * from the trie because this function should only be called from
 * <code>cgul_trie__delete()</code> or under carefully controlled
 * circumstances.  Because this method is called from the destructor,
 * it is intentional that a <code>cgul_exception</code> instance is
 * not passed to this method.  On return, <code>trie->root</code> will
 * be <code>NULL</code>.
 *
 * @param[in] trie <code>cgul_trie</code> instance
 */
static void
cgul_trie__free_all_nodes(cgul_trie_t trie)
{
    cgul_trie_node_t node = NULL;
    cgul_trie_node_t parent = NULL;

    assert(trie);

    /*
     * Perform a post-order traversal of all the nodes in the trie
     * deleting each node in turn.  (Note that we cannot use the
     * embedded older/younger linked list because it only links the
     * nodes associated with a key, but in a Ternary Search Trie there
     * could be many more nodes associated with characters that are
     * not the terminus of a key.)
     */

    node = trie->root;
    while (node) {

        /* Navigate down to a leaf node.  Free it and update the
         * parents pointer for that node to be NULL.  After the
         * parent's left, middle, and right subtrees have been freed,
         * the parent will become a leaf node, and the process will
         * continue. */
        if (node->left) {
            node = node->left;
        } else if (node->middle) {
            node = node->middle;
        } else if (node->right) {
            node = node->right;
        } else {
            /* Access node->parent while "node" is still valid. */
            parent = node->parent;
            /* Clear the parent's pointer to "node" so that the parent
             * can eventually become a leaf node. */
            if (parent) {
                if (node == parent->left) {
                    parent->left = NULL;
                } else if (node == parent->middle) {
                    parent->middle = NULL;
                } else if (node == parent->right) {
                    parent->right = NULL;
                }
            }
            /* Free the node. */
            free(node);
            /* Move up to the parent. */
            node = parent;
        }
    }

    trie->root = NULL;

    return;
}

/************************************************************************
 * Public Methods
 ************************************************************************/

cgul_trie_t
cgul_trie__new(cgul_exception_t* cex)
{
    cgul_trie_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /*
     * Allocate the trie.
     */
    rv = (cgul_trie_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Create the cache used to create the nodes. */
    rv->cache = cgul_cache__new(
        cex,
        (cgul_cache__constructor_t)cgul_trie_node__new,
        (cgul_cache__destructor_t)cgul_trie_node__delete,
        (cgul_cache__restructor_t)cgul_trie_node__initialize);
    if (*cex) {
        goto out;
    }

    /* Allocate the root node.
     *
     * Remember that root->ch is ignored for the root node because the
     * root node maps to the empty string.
     *
     * Also remember that if root->key is NULL, it means the current
     * node does not terminate any string that has been inserted into
     * the trie. */
    rv->root = (cgul_trie_node_t)cgul_cache__get(cex, rv->cache);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (*cex && rv) {
        cgul_trie__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_trie_t
cgul_trie__new_with_indexing(cgul_exception_t* cex)
{
    cgul_trie_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /*
     * Allocate the trie.
     */
    rv = (cgul_trie_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Create the cache used to create the nodes. */
    rv->cache = cgul_cache__new(
        cex,
        (cgul_cache__constructor_t)cgul_trie_node_with_indexing__new,
        (cgul_cache__destructor_t)cgul_trie_node_with_indexing__delete,
        (cgul_cache__restructor_t)cgul_trie_node_with_indexing__initialize);
    if (*cex) {
        goto out;
    }

    /* Enable indexing. */
    rv->is_indexed = 1;

    /* Allocate the root node.
     *
     * Remember that root->ch is ignored for the root node because the
     * root node maps to the empty string.
     *
     * Also remember that if root->key is NULL, it means the current
     * node does not terminate any string that has been inserted into
     * the trie. */
    rv->root = (cgul_trie_node_t)cgul_cache__get(cex, rv->cache);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (*cex && rv) {
        cgul_trie__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_trie__delete(cgul_trie_t trie)
{
    if (trie) {

        /* Free all the nodes in the trie. */
        cgul_trie__free_all_nodes(trie);

        /* Delete the cache only after freeing the nodes. */
        if (trie->cache) {
            cgul_cache__delete(trie->cache);
            trie->cache = NULL;
        }

        /* Free the trie itself. */
        free(trie);
    }
}

void
cgul_trie__free_keys(cgul_trie_t trie)
{
    cgul_trie_node_t node = NULL;
    cgul_exception_t local = NULL;

    if (!trie) {
        goto out;
    }

    /* Iterate over each node that holds a key and free the key. */
    node = cgul_trie__get_youngest(&local, trie);
    for ( ; node ; node = node->older) {
        if (node->key) {
            free((void*)node->key);
        }
    }

 out:
    return;
}

void
cgul_trie__free_values(cgul_trie_t trie)
{
    cgul_trie_node_t node = NULL;
    cgul_exception_t local = NULL;

    if (!trie) {
        goto out;
    }

    /* Iterate over each node that holds a value and free the value. */
    node = cgul_trie__get_youngest(&local, trie);
    for ( ; node ; node = node->older) {
        if (node->value) {
            free(node->value);
        }
    }

 out:
    return;
}

unsigned long int
cgul_trie__get_cache_size(cgul_exception_t* cex,
                          cgul_trie_t trie)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    rv = cgul_cache__get_size(cex, trie->cache);

 out:
    return rv;
}

void
cgul_trie__set_cache_size(cgul_exception_t* cex,
                          cgul_trie_t trie,
                          unsigned long int size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    cgul_cache__set_size(cex, trie->cache, size);

 out:
    return;
}

unsigned long int
cgul_trie__get_cache_reserve(cgul_exception_t* cex,
                             cgul_trie_t trie)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    rv = cgul_cache__get_reserve(cex, trie->cache);

 out:
    return rv;
}

void
cgul_trie__set_cache_reserve(cgul_exception_t* cex,
                             cgul_trie_t trie,
                             unsigned long int reserve)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    cgul_cache__set_reserve(cex, trie->cache, reserve);

 out:
    return;
}

int
cgul_trie__is_empty(cgul_exception_t* cex,
                    cgul_trie_t trie)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    rv = (trie->size == 0);

 out:
    return rv;
}

void
cgul_trie__balance(cgul_exception_t* cex,
                   cgul_trie_t trie,
                   int maintain_relative_ages)
{
    unsigned long int i = 0;
    unsigned long int trie_size = 0;
    const char** keys = NULL;
    void** values = NULL;
    cgul_trie_t tmp = NULL;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /*
     * NOTE: A temporary trie is used so that no memory is leaked if
     * an exception is thrown because the original, unmodified trie
     * still holds pointers to everything.
     */

    /* Get the size of the trie. */
    trie_size = cgul_trie__get_size(cex, trie);

    /* Allocate. */
    tmp = cgul_trie__new(cex);
    keys = (const char**)cgul_libc__calloc(cex, trie_size, sizeof(*keys));
    values = (void**)cgul_libc__calloc(cex, trie_size, sizeof(*values));
    if (*cex) {
        goto out;
    }

    /* Copy the keys and values from the trie to the arrays. */
    node = cgul_trie__get_oldest(cex, trie);
    for (i = 0 ; node ; ++i, node = node->younger) {
        keys[i] = node->key;
        values[i] = node->value;
    }

    /* Insert the key/value pairs into the temporary trie. */
    cgul_trie__insert_from_arrays(
        cex, tmp, keys, values, trie_size, maintain_relative_ages);
    if (*cex) {
        goto out;
    }

    /* Success. */
    cgul_trie__swap(cex, trie, tmp);

 out:

    /* Clean up. */
    if (values) {
        free(values);
    }
    if (keys) {
        free(keys);
    }
    if (tmp) {
        cgul_trie__delete(tmp);
    }

    return;
}

int
cgul_trie__insert(cgul_exception_t* cex,
                  cgul_trie_t trie,
                  const char* key,
                  cgul_trie_node_t* n)
{
    int rv = 0;
    cgul_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Clear the user's pointer to the node that is to be returned.
     * We do this now so that NULL will be returned if this method
     * throws an exception. */
    if (n) {
        *n = NULL;
    }

    /* Find the node for this key. */
    curr = cgul_trie__find_terminus(cex,
                                    trie,
                                    trie->root,
                                    key,
                                    1 /* insert_missing_nodes */);
    if (*cex) {
        goto out;
    }
    if (!curr) {
        *cex = cgul_exception__new(
            "cgul_trie__insert: internal error: terminus unreachable");
        goto out;
    }

    /* If this is a node for a new key, the insert was a success.  We
     * know it is a new key if curr->key is still NULL. */
    if (!curr->key) {

        curr->key = key;
        curr->older = trie->youngest;

        if (trie->youngest) {
            trie->youngest->younger = curr;
        }

        trie->youngest = curr;

        if (!trie->oldest) {
            trie->oldest = curr;
        }

        /* Remember that there is one more key in the trie. */
        ++trie->size;
        if (trie->is_indexed) {
            cgul_trie__increment_parent_counts_by(cex, trie, curr, 1);
        }

        /* Success. */
        rv = 1;
    }

    /* Return the node to the user. */
    if (n) {
        *n = curr;
    }

 out:
    return rv;
}

int
cgul_trie__insert_from_arrays(cgul_exception_t* cex,
                              cgul_trie_t trie,
                              const char** keys,
                              void** values,
                              unsigned long int kv_size,
                              int maintain_relative_ages)
{
    int insert_rv = 0;
    unsigned long int rv = 0;
    unsigned long int i = 0;
    const char** keys_copy = NULL;
    void** values_copy = NULL;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "values" which is allowed to be NULL. */
    assert(trie);

    /* Sanity check. */
    if (!keys) {
        goto out;
    }

    /* Duplicate the "keys" array. */
    keys_copy = (const char**)cgul_trie__duplicate_pointers_array(
                                  cex, (void**)keys, kv_size);
    if (*cex) {
        goto out;
    }

    /* Duplicate the "values" array. */
    if (values) {
        values_copy = cgul_trie__duplicate_pointers_array(cex, values, kv_size);
        if (*cex) {
            goto out;
        }
    }

    /* Shuffle the duplicated arrays (which handles NULL arrays gracefully). */
    cgul_trie__shuffle_key_value_arrays(cex, keys_copy, values_copy, kv_size);
    if (*cex) {
        goto out;
    }

    /* Insert the key/value pairs into the trie. */
    for (i = 0 ; i < kv_size ; ++i) {

        /* Insert the key or get the node for the existing key. */
        insert_rv = cgul_trie__insert(cex, trie, keys_copy[i], &node);
        if (*cex) {
            goto out;
        }

        /* Insert or replace the value. */
        node->value = (values_copy ? values_copy[i] : NULL);

        /* If a new key was inserted, increment the return count. */
        if (insert_rv) {
            ++rv;
        }
    }

    /* Check whether to manipulate the older/younger linked list so
     * that it appears as though the key/value pairs were inserted in
     * the order indicated by the "keys" and "values" arrays (instead
     * of in the actualy random order of insertion). */
    if (maintain_relative_ages) {

        /* Work front-to-back making each key the youngest.  When you
         * reach the back of the array, all keys will have been
         * touched in the correct order. */
        for (i = 0 ; i < kv_size ; ++i) {

            /* Get the node for this key. */
            node = cgul_trie__find(cex, trie, keys[i]);
            if (!node) {
                *cex = cgul_exception__new(
                    "cgul_trie__insert_from_arrays: Internal Error: "
                    "Unable to find key just inserted!");
                goto out;
            }

            /* Make this node the youngest.  The cumulative effect is
             * to maintain the relative ages of the nodes. */
            cgul_trie__set_youngest(cex, trie, node);
        }
    }

 out:

    /* Clean up. */
    if (values_copy) {
        free(values_copy);
    }
    if (keys_copy) {
        free(keys_copy);
    }

    return rv;
}

cgul_trie_node_t
cgul_trie__find(cgul_exception_t* cex,
                cgul_trie_t trie,
                const char* key)
{
    cgul_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Find the terminating node which may or may not be associated
     * with a key so we check below to make sure it is associated with
     * a key; otherwise, the key is not in the trie.*/
    rv = cgul_trie__find_terminus(cex,
                                  trie,
                                  trie->root,
                                  key,
                                  0 /* insert_missing_nodes */);
    if (rv && !rv->key) {
        rv = NULL;
    }

 out:
    return rv;
}

/* Each node keeps tracks of the number of keys in its left subtree
 * (lcount), middle subtree (mcount), and right subtree (rcount).
 * This lets you find any node in logarithmic time by an index into
 * its sorted position (assuming the best-case scenario where the trie
 * is balanced). */
cgul_trie_node_t
cgul_trie__find_at(cgul_exception_t* cex,
                   cgul_trie_t trie,
                   unsigned long int index)
{
    unsigned long int lesser_count = 0;
    unsigned long int partition_count = 0;
    cgul_trie_node_with_indexing_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Make sure the trie is using nodes from the
     * cgul_trie_node_with_indexing class instead of from the base
     * cgul_trie_node class. */
    if (!trie->is_indexed) {
        *cex = cgul_exception__new(
                  "cgul_trie__find_at: Indexing not supported for this "
                  "trie.\ncgul_trie object not created with "
                  "cgul_trie__new_with_indexing().");
        goto out;
    }

    /*
     * The user passes in a C-style index.  So, if the user passed in
     * 5 for index (for example), he or she is looking for the 6th
     * element in sort order.  We can find the 6th element by finding
     * the node with 5 lesser children.  It is a little more
     * complicated because going down the middle or right subtrees
     * "accumulates" children whereas going left does not.
     *
     * So the way this works is that "lesser_count" is the total
     * number of keys present in the trie that are less than the key
     * for the current node (or if rv->key is NULL, the key that would
     * be present).  The trick is to keep "lesser_count" up to date
     * after each transition to a new node.
     *
     * When you take the left branch of the *current* node, you just
     * subtract out the nodes that are not along the left branch of
     * the *new* node.  The nodes that are along the left branch of
     * the *new* node have already been counted (when you took a
     * middle or right branch earlier).
     *
     * When you take the middle branch, add 1 for the *current* node
     * if rv->key is not NULL (because you are moving beyond that
     * key).  Then, just add the nodes that are along the left branch
     * of the *new* node.
     *
     * When you take the right branch, add 1 for the *current* node if
     * rv->key is not NULL (because you are moving beyond that key and
     * it hasn't been counted yet).  Add the *current* node's mcount
     * (again because you are moving beyond the middle branch and it
     * hasn't been counted yet).  Then, just add the nodes that are
     * along the left branch of the *new* node.
     *
     * You know you are done when the current node has a key and
     * "lesser_count == index".
     *
     * See cgul_rbtree__find_at() for an implementation that is easier
     * to understand because it is for a binary tree instead of a
     * ternary tree.
     */

    if (index < trie->size) {

        rv = (cgul_trie_node_with_indexing_t)trie->root;
        assert(rv);
        lesser_count = rv->lcount;

        for (;;) {

            /* Growing the partition until it includes the node for
             * our index. */
            partition_count = 0;

            /* Left Branch */
            partition_count += lesser_count;
            if (index < partition_count) {
                /* Move to next node. */
                rv = (cgul_trie_node_with_indexing_t)rv->base.left;
                assert(rv);
                /* for next node - subtract count of not lesser keys */
                lesser_count -= (rv->base.key ? 1 : 0);
                lesser_count -= rv->mcount;
                lesser_count -= rv->rcount;
                continue;
            }

            /* Current Node */
            partition_count += (rv->base.key ? 1 : 0);
            if (index < partition_count) {
                /* Success. */
                break;
            }

            /* Middle Branch */
            partition_count += rv->mcount;
            if (index < partition_count) {
                /* for curr node - add one if curr node holds a key */
                lesser_count += (rv->base.key ? 1 : 0);
                /* Move to next node. */
                rv = (cgul_trie_node_with_indexing_t)rv->base.middle;
                assert(rv);
                /* for next node - add count of lesser keys */
                lesser_count += rv->lcount;
                continue;
            }

            /* Right Branch */
            partition_count += rv->rcount;
            if (index < partition_count) {
                /* for curr node - add one if curr node holds a key + mcount */
                lesser_count += (rv->base.key ? 1 : 0);
                lesser_count += rv->mcount;
                /* Move to next node. */
                rv = (cgul_trie_node_with_indexing_t)rv->base.right;
                assert(rv);
                /* for next node - add count of lesser keys */
                lesser_count += rv->lcount;
                continue;
            }

            /* Not reached. */
            assert(0);
        }
    }

 out:
    return (cgul_trie_node_t)rv;
}

int
cgul_trie__find_rank(cgul_exception_t* cex,
                     cgul_trie_t trie,
                     const char* key,
                     unsigned long int* rank)
{
    int rv = 0;
    size_t i = 0;
    unsigned long int lesser_count = 0;
    cgul_trie_node_with_indexing_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Clear the user's return parameter. */
    if (rank) {
        *rank = 0;
    }

    /* Make sure the trie is using nodes from the
     * cgul_trie_node_with_indexing class instead of from the base
     * cgul_trie_node class. */
    if (!trie->is_indexed) {
        *cex = cgul_exception__new(
                  "cgul_trie__find_rank: "
                  "Indexing not supported for this trie.\n"
                  "cgul_trie object not created with "
                  "cgul_trie__new_with_indexing().");
        goto out;
    }

    /* Sanity check. */
    if (!key) {
        goto out;
    }

    /*
     * Special Case: Empty strings map directly to the root node.  We
     * cannot enter the while() loop below because the character in
     * trie->root is not valid for the root node (which represents the
     * empty string so it doesn't really have a character).
     */
    if (key[0] == 0) {
        if (trie->root->key) {
            /* Success. */
            if (rank) {
                *rank = 0;
            }
            rv = 1;
        }
        goto out;
    }

    /*
     * See comments in cgul_trie__find_at() for how "lesser_count"
     * works.  Briefly, "lesser_count" is the number of keys less than
     * the key that is (or would be (if curr->key is NULL))
     * represented by the current node.
     */

    /* Start with the root node. */
    curr = (cgul_trie_node_with_indexing_t)trie->root;
    lesser_count = curr->lcount;

    /* We know from the "Special Case" above that the root node is not
     * a match for "key", but the root node represent the empty string
     * which is a prefix for all other strings.  Thus, we need to take
     * the middle branch. */
    lesser_count += (curr->base.key ? 1 : 0);
    curr = (cgul_trie_node_with_indexing_t)curr->base.middle;
    lesser_count += curr->lcount;

    i = 0;
    while (curr && key[i]) {

        /*
         * Navigate the trie.
         */

        /* The left and right branches are for other characters that
         * might be at the some position in "key."  The middle branch
         * is for characters at the next position in "key", and we
         * only take the middle branch when we have a match at the
         * current position in "key." */
        if (key[i] < curr->base.ch) {

            curr = (cgul_trie_node_with_indexing_t)curr->base.left;
            if (curr) {
                lesser_count -= (curr->base.key ? 1 : 0);
                lesser_count -= curr->mcount;
                lesser_count -= curr->rcount;
            }

        } else if (key[i] > curr->base.ch) {

            lesser_count += (curr->base.key ? 1 : 0);
            lesser_count += curr->mcount;
            curr = (cgul_trie_node_with_indexing_t)curr->base.right;
            if (curr) {
                lesser_count += curr->lcount;
            }

        } else {

            /* If this is the last character in "key", "curr" points
             * to the terminating node. */
            if (key[i + 1] == 0) {
                if (curr->base.key) {
                    /* Success. */
                    if (rank) {
                        *rank = lesser_count;
                    }
                    rv = 1;
                }
                goto out;
            }

            /* Otherwise, take the middle branch and move to the next
             * character in the "key" because we have a match for the
             * *current character in "key." */
            ++i;
            lesser_count += (curr->base.key ? 1 : 0);
            curr = (cgul_trie_node_with_indexing_t)curr->base.middle;
            if (curr) {
                lesser_count += curr->lcount;
            }

        }

    }

 out:
    return rv;
}

cgul_trie_node_t
cgul_trie__find_floor(cgul_exception_t* cex,
                      cgul_trie_t trie,
                      const char* search_key)
{
    size_t i = 0;
    cgul_trie_node_t rv = NULL;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t target_subtree = NULL;
    enum {

        TODO_NOTHING,

        /**
         * Handle the case for nodes where the path takes the right
         * branch to skip the entire subtree rooted at the node
         * meaning all the keys in that subtree are less than the
         * search key.
         */
        TODO_START_WITH_MIDDLE,

        /**
         * Handle the case for nodes along the path of the search key
         * meaning only target_subtree->key or subtree->left are
         * guaranteed to hold keys less than or equal to the search
         * key.  (target_subtree->middle *could* also hold such keys,
         * but it gets checked in subsequent iterations of the main
         * loop.)
         */
        TODO_START_WITH_KEY,

    } todo = TODO_NOTHING;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Sanity check. */
    if (!search_key) {
        goto out;
    }

    /*
     * Special Case: Empty strings map directly to the root node.
     * Matching the "empty" character at the front of every string is
     * done here by hand.
     */
    if (trie->root->key) {
        target_subtree = trie->root;
        todo = TODO_START_WITH_KEY;
    }

    /* Take the middle branch.  (The root node should only ever have a
     * middle branch because the empty string is a prefix of all other
     * strings.) */
    assert(!trie->root->left);
    assert(!trie->root->right);
    curr = trie->root->middle;

    /* Descend the tree under the control of the "search_key". */
    while (curr && search_key[i]) {

        if (search_key[i] < curr->ch) {
            curr = curr->left;
        } else if (search_key[i] > curr->ch) {
            /* Because the search key is larger than the current node,
             * the current node will hold a key less than the search
             * key. */
            if (curr->middle || curr->key || curr->left) {
                target_subtree = curr;
                todo = TODO_START_WITH_MIDDLE;
            }
            curr = curr->right;
        } else {
            /* Matching character found so this node becomes our
             * target subtree because it is the largest key found so
             * far that is less than or equal to the search key. */
            if (curr->key || curr->left) {
                target_subtree = curr;
                todo = TODO_START_WITH_KEY;
            }
            /* Move to next character. */
            ++i;
            curr = curr->middle;
        }

    }

    /* At this point, we know the target subtree holds the largest key
     * less than or equal to the search key.  We just need to find it. */
    if (target_subtree) {

        /* See comments above for what the possible values of "todo" mean. */
        switch (todo) {

        case TODO_START_WITH_MIDDLE:
            if (target_subtree->middle) {
                rv = cgul_trie_node__get_largest_node_in_subtree(
                         cex, target_subtree->middle);
                break;
            }
            /* Fall through! */

        case TODO_START_WITH_KEY:
            if (target_subtree->key) {
                rv = target_subtree;
                break;
            }
            if (target_subtree->left) {
                rv = cgul_trie_node__get_largest_node_in_subtree(
                         cex, target_subtree->left);
                break;
            }
            /* Not reached. */
            assert(0);
            break;

        default:
            /* Not reached. */
            assert(0);
            break;
        }

    }

 out:
    return rv;
}

cgul_trie_node_t
cgul_trie__find_ceiling(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        const char* search_key)
{
    size_t i = 0;
    cgul_trie_node_t rv = NULL;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t target_subtree = NULL;
    enum {

        TODO_NOTHING,

        /**
         * Handle the case for nodes where the path takes the left
         * branch to skip the entire subtree rooted at the node
         * meaning all the keys in that subtree are greater than the
         * search key.
         */
        TODO_START_WITH_KEY,

        /**
         * Handle the case for when the search key is a prefix for a
         * key in the tree.  We run out of characters in the search
         * key so there will be no subsequent iterations of the main
         * loop to explore the middle subtree further so we have to
         * remember to do it when the loop terminates (which will
         * happen almost immediately given that all the characters in
         * the search tree have been exhausted).
         */
        TODO_START_WITH_MIDDLE,

        /**
         * Handle the case for nodes along the path of the search key
         * meaning only subtree->right is guaranteed to hold keys
         * greater than or equal to the search key.
         * (target_subtree->middle *could* also hold such keys, but it
         * gets checked in subsequent iterations of the main loop.)
         */
        TODO_START_WITH_RIGHT,

    } todo = TODO_NOTHING;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Sanity check. */
    if (!search_key) {
        goto out;
    }

    /*
     * Special Case: Empty strings map directly to the root node.  The
     * empty string is a prefix of all strings.  In other words, the
     * prefix is less than all other strings so it can only be a
     * ceiling for itself.
     */
    if (trie->root->key) {
        if (search_key[0] == 0) {
            rv = trie->root;
            goto out;
        }
    }

    /* If the search key is empty but the trie isn't, the smallest key
     * in the trie is the ceiling.  For all other keys, we will enter
     * the while() loop below which will do the right thing. */
    if ((trie->size > 0) && (search_key[i] == 0)) {
        target_subtree = trie->root;
        todo = TODO_START_WITH_KEY;
    }

    /* Take the middle branch.  (The root node should only ever have a
     * middle branch because the empty string is a prefix of all other
     * strings.) */
    assert(!trie->root->left);
    assert(!trie->root->right);
    curr = trie->root->middle;

    /* Descend the tree under the control of the "search_key". */
    while (curr && search_key[i]) {

        if (search_key[i] < curr->ch) {
            /* Because the search key is smaller than the current
             * node, the current node will hold a key greater than the
             * search key. */
            if (curr->key || curr->middle || curr->right) {
                target_subtree = curr;
                todo = TODO_START_WITH_KEY;
            }
            curr = curr->left;
        } else if (search_key[i] > curr->ch) {
            curr = curr->right;
        } else {
            /* Matching character found.  If we've reached the end of
             * the search key, and there is an exact match, we have
             * found our node. */
            if ((search_key[i + 1] == 0) && curr->key) {
                rv = curr;
                goto out;
            }
            /* If we've reached the end of the search key and there
             * isn't an exact match we have to continue the search
             * after the loop terminates (which is almost immediately
             * because all of the characters in the search key have
             * been exhausted). */
            if ((search_key[i + 1] == 0) && (curr->middle || curr->right)) {
                target_subtree = curr;
                todo = TODO_START_WITH_MIDDLE;
            } else if (curr->right) {
                /* If you reach here, the matching character is part
                 * of a proper prefix.  Subsequent iterations of this
                 * loop are going to check the middle branch so at
                 * this point we only need to remember to check the
                 * right branch. */
                target_subtree = curr;
                todo = TODO_START_WITH_RIGHT;
            }
            /* Move to next character. */
            ++i;
            curr = curr->middle;
        }

    }

    /* At this point, we know the target subtree holds the smallest key
     * less than or equal to the search key.  We just need to find it. */
    if (target_subtree) {

        /* See comments above for what the possible values of "todo" mean. */
        switch (todo) {

        case TODO_START_WITH_KEY:
            if (target_subtree->key) {
                rv = target_subtree;
                break;
            }
            /* Fall through! */

        case TODO_START_WITH_MIDDLE:
            if (target_subtree->middle) {
                rv = cgul_trie_node__get_smallest_node_in_subtree(
                         cex, target_subtree->middle);
                break;
            }
            /* Fall through! */

        case TODO_START_WITH_RIGHT:
            if (target_subtree->right) {
                rv = cgul_trie_node__get_smallest_node_in_subtree(
                         cex, target_subtree->right);
                break;
            }
            /* Not reached. */
            assert(0);
            break;

        default:
            /* Not reached. */
            assert(0);
            break;
        }

    }

 out:
    return rv;
}

int
cgul_trie__find_closed_range_by_prefix(cgul_exception_t* cex,
                                       cgul_trie_t trie,
                                       const char* prefix,
                                       cgul_trie_node_t* begin_node,
                                       cgul_trie_node_t* end_node)
{
    int rv = 0;
    cgul_trie_node_t top = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "begin_node" or "end_node" which both can be NULL. */
    assert(trie);

    /* Clear the user's pointers. */
    if (begin_node) {
        *begin_node = NULL;
    }
    if (end_node) {
        *end_node = NULL;
    }

    /* Sanity check. */
    if (!prefix) {
        goto out;
    }

    /* Get the node (if any) that corresponds to the last matching
     * letter in the prefix. */
    top = cgul_trie__find_terminus(cex,
                                   trie,
                                   trie->root,
                                   prefix,
                                   0 /* insert_missing_nodes */);
    if (!top) {
        goto out;
    }

    /* Find the nodes for the smallest and largest keys under "top". */
    if (begin_node) {
        if (top->key) {
            *begin_node = top;
        } else if (top->middle) {
            *begin_node = 
                cgul_trie_node__get_smallest_node_in_subtree(cex, top->middle);
        } else {
            goto out;
        }
    }
    if (end_node) {
        /* Only the middle branch holds longer keys with the same
         * prefix.  If the middle branch is NULL, then top is the end
         * node too. */
        if (top->middle) {
            *end_node =
                cgul_trie_node__get_largest_node_in_subtree(cex, top->middle);
        } else {
            assert(top->key);
            *end_node = top;
        }
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
cgul_trie__find_half_opened_range_by_prefix(cgul_exception_t* cex,
                                            cgul_trie_t trie,
                                            const char* prefix,
                                            cgul_trie_node_t* begin_node,
                                            cgul_trie_node_t* end_node)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the endpoints for the *closed* range. */
    if (!cgul_trie__find_closed_range_by_prefix(
            cex, trie, prefix, begin_node, end_node))
    {
        goto out;
    }

    /* Get the right endpoint for the half-open range. */
    if (end_node && *end_node) {
        *end_node = cgul_trie_node__get_next(cex, *end_node);
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
cgul_trie__find_keys_by_prefix(cgul_exception_t* cex,
                               cgul_trie_t trie,
                               const char* prefix,
                               const char*** keys,
                               unsigned long int* keys_size)
{
    int rv = 0;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t largest = NULL;
    cgul_vector_t v = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "keys" or "keys_size" which both can be NULL. */
    assert(trie);

    /* Clear the user's output parameters. */
    if (keys) {
        *keys = NULL;
    }
    if (keys_size) {
        *keys_size = 0;
    }

    /* Sanity check. */
    if (!prefix) {
        goto out;
    }

    /* Allocate a vector. */
    v = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Find the nodes for the half-opened range of keys under "top". */
    if (!cgul_trie__find_half_opened_range_by_prefix(
            cex, trie, prefix, &curr, &largest))
    {
        goto out;
    }

    /* Iterate over the half-opened range of nodes collecting keys. */
    for ( ; curr != largest ; curr = cgul_trie_node__get_next(cex, curr)) {

        /* Append the key for the next node to the vector. */
        cgul_vector__push_back(cex, v, curr->key);
        if (*cex) {
            goto out;
        }

    }

    /* Success. */
    if (keys_size) {
        *keys_size = cgul_vector__get_size(cex, v);
    }
    if (keys) {
        *keys = (const char**)cgul_vector__take_array(cex, v);
        if (*cex) {
            goto out;
        }
    }
    rv = 1;

 out:

    /* Clean up. */
    if (v) {
        /* Do not free the keys from the vector even if an exception
         * is thrown because they are owned by the trie. */
        cgul_vector__delete(v);
    }

    return rv;
}

int
cgul_trie__find_values_by_prefix(cgul_exception_t* cex,
                                 cgul_trie_t trie,
                                 const char* prefix,
                                 void*** values,
                                 unsigned long int* values_size)
{
    int rv = 0;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t largest = NULL;
    cgul_vector_t v = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "values" or "values_size" which both can be NULL. */
    assert(trie);

    /* Clear the user's output parameters. */
    if (values) {
        *values = NULL;
    }
    if (values_size) {
        *values_size = 0;
    }

    /* Sanity check. */
    if (!prefix) {
        goto out;
    }

    /* Allocate a vector. */
    v = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Find the nodes for the half-opened range of keys under "top". */
    if (!cgul_trie__find_half_opened_range_by_prefix(
            cex, trie, prefix, &curr, &largest))
    {
        goto out;
    }

    /* Iterate over the half-opened range of nodes collecting values. */
    for ( ; curr != largest ; curr = cgul_trie_node__get_next(cex, curr)) {

        /* Append the value for the next node to the vector. */
        cgul_vector__push_back(cex, v, curr->value);
        if (*cex) {
            goto out;
        }

    }

    /* Success. */
    if (values_size) {
        *values_size = cgul_vector__get_size(cex, v);
    }
    if (values) {
        *values = cgul_vector__take_array(cex, v);
        if (*cex) {
            goto out;
        }
    }
    rv = 1;

 out:

    /* Clean up. */
    if (v) {
        /* Do not free the values from the vector even if an exception
         * is thrown because they are owned by the trie. */
        cgul_vector__delete(v);
    }

    return rv;
}

int
cgul_trie__find_nodes_by_prefix(cgul_exception_t* cex,
                                cgul_trie_t trie,
                                const char* prefix,
                                cgul_trie_node_t** nodes,
                                unsigned long int* nodes_size)
{
    int rv = 0;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t largest = NULL;
    cgul_vector_t v = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "nodes" or "nodes_size" which both can be NULL. */
    assert(trie);

    /* Clear the user's output parameters. */
    if (nodes) {
        *nodes = NULL;
    }
    if (nodes_size) {
        *nodes_size = 0;
    }

    /* Sanity check. */
    if (!prefix) {
        goto out;
    }

    /* Allocate a vector. */
    v = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Find the nodes for the half-opened range of keys under "top". */
    if (!cgul_trie__find_half_opened_range_by_prefix(
            cex, trie, prefix, &curr, &largest))
    {
        goto out;
    }

    /* Iterate over the half-opened range of nodes collecting nodes. */
    for ( ; curr != largest ; curr = cgul_trie_node__get_next(cex, curr)) {

        /* Append the node for the next node to the vector. */
        cgul_vector__push_back(cex, v, curr);
        if (*cex) {
            goto out;
        }

    }

    /* Success. */
    if (nodes_size) {
        *nodes_size = cgul_vector__get_size(cex, v);
    }
    if (nodes) {
        *nodes = (cgul_trie_node_t*)cgul_vector__take_array(cex, v);
        if (*cex) {
            goto out;
        }
    }
    rv = 1;

 out:

    /* Clean up. */
    if (v) {
        /* Do not free the nodes from the vector even if an exception
         * is thrown because they are owned by the trie. */
        cgul_vector__delete(v);
    }

    return rv;
}

const char*
cgul_trie__find_longest_prefix_of(cgul_exception_t* cex,
                                  cgul_trie_t trie,
                                  const char* complete_string)
{
    size_t i = 0;
    const char* rv = NULL;
    cgul_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Sanity check. */
    if (!complete_string) {
        goto out;
    }

    /*
     * Special Case: Empty strings map directly to the root node.
     * Matching the "empty" character at the front of every string is
     * done here by hand.
     */
    if (trie->root->key) {
        rv = trie->root->key;
    }

    /* Take the middle branch.  (The root node should only ever have a
     * middle branch because the empty string is a prefix of all other
     * strings.) */
    assert(!trie->root->left);
    assert(!trie->root->right);
    curr = trie->root->middle;

    /* Descend the tree under the control of the "complete_string". */
    while (curr && complete_string[i]) {

        if (complete_string[i] < curr->ch) {
            curr = curr->left;
        } else if (complete_string[i] > curr->ch) {
            curr = curr->right;
        } else {
            /* Matching character found.  Check if this prefix of the
             * "complete_string" has been inserted into the tree.  If
             * so, remember it as the longest prefix. */
            if (curr->key) {
                rv = curr->key;
            }
            /* Move to next character. */
            ++i;
            curr = curr->middle;
        }

    }

 out:
    return rv;
}

cgul_trie_node_t
cgul_trie__get_front(cgul_exception_t* cex,
                     cgul_trie_t trie)
{
    cgul_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    rv = cgul_trie_node__get_smallest_node_in_subtree(cex, trie->root);

 out:
    return rv;
}

cgul_trie_node_t
cgul_trie__get_back(cgul_exception_t* cex,
                    cgul_trie_t trie)
{
    cgul_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    rv = cgul_trie_node__get_largest_node_in_subtree(cex, trie->root);

 out:
    return rv;
}

cgul_trie_node_t
cgul_trie__get_oldest(cgul_exception_t* cex,
                      cgul_trie_t trie)
{
    cgul_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    rv = trie->oldest;

 out:
    return rv;
}

void
cgul_trie__set_oldest(cgul_exception_t* cex,
                      cgul_trie_t trie,
                      cgul_trie_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(trie->oldest);
    assert(n);

    /* There is nothing to do if n is already the oldest. */
    if (n == trie->oldest) {
        goto out;
    }

    /* If n is currently the youngest, it will not be when this method
     * returns.  In this case, we need to remember which node is the
     * new youngest node. */
    if (n == trie->youngest) {
        trie->youngest = n->older;
    }

    /* Carefully remove n from the list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Insert n as the oldest node.  We assert trie->oldest above
     * because any non-empty trie has an oldest node, and because n is
     * a node in t, t is non-empty. */
    trie->oldest->older = n;
    n->younger = trie->oldest;
    n->older = NULL;

    /* Remember that n is now the oldest node. */
    trie->oldest = n;

 out:
    return;
}

cgul_trie_node_t
cgul_trie__get_youngest(cgul_exception_t* cex,
                        cgul_trie_t trie)
{
    cgul_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    rv = trie->youngest;

 out:
    return rv;
}

void
cgul_trie__set_youngest(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        cgul_trie_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(trie->youngest);
    assert(n);

    /* There is nothing to do if n is already the youngest. */
    if (n == trie->youngest) {
        goto out;
    }

    /* If n is currently the oldest, it will not be when this method
     * returns.  In this case, we need to remember which node is the
     * new oldest node. */
    if (n == trie->oldest) {
        trie->oldest = n->younger;
    }

    /* Carefully remove n from the list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Insert n as the youngest node.  We assert trie->youngest above
     * because any non-empty trie has an youngest node, and because n
     * is a node in t, t is non-empty. */
    trie->youngest->younger = n;
    n->older = trie->youngest;
    n->younger = NULL;

    /* Remember that n is now the youngest node. */
    trie->youngest = n;

 out:
    return;
}

int
cgul_trie__remove(cgul_exception_t* cex,
                  cgul_trie_t trie,
                  const char* key_in,
                  char** key_out,
                  void** value_out)
{
    int rv = 0;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Sanity check. */
    if (!key_in) {
        goto out;
    }

    /* Get the node for this key. */
    node = cgul_trie__find_terminus(cex,
                                    trie,
                                    trie->root,
                                    key_in,
                                    0 /* insert_missing_nodes */);
    if (*cex) {
        goto out;
    }
    if (!node) {
        goto out;
    }

    /* Remove the node. */
    cgul_trie__remove_node(cex, trie, node, key_out, value_out);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

void
cgul_trie__remove_node(cgul_exception_t* cex,
                       cgul_trie_t trie,
                       cgul_trie_node_t n,
                       char** key_out,
                       void** value_out)
{
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t parent = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(trie->cache);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Santy check. */
    if (!n) {
        goto out;
    }

    /* Remove the node from the younger/older linked list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }
    if (trie->oldest == n) {
        trie->oldest = n->younger;
    }
    if (trie->youngest == n) {
        trie->youngest = n->older;
    }

    /* Return the key and value to the caller. */
    if (key_out) {
        *key_out = (char*)n->key;
    }
    if (value_out) {
        *value_out = n->value;
    }

    /* Mark the node as no longer holding a key/value pair. */
    n->key = NULL;
    n->value = NULL;

    /* Remember that there is one less key in the trie.  Do this
     * before the while() loop partially removes the path back to the
     * root node. */
    --trie->size;
    if (trie->is_indexed) {
        /* Decrement counts by "incrementing" by -1. */
        cgul_trie__increment_parent_counts_by(cex, trie, n, -1);
    }

    /* Recurse up the trie removing empty nodes except for trie->root
     * which remains even in empty tries.  trie-root is skipped by
     * only working on nodes that have parents. */
    curr = n;
    while (curr && curr->parent) {

        /* If the current node is not empty, just leave it in place. */
        if (curr->left || curr->key || curr->middle || curr->right) {
            break;
        }

        /*
         * Otherwise, the current node is empty so delete it.
         */

        /* Dereference the current node while it is still valid. */
        parent = curr->parent;

        /* Clear the parents pointer to this node. */
        if (curr == parent->left) {
            parent->left = NULL;
        } else if (curr == parent->middle) {
            parent->middle = NULL;
        } else if (curr == parent->right) {
            parent->right = NULL;
        }

        /* Delete the current node. */
        cgul_cache__put(cex, trie->cache, curr);

        /* Recurse up the trie. */
        curr = parent;
    }

 out:
    return;
}

int
cgul_trie__remove_at(cgul_exception_t* cex,
                     cgul_trie_t trie,
                     unsigned long int index,
                     char** key_out,
                     void** value_out)
{
    int rv = 0;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Sanity check. */
    if (index >= trie->size) {
        goto out;
    }

    /* Get the node by index. */
    node = cgul_trie__find_at(cex, trie, index);
    if (*cex) {
        goto out;
    }
    if (!node) {
        goto out;
    }

    /* Remove the node. */
    cgul_trie__remove_node(cex, trie, node, key_out, value_out);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
cgul_trie__remove_front(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        char** key_out,
                        void** value_out)
{
    int rv = 0;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Get the front node. */
    node = cgul_trie__get_front(cex, trie);
    if (!node) {
        goto out;
    }

    /* Remove the front node. */
    cgul_trie__remove_node(cex, trie, node, key_out, value_out);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
cgul_trie__remove_back(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        char** key_out,
                        void** value_out)
{
    int rv = 0;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Get the back node. */
    node = cgul_trie__get_back(cex, trie);
    if (!node) {
        goto out;
    }

    /* Remove the back node. */
    cgul_trie__remove_node(cex, trie, node, key_out, value_out);

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
cgul_trie__remove_range(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        cgul_trie_node_t first,
                        cgul_trie_node_t last,
                        cgul_cache_t keys_cache,
                        cgul_cache_t values_cache)
{
    int rv = 0;
    struct cgul_trie__clear_cb_data data;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    data.keys_cache = keys_cache;
    data.values_cache = values_cache;
    data.rv = 0;

    cgul_trie__traverse_range(cex,
                              trie,
                              first,
                              last,
                              &cgul_trie__clear_cb,
                              &data);

    rv = data.rv;

 out:
    return rv;
}

void
cgul_trie__clear(cgul_exception_t* cex,
                 cgul_trie_t trie,
                 cgul_cache_t keys_cache,
                 cgul_cache_t values_cache)
{
    struct cgul_trie__clear_cb_data data;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    data.keys_cache = keys_cache;
    data.values_cache = values_cache;
    data.rv = 0;

    cgul_trie__traverse(cex,
                        trie,
                        &cgul_trie__clear_cb,
                        &data);

 out:
    return;
}

unsigned long int
cgul_trie__get_size(cgul_exception_t* cex,
                    cgul_trie_t trie)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    rv = trie->size;

 out:
    return rv;
}

void
cgul_trie__swap(cgul_exception_t* cex,
                cgul_trie_t trie1,
                cgul_trie_t trie2)
{
    struct cgul_trie tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie1);
    assert(trie2);

    /* Nothing to swap if they point to the same objects. */
    if (trie1 == trie2) {
        goto out;
    }

    /* A shallow copy, by design, is exactly what is needed. */
    tmp = *trie1;
    *trie1 = *trie2;
    *trie2 = tmp;

 out:
    return;
}

void
cgul_trie__foldl_keys(cgul_exception_t* cex,
                      cgul_trie_t trie,
                      cgul_trie__fold_key_t f,
                      void* data)
{
    int f_rv = 0;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(f);

    /* Iterate over each node in the tree. */
    curr = cgul_trie__get_front(cex, trie);
    for ( ; curr ; curr = next) {
        /* Get the next node now in case f deletes curr. */
        next = cgul_trie_node__get_next(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_trie__foldr_keys(cgul_exception_t* cex,
                      cgul_trie_t trie,
                      cgul_trie__fold_key_t f,
                      void* data)
{
    int f_rv = 0;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(f);

    /* Iterate over each node in the tree. */
    curr = cgul_trie__get_back(cex, trie);
    for ( ; curr; curr = prev) {
        /* Get the prev node now in case f deletes curr. */
        prev = cgul_trie_node__get_prev(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_trie__foldl_values(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        cgul_trie__fold_value_t f,
                        void* data)
{
    int f_rv = 0;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(f);

    /* Iterate over each node in the tree. */
    curr = cgul_trie__get_front(cex, trie);
    for ( ; curr ; curr = next) {
        /* Get the next node now in case f deletes curr. */
        next = cgul_trie_node__get_next(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_trie__foldr_values(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        cgul_trie__fold_value_t f,
                        void* data)
{
    int f_rv = 0;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(f);

    /* Iterate over each node in the tree. */
    curr = cgul_trie__get_back(cex, trie);
    for ( ; curr; curr = prev) {
        /* Get the prev node now in case f deletes curr. */
        prev = cgul_trie_node__get_prev(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_trie__foldl_pairs(cgul_exception_t* cex,
                       cgul_trie_t trie,
                       cgul_trie__fold_pair_t f,
                       void* data)
{
    int f_rv = 0;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(f);

    /* Iterate over each node in the tree. */
    curr = cgul_trie__get_front(cex, trie);
    for ( ; curr ; curr = next) {
        /* Get the next node now in case f deletes curr. */
        next = cgul_trie_node__get_next(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_trie__foldr_pairs(cgul_exception_t* cex,
                       cgul_trie_t trie,
                       cgul_trie__fold_pair_t f,
                       void* data)
{
    int f_rv = 0;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);
    assert(f);

    /* Iterate over each node in the tree. */
    curr = cgul_trie__get_back(cex, trie);
    for ( ; curr ; curr = prev) {
        /* Get the prev node now in case f deletes curr. */
        prev = cgul_trie_node__get_prev(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_trie__traverse(cgul_exception_t* cex,
                    cgul_trie_t trie,
                    cgul_trie__traverse_t f,
                    void* data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_trie__traverse_range(cex, trie, NULL, NULL, f, data);

 out:
    return;
}

void
cgul_trie__traverse_range(cgul_exception_t* cex,
                          cgul_trie_t trie,
                          cgul_trie_node_t first,
                          cgul_trie_node_t last,
                          cgul_trie__traverse_t f,
                          void* data)
{
    int f_rv = 0;
    cgul_trie_node_t curr = NULL;
    cgul_trie_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert(first) nor assert(last) because NULL values
     * indicate that we should begin or end with the first or last
     * node of the entire tree respectively. */
    assert(trie);
    assert(f);

    if (first == NULL) {
        first = cgul_trie__get_front(cex, trie);
    }

    if (last == NULL) {
        last = cgul_trie__get_back(cex, trie);
    }

    /* In order to provide "f" with a context in which it is safe for
     * "f" to delete curr, you have to be careful to dereference
     * curr before calling "f". */
    for (curr = first ; curr ; curr = next) {
        next = cgul_trie_node__get_next(cex, curr);
        f_rv = (*f)(cex, trie, curr, data);
        if (*cex) {
            goto out;
        }
        /* We are done if f returned false or if we just processed the
         * last node. */
        if (!f_rv || (curr == last)) {
            break;
        }
    }

 out:
    return;
}
