#include "cgul_wstring_p.h"
#include "cgul_defensive.h"
#include "cgul_libc.h"
#include "cgul_unicode.h"

/* The default capacity for each newly allocated string.  This is the
 * number of characters _not_ including the NUL terminator.  When
 * realloc() is called one is added to the capacity to make room for
 * the NUL terminator. */
static const size_t CGUL_WSTRING__DEFAULT_MINIMUM_CAPACITY = 128;

/* Comments are in cgul_wstring_p.h. */
void
cgul_wstring__set_capacity(cgul_exception_t* cex,
                           cgul_wstring_t ws,
                           size_t capacity)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Sanity check. */
    if (capacity == 0) {
        capacity = ws->minimum_capacity;
    }

    /* Make sure the capacity is at least large enough to hold the
     * current number of elements plus 1 for the NUL terminator. */
    capacity = CGUL_MAX(capacity, ws->length + 1);

#if 0
    fprintf(stderr, "cgul_wstring__set_capacity = %lu\n", capacity);
    fflush(stderr);
#endif

    /* realloc */
    ws->value = (cgul_wchar_t*)cgul_libc__realloc(
                    cex,
                    ws->value,
                    sizeof(*ws->value) * capacity);
    if (*cex) {
        goto out;
    }

    /* Rememeber the new capacity. */
    ws->capacity = capacity;

    /* Clear the extra elements. */
    if (capacity > ws->length) {
        memset(ws->value + ws->length,
               0,
               sizeof(*ws->value) * (ws->capacity - ws->length));
    }

 out:
    return;
}

/* Comments are in cgul_wstring_p.h. */
void
cgul_wstring__decrease_capacity_if_necessary(cgul_exception_t* cex,
                                             cgul_wstring_t ws)
{
    size_t actual_new_capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* We only have excess capacity when 1/4 capacity is greater than
     * or equal to the size.  So, do nothing when 1/4 capacity is less
     * than the size. */
    if ((ws->capacity >> 2) < (ws->length + 1)) {
        goto out;
    }

    /* Keep halving the capacity to get the best fit.  It should be at
     * least double the size to avoid thrashing the allocator.  We
     * know from the if() above that we can start iterating with 1/2
     * of ws->capacity. */
    actual_new_capacity = ws->capacity >> 1;
    while (   ((actual_new_capacity >> 1) > 0)
           && ((actual_new_capacity >> 1) >= ((ws->length + 1) << 1)))
    {
        actual_new_capacity >>= 1;
    }

    /* The actual new capacity should never drop below the minimum
     * capacity. */
    actual_new_capacity = CGUL_MAX(actual_new_capacity, ws->minimum_capacity);
    assert(actual_new_capacity > 0);

    /* If the capacity is already less than or equal to the actual new
     * capacity, do nothing. */
    if (ws->capacity <= actual_new_capacity) {
        goto out;
    }

    /* Sanity check to make sure we are not going to accidentally
     * reduce the capacity below the necessary size. */
    if (actual_new_capacity <= (ws->length + 1)) {
        goto out;
    }

    /* Set the reduced capacity. */
    cgul_wstring__set_capacity(cex, ws, actual_new_capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Wrapper around <code>cgul_wstring__increase_capacity_if_necessary()</code>
 * that performs a quick check to determine if there is already enough
 * excess capacity to accommodate the extra size.  If not, the real
 * method is called.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] extra_size extra size in characters excluding NUL-terminator
 */
#define CGUL_WSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, ws, extra_size) \
do {                                                                    \
    size_t excess_capacity = (ws)->capacity - (ws)->length - 1;         \
    size_t local_extra_size = (extra_size);                             \
    if (excess_capacity < local_extra_size) {                           \
        cgul_wstring__increase_capacity_if_necessary(                   \
            (cex), (ws), local_extra_size);                             \
    }                                                                   \
} while (0)

/* Comments are in cgul_wstring_p.h. */
void
cgul_wstring__increase_capacity_if_necessary(cgul_exception_t* cex,
                                             cgul_wstring_t ws,
                                             size_t extra_size)
{
    size_t needed_capacity = 0;
    size_t target_capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Sanity check. */
    if (extra_size == 0) {
        goto out;
    }

    /*
     * NOTE:
     * ====
     *
     * It is more efficient (on average) to call this method
     * through the CGUL_WSTRING__INCREASE_CAPACITY_IF_NECESSARY() macro
     * defined above.
     */

    /* Add 1 to extra_size to account for the NUL terminator. */
    extra_size = cgul_defensive__add_size_type(cex, extra_size, 1);
    if (*cex) {
        goto out;
    }

    /* Calculate the needed capacity. */
    needed_capacity =
        cgul_defensive__add_size_type(cex, ws->length, extra_size);
    if (*cex) {
        goto out;
    }

    /* If we need more capacity, realloc() it.*/
    if (needed_capacity > ws->capacity) {

        /* Try to double the capacity to allow room to grow.  The test
         * against zero is used to detect overflow. */
        target_capacity = ws->capacity;
        while (   ((target_capacity << 1) > 0)
               && ((target_capacity << 1) <= (needed_capacity << 1)))
        {
            target_capacity <<= 1;
        }

        /* If there is overflow, just use needed_capacity as-is which
         * doesn't allow for room to grow without reallocating but is
         * large enough to hold the new elements.  Also, we know that
         * needed_capacity did not overflow when calculated above
         * because an exception would have been thrown. */
        if (target_capacity < needed_capacity) {
            target_capacity = needed_capacity;
        }

        /* Set the new capacity. */
        cgul_wstring__set_capacity(cex, ws, target_capacity);
    }

 out:
    return;
}

/**
 * Initialize <code>ws</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws cgul wide-character string
 * @param[in] value initial value
 */
static void
cgul_wstring__initialize(cgul_exception_t* cex,
                         cgul_wstring_t ws,
                         const cgul_wchar_t* value)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Free the old underlying array if any. */
    if (ws->value) {
        free(ws->value);
    }

    /* Initialize all member variables. */
    ws->value = NULL;
    ws->length = 0;
    ws->capacity = 0;
    ws->minimum_capacity = CGUL_WSTRING__DEFAULT_MINIMUM_CAPACITY;

    /* Initialize the capacity. */
    cgul_wstring__set_capacity(cex, ws, ws->minimum_capacity);
    if (*cex) {
        goto out;
    }

    /* Now that it is safe to call methods, use set_value() to set the value. */
    cgul_wstring__set_value(cex, ws, value);

 out:
    return;
}

cgul_wstring_t
cgul_wstring__new(cgul_exception_t* cex)
{
    cgul_wstring_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_wstring_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    cgul_wstring__initialize(cex, rv, CGUL_WCHAR__EMPTY_STRING);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_wstring__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_wstring_t
cgul_wstring__new_from_wcstring(cgul_exception_t* cex,
                                const cgul_wchar_t* value)
{
    cgul_wstring_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(value);

    /* Create an empty string object. */
    rv = cgul_wstring__new(cex);
    if (*cex) {
        goto out;
    }

    /* Set the value. */
    cgul_wstring__set_value(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_wstring__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_wstring__delete(cgul_wstring_t ws)
{
    if (ws) {
        if (ws->value) {
            free(ws->value);
        }
        free(ws);
    }
}

cgul_wchar_t*
cgul_wstring__get_value(cgul_exception_t* cex,
                        cgul_wstring_t ws)
{
    cgul_wchar_t* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    rv = ws->value;

 out:
    return rv;
}

cgul_wchar_t*
cgul_wstring__take_value(cgul_exception_t* cex, cgul_wstring_t ws)
{
    cgul_wchar_t* rv= NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(ws->value);

    /* The client is taking ownership of ws->value. */
    rv = ws->value;
    ws->value = NULL;

    /* Reset this object. */
    cgul_wstring__initialize(cex, ws, CGUL_WCHAR__EMPTY_STRING);

 out:
    return rv;
}

void
cgul_wstring__set_value(cgul_exception_t* cex,
                        cgul_wstring_t ws,
                        const cgul_wchar_t* value)
{
    int is_interior_ptr = 0;
    size_t i = 0;
    size_t value_length = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(ws->value);
    assert(value);

    /* Do not copy the string to itself. */
    if (ws->value == value) {
        goto out;
    }

    /* Determine if "value" is a pointer to a substring inside of ws->value. */
    is_interior_ptr = (   (value >= ws->value)
                       && (value <= ws->value + ws->length));

    /* Make sure the underlying buffer is large enough but only if
     * "value" is not an interior pointer.  Without this extra check,
     * the realloc() call could invalidate "value", and in this case
     * there is no reason to call realloc() because we know ws->value
     * is large enough to hold any of its own substrings. */
    if (!is_interior_ptr) {
        value_length = cgul_wchar__wcslen(cex, value);
        CGUL_WSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, ws, value_length);
        if (*cex) {
            goto out;
        }
    }

    /* Copy "value" to "ws->value" including the NUL terminator. */
    for (i = 0 ; i <= value_length ; ++i) {
        ws->value[i] = value[i];
    }
    ws->length = value_length;

    /* See if the capacity can be decreased.  (Note that we *increase*
     * the capacity if necessary before attempting to copy, but
     * sometimes we are setting the string to a shorter value.) */
    cgul_wstring__decrease_capacity_if_necessary(cex, ws);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

cgul_wchar_t
cgul_wstring__get_value_at(cgul_exception_t* cex,
                           cgul_wstring_t ws,
                           size_t index)
{
    cgul_wchar_t rv = CGUL_WCHAR__NUL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    if (index >= ws->length) {
        *cex = cgul_exception__new("cgul_wstring__get_value_at: "
                                   "index out of bounds: %lu",
                                   (unsigned long)index);
        goto out;
    }

    rv = ws->value[index];

 out:
    return rv;
}

extern void
cgul_wstring__set_value_at(cgul_exception_t* cex,
                           cgul_wstring_t ws,
                           size_t index,
                           cgul_wchar_t wc)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Verify that "index" is within bounds. */
    if (index >= ws->length) {
        *cex = cgul_exception__new("cgul_wstring__set_value_at: "
                                   "index out of bounds: %lu",
                                   (unsigned long)index);
        goto out;
    }

    ws->value[index] = wc;

 out:
    return;
}

/* This method is part of the "cacheable" interface defined by
 * "cgul_cache".  It is required to not throw any exception. */
void
cgul_wstring__clear(cgul_wstring_t ws)
{
    cgul_exception_t local = NULL;
    cgul_wstring__set_length(&local, ws, 0);
}

signed char
cgul_wstring__to_schar(cgul_exception_t* cex,
                       cgul_wstring_t ws,
                       int base)
{
    signed char rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__wcstombs(cex, ws->value, ws->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_schar(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

unsigned char
cgul_wstring__to_uchar(cgul_exception_t* cex,
                       cgul_wstring_t ws,
                       int base)
{
    unsigned char rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__wcstombs(cex, ws->value, ws->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_uchar(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

short int
cgul_wstring__to_short(cgul_exception_t* cex,
                       cgul_wstring_t ws,
                       int base)
{
    short int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__wcstombs(cex, ws->value, ws->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_short(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

unsigned short int
cgul_wstring__to_ushort(cgul_exception_t* cex,
                        cgul_wstring_t ws,
                        int base)
{
    unsigned short int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__wcstombs(cex, ws->value, ws->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_ushort(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

int
cgul_wstring__to_int(cgul_exception_t* cex,
                     cgul_wstring_t ws,
                     int base)
{
    int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__wcstombs(cex, ws->value, ws->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_int(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

unsigned int
cgul_wstring__to_unsigned(cgul_exception_t* cex,
                          cgul_wstring_t ws,
                          int base)
{
    unsigned int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__wcstombs(cex, ws->value, ws->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_unsigned(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

long int
cgul_wstring__to_long(cgul_exception_t* cex,
                      cgul_wstring_t ws,
                      int base)
{
    long int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__wcstombs(cex, ws->value, ws->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_long(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

unsigned long int
cgul_wstring__to_unsigned_long(cgul_exception_t* cex,
                               cgul_wstring_t ws,
                               int base)
{
    unsigned long int rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__wcstombs(cex, ws->value, ws->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_unsigned_long(cex, tmp, base);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

double
cgul_wstring__to_double(cgul_exception_t* cex,
                        cgul_wstring_t ws)
{
    double rv = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert to UTF-8. */
    cgul_unicode__wcstombs(cex, ws->value, ws->length, /* skip_bom */ 1, tmp);
    if (*cex) {
        goto out;
    }

    /* Do the numberic conversion. */
    rv = cgul_string__to_double(cex, tmp);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

void
cgul_wstring__append(cgul_exception_t* cex,
                     cgul_wstring_t ws,
                     const cgul_wchar_t* value)
{
    size_t value_length = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(value);

    /* Calculate the string length of "value". */
    value_length = cgul_wchar__wcslen(cex, value);

    /* Add capacity if necessary. */
    CGUL_WSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, ws, value_length);
    if (*cex) {
        goto out;
    }

    /* Because we have already calculated the length of "value", we
     * can use memcpy(). */
    memcpy(&ws->value[ws->length], value, value_length * sizeof(cgul_wchar_t));

    /* Calculate the new length. */
    ws->length += value_length;

    /* Null-terminate the string. */
    ws->value[ws->length] = CGUL_WCHAR__NUL;

 out:
    return;
}

void
cgul_wstring__append_block(cgul_exception_t* cex,
                           cgul_wstring_t ws,
                           const void* block,
                           size_t block_size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(ws->value);
    assert(block);

    CGUL_WSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, ws, block_size);
    if (*cex) {
        goto out;
    }

    memcpy(&ws->value[ws->length], block, block_size * sizeof(cgul_wchar_t));
    ws->length += block_size;
    ws->value[ws->length] = CGUL_WCHAR__NUL;

 out:
    return;
}

void
cgul_wstring__append_char(cgul_exception_t* cex,
                          cgul_wstring_t ws,
                          cgul_wchar_t c)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(ws->value);

    CGUL_WSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, ws, 1);
    if (*cex) {
        goto out;
    }

    ws->value[ws->length] = c;
    ++ws->length;
    ws->value[ws->length] = CGUL_WCHAR__NUL;

 out:
    return;
}

void
cgul_wstring__insert(cgul_exception_t* cex,
                     cgul_wstring_t ws,
                     size_t index,
                     const cgul_wchar_t* value,
                     size_t value_length)
{
    size_t i = 0;
    size_t i_max = 0;
    size_t src = 0;
    size_t dst = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(value);

    /* Verify that "index" is within bounds. */
    if (index > ws->length) {
        *cex = cgul_exception__new("cgul_wstring__insert: "
                                   "index out of bounds: %lu > %lu",
                                   (unsigned long)index,
                                   (unsigned long)ws->length);
        goto out;
    }

    /* Add extra capacity if necessary to accommodate the extra characters. */
    CGUL_WSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, ws, value_length);
    if (*cex) {
        goto out;
    }

    /* Shift the characters after and including the insertion point
     * down to the new end of the wstring.  Notice that this includes
     * shifting the NUL terminator. */
    src = ws->length;
    dst = ws->length + value_length;
    i_max = ws->length - index;
    for (i = 0 ; i <= i_max ; ++i, --src, --dst) {
        ws->value[dst] = ws->value[src];
    }

    /* Insert the characters from "value". */
    for (i = 0 ; i < value_length ; ++i) {
        ws->value[index + i] = value[i];
    }

    /* Calculate the new length. */
    ws->length += value_length;

    /* Null-terminate the wide-character string. */
    ws->value[ws->length] = CGUL_WCHAR__NUL;

 out:
    return;
}

void
cgul_wstring__insert_char(cgul_exception_t* cex,
                          cgul_wstring_t ws,
                          size_t index,
                          cgul_wchar_t value)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(value);

    /* Have cgul_wstring__insert() do the work. */
    cgul_wstring__insert(cex, ws, index, &value, 1);

 out:
    return;
}

void
cgul_wstring__pad_right(cgul_exception_t* cex,
                        cgul_wstring_t ws,
                        size_t width,
                        cgul_wchar_t pad)
{
    size_t i = 0;
    size_t j = 0;
    size_t pad_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(ws->value);

    if (ws->length < width) {

        /* Calculate the size of the padding. */
        pad_size = width - ws->length;

        /* Add capacity if necessary. */
        CGUL_WSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, ws, pad_size);
        if (*cex) {
            goto out;
        }

        /* Shift the string right. */
        for (i = ws->length, j = width ; i > 0 ; --i, --j) {
            ws->value[j - 1] = ws->value[i - 1];
        }

        /* Insert the padding and adjust ws->length along the way. */
        for (i = 0 ; i < pad_size ; ++i) {
            ws->value[i] = pad;
        }
        ws->length = width;
        ws->value[ws->length] = CGUL_WCHAR__NUL;
    }

 out:
    return;
}

void
cgul_wstring__pad_left(cgul_exception_t* cex,
                       cgul_wstring_t ws,
                       size_t width,
                       cgul_wchar_t pad)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(ws->value);

    if (ws->length < width) {
        size_t i = 0;
        size_t pad_size = width - ws->length;

        CGUL_WSTRING__INCREASE_CAPACITY_IF_NECESSARY(cex, ws, pad_size);
        if (*cex) {
            goto out;
        }

        /* Insert the padding and adjust ws->length along the way. */
        for ( i = 0 ; i < pad_size ; ++i ) {
            ws->value[ws->length++] = pad;
        }
        ws->value[ws->length] = CGUL_WCHAR__NUL;
    }

 out:
    return;
}

void
cgul_wstring__trim(cgul_exception_t* cex,
                   cgul_wstring_t ws)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Trim the string.  Because padding right does not require a
     * memmove(), it is done first in the hope that it will reduce the
     * amount of work memmove() has to do for trim_left(). */
    cgul_wstring__trim_right(cex, ws);
    cgul_wstring__trim_left(cex, ws);

 out:
    return;
}

void
cgul_wstring__trim_left(cgul_exception_t* cex,
                        cgul_wstring_t ws)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Find the first character in ws that is not white space. */
    for (i = 0 ; i < ws->length ; ++i) {
        if (!cgul_wchar__isspace(cex, ws->value[i])) {
            break;
        }
    }

    /* Nothing to do if there is no leading white space. */
    if (i == 0) {
        goto out;
    }

    /* Trim the string. */
    memmove(ws->value, &ws->value[i], (ws->length - i) * sizeof(cgul_wchar_t));
    ws->length -= i;
    ws->value[ws->length] = CGUL_WCHAR__NUL;

    /* Decrease capacity if necessary. */
    cgul_wstring__decrease_capacity_if_necessary(cex, ws);

 out:
    return;
}

void
cgul_wstring__trim_right(cgul_exception_t* cex,
                         cgul_wstring_t ws)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Iterate in reverse order over the characters in ws.  Note that
     * "i" is always one beyond the character we are checking making
     * it easier to test to make sure we do not read off the front of
     * the string. */
    for (i = ws->length ; i > 0 ; --i) {
        if (!cgul_wchar__isspace(cex, ws->value[i - 1])) {
            break;
        }
    }

    /* Nothing to do if there is no trailing white space. */
    if (i == ws->length) {
        goto out;
    }

    /* Trim the string. */
    ws->length = i;
    ws->value[ws->length] = CGUL_WCHAR__NUL;

    /* Decrease capacity if necessary. */
    cgul_wstring__decrease_capacity_if_necessary(cex, ws);

 out:
    return;
}

void
cgul_wstring__reverse(cgul_exception_t* cex,
                      cgul_wstring_t ws)
{
    size_t i = 0;
    size_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(ws->value);

    /* If the string is empty or has only one charcter, there is
     * nothing to reverse. */
    if (ws->length <= 1) {
        goto out;
    }

    for (j = ws->length - 1 ; i < j ; ++i, --j) {
        cgul_wchar_t tmp = ws->value[i];
        ws->value[i] = ws->value[j];
        ws->value[j] = tmp;
    }

 out:
    return;
}

size_t
cgul_wstring__get_length(cgul_exception_t* cex,
                         cgul_wstring_t ws)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    rv = ws->length;

 out:
    return rv;
}

size_t
cgul_wstring__set_length(cgul_exception_t* cex,
                         cgul_wstring_t ws,
                         size_t length)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Throw an exception if the user tries to increase the length. */
    if (length > ws->length) {
        *cex = cgul_exception__new("cgul_wstring__set_length: "
                                   "attempt to increase length");
        goto out;
    }

    /* Return the previous length. */
    rv = ws->length;

    /* Truncate value. */
    ws->value[length] = CGUL_WCHAR__NUL;

    /* Set the new length. */
    ws->length = length;

    /* Decrease capacity if necessary. */
    cgul_wstring__decrease_capacity_if_necessary(cex, ws);

 out:
    return rv;
}

void
cgul_wstring__externally_adjusted(cgul_exception_t* cex,
                                  cgul_wstring_t ws)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);
    assert(ws->value);

    ws->length = cgul_wchar__wcslen(cex, ws->value);
    cgul_wstring__decrease_capacity_if_necessary(cex, ws);

 out:
    return;
}

void
cgul_wstring__trim_capacity(cgul_exception_t* cex,
                            cgul_wstring_t ws)
{
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Use realloc() to trim the buffer. */
    ws->value =
        (cgul_wchar_t*)cgul_libc__realloc(
                           cex,
                           ws->value,
                           sizeof(cgul_wchar_t) * (ws->length + 1));
    if (*cex) {
        goto out;
    }

    /* Remember the new capacity. */
    ws->capacity = ws->length + 1;

 out:
    return;
}

size_t
cgul_wstring__get_minimum_capacity(cgul_exception_t* cex,
                                   cgul_wstring_t ws)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Return the minimum capacity to the user. */
    rv = ws->minimum_capacity;

 out:
    return rv;
}

void
cgul_wstring__set_minimum_capacity(cgul_exception_t* cex,
                                   cgul_wstring_t ws,
                                   size_t minimum_capacity)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws);

    /* Set the new minimum_capacity. */
    if (minimum_capacity == 0) {
        ws->minimum_capacity = CGUL_WSTRING__DEFAULT_MINIMUM_CAPACITY;
    } else {
        ws->minimum_capacity = minimum_capacity;
    }

    /* Realloc the underlying buffer (if necessary). */
    cgul_wstring__set_capacity(cex, ws, ws->minimum_capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

int
cgul_wstring__compare(cgul_exception_t* cex,
                      cgul_wstring_t ws1,
                      cgul_wstring_t ws2)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws1);
    assert(ws2);

    /* Compare the strings. */
    rv = cgul_wchar__wcscmp(ws1->value, ws2->value);

 out:
    return rv;
}

void
cgul_wstring__swap(cgul_exception_t* cex,
                   cgul_wstring_t ws1,
                   cgul_wstring_t ws2)
{
    struct cgul_wstring tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(ws1);
    assert(ws2);

    /* Nothing to swap if they point to the same objects. */
    if (ws1 == ws2) {
        goto out;
    }

    memcpy((void*)&tmp, (void*)ws1,  sizeof(tmp));
    memcpy((void*)ws1,  (void*)ws2,  sizeof(tmp));
    memcpy((void*)ws2,  (void*)&tmp, sizeof(tmp));

 out:
    return;
}
