/*
 * You can use this program to test the cgul_bitwise_trie class.  The code is
 * set up to read a command string (like "a" for add, "r" for remove,
 * or "pc" for print chronologically) followed by an integer for the
 * value you would like to insert or remove.
 *
 * You can use "test" as the command.  This will cause this code to
 * randomly insert TEST_POOL_SIZE integers into the test trie.  It
 * will then randomly remove each integer.
 *
 * You need to type "q" for quit.
 */

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include "cgul.h"
#include "cgul_rbtree.h"
#include "cgul_bitwise_trie_p.h"
#include "cgul_bitwise_trie_node_p.h"

#define TEST_POOL_SIZE (10240)

static int
myintcmp(const void* k1, const void* k2)
{
    return (long int)k1 - (long int)k2;
}

/**
 * Get the next word from the input file <code>f</code> and assign it
 * to the string <code>word</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f input file
 * @param[in,out] word next word
 */
static int
get_next_word(cgul_exception_t* cex,
              FILE* f,
              cgul_string_t word)
{
    int c;

    assert(cex);
    if (*cex) {
        goto out;
    }

    for (;;) {
        c = fgetc(f);
        if (c == EOF) {
            return 0;
        }
        if (!isspace(c)) {
            break;
        }
    }
    cgul_string__set_value(cex, word, "");
    cgul_string__append_char(cex, word, (char)c);
    for (;;) {
        c = fgetc(f);
        if (c == EOF) {
            return 0;
        }
        if (isspace(c)) {
            break;
        }
        cgul_string__append_char(cex, word, (char)c);
    }

 out:
    return 1;
}

static int
get_next_ulong(cgul_exception_t* cex,
               FILE* f,
               cgul_string_t w,
               unsigned long* n)
{
    int rv = 0;
    unsigned long tmp = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_word(cex, stdin, w)) {
        tmp = cgul_string__to_unsigned_long(cex, w, 0 /* base */);
        if (*cex) {
            goto out;
        }
        *n = tmp;
        rv = 1;
    }

 out:
    return rv;
}

static void
print_trie_node_with_ints(cgul_exception_t* cex,
                          cgul_bitwise_trie_node_t n,
                          int is_root)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    if (is_root) {
        printf("<root>");
    } else {
        printf("%-6ld", (long int)n->index);
    }
    if (n->children[0]) {
        printf(" | l = %-6ld", (long int)n->children[0]->index);
    } else {
        printf(" | l = %-6s", "NULL");
    }
    if (n->children[1]) {
        printf(" | r = %-6ld", (long int)n->children[1]->index);
    } else {
        printf(" | r = %-6s", "NULL");
    }
    puts("");

 out:
    return;
}

#if 0
static void
print_trie_with_ints_for_keys(cgul_exception_t* cex,
                              cgul_bitwise_trie_t t)
{
    cgul_bitwise_trie_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* The root node is a dummy node because it does not correspond to
     * any index value; however, its left and right children are
     * valid.  See comments at the top of src/cgul_bitwise_trie.c. */
    print_trie_node_with_ints(cex, t->root, 1 /* is_root */);

    n = cgul_bitwise_trie__get_oldest(cex, t);
    for ( ; n ; n = cgul_bitwise_trie_node__get_younger(cex, n)) {
        print_trie_node_with_ints(cex, n, 0 /* is_root */);
    }

 out:
    return;
}
#endif

static void
print_chronologically_trie_with_ints_for_keys(cgul_exception_t* cex,
                                              cgul_bitwise_trie_t t)
{
    cgul_bitwise_trie_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* The root node is a dummy node because it does not correspond to
     * any index value; however, its left and right children are
     * valid.  See comments at the top of src/cgul_bitwise_trie.c. */
    print_trie_node_with_ints(cex, t->root, 1 /* is_root */);

    n = cgul_bitwise_trie__get_oldest(cex, t);
    for ( ; n != NULL ; n = cgul_bitwise_trie_node__get_younger(cex, n)) {
        print_trie_node_with_ints(cex, n, 0 /* is_root */);
    }

    if (t->oldest) {
        printf("oldest = %ld\n", (long int)t->oldest->index);
    }

    if (t->youngest) {
        printf("youngest = %ld\n", (long int)t->youngest->index);
    }

 out:
    return;
}

static void
find_ulong(cgul_exception_t* cex,
           cgul_bitwise_trie_t t,
           cgul_string_t s)
{
    unsigned long int i = 0;
    unsigned long int* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_ulong(cex, stdin, s, &i)) {
        value = (unsigned long*)cgul_bitwise_trie__at(cex, t, i, NULL);
        if (!value) {
            fprintf(stderr, "Warning: Unable to find: %lu\n", i);
            goto out;
        }
        cgul_libc__printf(cex, "found: %lu\n", *value);
    }

 out:
    return;
}

static void
assign_int(cgul_exception_t* cex,
           cgul_bitwise_trie_t t,
           cgul_string_t s)
{
    unsigned long int i = 0;
    unsigned long int* value = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_ulong(cex, stdin, s, &i)) {

        /* Allocate value to assign to the trie. */
        value = cgul_libc__malloc(cex, sizeof(*value));
        if (*cex) {
            goto out;
        }
        *value = i;

        /* Assign. */
        cgul_bitwise_trie__assign(cex, t, i, value);
        if (*cex) {
            goto out;
        }

        /* The trie now owns the value. */
        value = NULL;

        /* Print result. */
        cgul_libc__printf(cex, "assigned: %lu\n", i);
    }

 out:

    /* Clean up. */
    if (value) {
        free(value);
    }

    return;
}

static void
remove_int(cgul_exception_t* cex,
           cgul_bitwise_trie_t t,
           cgul_string_t s)
{
    unsigned long int i = 0;
    unsigned long int* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_ulong(cex, stdin, s, &i)) {
        if (!cgul_bitwise_trie__remove(cex, t, i, (void**)&value)) {
            fprintf(stderr,
                    "Warning: Unable to remove missing index: %lu\n",
                    i);
            goto out;
        }
        free(value);
    }

 out:
    return;
}

static void
reload_pool(cgul_exception_t* cex, cgul_rbtree_t pool)
{
    int insert_rv = 0;
    unsigned long int i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    for (i = 0 ; i < TEST_POOL_SIZE ; ++i) {
        insert_rv = cgul_rbtree__insert(cex, pool, (void*)i, NULL);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            cgul_libc__fprintf(cex, stderr, "***\n");
            cgul_libc__fprintf(cex, stderr,
                 "*** cgul_rbtree__insert() failed while initializing pool\n");
            cgul_libc__fprintf(cex, stderr, "***\n");
            exit(1);
        }
    }

 out:
    return;
}

static void
verify_trie_is_full(cgul_exception_t* cex,
                    cgul_bitwise_trie_t t)
{
    int exists = 0;
    unsigned long int i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    for (i = 0 ; i < TEST_POOL_SIZE ; ++i) {
        cgul_bitwise_trie__at(cex, t, i, &exists);
        if (!exists) {
            cgul_libc__fprintf(cex, stderr,
                           "*** Error: trie did not verify as being full.\n");
            exit(1);
        }
    }

 out:
    return;
}

/* This function randomly removes a node from pool and inserts the
 * same key into t. */
static void
randomly_fill_test_trie(cgul_exception_t* cex,
                        cgul_bitwise_trie_t t,
                        cgul_rbtree_t pool,
                        cgul_random_cmwc32_t prng)
{
    cgul_uint32_t i = 0;
    cgul_uint32_t next = 0;
    long unsigned int value = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (cgul_rbtree__get_size(cex, pool) != TEST_POOL_SIZE) {
        fprintf(stderr, "*** Error: pool is not full.\n");
        exit(1);
    }

    if (cgul_bitwise_trie__get_size(cex, t) != 0) {
        fprintf(stderr, "*** Error: t is not empty.\n");
        exit(1);
    }

    for (i = TEST_POOL_SIZE ; i > 0 ; --i) {

        /* The idea is to have prng generate a number between 0 and
         * pool->size - 1.  We then use remove_at() (not remove()) to
         * remove the node at the index returned by prng and insert
         * the same key into t.  We then have prng generate the next
         * random number taking into account that we just removed a
         * node. */
        next = cgul_random_cmwc32__next_in_unit(cex, prng) * i;

#if 0
        printf("insert %d\n", next);
#endif

        /* Remove the value at the "next" index from the pool */
        if (!cgul_rbtree__remove_at(cex, pool, next, (void**)&value, NULL)) {
            fprintf(stderr, "*** Error: Unable to remove pool[%u]\n", next);
            exit(1);
        }

        /* Assign that same value to the trie. */
        cgul_bitwise_trie__assign(cex, t, value, NULL);
        if (*cex) {
            goto out;
        }
    }


    if (cgul_rbtree__get_size(cex, pool) != 0) {
        fprintf(stderr, "*** Error: pool is not empty.\n");
        exit(1);
    }

    if (cgul_bitwise_trie__get_size(cex, t) != TEST_POOL_SIZE) {
        fprintf(stderr, "*** Error: t is not full.\n");
        exit(1);
    }

 out:
    return;
}

/* This function randomly removes a node from pool and inserts the
 * same key into t. */
static void
randomly_empty_test_trie(cgul_exception_t* cex,
                         cgul_bitwise_trie_t t,
                         cgul_rbtree_t pool,
                         cgul_random_cmwc32_t prng)
{
    cgul_uint32_t i = 0;
    cgul_uint32_t next = 0;
    unsigned long int value = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (cgul_rbtree__get_size(cex, pool) != TEST_POOL_SIZE) {
        fprintf(stderr, "*** Error: pool is not full.\n");
        exit(1);
    }

    if (cgul_bitwise_trie__get_size(cex, t) != TEST_POOL_SIZE) {
        fprintf(stderr, "*** Error: t is not full.\n");
        exit(1);
    }

    for (i = TEST_POOL_SIZE ; i > 0 ; --i) {

        /* The idea is to have prng generate a number between 0 and
         * pool->size - 1.  We then use remove_at() (not remove()) to
         * remove the node at the index returned by prng and insert
         * the same key into t.  We then have prng generate the next
         * random number taking into account that we just removed a
         * node. */
        next = cgul_random_cmwc32__next_in_unit(cex, prng) * i;

#if 0
        printf("remove %d\n", next);
#endif

        /* Remove the value at the "next" index from the pool */
        if (!cgul_rbtree__remove_at(cex, pool, next, (void**)&value, NULL)) {
            fprintf(stderr,
                    "*** Error: Unable to find pool[%u].  Exiting!\n",
                    next);
            exit(1);
        }

        /* Remove that same value from the trie. */
        if (!cgul_bitwise_trie__remove(cex, t, value, NULL)) {
            fprintf(stderr,
                    "*** Error: Unable to remove index from t: %lu\n",
                    value);
            exit(1);
        }
    }

    if (cgul_rbtree__get_size(cex, pool) != 0) {
        fprintf(stderr, "*** Error: pool is not empty.\n");
        exit(1);
    }

    if (cgul_bitwise_trie__get_size(cex, t) != 0) {
        fprintf(stderr, "*** Error: t is not empty.\n");
        exit(1);
    }

 out:
    return;
}

static void
test_trie(cgul_exception_t* cex,
          cgul_random_cmwc32_t prng)
{
    unsigned long count = 0;
    cgul_rbtree_t pool = NULL;
    cgul_bitwise_trie_t t = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    pool = cgul_rbtree__new_with_indexing(cex, &myintcmp);
    t = cgul_bitwise_trie__new(cex);

    /*
     * Run the main loop.
     */
    for (count = 1; count <= 100 ; ++count) {
        printf("Starting loop %lu\n", count);
        printf("    reload_pool()\n");
        reload_pool(cex, pool);
        printf("    randomly_fill_test_trie()\n");
        randomly_fill_test_trie(cex, t, pool, prng);
        printf("    verify_trie_is_full(t)\n");
        verify_trie_is_full(cex, t);
        printf("    reload_pool()\n");
        reload_pool(cex, pool);
        printf("    randomly_empty_test_trie()\n");
        randomly_empty_test_trie(cex, t, pool, prng);
        printf("    Finished loop %lu\n", count);
    }

 out:

    if (t) {
        cgul_bitwise_trie__delete(t);
    }
    if (pool) {
        cgul_rbtree__delete(pool);
    }

    return;
}

int main(int argc, char* argv[])
{
    int rv = 0;
    char* cmd = NULL;
    cgul_uint32_t seed = 0;
    cgul_bitwise_trie_t t = NULL;
    cgul_string_t s = NULL;
    cgul_random_cmwc32_t prng = NULL;
    cgul_exception_t local = NULL;

    /* Generate and print the seed for the PRNG. */
#if 1
    seed = (cgul_uint32_t)time(NULL);
#else
    seed = 1357531434;
#endif
    printf("PRNG Seed: %lu\n", (unsigned long)seed);

    /* Allocate. */
    t = cgul_bitwise_trie__new(&local);
    s = cgul_string__new(&local);
    prng = cgul_random_cmwc32__new_from_seed(&local, seed);
    if (local) {
        goto out;
    }

    while (get_next_word(&local, stdin, s)) {

        /* Display any exceptions from last command. */
        if (local) {
            const char* msg = cgul_exception__get_message(local);
            fprintf(stderr, "*** Error: %s\n", msg);
            cgul_exception__delete(local);
            local = NULL;
            continue;
        }

        cmd = cgul_string__get_value(&local, s);
        if (cgul_libc__strcmp(&local, "a", cmd) == 0) {
            assign_int(&local, t, s);
            /* print_trie_with_ints_for_keys(t); */
            continue;
        }
        if (cgul_libc__strcmp(&local, "f", cmd) == 0) {
            find_ulong(&local, t, s);
            /* print_trie_with_ints_for_keys(t); */
            continue;
        }
#if 0
        /* This is currently commented out because the trie does not
         * keep the indexes in any easily sortable order. */
        if (cgul_libc__strcmp(&local, "p", cmd) == 0) {
            print_trie_with_ints_for_keys(&local, t);
            continue;
        }
#endif
        if (cgul_libc__strcmp(&local, "pc", cmd) == 0) {
            print_chronologically_trie_with_ints_for_keys(&local, t);
            continue;
        }
        if (cgul_libc__strcmp(&local, "r", cmd) == 0) {
            remove_int(&local, t, s);
            /* print_trie_with_ints_for_keys(&local, t); */
            continue;
        }
        if (cgul_libc__strcmp(&local, "test", cmd) == 0) {
            test_trie(&local, prng);
            continue;
        }
        if (   (cgul_libc__strcmp(&local, "q", cmd) == 0)
            || (cgul_libc__strcmp(&local, "quit", cmd) == 0))
        {
            break;
        }

        cgul_libc__fprintf(&local,
                           stderr,
                           "Warning: Unknown command: %s\n",
                           cmd);
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    if (prng) {
        cgul_random_cmwc32__delete(prng);
    }
    if (s) {
        cgul_string__delete(s);
    }
    if (t) {
        cgul_bitwise_trie__delete(t);
    }

    return rv;
}
