/*
 * This program tests whether cgul_csv works with character-oriented
 * or line-oriented devices like the terminal.  To test with this
 * program, run it from the terminal and copy one line at a time of a
 * CSV file to its standard input.  It should immediately print
 * results once the current CSV record is complete.
 */

#include "cgul.h"

/**
 * Print the rows and columns in the CSV file <code>csv</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] csv CSV file
 * @param[in] fout output file
 */
static void
print_csv(cgul_exception_t* cex,
          cgul_csv_t csv,
          FILE* fout)
{
    unsigned long int i = 0;
    unsigned long int n = 0;
    const char* field = NULL;
    const char* header = NULL;
    cgul_rbtree_t headers = NULL;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the headers for the cgul_csv instance. */
    headers = cgul_csv__get_headers(cex, csv);

    /* Iterate over each row in the csv file. */
    while (cgul_csv__load_next(cex, csv)) {
        if (*cex) {
            goto out;
        }

        /* Iterate over each field. */
        n = cgul_csv__get_number_of_fields(cex, csv);
        for (i = 0 ; i < n ; ++i) {
            field = cgul_csv__lookup(cex, csv, i);
            cgul_libc__fprintf(cex, fout, "field[%lu] = %s\n", i, field);
            if (*cex) {
                goto out;
            }
        }

        /* Iterate again but use the headers to do the lookup. */
        node = cgul_rbtree__get_front(cex, headers);
        for ( ; node ; node = cgul_rbtree_node__get_next(cex, node)) {
            header = cgul_rbtree_node__get_key(cex, node);
            field = cgul_csv__lookup_by_header(cex, csv, header);
            if (field) {
                cgul_libc__fprintf(
                    cex, fout, "field[\"%s\"] = %s\n", header, field);
            }
        }

        /* Print separator. */
        cgul_libc__fprintf(cex, fout, "=====\n");
        if (*cex) {
            goto out;
        }

    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_csv_t csv = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    csv = cgul_csv__new_from_file(&local, stdin, 0 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /* Print the CSV file. */
    print_csv(&local, csv, stdout);
    if (local) {
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (csv) {
        cgul_csv__delete(csv);
    }

    return rv;
}
