/*
 * This program reorders a CSV file by sorting the columns
 * alphabetically based on the headers for each column.
 */

#include "cgul.h"

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    FILE* fout = NULL;
    const char* header = NULL;
    const char* field = NULL;
    cgul_csv_t csv = NULL;
    cgul_rbtree_t headers = NULL;
    cgul_rbtree_node_t node = NULL;
    cgul_string_t row = NULL;
    cgul_exception_t local = NULL;

    /* Validate the command line. */
    if (argc != 3) {
        fprintf(stderr, "Usage: main_csv_sort <in.csv> <out.csv>\n");
        goto out;
    }

    /* Allocate the "row" string. */
    row = cgul_string__new(&local);

    /* Open the input csv file. */
    csv = cgul_csv__new_from_fname(&local, argv[1], 1 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /* Open the output csv file. */
    fout = cgul_libc__fopen(&local, argv[2], "w");
    if (local) {
        goto out;
    }

    /* Get the tree that holds the sorted headers for the csv file. */
    headers = cgul_csv__get_headers(&local, csv);

    /* Iterate over each row in the csv file. */
    while (cgul_csv__load_next(&local, csv)) {

        /* Reset the "row" string. */
        cgul_string__set_value(&local, row, "");

        /* Iterate over the sorted headers. */
        node = cgul_rbtree__get_front(&local, headers);
        for ( ; node ; node = cgul_rbtree_node__get_next(&local, node)) {

            /* Get the header. */
            header = cgul_rbtree_node__get_key(&local, node);

            /* Get the field associated with this header. */
            field = cgul_csv__lookup_by_header(&local, csv, header);
            if (!field) {
                local = cgul_exception__new("Internal Error: Missing header!");
                goto out;
            }

            /* Append the comma that separates each field, but do not
             * append a comma if this is the first field. */
            if (cgul_string__get_length(&local, row) > 0) {
                cgul_string__append_char(&local, row, ',');
            }

            /* Append the quoted field. */
            cgul_csv__quote(&local, field, row);

            /* Check for errors. */
            if (local) {
                goto out;
            }
        }

        /* Append the EOL character. */
        cgul_string__append_char(&local, row, '\n');

        /* Write the row to the output csv file. */
        cgul_libc__fputs(&local, cgul_string__get_value(&local, row), fout);

        /* Check for errors. */
        if (local) {
            goto out;
        }
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (fout) {
        fclose(fout);
    }
    if (csv) {
        cgul_csv__delete(csv);
    }
    if (row) {
        cgul_string__delete(row);
    }

    return rv;
}
