#include <cassert>
#include <iostream>
#include "cgul_cxx.h"

using namespace std;

/* In order to avoid problems associated with using string literals
 * for the rendezvous string, we use the motif trick of assigning the
 * string literals to variables.  We then use the variable exclusively
 * and get compiler support for detecting most known problems. */
static const char* all_rendezvous = "all";
static const char* foo_rendezvous = "foo_rendezvous";
static const char* bar_rendezvous = "bar_rendezvous";
static const char* reenter_and_remove_rendezvous =
                                    "reenter_and_remove_rendezvous";
static const char* reenter_and_add_then_send_rendezvous =
                                    "reenter_and_add_then_send_rendezvous";

/* We need to hold on to the listeners so we can remove them. */
static cgul_event_router_cxx::listener_t foo_listener_1;
static cgul_event_router_cxx::listener_t foo_listener_2;
static cgul_event_router_cxx::listener_t bar_listener_1;
static cgul_event_router_cxx::listener_t bar_listener_2;
static cgul_event_router_cxx::listener_t all_listener_1;
static cgul_event_router_cxx::listener_t all_listener_2;

static int
foo(const char* rendezvous,
    const void* event,
    const void* data)
{
    cout << "foo" << endl;
    return 1;
}

static int
bar(const char* rendezvous,
    const void* event,
    const void* data)
{
    cout << "bar" << endl;
    return 1;
}

static int
all(const char* rendezvous,
    const void* event,
    const void* data)
{
    cout << "all" << endl;
    return 1;
}

static int
reenter_and_remove(const char* rendezvous,
                   const void* event,
                   const void* data)
{
    try {
        /* Get the event router instance. */
        cgul_event_router_cxx* er = cgul_event_router_cxx::get_instance();

        /* Try to reenter the cgul_event_router class by removing the foo
         * and bar events. */
        er->remove_listener(foo_listener_1);
        foo_listener_1 = NULL;
        er->remove_listener(bar_listener_1);
        bar_listener_1 = NULL;

    /* We cannot let exceptions escape from the event listeners. */
    } catch (const cgul_exception_cxx& e) {
        cerr << "main_event_router: reenter_and_remove: "
             << e.get_message();
    }

    return 1;
}

static int
reenter_and_add_then_send(const char* rendezvous,
                          const void* event,
                          const void* data)
{
    try {
        /* Get the event router instance. */
        cgul_event_router_cxx* er = cgul_event_router_cxx::get_instance();

        /* Try to reenter the cgul_event_router class by adding all the
         * events. */
        foo_listener_1 = er->add_listener(foo_rendezvous, &foo, NULL);
        bar_listener_1 = er->add_listener(bar_rendezvous, &bar, NULL);
        all_listener_2 = er->add_listener(all_rendezvous, &all, NULL);

        /* Now try to send the events. */
        er->send_event(foo_rendezvous, NULL);
        er->send_event(bar_rendezvous, NULL);

    /* We cannot let exceptions escape from the event listeners. */
    } catch (const cgul_exception_cxx& e) {
            cerr << "main_event_router: reenter_and_add_then_send: "
                 << e.get_message();
    }

    return 1;
}

static int
test_adding_and_removing_many_listeners()
{
    char buf[64];
    int i = 0;
    cgul_event_router_cxx::listener_t listener = NULL;

    /* Get the event router instance. */
    cgul_event_router_cxx* er = cgul_event_router_cxx::get_instance();
    for (i = 0 ; i < 100 ; ++i) {
        sprintf(buf, "test_removing_listeners: %d", i);
        listener = er->add_listener(buf, foo, NULL);
        er->remove_listener(listener);
        listener = NULL;
    }

    return 1;
}

int main(int argc, char* argv[])
{
    int rv = 0;
    cgul_event_router_cxx* er = 0;

    try {

        /* Instantiate the singleton. */
        er = er->get_instance();

        /* Try sending events when there are no event listeners.  This
         * should do nothing.*/
        puts("*** Send foo and bar events with no listeners:");
        er->send_event(foo_rendezvous, NULL);
        er->send_event(bar_rendezvous, NULL);

        /* So lets add some event listeners.  Notice that all is listening
         * for the special "all" event.  Which means it should receive all
         * events without have to register for each one individually. */
        puts("*** Add all listeners and send foo and bar events:");
        all_listener_1 = er->add_listener(all_rendezvous, &all, NULL);
        foo_listener_1 = er->add_listener(foo_rendezvous, &foo, NULL);
        bar_listener_1 = er->add_listener(bar_rendezvous, &bar, NULL);
        er->add_listener(reenter_and_remove_rendezvous,
                         &reenter_and_remove,
                         NULL);
        er->add_listener(reenter_and_add_then_send_rendezvous,
                         &reenter_and_add_then_send,
                         NULL);
        er->send_event(foo_rendezvous, NULL);
        er->send_event(bar_rendezvous, NULL);


        /* Try to add the same event listeners for the same rendezvous
         * string.  It is important that this succeeds because we need it
         * in C++ code when using a single private "static" method to
         * forward the callback to a member method for multiple
         * objects. */
        puts("*** Add foo and bar listeners again and "
             "resend foo and bar events:");
        foo_listener_2 = er->add_listener(foo_rendezvous, &foo, NULL);
        bar_listener_2 = er->add_listener(bar_rendezvous, &bar, NULL);
        er->send_event(foo_rendezvous, NULL);
        er->send_event(bar_rendezvous, NULL);

        /* Try to remove the foo_listener. */
        puts("*** Remove both foo listeners, one bar listener, "
             "and resend foo and bar events:");
        er->remove_listener(foo_listener_1);
        foo_listener_1 = NULL;
        er->remove_listener(foo_listener_2);
        foo_listener_2 = NULL;
        er->remove_listener(bar_listener_2);
        bar_listener_2 = NULL;
        er->send_event(foo_rendezvous, NULL);
        er->send_event(bar_rendezvous, NULL);

        /* Try to remove the all_listener. */
        puts("*** Remove the \"all\": listener and resend foo and bar events:");
        er->remove_listener(all_listener_1);
        all_listener_1 = NULL;
        er->send_event(foo_rendezvous, NULL);
        er->send_event(bar_rendezvous, NULL);

        /* Try to re-add the foo_event listener.  */
        puts("*** Re-add foo listener and resend foo and bar events:");
        foo_listener_1 = er->add_listener(foo_rendezvous, &foo, NULL);
        er->send_event(foo_rendezvous, NULL);
        er->send_event(bar_rendezvous, NULL);

        /* Try to re-add the all_rendezvous listener.  */
        puts("*** Re-add the \"all\" event and resend foo and bar events:");
        all_listener_1 = er->add_listener(all_rendezvous, &all, NULL);
        er->send_event(foo_rendezvous, NULL);
        er->send_event(bar_rendezvous, NULL);

        /* Send the "reenter_and_remove_event" event which will cause the
         * event listener itself to try to reenter the cgul_event_router
         * class by attempting to remove listeners. */
        puts("*** Send reenter_and_remove_event to re-entrantly "
             "remove foo and bar:");
        er->send_event(reenter_and_remove_rendezvous, NULL);

        /* Send the foo_event and bar_event now that
         * reenter_and_remove_event should have removed those
         * listeners.  */
        puts("*** Sending foo and bar events:");
        er->send_event(foo_rendezvous, NULL);
        er->send_event(bar_rendezvous, NULL);

        /* Send the "reenter_and_add_then_send_event" which will cause the
         * event listener itself to try to reenter the cgul_event_router
         * class by attempting to add listeners and then attempting to
         * send events to those listeners. */
        puts("*** Send reeenter_and_add_then_send_event:");
        er->send_event(reenter_and_add_then_send_rendezvous, NULL);

        /* Test adding and removing many listeners. */
        puts("*** Test add and removing many listeners:");
        test_adding_and_removing_many_listeners();

        puts("*** Done!");

    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message();
        rv = 1;
    }

    /* Free the event router. */
    cgul_event_router_cxx::delete_instance();

    return rv;
}
