/**
 * Test the hash function and comparision function when using custom
 * keys in a <code>cgul_hash</code> object.
 *
 * @file
 * @brief test program for custom hash keys
 */

#include <assert.h>
#include "cgul.h"

/**
 * Data structure to use as custom keys.
 */
typedef struct point {
    int x;
    int y;
} *point_t;

/**
 * Hash function.  This is an adaptor for the
 * <code>cgul_hash__hash_block()</code> function.  The purpose of this
 * function is to provide the size of the block
 * <code>cgul_hash__hash_block()</code> can hash the block.
 *
 * @param[in] key key
 * @param[in] data client data
 * @return hash of <code>key</code>
 */
static unsigned int
hf(void* key,
   void* data)
{
    return cgul_hash__hash_block(key, sizeof(point_t*));
}

/**
 * Comparison function for <code>point_t</code> objects.  This is
 * needed so the hash can find a particular key when the hash for two
 * different keys collide.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return comparison of <code>key1</code> with <code>key2</code>
 */
static int
cf(const void* key1,
   const void* key2)
{
    int rv = 0;
    point_t p1 = (point_t)key1;
    point_t p2 = (point_t)key2;
    if (p1->x < p2->x) {
        rv = -1;
    } else if (p1->x > p2->x) {
        rv = 1;
    } else if (p1->y < p2->y) {
        rv = -1;
    } else if (p1->y > p2->y) {
        rv = 1;
    }
    return rv;
}

/**
 * Print a point to <code>f</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] p point
 */
static void
print_point(cgul_exception_t* cex,
            point_t p)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(p);

    printf("(%d, %d)\n", p->x, p->y);

 out:
    return;
}

/**
 * This function loads keys into the hash <code>h</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] h hash
 */
static void
load_hash(cgul_exception_t* cex,
          cgul_hash_t h)
{
    int i = 0;
    int j = 0;
    int insert_rv = 0;
    point_t p = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Insert points into h. */
    for (i = 0 ; i < 5 ; ++i) {
        for (j = 0 ; j < 5 ; ++j) {

            /* Allocate a new point. */
            p = (point_t)cgul_libc__calloc(cex, 1, sizeof(*p));
            if (*cex) {
                goto out;
            }

            /* Initialize the point. */
            p->x = 5 - i;
            p->y = 5 - j;

            /* Insert the point as a key in h. */
            insert_rv = cgul_hash__insert(cex, h, (void*)p, NULL);
            if (*cex) {
                goto out;
            }

            /* Free the point if it is a duplicate of one already in
             * the hash. */
            if (!insert_rv) {
                free(p);
            }

            /* Either way, we no longer own p. */
            p = NULL;
        }
    }

 out:

    /* Clean up. */
    if (p) {
        free(p);
    }

    return;
}

/**
 * This function prints the keys the hash <code>h</code> in insertion
 * order.
 *
 * @param[in] cex c-style exception
 * @param[in] h hash
 */
static void
print_in_insertion_order(cgul_exception_t* cex,
                         cgul_hash_t h)
{
    point_t p = NULL;
    cgul_hash_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    n = cgul_hash__get_oldest(cex, h);
    for ( ; n ; n = cgul_hash_node__get_younger(cex, n)) {
        p = (point_t)cgul_hash_node__get_key(cex, n);
        print_point(cex, p);
    }

 out:
    return;
}

/**
 * This function prints the keys the hash <code>h</code> in sort order
 * from least to greatest.
 *
 * @param[in] cex c-style exception
 * @param[in] h hash
 */
static void
print_in_sort_order(cgul_exception_t* cex,
                    cgul_hash_t h)
{
    size_t i = 0;
    point_t* v = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    /* Get an array of keys. */
    v = (point_t*)cgul_hash__get_sorted_keys(cex, h, NULL);

    /* Iterate over the array. */
    for (i = 0 ; v[i] ; ++i) {
        print_point(cex, (point_t)v[i]);
    }

 out:

    /* Clean up. */
    if (v) {
        free(v);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    point_t p = NULL;
    cgul_hash_t h = NULL;
    cgul_exception_t local = NULL;

    /* Allocate h. */
    h = cgul_hash__new_custom(&local, hf, NULL, cf);
    if (local) {
        goto out;
    }

    /* Load custom keys into the hash. */
    load_hash(&local, h);

    /* Print the points in insertion order. */
    print_in_insertion_order(&local, h);

    puts("========");

    /* Print the points in sort order from least to greatest. */
    print_in_sort_order(&local, h);

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (h) {
        /* Free each point in the hash. */
        while (cgul_hash__remove_back(&local, h, (void**)&p, NULL)) {
            free(p);
        }
        /* Delete the hash itself. */
        cgul_hash__delete(h);
    }

    return rv;
}
