#include "cgul.h"

/**
 * This function prints the key/value pair stored in the node <code>n</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] h hash
 * @param[in] n node
 * @param[in] data client data
 */
static int
print_pair(cgul_exception_t* cex,
           cgul_hash_t h,
           cgul_hash_node_t n,
           void* data)
{
    int rv = 0;
    char* key = NULL;
    int* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    key = (char*)cgul_hash_node__get_key(cex, n);
    value = (int*)cgul_hash_node__get_value(cex, n);
    printf("key=%s  value=%d\n", key, *value);
    rv = 1;

 out:
    return rv;
}

/**
 * This function calls <code>free()</code> on the key and on the value
 * stored in the node <code>n</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] h hash
 * @param[in] n node
 * @param[in] data client data
 */
static int
free_pair(cgul_exception_t* cex,
          cgul_hash_t h,
          cgul_hash_node_t n,
          void* data)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    free((void*)cgul_hash_node__get_key(cex, n));
    free(cgul_hash_node__get_value(cex, n));
    rv = 1;

 out:
    return rv;
}

/**
 * Initialize the hash <code>h</code> by loading key/value pairs.
 *
 * @param[in,out] cex c-style exception
 * @param[in] h hash
 */
static void
init_hash(cgul_exception_t* cex,
          cgul_hash_t h)
{
    int i = 0;
    int insert_rv = 0;
    char* key = NULL;
    int* value = NULL;
    cgul_hash_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);

    for (i = 0 ; i < 5 ; ++i) {

        /* Allocate. */
        key = cgul_libc__calloc(cex, 1, sizeof(char) * 2);
        value = cgul_libc__calloc(cex, 1, sizeof(int));
        if (*cex) {
            goto out;
        }

        /* Assign values. */
        *key = 'a' + i;
        *value = i;

        /* Insert into hash. */
        insert_rv = cgul_hash__insert(cex, h, key, &n);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            *cex = cgul_exception__new("key already exists: %s", key);
            goto out;
        }
        cgul_hash_node__set_value(cex, n, value);

        /* The hash now owns the key and value. */
        key = NULL;
        value = NULL;
    }

 out:

    /* Clean up. */
    if (key) {
        free(key);
    }
    if (value) {
        free(value);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_hash_t h = NULL;
    cgul_hash_node_t first = NULL;
    cgul_hash_node_t last = NULL;
    cgul_exception_t local = NULL;

    /* Create a new hash that holds a range of values. */
    h = cgul_hash__new(&local);
    if (local) {
        goto out;
    }

    /* Initialize hash. */
    init_hash(&local, h);
    if (local) {
        goto out;
    }

    /* Print the hash. */
    cgul_hash__traverse(&local, h, print_pair, NULL);

    puts("=====");

    /* Print a range of the hash. */
    first = cgul_hash__get_front(&local, h);
    if (first) {
        first = cgul_hash_node__get_next(&local, first);
    }
    last = cgul_hash__get_back(&local, h);
    if (last) {
        last = cgul_hash_node__get_prev(&local, last);
    }
    cgul_hash__traverse_range(&local, h, first, last, print_pair, NULL);

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (h) {
        cgul_hash__traverse(&local, h, free_pair, NULL);
        cgul_hash__delete(h);
    }

    return rv;
}
