#include "cgul.h"

/**
 * This function prints the value stored in the node <code>n</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] l list
 * @param[in] n node
 * @param[in] data client data
 */
static int
print_value(cgul_exception_t* cex,
            cgul_list_t l,
            cgul_list_node_t n,
            void* data)
{
    int rv = 0;
    int* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    value = (int*)cgul_list_node__get_value(cex, n);
    printf("%d\n", *value);
    rv = 1;

 out:
    return rv;
}

/**
 * This function calls <code>free()</code> on the value stored in the
 * node <code>n</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] l list
 * @param[in] n node
 * @param[in] data client data
 */
static int
free_value(cgul_exception_t* cex,
           cgul_list_t l,
           cgul_list_node_t n,
           void* data)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    free(cgul_list_node__get_value(cex, n));
    rv = 1;

 out:
    return rv;
}

/**
 * This function returns a list of integers in the half-open range
 * <code>[0, upper)</code>.  The caller is responsible for calling
 * <code>free()</code> on each element in the list and for calling
 * <code>cgul_list__delete()</code> on the list itself.
 *
 * @param[in,out] cex c-style exception
 * @param[in] upper exclusive upper bound on the range
 * @return list holding integers from 0 to <code>upper</code>
 */
static cgul_list_t
range(cgul_exception_t* cex,
      int upper)
{
    int i = 0;
    int* value = NULL;
    cgul_list_t rv = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the return value. */
    rv = cgul_list__new(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate from 0 to upper. */
    for (i = 0 ; i < upper ; ++i) {

        /* Allocate a new int to hold this value. */
        value = cgul_libc__malloc(cex, sizeof(*value));
        if (*cex) {
            goto out;
        }

        /* Assign the value. */
        *value = i;

        /* Push the value on to the stack. */
        cgul_list__push_back(cex, rv, value);
        if (*cex) {
            goto out;
        }

        /* The list now owns value. */
        value = NULL;
    }

 out:

    /* Clean up. */
    if (value) {
        free(value);
        value = NULL;
    }
    if (*cex && rv) {
        cgul_list__traverse(&local, rv, free_value, NULL);
        cgul_list__delete(rv);
        rv = NULL;
    }

    return rv;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_list_t l = NULL;
    cgul_list_node_t first = NULL;
    cgul_list_node_t last = NULL;
    cgul_exception_t local = NULL;

    /* Create a new list that holds a range of values. */
    l = range(&local, 5);
    if (local) {
        goto out;
    }

    /* Print the list. */
    cgul_list__traverse(&local, l, print_value, NULL);

    puts("=====");

    /* Print a range of the list. */
    first = cgul_list__get_front(&local, l);
    if (first) {
        first = cgul_list_node__get_next(&local, first);
    }
    last = cgul_list__get_back(&local, l);
    if (last) {
        last = cgul_list_node__get_prev(&local, last);
    }
    cgul_list__traverse_range(&local, l, first, last, print_value, NULL);

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (l) {
        cgul_list__traverse(&local, l, free_value, NULL);
        cgul_list__delete(l);
    }

    return rv;
}
