/**
 * MicroXML Identity Transformation <p>
 *
 * This program will work with UTF-8, UTF-16, and UTF-32 (provided the
 * UTF-16 and UTF-32) MicroXML files have a Unicode Byte-Order Mark
 * (BOM).  If "test.xml" is UTF-8 MicroXML file without a BOM, you can
 * test UTF-16 and UTF-32 as follows: <p>
 *
 * <pre>
 *     iconv -f utf-8 -t utf-16 < test.xml | ./main_microxml -
 *     iconv -f utf-8 -t utf-32 < test.xml | ./main_microxml -
 * </pre>
 *
 * You can test UTF-8 (with BOM), UTF-16LE, UTF-16BE, UTF-32LE, and
 * UTF-32BE as follows provided "test.xml" is UTF-8 MicroXML file
 * without a BOM: <p>
 *
 * <pre>
 *     { printf '\xef\xbb\xbf';
 *       cat < test.xml ; } | ./main_microxml -
 *     { printf '\xfe\xff';
 *       iconv -f utf-8 -t utf-16be < test.xml ; } | ./main_microxml -
 *     { printf '\xff\xfe';
 *       iconv -f utf-8 -t utf-16le < test.xml ; } | ./main_microxml -
 *     { printf '\x00\x00\xfe\xff';
 *       iconv -f utf-8 -t utf-32be < test.xml ; } | ./main_microxml -
 *     { printf '\xff\xfe\x00\x00';
 *       iconv -f utf-8 -t utf-32le < test.xml ; } | ./main_microxml -
 * </pre>
 *
 * Note that the _output_ from the UTF-16 and UTF-32 will be in UTF-8
 * format simply because that is how this program has coded the
 * output.  The important thing is that the program be able to read
 * the UTF-16 and UTF-32 input. <p>
 *
 * @file
 */

#include "cgul.h"

/* Forward References */
static void
main__print_token(cgul_exception_t* cex,
                  cgul_microxml_t uxml,
                  FILE* fout);

/**
 * Print the element attributes <code>attrs</code> to the output file
 * <code>fout</code>.  The caller is responsible for making sure a
 * start element token is currently loaded.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml MicroXML tokenizer
 * @param[in] fout output file
 */
static void
main__print_element_attributes(cgul_exception_t* cex,
                               cgul_microxml_t uxml,
                               FILE* fout)
{
    char* escaped_value = NULL;
    cgul_hash_t attrs = NULL;
    cgul_hash_node_t attr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the token attributes. */
    attrs = cgul_microxml__get_attributes(cex, uxml);
    if (*cex) {
        goto out;
    }

    /* Iterate over the token attributes. */
    attr = cgul_hash__get_front(cex, attrs);
    for ( ; attr ; attr = cgul_hash_node__get_next(cex, attr)) {

        /* Get the key/value pair for this attribute. */
        const char* key = (const char*)cgul_hash_node__get_key(cex, attr);
        cgul_string_t tmp = (cgul_string_t)cgul_hash_node__get_value(cex, attr);
        const char* value = cgul_string__get_value(cex, tmp);

        /* Escape the value. */
        ///escaped_value = cgul_microxml__escape(cex, value);
        if (*cex) {
            goto out;
        }

        /* Print the attribute. */
        fprintf(fout, " %s=\"%s\"", key,
                (escaped_value ? escaped_value : value));

        /* Free the escaped value. */
        free(escaped_value);
        escaped_value = NULL;
    }

 out:

    /* Clean up. */
    if (escaped_value) {
        free(escaped_value);
    }

    return;
}

/**
 * Recursively print the current element to the output file
 * <code>fout</code>.  The caller is responsible for making sure a
 * start element token is currently loaded.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml MicroXML tokenizer
 * @param[in] fout output file
 */
static void
main__print_element(cgul_exception_t* cex,
                    cgul_microxml_t uxml,
                    FILE* fout)
{
    const char* start_tag_name = NULL;
    const char* end_tag_name = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Print the token name.  Note that because cgul_microxml interns
     * the name it is valid until uxml is deleted or reset so it is
     * safe to hold a pointer to the name even while iterating over
     * other elements. */
    start_tag_name = cgul_microxml__get_element_name(cex, uxml);
    if (*cex) {
        goto out;
    }
    fprintf(fout, "<%s", start_tag_name);

    /* Print the token attributes. */
    main__print_element_attributes(cex, uxml, fout);

    /* Close the tag. */
    if (cgul_microxml__is_empty_element(cex, uxml)) {
        fprintf(fout, "/>");
    } else {
        fprintf(fout, ">");
    }

    /* Iterate over the child content until we reach the ELEMENT_STOP
     * token for this element. */
    for (;;) {

        /* Read the next token. */
        cgul_microxml__read_next_token(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* End of Children */
        if(cgul_microxml__is_element_end(cex, uxml)) {
            break;
        }

        /* Recursively print the child token. */
        main__print_token(cex, uxml, fout);
        if (*cex) {
            goto out;
        }
    }

    /* Print the end tag. */
    if (!cgul_microxml__is_empty_element(cex, uxml)) {

        /* Get the name again but this time for the ELEMENT_STOP token. */
        end_tag_name = cgul_microxml__get_element_name(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* If we programmed the recursion correctly, "start_tag_name"
         * will be the same as "end_tag_name". */
        assert(strcmp(start_tag_name, end_tag_name) == 0);

        /* Print the tag. */
        fprintf(fout, "</%s>", end_tag_name);
    }

 out:
    return;
}

/**
 * Recursively print the current MicroXML token to the output file
 * <code>fout</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml MicroXML tokenizer
 * @param[in] fout output file
 */
static void
main__print_token(cgul_exception_t* cex,
                  cgul_microxml_t uxml,
                  FILE* fout)
{
    char* escaped_value = NULL;
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    if (cgul_microxml__is_document_start(cex, uxml)) {
        /* document start - empty */

    } else if (cgul_microxml__is_document_end(cex, uxml)) {
        /* document end - empty */

    } else if (cgul_microxml__is_element_start(cex, uxml)) {
        /* element start */
        main__print_element(cex, uxml, fout);
        if (*cex) {
            goto out;
        }

    } else if (cgul_microxml__is_element_end(cex, uxml)) {
        /* element end - empty (handled as part of element start) */

    } else if (cgul_microxml__is_text(cex, uxml)) {
        /* text */
        cgul_string_t s = cgul_microxml__get_text(cex, uxml);
        const char* text = cgul_string__get_value(cex, s);
        ///escaped_value = cgul_microxml__escape(cex, text);
        if (*cex) {
            goto out;
        }
        fprintf(fout, "%s", (escaped_value ? escaped_value : text));

    } else if (cgul_microxml__is_comment(cex, uxml)) {
        /* comment */
        cgul_string_t s = cgul_microxml__get_comment(cex, uxml);
        const char* comment = cgul_string__get_value(cex, s);
        if (*cex) {
            goto out;
        }
        fprintf(fout, "<!--%s-->", comment);

    } else {
        /* Error. */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected token",
            cgul_microxml__get_presentation_name(cex, uxml),
            cgul_microxml__get_line_number(cex, uxml),
            cgul_microxml__get_column_number(cex, uxml));
    }

 out:

    /* Clean up. */
    if (escaped_value) {
        free(escaped_value);
    }
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_block_reader_file_t br = NULL;
    cgul_stream_reader_t sr = NULL;
    cgul_microxml_t uxml = NULL;
    cgul_exception_t local = NULL;

    /* Validate command line. */
    if (argc != 2) {
        fprintf(stderr, "Usage: main_microxml <xml>\n");
        goto out;
    }

    /* Allocate. */
    br = cgul_block_reader_file__new(&local);
    sr = cgul_stream_reader__new(&local);
    uxml = cgul_microxml__new_from_stream_reader(&local, sr);
    if (local) {
        goto out;
    }

    /* Set up the MicroXML parser. */
    ///cgul_microxml__set_maximum_entity_length(&local, uxml, 6);
    ///cgul_microxml__set_maximum_text_length(&local, uxml, 1);
    ///cgul_microxml__set_skip_white_space(&local, uxml, 1);
    cgul_microxml__set_skip_comments(&local, uxml, 0);
    ///cgul_microxml__set_maximum_comment_length(&local, uxml, 1);
    cgul_microxml__set_skip_prolog_or_epilog_tokens(&local, uxml, 0);

    /* 10% can be gained by disabling line and column number
     * tracking, but this negatively affects the quality of the error
     * messages. */
    ///cgul_stream_reader__set_line_and_column_tracking(&local, sr, 0);

    /* For testing it is nice to be able to set a reduced block size,
     * but reducing the block size only works if the block reader has
     * not yet been added to the stream reader. */
    ///cgul_stream_reader__set_block_size(&local, sr, 17);

    /* Now that cgul_stream_reader__set_block_size() has had a chance
     * to be called, we can add the block reader. */
    cgul_stream_reader__add_block_reader(
        &local, sr, cgul_block_reader_file__as_block_reader(&local, br));

    /* Open the file. */
    if (strcmp(argv[1], "-") == 0) {
        cgul_block_reader_file__open_file(&local, br, stdin);
    } else {
        cgul_block_reader_file__open_fname(&local, br, argv[1]);
    }

    /* Skip the document start token. */
    cgul_microxml__read_next_token(&local, uxml);
    if (local) {
        goto out;
    }
    if (!cgul_microxml__is_document_start(&local, uxml)) {
        local = cgul_exception__new("expected document start");
        goto out;
    }

    /* Iterate over the top-level MicroXML tokens.  Note that there
     * can be multiple top-level tokens if reporting prolog or epilog
     * tokens (leading or trailing white space or comments) is
     * enabled. */
    for (;;) {

        /* Read the next top-level MicroXML token. */
        cgul_microxml__read_next_token(&local, uxml);
        if (local) {
            goto out;
        }

        /* EOF */
        if (cgul_microxml__is_document_end(&local, uxml)) {
            break;
        }

        /* Recursively print this top-level MicroXML token. */
        main__print_token(&local, uxml, stdout);
        if (local) {
            goto out;
        }
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (uxml) {
        cgul_microxml__delete(uxml);
    }
    if (sr) {
        cgul_stream_reader__delete(sr);
    }
    if (br) {
        cgul_block_reader_file__delete(br);
    }

    return rv;
}
