#include <assert.h>
#include <string.h>
#include "cgul.h"

static int
compare_cgul_strings(cgul_string_t s1,
                     cgul_string_t s2)
{
    cgul_exception_t local = NULL;
    return strcmp(cgul_string__get_value(&local, s1),
                  cgul_string__get_value(&local, s2));
}

static int
remove_node_cb(cgul_exception_t* cex,
               cgul_multimap_t mm,
               cgul_multimap_node_t n,
               void* data)
{
    int rv = 0;
    cgul_string_t key = NULL;
    cgul_string_t value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_multimap__remove_node(cex, mm, n, (void**)&key, (void**)&value);
    if (key) {
        cgul_string__delete(key);
    }
    cgul_string__delete(value);

    rv = 1;

 out:
    return rv;
}

static int
print_node_cb(cgul_exception_t* cex,
              cgul_multimap_t mm,
              cgul_multimap_node_t n,
              void* data)
{
    int rv = 0;
    cgul_string_t key = NULL;
    cgul_string_t value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    key = cgul_multimap_node__get_key(cex, n);
    value = cgul_multimap_node__get_value(cex, n);
    cgul_libc__printf(cex,
                      "%s : %s\n",
                      cgul_string__get_value(cex, key),
                      cgul_string__get_value(cex, value));

    rv = 1;

 out:
    return rv;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    unsigned long int i = 0;
    cgul_string_t key = NULL;
    cgul_string_t test_key = NULL;
    cgul_string_t value = NULL;
    cgul_cache_t cache = NULL;
    cgul_multimap_t mm = NULL;
    cgul_exception_t local = NULL;

    /* We need a cache to test removals that automatically put keys
     * and values back on their caches. */
    cache = cgul_cache__new(&local,
                            (cgul_cache__constructor_t)&cgul_string__new,
                            (cgul_cache__destructor_t)&cgul_string__delete,
                            (cgul_cache__restructor_t)&cgul_string__clear);
    if (local) {
        goto out;
    }

    /* Create a multimap. */
    mm = cgul_multimap__new(&local,
                            (cgul_multimap__compare_t)compare_cgul_strings);
    if (local) {
        goto out;
    }

    /* Create test key so we can query the multimap. */
    test_key = cgul_string__new(&local);
    if (local) {
        goto out;
    }

    /* Initialize the multimap so we can have some nodes to play with. */
    for (i = 0 ; i < 30 ; ++i) {

        /* Initialize the key. */
        key = (cgul_string_t)cgul_cache__get(&local, cache);
        cgul_string__append_sprintf(&local, key, "%lu", i % 10);

        /* Initialize the value. */
        value = (cgul_string_t)cgul_cache__get(&local, cache);
        cgul_string__append_sprintf(&local, value, "%lu", i);

        /* Insert the key/value pair. */
        if (cgul_multimap__insert(&local, mm, key, value, NULL) > 1) {
            /* cgul_multimap does not keep identical keys. */
            cgul_cache__put(&local, cache, key);
        }

        /* Check for errors. */
        if (local) {
            goto out;
        }
    }

    printf("cgul_multimap__get_size() = %lu\n",
           cgul_multimap__get_size(&local, mm));

#if 0
    {
        /* Remove the "8" key using remove(). */
        cgul_string__set_value(&local, test_key, "8");
        cgul_multimap__remove(&local, mm, test_key, cache, cache);
    }
#endif

#if 1
    {
        /* Remove the "8" key using traverse_range(). */
        cgul_multimap_node_t first = NULL;
        cgul_multimap_node_t last = NULL;
        cgul_string__set_value(&local, test_key, "8");
        cgul_multimap__find(&local, mm, test_key, 1, &first);
        cgul_multimap__find(&local, mm, test_key, 0, &last);
        /* Must test "first" and "last" to make sure they are not NULL
         * because the convention is that NULL values mean the first
         * and lost node of the entire multimap. */
        if (first && last) {
            cgul_multimap__traverse_range(&local,
                                          mm,
                                          first,
                                          last,
                                          1,
                                          &remove_node_cb,
                                          NULL);
        }
    }
#endif

#if 0
    {
        /* Remove the "8" key using iteration. */
        unsigned long int count = 0;
        void* key_out = NULL;
        void* value_out = NULL;
        cgul_multimap_node_t curr = NULL;
        cgul_multimap_node_t next = NULL;
        cgul_string__set_value(&local, test_key, "8");
        count = cgul_multimap__find(&local, mm, test_key, 0, &curr);
        if (curr) {
            for (i = 0 ; i < count ; ++i, curr = next) {
                next = cgul_multimap_node__get_next(&local, curr, 0);
                cgul_multimap__remove_node(&local, mm, curr,
                                           &key_out, &value_out);
                if (key_out) {
                    cgul_string__delete(key_out);
                }
                cgul_string__delete(value_out);
            }
        }
    }
#endif

#if 1
    {
        /* Test find_floor() using the "8" key which was removed above. */
        printf("===== find_floor() \n");
        unsigned long int i = 0;
        unsigned long int count = 0;
        cgul_multimap_node_t curr = NULL;
        cgul_string__set_value(&local, test_key, "8");
        count = cgul_multimap__find_floor(&local, mm, test_key, 0, &curr);
        for (i = 0 ; curr && (i < count) ; ++i) {
            /* Get the key/value pair. */
            cgul_string_t key = cgul_multimap_node__get_key(&local, curr);
            cgul_string_t value = cgul_multimap_node__get_value(&local, curr);
            /* Print the key/value pair. */
            cgul_libc__printf(&local,
                              "%s : %s\n",
                              cgul_string__get_value(&local, key),
                              cgul_string__get_value(&local, value));
            /* Get the next node. */
            curr = cgul_multimap_node__get_next(&local, curr, 0);
        }
    }
#endif

#if 1
    {
        /* Test find_ceiling() using the "8" key which was removed above. */
        printf("===== find_ceiling() \n");
        unsigned long int i = 0;
        unsigned long int count = 0;
        cgul_multimap_node_t curr = NULL;
        cgul_string__set_value(&local, test_key, "8");
        count = cgul_multimap__find_ceiling(&local, mm, test_key, 0, &curr);
        for (i = 0 ; curr && (i < count) ; ++i) {
            /* Get the key/value pair. */
            cgul_string_t key = cgul_multimap_node__get_key(&local, curr);
            cgul_string_t value = cgul_multimap_node__get_value(&local, curr);
            /* Print the key/value pair. */
            cgul_libc__printf(&local,
                              "%s : %s\n",
                              cgul_string__get_value(&local, key),
                              cgul_string__get_value(&local, value));
            /* Get the previous node. */
            curr = cgul_multimap_node__get_prev(&local, curr, 0);
        }
        printf("=====\n");
    }
#endif

    printf("cgul_multimap__get_size() = %lu\n",
           cgul_multimap__get_size(&local, mm));

#if 0
    {
        cgul_multimap_node_t node = NULL;
        /* Find a key and print all the values associated with it. */
        cgul_string__set_value(&local, test_key, "7");
        count = cgul_multimap__find(&local, mm, test_key, 0, &node);
        for (i = 0 ; i < count ; ++i) {
            /* Get the key/value pair. */
            key = cgul_multimap_node__get_key(&local, node);
            value = cgul_multimap_node__get_value(&local, node);
            /* Print the key/value pair. */
            cgul_libc__printf(&local,
                              "%s : %s\n",
                              cgul_string__get_value(&local, key),
                              cgul_string__get_value(&local, value));
            /* Move to the next value for this key. */
            node = cgul_multimap_node__get_next(&local, node, 0);
        }
    }
#endif

#if 0
    {
        cgul_multimap_node_t node = NULL;
        /* Walk the tree printing each node. */
        node = cgul_multimap__get_front(&local, mm, 0);
        for ( ; node ; node = cgul_multimap_node__get_next(&local, node, 0)) {
            key = cgul_multimap_node__get_key(&local, node);
            value = cgul_multimap_node__get_value(&local, node);
            cgul_libc__printf(&local,
                              "%s : %s\n",
                              cgul_string__get_value(&local, key),
                              cgul_string__get_value(&local, value));
        }
    }
#endif

#if 1
    {
        /* Walk the tree printing each node. */
        cgul_multimap__traverse(&local, mm, 0, &print_node_cb, NULL);
    }
#endif

 out:

    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    if (test_key) {
        cgul_string__delete(test_key);
        test_key = NULL;
    }

    if (mm) {
        cgul_multimap__clear(&local, mm, cache, cache);
        cgul_multimap__delete(mm);
    }

    if (cache) {
        cgul_cache__delete(cache);
        cache = NULL;
    }

    return rv;
}
