#include <cstring>
#include <iostream>
#include <stdexcept>
#include "cgul_cxx.h"

using namespace std;

static int
remove_node_cb(cgul_multimap_cxx* mm,
               cgul_multimap_node_cxx* node,
               void* data)
{
    void* key = NULL;
    void* value = NULL;
    mm->remove_node(node, (void**)&key, (void**)&value);
    if (key) {
        free(key);
    }
    free(value);
    return 1;
}

static int
print_node_cb(cgul_multimap_cxx* mm,
              cgul_multimap_node_cxx* node,
              void* data)
{
    cout << (char*)node->get_key()
         << " : " << (char*)node->get_value() << endl;
    return 1;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    unsigned long int i = 0;

    try {

        /* Create the multimap. */
        cgul_multimap_cxx mm((cgul_multimap_cxx::compare_t)strcmp);

        /* Create objects to build up the keys and values. */
        cgul_string_cxx key;
        cgul_string_cxx value;

        /* Initialize the multimap. */
        for (i = 0 ; i < 30 ; ++i) {
            key.append_sprintf("%lu", i % 10);
            value.append_sprintf("%lu", i);
            if (mm.insert(key.get_value(), value.get_value()) == 1) {
                key.take_value();
            } else {
                key = "";
            }
            value.take_value();
        }

#if 0
        /* Print all the values for "7". */
        unsigned long count = mm.find((void*)"7", 0, &node);
        for (i = 0 ; i < count ; ++i, node = node->get_next(0)) {
            char* key = (char*)node->get_key();
            char* value = (char*)node->get_value();
            cout << key << " : " << value << endl;
        }
#endif

#if 0
        /* Remove the "8" key using iteration.  This shows that you
         * need to be careful when you remove keys while iterating
         * instead of calling functions like remove(), remove_range(),
         * traverse(), or traverse_range(). */
        {
            void* key_out = NULL;
            void* value_out = NULL;
            cgul_multimap_node_cxx* curr = NULL;
            cgul_multimap_node_cxx* next = NULL;
            unsigned long count = mm.find((void*)"8", 1, &curr);
            if (curr) {
                for (i = 0 ; i < count ; ++i, curr = next) {
                    next = curr->get_next(1);
                    mm.remove_node(curr, &key_out, &value_out);
                    if (key_out) {
                        free(key_out);
                    }
                    free(value_out);
                }
            }
        }
#endif

#if 1
        {
            /* Remove the "8" key using traverse_range(). */
            cgul_multimap_node_cxx* first = NULL;
            cgul_multimap_node_cxx* last = NULL;
            mm.find((void*)"8", 1, &first);
            mm.find((void*)"8", 0, &last);
            /* Must test "first" and "last" to make sure they are not
             * NULL because the convention is that NULL values mean
             * the first and lost node of the entire multimap. */
            if (first && last) {
                mm.traverse_range(first, last, 1, &remove_node_cb, NULL);
            }
        }
#endif

#if 0
        /* Iterate over the multimap. */
        cgul_multimap_node_cxx* node = mm.get_front(0);
        for ( ; node ; node = node->get_next(0)) {
            char* key = (char*)node->get_key();
            char* value = (char*)node->get_value();
            cout << key << " : " << value << endl;
        }
#endif

#if 1
    {
        /* Test find_floor() using the "8" key which was removed above. */
        printf("===== find_floor() \n");
        unsigned long int i = 0;
        unsigned long int count = 0;
        cgul_multimap_node_cxx* curr = NULL;
        key.set_value("8");
        count = mm.find_floor(key.get_value(), 0, &curr);
        for (i = 0 ; curr && (i < count) ; ++i) {
            /* Get the key/value pair. */
            const char* key = (const char*)curr->get_key();
            const char* value = (const char*)curr->get_value();
            /* Print the key/value pair. */
            cgul_libc_cxx::printf("%s : %s\n", key, value);
            /* Get the next node. */
            curr = curr->get_next(0);
        }
    }
#endif

#if 1
    {
        /* Test find_ceiling() using the "8" key which was removed above. */
        printf("===== find_ceiling() \n");
        unsigned long int i = 0;
        unsigned long int count = 0;
        cgul_multimap_node_cxx* curr = NULL;
        key.set_value("8");
        count = mm.find_ceiling(key.get_value(), 0, &curr);
        for (i = 0 ; curr && (i < count) ; ++i) {
            /* Get the key/value pair. */
            const char* key = (const char*) curr->get_key();
            const char* value = (const char*) curr->get_value();
            /* Print the key/value pair. */
            cgul_libc_cxx::printf("%s : %s\n", key, value);
            /* Get the previous node. */
            curr = curr->get_prev(0);
        }
        printf("=====\n");
    }
#endif

#if 1
        {
            /* Print remaining nodes. */
            mm.traverse(0, &print_node_cb, NULL);
        }
#endif

        /* Clear the multimap. */
        void* key_out = NULL;
        void* value_out = NULL;
        while (mm.remove_front(1, &key_out, &value_out)) {
            if (key_out) {
                free(key_out);
            }
            free(value_out);
        }

    } catch (const exception& e) {
        cout << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cout << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    return rv;
}
