/**
 * This program can be used either to generate PRNs in binary on
 * <code>stdout</code> or to time the generation of 100M PRNs without
 * writing them to <code>stdout</code> as follows: <p>
 *
 * <pre>
 *      PRN Source:  main_random_xorshift32_bytes
 *     Timing Test:  main_random_xorshift32_bytes -time
 * </pre>
 */

#include "cgul.h"

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    int is_timing_test = 0;
    size_t i = 0;
    size_t buf_count = 4096;  /* Count of 32-bit values in "buf." */
    cgul_uint32_t* buf = NULL;
    cgul_random_xorshift32_t prng = NULL;
    cgul_exception_t local = NULL;

    /* Verify the command line. */
    if (argc == 2) {
        if ((strcmp(argv[1], "-t") == 0)
            || (strcmp(argv[1], "-time") == 0)
            || (strcmp(argv[1], "--time") == 0))
        {
            is_timing_test = 1;
        } else {
            fprintf(stderr, "Usage: main_random_xorshift32_bytes [-time]\n");
            goto out;
        }
    } else if (argc > 2) {
        fprintf(stderr, "Usage: main_random_xorshift32_bytes [-time]\n");
        goto out;
    }

    /* Allocate. */
    prng = cgul_random_xorshift32__new(&local);
    if (local) {
        goto out;
    }

    /* If this is a timing test, generate 100M random unsigned longs
     * as quickly as possible. */
    if (is_timing_test) {
        for (i = 0 ; i < 100000000 ; ++i) {
            cgul_random_xorshift32__next(&local, prng);
        }
        goto out;
    }

    /*
     * If reached, this is not a timing test.  Rather, the PRNs should
     * be written in binary to stdout.
     */

    /* Allocate. */
    buf = cgul_libc__calloc(&local, buf_count, sizeof(*buf));
    if (local) {
        goto out;
    }

    /* Write the PRNs directly as binary to stdout not worrying about
     * byte order.  A buffer is used to reduce the frequence at which
     * fwrite() is called. */
    for (;;) {
        /* Write the buffer if it is full. */
        if (i == buf_count) {
            if (fwrite(buf, sizeof(*buf), buf_count, stdout) != buf_count) {
                local = cgul_exception__new("fwrite: %m");
                goto out;
            }
            /* Reset i. */
            i = 0;
        }
        /* Get the next PRN. */
        buf[i++] = cgul_random_xorshift32__next(&local, prng);
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (buf) {
        free(buf);
    }
    if (prng) {
        cgul_random_xorshift32__delete(prng);
    }

    return rv;
}
