/*
 * Copy UTF-8, UTF-16, or UTF-32 on stdin to stdout without EOL
 * normalization.  If "test.xml" is UTF-8 MicroXML file without a BOM,
 * you can test UTF-16 and UTF-32 as follows: <p>
 *
 * <pre>
 *     iconv -f utf-8 -t utf-16 < test.xml | ./main_stream_reader -
 *     iconv -f utf-8 -t utf-32 < test.xml | ./main_stream_reader -
 * </pre>
 *
 * You can test UTF-8 (with BOM), UTF-16LE, UTF-16BE, UTF-32LE, and
 * UTF-32BE as follows provided "test.xml" is UTF-8 MicroXML file
 * without a BOM: <p>
 *
 * <pre>
 *     { printf '\xef\xbb\xbf';
 *       cat < test.xml ; } | ./main_stream_reader -
 *     { printf '\xfe\xff';
 *       iconv -f utf-8 -t utf-16be < test.xml ; } | ./main_stream_reader -
 *     { printf '\xff\xfe';
 *       iconv -f utf-8 -t utf-16le < test.xml ; } | ./main_stream_reader -
 *     { printf '\x00\x00\xfe\xff';
 *       iconv -f utf-8 -t utf-32be < test.xml ; } | ./main_stream_reader -
 *     { printf '\xff\xfe\x00\x00';
 *       iconv -f utf-8 -t utf-32le < test.xml ; } | ./main_stream_reader -
 * </pre>
 *
 * Note that the _output_ from the UTF-16 and UTF-32 will be in UTF-8
 * format simply because that is how this program has coded the
 * output.  The important thing is that the program be able to read
 * the UTF-16 and UTF-32 input. <p>

 */

#include "cgul.h"

#ifdef WIN32
#include <fcntl.h>
#include <io.h>
#endif

/**
 * Force the standard input <code>stdin</code> and standard output
 * <code>stdout</code> to be in binary mode (instead of text mode).
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 */
static void
main__force_binary_mode(cgul_exception_t* cex)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

#ifdef WIN32
    /* Convert stdin and stdout to binary mode. */
    if (_setmode(_fileno(stdin), _O_BINARY) < 0) {
        *cex = cgul_exception__new("_setmode(stdin): %m");
        goto out;
    }
    if (_setmode(_fileno(stdout), _O_BINARY) < 0) {
        *cex = cgul_exception__new("_setmode(stdout): %m");
        goto out;
    }
#endif

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    int i = 0;
    char utf8[CGUL_WCHAR__MB_LEN_MAX + 1] = {0};
    cgul_wchar_t utf32 = CGUL_WCHAR__NUL;
    cgul_block_reader_file_t brfile = NULL;
    cgul_stream_reader_t sr = NULL;
    cgul_exception_t local = NULL;

    /* Force stdin and stdout to binary mode. */
    main__force_binary_mode(&local);

    /* Allocate. */
    brfile = cgul_block_reader_file__new_from_file(&local, stdin);
    sr = cgul_stream_reader__new_from_block_reader(&local,
             cgul_block_reader_file__as_block_reader(&local, brfile));
    if (local) {
        goto out;
    }

    /* Configure the stream reader. */
    cgul_stream_reader__set_normalize_eol(&local, sr, 0);
    cgul_stream_reader__set_line_and_column_tracking(&local, sr, 0);

    /* Iterate. */
    for (;;) {

        /* Read the next character. */
        if (!cgul_stream_reader__read(&local, sr, &utf32)) {
            if (local) {
                goto out;
            }
            /* EOF */
            break;
        }

        /* Convert the next character to UTF-8. */
        cgul_unicode__wctomb(&local, utf32, utf8);
        if (local) {
            goto out;
        }

        /* Print the UTF-8 character one byte at a time. */
        for (i = 0 ; utf8[i] ; ++i) {
            putc(utf8[i], stdout);
        }
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (sr) {
        cgul_stream_reader__delete(sr);
    }
    if (brfile) {
        cgul_block_reader_file__delete(brfile);
    }

    return rv;
}
