/*
 * This program reads a dictionary file with one word or phrase per
 * line.  The dictionary file is a text file where each line of text
 * corresponds to one key in the trie.  The program then reads
 * prefixes from standard input and prints the dictionary entries that
 * have the same prefix.  The program can be run as follows:
 *
 *     ./main_trie_auto_complete <dictionary>
 *
 * To quit, type the following:
 *
 *     quit
 */

#include "cgul.h"

/**
 * Print this program's usage message to the file <code>f</code>.  The
 * base name of <code>progname</code> is printed as part of the usage
 * message.
 *
 * @param[in] cex c-style exception
 * @param[in] progname program name
 * @param[in] fout output file
 */
static void
main__print_usage(cgul_exception_t* cex,
                  const char* progname,
                  FILE* fout)
{
    char* bname = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(progname);
    assert(fout);

    /* Get the base name for the program. */
    bname = cgul_libc__basename(cex, progname);
    if (*cex) {
        goto out;
    }

    fprintf(fout, "\n");
    fprintf(fout, "Usage: %s <dictionary>\n", bname);
    fprintf(fout, "\n");

 out:

    /* Clean up. */
    if (bname) {
        free(bname);
    }

    return;
}

#if 1
/**
 * Insert the keys in the <code>keys</code> array into the trie
 * <code>trie</code> freeing the duplicate keys from the
 * <code>keys</code> array and setting their place in the
 * <code>keys</code> array to <code>NULL</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys array of keys to insert into the trie
 * @param[in] key_count number of keys
 */
static void
main__insert_keys_freeing_duplicates(cgul_exception_t* cex,
                                     cgul_trie_t trie,
                                     char** keys,
                                     unsigned long int key_count)
{
    unsigned long int i = 0;
    cgul_trie_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Sanity check. */
    if (!keys || (key_count == 0)) {
        goto out;
    }

    /* Log. */
    printf("Inserting keys ... ");
    fflush(stdout);

    /* Insert the keys into the trie.  This transfers ownership of all
     * the keys that get inserted to the trie; however, for duplicate
     * keys, ownership stays with the "lines" array. */
    cgul_trie__insert_from_arrays(
        cex,                  /* c-style exception */
        trie,                 /* trie */
        (const char**)keys,   /* keys array */
        NULL,                 /* values array */
        key_count,            /* number of key/value pairs */
        1);                   /* maintain_relative_ages */
    if (*cex) {
        goto out;
    }

    /* Log. */
    printf("Freeing duplicate keys ... ");
    fflush(stdout);

    /* Free duplicate keys from the "keys" array that are not stored
     * in the trie.  This occurs because the trie only stores the
     * first copy of the key when the user tries to other copies. */
    for (i = 0 ; i < key_count ; ++i) {
        node = cgul_trie__find(cex, trie, keys[i]);
        if (!node ||
            (keys[i] != cgul_trie_node__get_key(cex, node)))
        {
            free(keys[i]);
        }
    }

 out:
    return;
}

#else

/**
 * Insert the keys in the <code>keys</code> array into the trie
 * <code>trie</code> freeing the duplicate keys from the
 * <code>keys</code> array and setting their place in the
 * <code>keys</code> array to <code>NULL</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys array of keys to insert into the trie
 * @param[in] key_count number of keys
 */
static void
main__insert_keys_freeing_duplicates(cgul_exception_t* cex,
                                     cgul_trie_t trie,
                                     char** keys,
                                     unsigned long int key_count)
{
    int insert_rv = 0;
    unsigned long int i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Sanity check. */
    if (!keys || (key_count == 0)) {
        goto out;
    }

    /* Log. */
    printf("Inserting keys (and freeing duplicates) ... ");
    fflush(stdout);

    /* Insert the keys one at a time freeing duplicates as they are
     * detected as part of the insertion process.*/
    for (i = 0 ; i < key_count ; ++i) {

        /* Try to insert the key. */
        insert_rv = cgul_trie__insert(cex, trie, keys[i], NULL);
        if (*cex) {
            goto out;
        }

        /* Free duplicates. */
        if (!insert_rv) {
            free(keys[i]);
            keys[i] = NULL;
        }

    }

    /* Log. */
    printf("balancing ... ");
    fflush(stdout);

    /* Balance the trie to avoid linear searches if the keys are
     * inserted in sort order. */
    cgul_trie__balance(cex, trie, 1 /* maintain_relative_ages */);

 out:
    return;
}
#endif

/**
 * Load the dictionary in the file <code>fname</code> into the trie
 * <code>trie</code>.  The dictionary file is a text file where each
 * line in the text corresponds to one key in the trie.  If an error
 * occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] fname file name for the dictionary
 */
static void
main__load_dictionary(cgul_exception_t* cex,
                      cgul_trie_t trie,
                      const char* fname)
{
    cgul_trie_t tmp = NULL;
    char** lines = NULL;
    unsigned long int line_count = 0;
    cgul_crlf_file_t crlf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Log. */
    printf("Loading dictionary ... ");
    fflush(stdout);

    /* Allocate. */
    tmp = cgul_trie__new(cex);
    crlf = cgul_crlf_file__new_from_fname(
               cex, fname, 1 /* is_block_buffered */);
    if (*cex) {
        goto out;
    }

    /* Get the list of lines from the dictionary file that will be
     * used as the keys for the trie. */
    cgul_crlf_file__get_lines(cex, crlf, &lines, &line_count);
    if (*cex) {
        goto out;
    }

    /* Insert the lines into the trie as keys.  Any duplicate lines in
     * the "lines" array will be freed.  This way only the trie holds
     * lines that still need to be freed. */
    main__insert_keys_freeing_duplicates(cex, tmp, lines, line_count);
    if (*cex) {
        goto out;
    }

    /* Success. */
    cgul_trie__swap(cex, trie, tmp);

    /* Log. */
    printf("Done.\n");
    fflush(stdout);

 out:

    /* Clean up. */
    if (lines) {
        free(lines);
    }
    if (crlf) {
        cgul_crlf_file__delete(crlf);
    }
    if (tmp) {
        cgul_trie__delete(tmp);
    }

    return;
}

/**
 * Print to standard output all the keys in the trie <code>trie</code>
 * having prefix <code>prefix</code>.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie_cxx</code> instance
 * @param[in] prefix prefix
 */
static void
main__handle_prefix_request(cgul_exception_t* cex,
                            cgul_trie_t trie,
                            const char* prefix)
{
    const char* key = NULL;
    cgul_trie_node_t start = NULL;
    cgul_trie_node_t stop = NULL;
    cgul_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Sanity check. */
    if (!prefix) {
        goto out;
    }

    /* Get iterators over the result set for this prefix. */
    cgul_trie__find_half_opened_range_by_prefix(
        cex, trie, prefix, &start, &stop);
    if (*cex) {
        goto out;
    }

    /* Iterate over the result set. */
    for (curr = start;
         curr != stop;
         curr = cgul_trie_node__get_next(cex, curr))
    {
        /* Print the key to standard output. */
        key = cgul_trie_node__get_key(cex, curr);
        puts(key);
    }

 out:
    return;
}

/**
 * Loop requesting the user to enter a prefix which is used to query
 * the trie <code>trie</code> to get a list of dictionary words that
 * match the prefix.  The list is then printed to standard output.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 */
static void
main__loop(cgul_exception_t* cex,
           cgul_trie_t trie)
{
    int get_line_rv = 0;
    cgul_string_t prefix = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    prefix = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    for (;;) {

        /* Prompt the user to enter a prefix. */
        puts("=====");
        printf("Enter prefix: ");
        fflush(stdout);

        /* Get the user's response. */
        cgul_string__set_value(cex, prefix, "");
        get_line_rv = cgul_get_line_relaxed(cex, stdin, prefix);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (!get_line_rv) {
            goto out;
        }

        /* quit */
        if (strcmp("quit", cgul_string__get_value(cex, prefix)) == 0) {
            goto out;
        }

        /* Handle prefix request. */
        main__handle_prefix_request(
            cex, trie, cgul_string__get_value(cex, prefix));
        if (*cex) {
            goto out;
        }
    }

 out:

    /* Clean up. */
    if (prefix) {
        cgul_string__delete(prefix);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_trie_t trie = NULL;
    cgul_exception_t local = NULL;

    /* Validate command-line arguments. */
    if (argc != 2) {
        main__print_usage(&local, argv[0], stderr);
        goto out;
    }

    /* Allocate. */
    trie = cgul_trie__new(&local);
    if (local) {
        goto out;
    }

    /* Load the dictionary into the trie. */
    main__load_dictionary(&local, trie, argv[1]);
    if (local) {
        goto out;
    }

    /* Enter the main loop. */
    main__loop(&local, trie);
    if (local) {
        goto out;
    }


 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (trie) {
        cgul_trie__free_keys(trie);
        cgul_trie__delete(trie);
    }

    return rv;
}
