#include "cgul.h"

/**
 * Insert the key/value pair (<code>key</code>, <code>value</code>)
 * into the trie <code>trie</code>.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in,out] trie trie
 * @param[in] key key to insert
 * @param[in] value value to associate with the key
 */
static void
main__insert_kvpair(cgul_exception_t* cex,
                    cgul_trie_t trie,
                    const char* key,
                    int value)
{
    int insert_rv = 0;
    char* key_ptr = NULL;
    int* value_ptr = NULL;
    cgul_trie_node_t node = NULL;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!key) {
        goto out;
    }

    /* Duplicate the string. */
    key_ptr = cgul_libc__strdup(cex, key);
    if (*cex) {
        goto out;
    }

    /* Duplicate the value. */
    value_ptr = (int*)cgul_libc__calloc(cex, 1, sizeof(*value_ptr));
    if (*cex) {
        goto out;
    }
    *value_ptr = value;

    /* Insert the key into the trie. */
    insert_rv = cgul_trie__insert(cex, trie, key_ptr, &node);
    if (*cex) {
        goto out;
    }
    if (!insert_rv) {
        *cex = cgul_exception__new(
            "main__insert_kvpair: key already exists: %s", key_ptr);
        goto out;
    }

    /* The trie now owns "key_ptr". */
    key_ptr = NULL;

    /* Set the value associated with the key. */
    cgul_trie_node__set_value(cex, node, value_ptr);

    /* The trie now owns "value_ptr". */
    value_ptr = NULL;

 out:

    /* Clean up. */
    if (value_ptr) {
        free(value_ptr);
    }
    if (key_ptr) {
        free(key_ptr);
    }

    return;
}

/**
 * Print all the key/value pairs in the trie <code>trie</code> in
 * chronological (i.e., insertion) order.
 *
 * @param[in] cex c-style exception
 * @param]in] trie <code>cgul_trie</code> instance
 */
static void
main__print_chrono_kvpairs(cgul_exception_t* cex,
                           cgul_trie_t trie)
{
    const char* key = NULL;
    int* value = NULL;
    cgul_trie_node_t curr = NULL;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!trie) {
        goto out;
    }

    /* Iterate over all the key/value pairs printing each. */
    curr = cgul_trie__get_oldest(cex, trie);
    for ( ; curr ; curr = cgul_trie_node__get_younger(cex, curr)) {
        key = cgul_trie_node__get_key(cex, curr);
        value = (int*)cgul_trie_node__get_value(cex, curr);
        if (value) {
            printf("%s -> %d\n", key, *value);
        } else {
            printf("%s -> %p\n", key, value);
        }
    }

 out:
    return;
}

/**
 * Print all the key/value pairs in the trie <code>trie</code> in sort
 * order.
 *
 * @param[in] cex c-style exception
 * @param]in] trie <code>cgul_trie</code> instance
 */
static void
main__print_sorted_kvpairs(cgul_exception_t* cex,
                           cgul_trie_t trie)
{
    const char* key = NULL;
    int* value = NULL;
    cgul_trie_node_t curr = NULL;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!trie) {
        goto out;
    }

    /* Iterate over all the key/value pairs printing each. */
    curr = cgul_trie__get_front(cex, trie);
    for ( ; curr ; curr = cgul_trie_node__get_next(cex, curr)) {
        key = cgul_trie_node__get_key(cex, curr);
        value = (int*)cgul_trie_node__get_value(cex, curr);
        if (value) {
            printf("%s -> %d\n", key, *value);
        } else {
            printf("%s -> %p\n", key, value);
        }
    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_trie_t trie = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    trie = cgul_trie__new(&local);
    if (local) {
        goto out;
    }

    /* Append key/value pairs. */
    main__insert_kvpair(&local, trie, "zero",  0);
    main__insert_kvpair(&local, trie, "one",   1);
    main__insert_kvpair(&local, trie, "two",   2);
    main__insert_kvpair(&local, trie, "three", 3);
    main__insert_kvpair(&local, trie, "four",  4);
    main__insert_kvpair(&local, trie, "five",  5);
    main__insert_kvpair(&local, trie, "six",   6);
    main__insert_kvpair(&local, trie, "seven", 7);
    main__insert_kvpair(&local, trie, "eight", 8);
    main__insert_kvpair(&local, trie, "nine",  9);
    if (local) {
        goto out;
    }

    /* Print the key/value pairs in chronological (i.e., insertion) order. */
    printf("===== Original Chronological Order\n");
    main__print_chrono_kvpairs(&local, trie);
    if (local) {
        goto out;
    }

    /*
     * Balance the trie.
     *
     * NOTE: You can change 1 to 0 (below) to verify that the trie
     * structure actually changes.  The fact that the chronological
     * order doesn't appear to change when maintaining the relative
     * node ages is because extra work takes place to adjust the
     * older/younger list.
     */
    cgul_trie__balance(&local, trie, 1 /* maintain_relative_ages */);
    if (local) {
        goto out;
    }

    /* Print the key/value pairs in chronological (i.e., insertion) order. */
    printf("===== New Chronological Order\n");
    main__print_chrono_kvpairs(&local, trie);
    if (local) {
        goto out;
    }

    /* Print the key/value pairs in sort order. */
    printf("===== Sort Order (by key)\n");
    main__print_sorted_kvpairs(&local, trie);
    if (local) {
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (trie) {
        cgul_trie__free_values(trie);
        cgul_trie__free_keys(trie);
        cgul_trie__delete(trie);
    }

    return rv;
}
