#include "cgul.h"

/**
 * Insert a copy of the key <code>key</code> into the trie
 * <code>trie</code>.  The caller is responsible for calling
 * <code>free()</code> on the key stored in the trie.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] key
 */
static void
main__insert_key(cgul_exception_t* cex,
                 cgul_trie_t trie,
                 const char* key)
{
    int insert_rv = 0;
    char* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Duplicate the key. */
    tmp = cgul_libc__strdup(cex, key);
    if (*cex) {
        goto out;
    }

    /* Insert the duplicated key. */
    insert_rv = cgul_trie__insert(cex, trie, tmp, NULL);
    if (*cex) {
        goto out;
    }
    if (insert_rv) {
        /* Transfer ownership of the duplicated key to the trie. */
        tmp = NULL;
    }

 out:

    /* Clean up. */
    if (tmp) {
        free(tmp);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    char* key_out = NULL;
    const char* floor_key = NULL;
    cgul_trie_t trie = NULL;
    cgul_trie_node_t node = NULL;
    cgul_exception_t local = NULL;

    /* Allocate*/
    trie = cgul_trie__new(&local);
    if (local) {
        goto out;
    }

    /* Make sure an empty trie returns NULL as the floor for the empty
     * string. */
    node = cgul_trie__find_floor(&local, trie, "");
    cgul_libc__printf(&local, "empty trie && floor(\"\") => node=%p\n", node);
    assert(!node);

    /* Make sure an empty trie returns NULL as the floor for a
     * non-empty string. */
    node = cgul_trie__find_floor(&local, trie, "foo");
    cgul_libc__printf(&local, "floor(\"foo\") => node=%p\n", node);
    assert(!node);

    /* Insert a key larger than "foo". */
    main__insert_key(&local, trie, "quux");
    if (local) {
        goto out;
    }

    /* Make sure the floor is still NULL for the empty string. */
    node = cgul_trie__find_floor(&local, trie, "");
    cgul_libc__printf(&local, "\"quux\" key && floor(\"\") node=%p\n", node);
    assert(!node);

    /* Make sure the floor is still NULL for a non-empty string. */
    node = cgul_trie__find_floor(&local, trie, "foo");
    cgul_libc__printf(&local, "floor(\"foo\") node=%p\n", node);
    assert(!node);

    /* Insert the empty string. */
    main__insert_key(&local, trie, "");
    if (local) {
        goto out;
    }

    /* Make sure the floor is the empty string. */
    node = cgul_trie__find_floor(&local, trie, "foo");
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "empty key too && floor(\"foo\") => floor=-->%s<--\n", floor_key);
    assert(floor_key[0] == 0);

    /* Insert a key smaller than "foo". */
    main__insert_key(&local, trie, "baz");
    if (local) {
        goto out;
    }

    /* Make sure the floor is the smaller key. */
    node = cgul_trie__find_floor(&local, trie, "foo");
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "\"baz\" key too && floor(\"foo\") => floor=-->%s<--\n", floor_key);
    assert(strcmp(floor_key, "baz") == 0);

    /* Insert key with a "foo" prefix.  Note that "foo" is still not
     * inserted at this point. */
    main__insert_key(&local, trie, "foobar");
    if (local) {
        goto out;
    }

    /* Make sure the floor that is the prefix can be found. */
    node = cgul_trie__find_floor(&local, trie, "foo");
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "\"foobar\" key too && floor(\"foo\") => floor=-->%s<--\n", floor_key);
    assert(strcmp(floor_key, "baz") == 0);

    /* Insert "foo". */
    main__insert_key(&local, trie, "foo");
    if (local) {
        goto out;
    }

    /* Make sure the floor is now "foo". */
    node = cgul_trie__find_floor(&local, trie, "foo");
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "\"foo\" key too && floor(\"foo\") => floor=-->%s<--\n", floor_key);
    assert(strcmp(floor_key, "foo") == 0);

    /* Make sure search keys that pass "foo" but that do not reach
     * "foobar" find "foo" as the floor. */
    node = cgul_trie__find_floor(&local, trie, "foob");  /* sic */
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "floor(\"foob\") => floor=-->%s<--\n", floor_key);
    assert(strcmp(floor_key, "foo") == 0);

    /* Ditto, but one character more. */
    node = cgul_trie__find_floor(&local, trie, "fooba");  /* sic */
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "floor(\"fooba\") => floor=-->%s<--\n", floor_key);
    assert(strcmp(floor_key, "foo") == 0);

    /* Make sure we can now find "foobar" exactly. */
    node = cgul_trie__find_floor(&local, trie, "foobar");
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "floor(\"foobar\") => floor=-->%s<--\n", floor_key);
    assert(strcmp(floor_key, "foobar") == 0);

    /* Make sure we can now find "foobar" when it is the prefix of a
     * longer string. */
    node = cgul_trie__find_floor(&local, trie, "foobarbaz");
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "floor(\"foobarbaz\") => floor=-->%s<--\n", floor_key);
    assert(strcmp(floor_key, "foobar") == 0);

    /* Make sure we can now find "foobar" when it is less than a
     * string that varies only in the last character. */
    node = cgul_trie__find_floor(&local, trie, "foobaz");
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "floor(\"foobarbaz\") => floor=-->%s<--\n", floor_key);
    assert(strcmp(floor_key, "foobar") == 0);

    /* Remove the "baz" key. */
    if (!cgul_trie__remove(&local, trie, "baz", &key_out, NULL)) {
        local = cgul_exception__new("Unable to remove key: baz");
        goto out;
    }
    free(key_out);
    key_out = NULL;

    /* Make sure the floor of "fo" [sic] is the empty string.  This
     * tests the to make sure the floor function doesn't pick up the
     * "b" node that is still in the tree but without a "left" or
     * "key" even though it has to pass through this node to reach the
     * "f" in "foo." */
    node = cgul_trie__find_floor(&local, trie, "fo");  /* sic */
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "\"bar\" key removed && floor(\"fo\") => floor=-->%s<--\n", floor_key);
    assert(strcmp(floor_key, "") == 0);

    /* Insert "fog". */
    main__insert_key(&local, trie, "fog");
    if (local) {
        goto out;
    }

    /* Make sure "fog" is the floor for "fom".  This tests passing
     * through the "fo" node which has a "left" subtree but no
     * "key". */
    node = cgul_trie__find_floor(&local, trie, "fom");
    assert(node);
    floor_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "\"fog\" key too && floor(\"fom\") => floor=-->%s<--\n", floor_key);
    assert(strcmp(floor_key, "fog") == 0);    

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (trie) {
        cgul_trie__free_keys(trie);
        cgul_trie__delete(trie);
    }

    return rv;
}
