/**
 * This program prints a range of Unicode characters in UTF-8 format.
 */

#include "cgul.h"

/**
 * The program name.
 */
static const char* progname = "main_unicode_range";

/**
 * Inclusive starting code point for the range of Unicode characters.
 */
static cgul_wchar_t main_start;

/**
 * Inclusive stopping code point for the range of Unicode characters.
 */
static cgul_wchar_t main_stop;

/**
 * Whether the user wants help.
 */
static int user_wants_help = 0;

/**
 * Whether the user wants a leading byte-order mark (BOM) printed.
 * This is convenient under Windows where a leading BOM marks the file
 * as UTF-8.
 */
static int user_wants_bom = 0;

/**
 * Usage message.
 */
static char usage_message[] = "\
Usage: %s [options] <start> <stop>\n\
\n\
    Options:\n\
\n\
            -b : Print a leading byte-order mark (BOM).\n\
\n\
            -h : Print help message.\n\
\n\
    Example:\n\
\n\
        %s -b 0 0x10ffff > unicode.txt\n\
";

/**
 * Print the usage message to the file <code>f</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 */
static void
print_usage(cgul_exception_t* cex,
            FILE* f)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    /* Print the usage message filling in the program name. */
    cgul_libc__fprintf(cex, f, "\n");
    cgul_libc__fprintf(cex, f, usage_message, progname, progname);
    cgul_libc__fprintf(cex, f, "\n\n");
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Convert the string <code>s</code> to a <code>cgul_wchar_t</code>
 * and return the result.  If an error occurs,
 * <code>CGUL_WCHAR__NUL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s string
 * @return converstion of <code>s</code> to a <code>cgul_wchar_t</code>
 */
static cgul_wchar_t
string_to_wchar(cgul_exception_t* cex,
                const char* s)
{
    cgul_wchar_t rv = CGUL_WCHAR__NUL;
    unsigned long int n = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Load s into tmp. */
    cgul_string__set_value(cex, tmp, s);
    if (*cex) {
        goto out;
    }

    /* Convert to an unsigned long. */
    n = cgul_string__to_unsigned_long(cex, tmp, 0 /* base */);
    if (*cex) {
        /* Quash the original exception. */
        cgul_exception__delete(*cex);
        /* Throw a more descriptive exception. */
        *cex = cgul_exception__new("unable to convert to unicode: %s", s);
        goto out;
    }

    /* The valid Unicode range is 0 - 0x10ffff. */
    if (n > 0x10ffff) {
        *cex = cgul_exception__new("out of range for a unicode: %s", s);
        goto out;
    }

    /* Success. */
    rv = (cgul_wchar_t)n;

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

/**
 * Parse the command-line options and set global values.
 *
 * @param[in,out] cex c-style exception
 * @param[in] argc argument count
 * @param[in] argv argument vector
 */
static void
parse_options(cgul_exception_t* cex,
              int argc,
              char** argv)
{
    int i = 1;
    int is_start_set = 0;
    int is_stop_set = 0;
    const char* arg = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate over the command-line arguments. */
    for (i = 1 ; i < argc ; ++i) {
        arg = argv[i];

        /* -bom */
        if (   (strcmp(arg, "-b") == 0)
            || (strcmp(arg, "-bom") == 0)
            || (strcmp(arg, "--bom") == 0))
        {
            user_wants_bom = 1;
            continue;
        }

        /* -help */
        if (   (strcmp(arg, "-h") == 0)
            || (strcmp(arg, "-help") == 0)
            || (strcmp(arg, "--help") == 0))
        {
            user_wants_help = 1;
            goto out;
        }

        /* Check for invalid flags. */
        if (arg[0] == '-') {
            *cex = cgul_exception__new("invalid argument: %s", arg);
            goto out;
        }

        /* Check for the starting code point. */
        if (!is_start_set) {
            main_start = string_to_wchar(cex, arg);
            if (*cex) {
                goto out;
            }
            is_start_set = 1;
            continue;
        }

        /* Check for the stopping code point. */
        if (!is_stop_set) {
            main_stop = string_to_wchar(cex, arg);
            if (*cex) {
                goto out;
            }
            is_stop_set = 1;
            continue;
        }

        /* Invalid command line. */
        *cex = cgul_exception__new("invalid argument: %s", arg);
        goto out;
    }

    /* Make sure both main_start and main_stop have been set. */
    if (!is_start_set || !is_stop_set) {
        *cex = cgul_exception__new("invalid command line");
        goto out;
    }

    /* make sure main_start is less than main_stop. */
    if (main_start >= main_stop) {
        *cex = cgul_exception__new(
                   "starting point is greater than stopping point");
        goto out;
    }

 out:
    return;
}

/**
 * Print the inclusive range <code>[start - stop]</code> of Unicode
 * characters to the file <code>f</code> encoded as UTF-8.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @param[in] start inclusive starting Unicode code point
 * @param[in] stop inclusive stopping Unicode code point
 */
static void
print_range(cgul_exception_t* cex,
            FILE* f,
            cgul_wchar_t start,
            cgul_wchar_t stop)
{
    cgul_wchar_t i = CGUL_WCHAR__NUL;
    char buf[CGUL_WCHAR__MB_LEN_MAX + 1] = {0};

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    /* Bounds check start. */
    if (start > 0x10ffff) {
        *cex = cgul_exception__new("print_range: start is out of bounds: %lu",
                                   (unsigned long)start);
        goto out;
    }

    /* Bounds check stop. */
    if (stop > 0x10ffff) {
        *cex = cgul_exception__new("print_range: stop is out of bounds: %lu",
                                   (unsigned long)stop);
        goto out;
    }

    /* Write the byte-order mark (BOM) to mark the resulting file as
     * UTF-8 under Windows. */
    if (user_wants_bom) {
        cgul_unicode__wctomb(cex, 0xfeff, buf);
        cgul_libc__fprintf(cex, f, "%s", buf);
        if (*cex) {
            goto out;
        }
    }

    /* Iterate over each Unicode character in the range. */
    for (i = start ; i <= stop ; ++i) {

        /* Skip the UTF-16 surrogate pair range. */
        if ((i >= 0xd800) && (i <= 0xdfff)) {
            fprintf(stderr,
                    "Warning: skipping UTF-16 surrogate pair range: "
                    "[0xd800 - 0xdfff]\n");
            i = 0xdfff;
            continue;
        }

        /* Convert UTF-32 to UTF-8. */
        cgul_unicode__wctomb(cex, i, buf);
        if (*cex) {
            goto out;
        }

        /* Print UTF-8 to file. */
        cgul_libc__fprintf(cex, f, "%#08lx: %s\n", (unsigned long)i, buf);
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_exception_t local = NULL;

    /* Parse the command line options. */
    parse_options(&local, argc, argv);
    if (local) {
        goto out;
    }

    /* Check if the user wants help. */
    if (user_wants_help) {
        print_usage(&local, stdout);
        goto out;
    }

    /* Print the range of Unicode characters. */
    print_range(&local, stdout, main_start, main_stop);
    if (local) {
        goto out;
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    return rv;
}
