#include <assert.h>
#include "cgul.h"

/**
 * This function uses <code>stanza</code> to lookup the key
 * <code>varname</code>.  If the lookup is successful, it appends the
 * result to <code>dst</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] varname variable name
 * @param[in,out] dst destination string
 * @param[in] stanza <code>cgul_stanza</code> object
 * @return whether the lookup was successful
 */
static int
do_lookup(cgul_exception_t* cex,
          const char* varname,
          cgul_string_t dst,
          cgul_stanza_t stanza)
{
    int rv = 0;
    const char* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Perform the lookup. */
    value = cgul_stanza__lookup(cex, stanza, varname);
    if (value == NULL) {
        goto out;
    }

    /* Perform the substitution by appending "value" to "dst". */
    cgul_string__append(cex, dst, value);
    if (*cex) {
        goto out;
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    const char* src = NULL;
    cgul_string_t dst = NULL;
    cgul_crlf_file_t crlf = NULL;
    cgul_stanza_t stanza = NULL;
    cgul_exception_t local = NULL;

    /* Allocate dst. */
    dst = cgul_string__new(&local);
    if (local) {
        goto out;
    }

    /* Read the mapping of variable names to values. */
    stanza = cgul_stanza__new_from_fname(&local,
                                         "main_varsub_mapping.txt",
                                         1 /* is_block_buffered */);
    cgul_stanza__all_one_stanza(&local, stanza, 1);
    cgul_stanza__load_next(&local, stanza);
    if (local) {
        goto out;
    }

    /* Open the text file. */
    crlf = cgul_crlf_file__new_from_fname(&local,
                                          "main_varsub_text.txt",
                                          1 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /* Iterate over each line in the text file. */
    while ((src = cgul_crlf_file__get_line(&local, crlf)) != NULL) {

        /* Clear dst. */
        cgul_string__set_value(&local, dst, "");

        /* Perform variable substitution. */
        cgul_varsub(&local,
                    src,
                    dst,
                    (cgul_varsub__lookup_t)do_lookup,
                    stanza);

        /* Print results. */
        cgul_libc__puts(&local, cgul_string__get_value(&local, dst));

        /* Check for exceptions. */
        if (local) {
            goto out;
        }
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (crlf) {
        cgul_crlf_file__delete(crlf);
    }
    if (stanza) {
        cgul_stanza__delete(stanza);
    }
    if (dst) {
        cgul_string__delete(dst);
    }

    return rv;
}
